/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:         
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "XMLParser.hpp"
#include "nmm/utils/xml/xml++.hpp"

#include "../apps/menu/automenu.hpp"
#include "../apps/mp3/mp3.hpp"
#include "../apps/osdmanager/osdinput.hpp"
#include "../apps/taskmgr/taskmgr.hpp"
#include "../apps/config/config.hpp"
#include "../apps/dvb/dvb.hpp"
#include "../apps/cd/cd.hpp"
#include "../apps/grab/grab.hpp"
#include "../apps/grabdvd/grabdvd.hpp"
#include "../apps/dvd/dvd.hpp"

#ifdef BUILD_VISCACAMERA
#include "../apps/camera/camera.hpp"
#endif 

#include "../apps/playlist/playlist.hpp"
#include "../apps/tvtimer/tvtimer.hpp"
#include "../apps/epg/epg.hpp"
#include "../apps/epg/search.hpp"
#include "../apps/epg/search_result.hpp"
#include "../apps/slideshow/SlideShowState.hpp"
#include "../apps/transcode/transcode.hpp"
#include "../apps/webserver/webserver.hpp"
#include "nmm/plugins/video/osd/manager/OSDElements/OSDElementComposit.hpp"


using namespace NMM;

XmlParser::XmlParser( string name /*= string("")*/ ) {
  xml      = NULL;
  rootnode = NULL;
  path = string("");

  if (string("") != name) {
    open( name );
  }
}

XmlParser::~XmlParser() {
  /* delete list of AutoMenu pointers */
  state_list.clear();
}

bool XmlParser::open( string name ) {
  //cout << "Trying to open " << name << endl;
  xml = new XMLTree( name );

  /* get pointer to XML root */
  rootnode = xml->root();
  if ( rootnode == NULL ) {
    delete xml;
    xml = NULL;
    return false;
  }

  if ( rootnode->name() != "configuration" ) {
    rootnode = NULL;
    delete xml;
    xml = NULL;
    return false;
  }

  parseResourcePath();

  return true;
}

const string XmlParser::getResourcePath() {
  return path;
}

void XmlParser::setResourcePath(const string _path) {
  path = _path;
}

void XmlParser::parseStdState( XMLNode* node, MMBoxState* state ) {
  if ( node && state ) {
    XMLProperty* prop;

    /* Try to find "background" attribute */
    prop = node->property( "background" );
    if ( prop ) state->setBackground( path+prop->value() );

    /* Try to find "needvideo" attribute */
    prop = node->property( "needvideo" );
    if ( prop ) {
      if ( prop->value() == "no" ) state->setLazyBackground( true );
      else state->setLazyBackground( false );
    }

    /* Try to find "needaudio" attribute */
    prop = node->property( "needaudio" );
    if ( prop ) {
      if ( prop->value() == "yes" ) state->setNeedAudio( true );
      else state->setNeedAudio( false );
    }

  }

}

state_list_t* XmlParser::parse( IBitmapReader* bmreader, MMBoxApplication* app ) {
  /* first delete all previously added States (only the pointer list) */
  state_list.clear();
  SMMBoxWebServer::getInstance().start(app);

  parseMenu( bmreader, app );
  parsePlaylist( bmreader, app );
  parseCDPlay( bmreader, app );
  parseCDGrabber( bmreader, app );
  parseDVDGrabber( bmreader, app );
  parseMP3Play( bmreader, app );
  parseDVBPlay( bmreader, app );
  parseEPG( bmreader, app);
  parseEPGSearch( bmreader, app);
  parseEPGSearchResult( bmreader, app );
  parseSlideShow( bmreader, app);
  parseTranscode( bmreader, app);
  parseDVDPlay( bmreader, app );

#ifdef BUILD_VISCACAMERA
  parseCamera( bmreader, app );
#endif

  parseTvTimer( bmreader, app );
  parseConfig( bmreader, app );
  parseOSD( bmreader, app );
  parseTaskMgr( bmreader, app );
  parseOSDSymbols( bmreader, app );
  parseSettings( app );

  return( &state_list );
}

bool XmlParser::parseMenu( IBitmapReader* bmreader, MMBoxApplication* app ) {

  /* check if we opened a valid XML file */
  if ( !rootnode ) return false;

  /* Get a list with all <menu> entries */
  XMLNodeList menu_list;
  menu_list = rootnode->children( "menu" );

  XMLNodeIterator iter, stop;
  XMLNodeIterator entry_iter, entry_stop;
  XMLNodeList     entry_list;

  AutoMenu* menustate;

  iter = menu_list.begin();
  stop = menu_list.end();

  /* Process each <menu> entry */
  while (iter != stop) {
    XMLNode* node = *iter;
    iter++;

    XMLProperty* prop;

    /* Try to find "id" attribute */
    string id;
    prop = node->property( "id" );
    if ( prop ) id = prop->value();

    /* Try to find "columns" attribute */
    string columns;
    prop = node->property( "columns" );
    if ( prop ) columns = prop->value();

    /* Create AutoMenu instance with id that was found in XML file */
    menustate = new AutoMenu( id, app );
    parseStdState( node, menustate );

    /* add AutoMenu to list */
    state_list.push_back( menustate );

    if ( columns != "" ) menustate->setStep( atoi( columns.c_str() ) );

    /* Get a list with all <entry> tags */
    entry_list = node->children( "entry" );

    entry_iter = entry_list.begin();
    entry_stop = entry_list.end();

    /* process all <entry> tags */
    while( entry_iter != entry_stop ) {
      XMLNode* enode = *entry_iter;
      entry_iter++;

      string e_id         = enode->getData();

      string e_index;
      prop = enode->property( "index" );
      if ( prop ) e_index = prop->value();

      string e_on;
      prop = enode->property( "on" );
      if ( prop ) e_on = path+prop->value();

      string e_off;
      prop = enode->property( "off" );
      if ( prop ) e_off = path+prop->value();

      string e_x;
      prop = enode->property( "x" );
      if ( prop ) e_x = prop->value();

      string e_y;
      prop = enode->property( "y" );
      if ( prop ) e_y = prop->value();

      GLOBAL_DEBUG_STREAM( e_id << " " << " " << e_index << " " << e_on << " " << e_off << "..." << endl );

      /* load bitmap for button ON state, the bitmap id is created by id, e_id and "ON" string */
      /* for example: MainMenuDvdPlayerON */
      GLOBAL_DEBUG_STREAM( "Loading Bitmap " << e_on << " with ID: " << id+e_id <<"ON" << endl );
      bmreader->loadImage( id+e_id+"ON"  , e_on );

      /* load bitmap for button OFF state, the bitmap id is created by id, e_id and "OFF" string */
      bmreader->loadImage( id+e_id+"OFF" , e_off );

      /* the buttons id is the tagdata, for example DvdPlayer */
      menustate->addButton( atoi( e_index.c_str() ),
      			    e_id,
			    atoi( e_x.c_str() ), atoi( e_y.c_str() ),
			    id+e_id+"OFF", id+e_id+"ON", id+e_id+"ON" );

    } // while( entry != entry_stop )
    menustate->setDimension( 720, 576 );

  } // while (iter != stop)

  return( true );
}

bool XmlParser::parsePlaylist( IBitmapReader* /*bmreader*/, MMBoxApplication* app ) {

  /* check if we opened a valid XML file */
  if ( !rootnode ) return false;

  /* Get a list with all <playlist> entries */
  XMLNodeList playlist_list;
  playlist_list = rootnode->children( "playlist" );

  XMLNodeIterator iter, stop;
  XMLNodeIterator entry_iter, entry_stop;
  XMLNodeList     entry_list;

  iter = playlist_list.begin();
  stop = playlist_list.end();

  /* Process each <playlist> entry */
  while (iter != stop) {
    XMLNode* node = *iter;
    iter++;

    XMLProperty* prop;

     /* Try to find "id" attribute */
    string str_pid;
    prop = node->property( "id" );
    if ( prop ) str_pid = prop->value();

    Playlist* playlist_state = new Playlist( str_pid, app );
    parseStdState( node, playlist_state );

     /* Get a list with all <win> tags */
    entry_list = node->children( "win" );

    entry_iter = entry_list.begin();
    entry_stop = entry_list.end();

    /* process all <entry> tags */
    while( entry_iter != entry_stop ) {
      XMLNode* enode = *entry_iter;
      entry_iter++;

      OSDElementComposit::win_properties_t props;

      int win;
      /* Try to find "id" attribute */
      string str_id;
      prop = enode->property( "id" );
      if ( prop ) str_id = prop->value();

      /* Try to find "x" attribute */
      string temp;
      prop = enode->property( "x" );
      if ( prop ) temp = prop->value();
      props.xpos = atoi( temp.c_str() );

      /* Try to find "y" attribute */
      prop = enode->property( "y" );
      if ( prop ) temp = prop->value();
      props.ypos = atoi( temp.c_str() );

      /* Try to find "width" attribute */
      prop = enode->property( "width" );
      if ( prop ) temp = prop->value();
      props.width = atoi( temp.c_str() );

      /* Try to find "height" attribute */
      prop = enode->property( "height" );
      if ( prop ) temp = prop->value();
      props.height = atoi( temp.c_str() );

      /* Try to find "win_active_color" attribute */
      string str_color;
      prop = enode->property( "win_active_color" );
      if ( prop ) str_color = prop->value();
      props.win_active_color = strToInt( str_color );

      /* Try to find "win_inactive_color" attribute */
      prop = enode->property( "win_inactive_color" );
      if ( prop ) str_color = prop->value();
      props.win_inactive_color = strToInt( str_color );

      /* Try to find "sel_active_color" attribute */
      prop = enode->property( "sel_active_color" );
      if ( prop ) str_color = prop->value();
      props.sel_active_color = strToInt( str_color );

      /* Try to find "sel_active_color2" attribute */
      prop = enode->property( "sel_active_color2" );
      if ( prop ) str_color = prop->value();
      props.sel_active_color2 = strToInt( str_color );

      /* Try to find "sel_inactive_color" attribute */
      prop = enode->property( "sel_inactive_color" );
      if ( prop ) str_color = prop->value();
      props.sel_inactive_color = strToInt( str_color );


      if ( str_id == "Selection" )     win = 0;//Playlist::SELECTION;
      else if ( str_id == "Track" )    win = 1;//Playlist::TRACK;
      else if ( str_id == "Playlist" ) win = 2;//Playlist::PLAYLIST;
      playlist_state->setWinProperties( win, &props );

    } // while( entry != entry_stop )

    /* add State to list */
    state_list.push_back( playlist_state );

  } // while (iter != stop)

  return( true );
}

bool XmlParser::parseTranscode( IBitmapReader* /*bmreader*/, MMBoxApplication* app ) {

  /* check if we opened a valid XML file */
  if ( !rootnode ) return false;

  /* Get a list with all <playlist> entries */
  XMLNodeList transcode_list;
  transcode_list = rootnode->children( "transcode" );

  XMLNodeIterator iter, stop;
  XMLNodeIterator entry_iter, entry_stop;
  XMLNodeList     entry_list;

  iter = transcode_list.begin();
  stop = transcode_list.end();

  /* Process each <playlist> entry */
  while (iter != stop) {
    XMLNode* node = *iter;
    iter++;

    XMLProperty* prop;

     /* Try to find "id" attribute */
    string str_pid;
    prop = node->property( "id" );
    if ( prop ) str_pid = prop->value();

    Transcode* transcode_state = new Transcode( str_pid, app );
    parseStdState( node, transcode_state );

    /* Get a list with all <showoption> entries */
    XMLNodeList showoption_list;
    showoption_list = node->children( "showoption" );
    iter = showoption_list.begin();
    stop = showoption_list.end();

    /* Process each <showoption> entry */
    while (iter != stop) {
      XMLNode* node = *iter;
      iter++;
      
      XMLProperty* prop;
      
      /* Try to find "id" attribute */
      string id = "";
      prop = node->property( "id" );
      if ( prop ) id = prop->value();
      
      string shortform = node->getData();
      
      transcode_state->showOption( id, shortform );
    }
    
     /* Get a list with all <win> tags */
    entry_list = node->children( "win" );

    entry_iter = entry_list.begin();
    entry_stop = entry_list.end();

    /* process all <entry> tags */
    while( entry_iter != entry_stop ) {
      XMLNode* enode = *entry_iter;
      entry_iter++;

      OSDElementComposit::win_properties_t props;

      int win;
      /* Try to find "id" attribute */
      string str_id;
      prop = enode->property( "id" );
      if ( prop ) str_id = prop->value();

      /* Try to find "x" attribute */
      string temp;
      prop = enode->property( "x" );
      if ( prop ) temp = prop->value();
      props.xpos = atoi( temp.c_str() );

      /* Try to find "y" attribute */
      prop = enode->property( "y" );
      if ( prop ) temp = prop->value();
      props.ypos = atoi( temp.c_str() );

      /* Try to find "width" attribute */
      prop = enode->property( "width" );
      if ( prop ) temp = prop->value();
      props.width = atoi( temp.c_str() );

      /* Try to find "height" attribute */
      prop = enode->property( "height" );
      if ( prop ) temp = prop->value();
      props.height = atoi( temp.c_str() );

      /* Try to find "win_active_color" attribute */
      string str_color;
      prop = enode->property( "win_active_color" );
      if ( prop ) str_color = prop->value();
      props.win_active_color = strToInt( str_color );

      /* Try to find "win_inactive_color" attribute */
      prop = enode->property( "win_inactive_color" );
      if ( prop ) str_color = prop->value();
      props.win_inactive_color = strToInt( str_color );

      /* Try to find "sel_active_color" attribute */
      prop = enode->property( "sel_active_color" );
      if ( prop ) str_color = prop->value();
      props.sel_active_color = strToInt( str_color );

      /* Try to find "sel_active_color2" attribute */
      prop = enode->property( "sel_active_color2" );
      if ( prop ) str_color = prop->value();
      props.sel_active_color2 = strToInt( str_color );

      /* Try to find "sel_inactive_color" attribute */
      prop = enode->property( "sel_inactive_color" );
      if ( prop ) str_color = prop->value();
      props.sel_inactive_color = strToInt( str_color );


      if ( str_id == "Selection" )     win = 0;//Playlist::SELECTION;
      else if ( str_id == "Track" )    win = 1;//Playlist::TRACK;
      else if ( str_id == "Playlist" ) win = 2;//Playlist::PLAYLIST;
      else if ( str_id == "GD")        win = 3;//TranscodeConfig:GD_win
      else if ( str_id == "Conf1")     win = 5;//TranscdoeConfig:Conf1_win
      else if ( str_id == "Conf2")     win = 6;//TranscdoeConfig:Conf3_win
      else if ( str_id == "Comment")   win = 4;//TransccodeConfig:comment_win
      else if ( str_id == "Summary")   win = 7;//TranscodeConfig:summ_win
      //else if ( str_id == "Input")     win = 9;//TranscodeConfig::input_win
      transcode_state->setWinProperties( win, &props );

    } // while( entry != entry_stop )

    /* add State to list */
    state_list.push_back( transcode_state );

  } // while (iter != stop)

  return( true );
}

bool XmlParser::parseCDPlay( IBitmapReader* /*bmreader*/, MMBoxApplication* app ) {

   /* check if we opened a valid XML file */
   if ( !rootnode ) return false;

   /* Get a list with all <playlist> entries */
   XMLNodeList cdplay_list;
   cdplay_list = rootnode->children( "cdplay" );

   XMLNodeIterator iter, stop;
   //XMLNodeIterator entry_iter, entry_stop;
   //XMLNodeList     entry_list;

   iter = cdplay_list.begin();
   stop = cdplay_list.end();

   /* Process each <cdplay> entry */
   while (iter != stop) {
     XMLNode* node = *iter;
     iter++;

     XMLProperty* prop;

      /* Try to find "id" attribute */
     string str_pid;
     prop = node->property( "id" );
     if ( prop ) str_pid = prop->value();

     CdState* cdplay_state = new CdState( str_pid, app );
     parseStdState( node, cdplay_state );

     /* add State to list */
     state_list.push_back( cdplay_state );

   } // while (iter != stop)

  return( true );
}

bool XmlParser::parseCDGrabber( IBitmapReader* /*bmreader*/, MMBoxApplication* app ) {

  /* check if we opened a valid XML file */
  if ( !rootnode ) return false;

  /* Get a list with all <playlist> entries */
  XMLNodeList cdgrab_list;
  cdgrab_list = rootnode->children( "cdgrab" );

  XMLNodeIterator iter, stop;
  //XMLNodeIterator entry_iter, entry_stop;
  //XMLNodeList     entry_list;

  iter = cdgrab_list.begin();
  stop = cdgrab_list.end();

  /* Process each <cdgrab> entry */
  while (iter != stop) {
    XMLNode* node = *iter;
    iter++;

    XMLProperty* prop;

     /* Try to find "id" attribute */
    string str_pid;
    prop = node->property( "id" );
    if ( prop ) str_pid = prop->value();

    GrabState* cdgrab_state = new GrabState( str_pid, app );
    parseStdState( node, cdgrab_state );

    /* add State to list */
    state_list.push_back( cdgrab_state );

  } // while (iter != stop)

  return( true );
}

bool XmlParser::parseDVDGrabber( IBitmapReader* /*bmreader*/, MMBoxApplication* app ) {

  /* check if we opened a valid XML file */
  if ( !rootnode ) return false;

  /* Get a list with all <dvdgrab> entries */
  XMLNodeList dvdgrab_list;
  dvdgrab_list = rootnode->children( "dvdgrab" );

  XMLNodeIterator iter, stop;
  //XMLNodeIterator entry_iter, entry_stop;
  //XMLNodeList     entry_list;

  iter = dvdgrab_list.begin();
  stop = dvdgrab_list.end();

  /* Process each <dvdgrab> entry */
  while (iter != stop) {
    XMLNode* node = *iter;
    iter++;

    XMLProperty* prop;

     /* Try to find "id" attribute */
    string str_pid;
    prop = node->property( "id" );
    if ( prop ) str_pid = prop->value();

    DVDGrabState* dvdgrab_state = new DVDGrabState( str_pid, app );
    parseStdState( node, dvdgrab_state );

    /* Get a list with all <showoption> entries */
    XMLNodeList showoption_list;
    showoption_list = node->children( "showoption" );
    iter = showoption_list.begin();
    stop = showoption_list.end();

    /* Process each <showoption> entry */
    while (iter != stop) {
      XMLNode* node = *iter;
      iter++;
      
      XMLProperty* prop;
      
      /* Try to find "id" attribute */
      string id = "";
      prop = node->property( "id" );
      if ( prop ) id = prop->value();
      
      string shortform = node->getData();
      
      dvdgrab_state->showOption( id, shortform );
    }
    
    /* add State to list */
    state_list.push_back( dvdgrab_state );

  } // while (iter != stop)

  return( true );
}

bool XmlParser::parseMP3Play( IBitmapReader* /*bmreader*/, MMBoxApplication* app ) {

  /* check if we opened a valid XML file */
  if ( !rootnode ) return false;

  /* Get a list with all <playlist> entries */
  XMLNodeList mp3play_list;
  mp3play_list = rootnode->children( "mp3play" );

  XMLNodeIterator iter, stop;
  //XMLNodeIterator entry_iter, entry_stop;
  //XMLNodeList     entry_list;

  iter = mp3play_list.begin();
  stop = mp3play_list.end();

  /* Process each <cdplay> entry */
  while (iter != stop) {
    XMLNode* node = *iter;
    iter++;

    XMLProperty* prop;

     /* Try to find "id" attribute */
    string str_pid;
    prop = node->property( "id" );
    if ( prop ) str_pid = prop->value();

    MP3State* mp3_state = new MP3State( str_pid, app );
    parseStdState( node, mp3_state );

    /* add State to list */
    state_list.push_back( mp3_state );

  } // while (iter != stop)

  return( true );
}

bool XmlParser::parseSlideShow( IBitmapReader* /* bmreader */, MMBoxApplication* app )
{
  XMLNodeList slideshow_list;
  slideshow_list = rootnode->children( "slideshow" );

  XMLNodeIterator iter, stop;

  iter = slideshow_list.begin();
  stop = slideshow_list.end();

  while (iter != stop) {
    XMLNode* node = *iter;
    iter++;

    XMLProperty* prop;

    string str_pid;
    prop = node->property( "id" );
    if ( prop ) str_pid = prop->value();

    SlideShowState* slideshowstate_state =
       new SlideShowState( str_pid, app );

    parseStdState( node, slideshowstate_state );

    state_list.push_back( slideshowstate_state );

  }

  return( true );
}



bool XmlParser::parseDVDPlay( IBitmapReader* /*bmreader*/, MMBoxApplication* app ) {
  
  /* check if we opened a valid XML file */
  if ( !rootnode ) return false;

  /* Get a list with all <playlist> entries */
  XMLNodeList dvdplay_list;
  dvdplay_list = rootnode->children( "dvdplay" );

  XMLNodeIterator iter, stop;
  //XMLNodeIterator entry_iter, entry_stop;
  //XMLNodeList     entry_list;

  iter = dvdplay_list.begin();
  stop = dvdplay_list.end();

  /* Process each <cdplay> entry */
  while (iter != stop) {
    XMLNode* node = *iter;
    iter++;

    XMLProperty* prop;

     /* Try to find "id" attribute */
    string str_pid;
    prop = node->property( "id" );
    if ( prop ) str_pid = prop->value();

    DvdState* dvd_state = new DvdState( str_pid, app );
    parseStdState( node, dvd_state );

    /* add State to list */
    state_list.push_back( dvd_state );

  } // while (iter != stop)

  return( true );
}

#ifdef BUILD_VISCACAMERA
bool XmlParser::parseCamera( IBitmapReader* /*bmreader*/, MMBoxApplication* app ) {
  
  /* check if we opened a valid XML file */
  if ( !rootnode ) return false;

  /* Get a list with all <playlist> entries */
  XMLNodeList camera_list;
  camera_list = rootnode->children( "camera" );

  XMLNodeIterator iter, stop;
  //XMLNodeIterator entry_iter, entry_stop;
  //XMLNodeList     entry_list;

  iter = camera_list.begin();
  stop = camera_list.end();

  /* Process each <cdplay> entry */
  while (iter != stop) {
    XMLNode* node = *iter;
    iter++;

    XMLProperty* prop;

     /* Try to find "id" attribute */
    string str_pid;
    prop = node->property( "id" );
    if ( prop ) str_pid = prop->value();

    CameraState* camera_state = new CameraState( str_pid, app );
    parseStdState( node, camera_state );

    /* add State to list */
    state_list.push_back( camera_state );

  } // while (iter != stop)

  return( true );
}
#endif

bool XmlParser::parseDVBPlay( IBitmapReader* /*bmreader*/, MMBoxApplication* app ) {

  /* check if we opened a valid XML file */
  if ( !rootnode ) return false;

  /* Get a list with all <playlist> entries */
  XMLNodeList dvbplay_list;
  dvbplay_list = rootnode->children( "dvbplay" );

  XMLNodeIterator iter, stop;
  //XMLNodeIterator entry_iter, entry_stop;
  //XMLNodeList     entry_list;

  iter = dvbplay_list.begin();
  stop = dvbplay_list.end();

  /* Process each <cdplay> entry */
  while (iter != stop) {
    XMLNode* node = *iter;
    iter++;

    XMLProperty* prop;

     /* Try to find "id" attribute */
    string str_pid;
    prop = node->property( "id" );
    if ( prop ) str_pid = prop->value();

    DvbState* dvb_state = new DvbState( str_pid, app );
    parseStdState( node, dvb_state );

    /* add State to list */
    state_list.push_back( dvb_state );

  } // while (iter != stop)

  return( true );
}

bool XmlParser::parseEPG( IBitmapReader* /* bmreader */, MMBoxApplication* app ){
/* check if we opened a valid XML file */
  if ( !rootnode ) return false;

  /* Get a list with all <playlist> entries */
  XMLNodeList epg_list;
  epg_list = rootnode->children( "epg" );

  XMLNodeIterator iter, stop;
  XMLNodeIterator entry_iter, entry_stop;
  XMLNodeList     entry_list;

  iter = epg_list.begin();
  stop = epg_list.end();

  /* Process each <playlist> entry */
  while (iter != stop) {
    XMLNode* node = *iter;
    iter++;

    XMLProperty* prop;

     /* Try to find "id" attribute */
    string str_pid;
    prop = node->property( "id" );
    if ( prop ) str_pid = prop->value();

    EPG* epg_state = new EPG( str_pid, app );
    parseStdState( node, epg_state );

     /* Get a list with all <win> tags */
    entry_list = node->children( "win" );

    entry_iter = entry_list.begin();
    entry_stop = entry_list.end();

    /* process all <entry> tags */
    while( entry_iter != entry_stop ) {
      XMLNode* enode = *entry_iter;
      entry_iter++;

      EPG::win_properties_t props;
      EPG::wintype_e        win;

      /* Try to find "id" attribute */
      string str_id;
      prop = enode->property( "id" );
      if ( prop ) str_id = prop->value();

      /* Try to find "x" attribute */
      string temp;
      prop = enode->property( "x" );
      if ( prop ) temp = prop->value();
      props.xpos = atoi( temp.c_str() );

      /* Try to find "y" attribute */
      prop = enode->property( "y" );
      if ( prop ) temp = prop->value();
      props.ypos = atoi( temp.c_str() );

      /* Try to find "width" attribute */
      prop = enode->property( "width" );
      if ( prop ) temp = prop->value();
      props.width = atoi( temp.c_str() );

      /* Try to find "height" attribute */
      prop = enode->property( "height" );
      if ( prop ) temp = prop->value();
      props.height = atoi( temp.c_str() );

      /* Try to find "win_active_color" attribute */
      string str_color;
      prop = enode->property( "win_active_color" );
      if ( prop ) str_color = prop->value();
      props.win_active_color = strToInt( str_color );

      /* Try to find "win_inactive_color" attribute */
      prop = enode->property( "win_inactive_color" );
      if ( prop ) str_color = prop->value();
      props.win_inactive_color = strToInt( str_color );

      /* Try to find "sel_active_color" attribute */
      prop = enode->property( "sel_active_color" );
      if ( prop ) str_color = prop->value();
      props.sel_active_color = strToInt( str_color );

      /* Try to find "sel_active_color2" attribute */
      prop = enode->property( "sel_active_color2" );
      if ( prop ) str_color = prop->value();
      props.sel_active_color2 = strToInt( str_color );

      /* Try to find "sel_inactive_color" attribute */
      prop = enode->property( "sel_inactive_color" );
      if ( prop ) str_color = prop->value();
      props.sel_inactive_color = strToInt( str_color );


      if ( str_id == "Channels" )          win = EPG::CHANNELS;
      else if ( str_id == "Schedule" )     win = EPG::SCHEDULE;
      else if ( str_id == "ExtendedInfo" ) win = EPG::EXTINFO;
      else if ( str_id == "Days" )         win = EPG::DAYS;
      epg_state->setWinProperties( win, &props );

    } // while( entry != entry_stop )

    /* add State to list */
    state_list.push_back( epg_state );

  } // while (iter != stop)

  return( true );
}//end parseEPG


bool XmlParser::parseEPGSearch( IBitmapReader* /* bmreader */, MMBoxApplication* app ){
/* check if we opened a valid XML file */
  if ( !rootnode ) return false;

  /* Get a list with all <playlist> entries */
  XMLNodeList epgsearch_list;
  epgsearch_list = rootnode->children( "epgsearch" );

  XMLNodeIterator iter, stop;
  XMLNodeIterator entry_iter, entry_stop;
  XMLNodeList     entry_list;

  iter = epgsearch_list.begin();
  stop = epgsearch_list.end();

  /* Process each <EPGSearch> entry */
  while (iter != stop) {
    XMLNode* node = *iter;
    iter++;

    XMLProperty* prop;

     /* Try to find "id" attribute */
    string str_pid;
    prop = node->property( "id" );
    if ( prop ) str_pid = prop->value();

    EPGSearch* epgsearch_state = new EPGSearch( str_pid, app );
    parseStdState( node, epgsearch_state );

     /* Get a list with all <win> tags */
    entry_list = node->children( "win" );

    entry_iter = entry_list.begin();
    entry_stop = entry_list.end();

    /* process all <entry> tags */
    while( entry_iter != entry_stop ) {
      XMLNode* enode = *entry_iter;
      entry_iter++;

      EPGSearch::win_properties_t props;
      EPGSearch::wintype_e        win;

      /* Try to find "id" attribute */
      string str_id;
      prop = enode->property( "id" );
      if ( prop ) str_id = prop->value();

      /* Try to find "x" attribute */
      string temp;
      prop = enode->property( "x" );
      if ( prop ) temp = prop->value();
      props.xpos = atoi( temp.c_str() );

      /* Try to find "y" attribute */
      prop = enode->property( "y" );
      if ( prop ) temp = prop->value();
      props.ypos = atoi( temp.c_str() );

      /* Try to find "width" attribute */
      prop = enode->property( "width" );
      if ( prop ) temp = prop->value();
      props.width = atoi( temp.c_str() );

      /* Try to find "height" attribute */
      prop = enode->property( "height" );
      if ( prop ) temp = prop->value();
      props.height = atoi( temp.c_str() );

      /* Try to find "win_active_color" attribute */
      string str_color;
      prop = enode->property( "win_active_color" );
      if ( prop ) str_color = prop->value();
      props.win_active_color = strToInt( str_color );

      /* Try to find "win_inactive_color" attribute */
      prop = enode->property( "win_inactive_color" );
      if ( prop ) str_color = prop->value();
      props.win_inactive_color = strToInt( str_color );

      /* Try to find "sel_active_color" attribute */
      prop = enode->property( "sel_active_color" );
      if ( prop ) str_color = prop->value();
      props.sel_active_color = strToInt( str_color );

      /* Try to find "sel_active_color2" attribute */
      prop = enode->property( "sel_active_color2" );
      if ( prop ) str_color = prop->value();
      props.sel_active_color2 = strToInt( str_color );

      /* Try to find "sel_inactive_color" attribute */
      prop = enode->property( "sel_inactive_color" );
      if ( prop ) str_color = prop->value();
      props.sel_inactive_color = strToInt( str_color );


      if ( str_id == "Bookmarks" )          win = EPGSearch::BOOKMARKS;
      else if ( str_id == "StartTime" )     win = EPGSearch::STARTTIME;
      else if ( str_id == "EndTime" )       win = EPGSearch::ENDTIME;
      else if ( str_id == "Phrase" )        win = EPGSearch::PHRASE;
      else if ( str_id == "Channel" )       win = EPGSearch::CHANNEL;
      else if ( str_id == "CategoryInput" ) win = EPGSearch::CATINPUT;
      else if ( str_id == "Category" )      win = EPGSearch::CATEGORY;
      epgsearch_state->setWinProperties( win, &props );

    } // while( entry != entry_stop )

    /* add State to list */
    state_list.push_back( epgsearch_state );

  } // while (iter != stop)

  return( true );
}//end parseEPGSearch

bool XmlParser::parseEPGSearchResult( IBitmapReader* /* bmreader */, MMBoxApplication* app ){
/* check if we opened a valid XML file */
  if ( !rootnode ) return false;

  /* Get a list with all <playlist> entries */
  XMLNodeList epgsearchresult_list;
  epgsearchresult_list = rootnode->children( "epgsearchresult" );

  XMLNodeIterator iter, stop;
  XMLNodeIterator entry_iter, entry_stop;
  XMLNodeList     entry_list;

  iter = epgsearchresult_list.begin();
  stop = epgsearchresult_list.end();

  /* Process each <EPGSearchResult> entry */
  while (iter != stop) {
    XMLNode* node = *iter;
    iter++;

    XMLProperty* prop;

     /* Try to find "id" attribute */
    string str_pid;
    prop = node->property( "id" );
    if ( prop ) str_pid = prop->value();

    EPGSearchResult* epgsearchresult_state = new EPGSearchResult( str_pid, app );
    parseStdState( node, epgsearchresult_state );

     /* Get a list with all <win> tags */
    entry_list = node->children( "win" );

    entry_iter = entry_list.begin();
    entry_stop = entry_list.end();

    /* process all <entry> tags */
    while( entry_iter != entry_stop ) {
      XMLNode* enode = *entry_iter;
      entry_iter++;

      EPGSearchResult::win_properties_t props;
      EPGSearchResult::wintype_e        win;

      /* Try to find "id" attribute */
      string str_id;
      prop = enode->property( "id" );
      if ( prop ) str_id = prop->value();

      /* Try to find "x" attribute */
      string temp;
      prop = enode->property( "x" );
      if ( prop ) temp = prop->value();
      props.xpos = atoi( temp.c_str() );

      /* Try to find "y" attribute */
      prop = enode->property( "y" );
      if ( prop ) temp = prop->value();
      props.ypos = atoi( temp.c_str() );

      /* Try to find "width" attribute */
      prop = enode->property( "width" );
      if ( prop ) temp = prop->value();
      props.width = atoi( temp.c_str() );

      /* Try to find "height" attribute */
      prop = enode->property( "height" );
      if ( prop ) temp = prop->value();
      props.height = atoi( temp.c_str() );

      /* Try to find "win_active_color" attribute */
      string str_color;
      prop = enode->property( "win_active_color" );
      if ( prop ) str_color = prop->value();
      props.win_active_color = strToInt( str_color );

      /* Try to find "win_inactive_color" attribute */
      prop = enode->property( "win_inactive_color" );
      if ( prop ) str_color = prop->value();
      props.win_inactive_color = strToInt( str_color );

      /* Try to find "sel_active_color" attribute */
      prop = enode->property( "sel_active_color" );
      if ( prop ) str_color = prop->value();
      props.sel_active_color = strToInt( str_color );

      /* Try to find "sel_active_color2" attribute */
      prop = enode->property( "sel_active_color2" );
      if ( prop ) str_color = prop->value();
      props.sel_active_color2 = strToInt( str_color );

      /* Try to find "sel_inactive_color" attribute */
      prop = enode->property( "sel_inactive_color" );
      if ( prop ) str_color = prop->value();
      props.sel_inactive_color = strToInt( str_color );


      if ( str_id == "ShortResult" )        win = EPGSearchResult::SHORTRESULT;
      else if ( str_id == "ExtResult" )     win = EPGSearchResult::EXTRESULT;
      epgsearchresult_state->setWinProperties( win, &props );

    } // while( entry != entry_stop )

    /* add State to list */
    state_list.push_back( epgsearchresult_state );

  } // while (iter != stop)

  return( true );
}//end parseEPGSearchResult


bool XmlParser::parseOSD( IBitmapReader* bmreader, MMBoxApplication* app ) {
  /* check if we opened a valid XML file */
  if ( !rootnode ) return false;

  /* Get a list with all <osd> entries */
  XMLNodeList osd_list;
  osd_list = rootnode->children( "osd" );

  XMLNodeIterator iter, stop;

  iter = osd_list.begin();
  stop = osd_list.end();

  /* Process each <osd> entry */
  while (iter != stop) {
    XMLNode* osd_node = *iter;
    iter++;
    XMLProperty* prop;

    /* Try to find "id" attribute */
    string str_pid;
    prop = osd_node->property( "id" );
    if ( prop ) str_pid = prop->value();

    OSDInputState* osd_state = new OSDInputState( str_pid, app );
    parseStdState( osd_node, osd_state );

    /* Get a list with all <button> entries */
    XMLNodeList button_list;
    button_list = osd_node->children( "button" );
    iter = button_list.begin();
    stop = button_list.end();

    /* Process each <button> entry */
    while (iter != stop) {
      XMLNode* node = *iter;
      iter++;

      XMLProperty* prop;

      /* Try to find "id" attribute */
      string id;
      prop = node->property( "id" );
      if ( prop ) id = prop->value();

      /* Try to find "bitmap" attribute */
      string bitmap;
      prop = node->property( "bitmap" );
      if ( prop ) bitmap = path+prop->value();

      bmreader->loadImage( id, bitmap );
    }

    /* add State to list */
    state_list.push_back( osd_state );
  }

  return true;

}

bool XmlParser::parseResourcePath() {
  /* check if we opened a valid XML file */
  if ( !rootnode ) return false;

  /* Get a list with all <osd> entries */
  XMLNodeList path_list = rootnode->children( "resourcepath" );

  XMLNodeIterator iter, stop;

  iter = path_list.begin();
  stop = path_list.end();

  /* Process each <osd> entry */
  while (iter != stop) {
    XMLNode* path_node = *iter;
    iter++;

    path = path_node->getData();
  }

  return true;

}

bool XmlParser::parseTvTimer( IBitmapReader* /*bmreader*/, MMBoxApplication* app) {

  /* check if we opened a valid XML file */
  if ( !rootnode ) return false;

  /* Get a list with all <tvtimer> entries */
  XMLNodeList tvtimer_list;
  tvtimer_list = rootnode->children( "tvtimer" );

  XMLNodeIterator iter, stop;
  XMLNodeIterator entry_iter, entry_stop;
  XMLNodeList     entry_list;

  iter = tvtimer_list.begin();
  stop = tvtimer_list.end();

  /* Process each <tvtimer> entry */
  while (iter != stop) {
    XMLNode* node = *iter;
    iter++;

    XMLProperty* prop;

     /* Try to find "id" attribute */
    string str_pid;
    prop = node->property( "id" );
    if ( prop ) str_pid = prop->value();

    TvTimer* tvtimer_state = new TvTimer( str_pid, app);
    parseStdState( node, tvtimer_state );


     /* Get a list with all <win> tags */
    entry_list = node->children( "win" );

    entry_iter = entry_list.begin();
    entry_stop = entry_list.end();

    /* process all <entry> tags */
    while( entry_iter != entry_stop ) {
      XMLNode* enode = *entry_iter;
      entry_iter++;

      TvTimer::win_properties_t props;
      TvTimer::winType_e        win;

      /* Try to find "id" attribute */
      string str_id;
      prop = enode->property( "id" );
      if ( prop ) str_id = prop->value();

      /* Try to find "x" attribute */
      string temp;
      prop = enode->property( "x" );
      if ( prop ) temp = prop->value();
      props.xpos = atoi( temp.c_str() );

      /* Try to find "y" attribute */
      prop = enode->property( "y" );
      if ( prop ) temp = prop->value();
      props.ypos = atoi( temp.c_str() );

      /* Try to find "width" attribute */
      prop = enode->property( "width" );
      if ( prop ) temp = prop->value();
      props.width = atoi( temp.c_str() );

      /* Try to find "height" attribute */
      prop = enode->property( "height" );
      if ( prop ) temp = prop->value();
      props.height = atoi( temp.c_str() );

      /* Try to find "win_active_color" attribute */
      string str_color;
      prop = enode->property( "win_active_color" );
      if ( prop ) str_color = prop->value();
      props.win_active_color = strToInt( str_color );

      /* Try to find "win_inactive_color" attribute */
      prop = enode->property( "win_inactive_color" );
      if ( prop ) str_color = prop->value();
      props.win_inactive_color = strToInt( str_color );

      /* Try to find "sel_active_color" attribute */
      prop = enode->property( "sel_active_color" );
      if ( prop ) str_color = prop->value();
      props.sel_active_color = strToInt( str_color );

      /* Try to find "sel_active_color2" attribute */
      prop = enode->property( "sel_active_color2" );
      if ( prop ) str_color = prop->value();
      props.sel_active_color2 = strToInt( str_color );

      /* Try to find "sel_inactive_color" attribute */
      prop = enode->property( "sel_inactive_color" );
      if ( prop ) str_color = prop->value();
      props.sel_inactive_color = strToInt( str_color );


      if ( str_id == "List" )          win = TvTimer::LIST;
      else if ( str_id == "Details" )         win = TvTimer::DETAILS;
      tvtimer_state->setWinProperties( win, &props );

    }
    /* add State to list */
    state_list.push_back( tvtimer_state );
    
  } // while (iter != stop)
  
  return( true );
}


bool XmlParser::parseConfig( IBitmapReader* /*bmreader*/, MMBoxApplication* app ) {

  /* check if we opened a valid XML file */
  if ( !rootnode ) return false;

  /* Get a list with all <tvtimer> entries */
  XMLNodeList config_list;
  config_list = rootnode->children( "config" );

  XMLNodeIterator iter, stop;

  iter = config_list.begin();
  stop = config_list.end();

  /* Process each <config> entry */
  while (iter != stop) {
    XMLNode* node = *iter;
    iter++;

    XMLProperty* prop;

     /* Try to find "id" attribute */
    string str_pid;
    prop = node->property( "id" );
    if ( prop ) str_pid = prop->value();

    ConfigState* config_state = new ConfigState( str_pid, app );
    parseStdState( node, config_state );

    /* Get a list with all <showoption> entries */
    XMLNodeList showoption_list;
    showoption_list = node->children( "showoption" );
    iter = showoption_list.begin();
    stop = showoption_list.end();

    /* Process each <showoption> entry */
    while (iter != stop) {
      XMLNode* node = *iter;
      iter++;
      
      XMLProperty* prop;
      
      /* Try to find "id" attribute */
      string id = "";
      prop = node->property( "id" );
      if ( prop ) id = prop->value();
      
      string shortform = node->getData();
      
      config_state->showOption( id, shortform );
    }

    /* add State to list */
    state_list.push_back( config_state );

  } // while (iter != stop)

  return( true );
}

bool XmlParser::parseTaskMgr( IBitmapReader* /*bmreader*/, MMBoxApplication* app ) {

  /* check if we opened a valid XML file */
  if ( !rootnode ) return false;

  /* Get a list with all <taskmgr> entries */
  XMLNodeList taskmgr_list;
  taskmgr_list = rootnode->children( "taskmgr" );

  XMLNodeIterator iter, stop;

  iter = taskmgr_list.begin();
  stop = taskmgr_list.end();

  /* Process each <taskmgr> entry */
  while (iter != stop) {
    XMLNode* node = *iter;
    iter++;

    XMLProperty* prop;

     /* Try to find "id" attribute */
    string str_pid;
    prop = node->property( "id" );
    if ( prop ) str_pid = prop->value();

    TaskMgrState* taskmgr_state = new TaskMgrState( str_pid, app );
    parseStdState( node, taskmgr_state );

    /* add State to list */
    state_list.push_back( taskmgr_state );
  } // while (iter != stop)

  return( true );
}

bool XmlParser::parseOSDSymbols( IBitmapReader* bmreader, MMBoxApplication* /*app*/ ) {

  /* check if we opened a valid XML file */
  if ( !rootnode ) return false;

  /* Get a list with all <osdsymbols> entries */
  XMLNodeList osdsymbols_list;
  osdsymbols_list = rootnode->children( "osdsymbols" );

  XMLNodeIterator iter, stop;

  iter = osdsymbols_list.begin();
  stop = osdsymbols_list.end();

  /* Process each <osdsymbols> entry */
  while (iter != stop) {
    XMLNode* node = *iter;
    iter++;

    /* Get a list with all <symbol> entries */
    XMLNodeList symbol_list;
    symbol_list = node->children( "symbol" );
    iter = symbol_list.begin();
    stop = symbol_list.end();

    /* Process each <symbol> entry */
    while (iter != stop) {
      XMLNode* node = *iter;
      iter++;
      XMLProperty* prop;

      /* Try to find "id" attribute */
      string str_pid;
      prop = node->property( "id" );
      if ( prop ) str_pid = prop->value();

      string bitmap = node->getData();

      bmreader->loadImage( str_pid , path+bitmap );
    } //symbol

  } // while (iter != stop)

  return( true );
}

bool XmlParser::parseSettings( MMBoxApplication* app ) {
  /* check if we opened a valid XML file */
  if ( !rootnode ) return false;
  
  /* Get a list with all <settings> entries */
  XMLNodeList settings_list;
  settings_list = rootnode->children( "settings" );
  
  XMLNodeIterator iter, stop;
  
  iter = settings_list.begin();
  stop = settings_list.end();
  
  /* Process each <settings> entry */
  while (iter != stop) {
    XMLNode* node = *iter;
    iter++;
    
    /* Get a list with all <option> entries */
    XMLNodeList option_list;
    option_list = node->children( "option" );
    iter = option_list.begin();
    stop = option_list.end();
    
    /* Process each <option> entry */
    while (iter != stop) {
      XMLNode* node = *iter;
      iter++;
      XMLProperty* prop;
      
      /* Try to find "id" attribute */
      string id = "";
      prop = node->property( "id" );
      if ( prop ) id = prop->value();
      
      /* Try to find "short" attribute */
      string short_val = "";
      prop = node->property( "short" );
      if ( prop ) short_val = prop->value();
      
      string values = node->getData();
      
      app->getMMBoxConfig()->addSetting( id, short_val, values );
    }
    
  } // while (iter != stop)
  
  return( true );
}

unsigned int XmlParser::strToInt( string s ) {
 unsigned int v = 0;

 for( int i=0; i<4; i++ ) {
   unsigned int l    = s.size();
   unsigned int pos  = s.find_last_of(' ');

   if ( pos != string::npos ) {
     string num = s.substr( pos+1, l );
     v = (v<<8) | atoi( num.c_str() );
     s = s.substr( 0, pos );
   } else {
     v = (v<<8) | atoi( s.c_str() );
     break;
   }
 }
 return( v );
}





