/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Christoph Wellner <chwellner@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#define EPGTIMEOFFSET 180

#include <iostream>
#include "tvtimer.hpp"

#include "nmm/base/proxy/tcp/TCPStrategy.hpp"
#include "nmm/base/ServiceProxy.hpp"

#include "nmm/interfaces/misc/tvguide/ProgramDescription.hpp"

#include "nmm/interfaces/video/tv/ITV.hpp"
#include "nmm/interfaces/file/IFileHandler.hpp"

#include "nmm/base/registry/ReferenceManager.hpp"

#include "nmm/utils/utils.hpp"     // for string + int
#include "nmm/base/utils/GraphURL.hpp"

#include "apps/mmbox/apps/playlist/playlist.hpp"
#include "apps/mmbox/apps/dvb/dvb.hpp"

#include "apps/mmbox/apps/webserver/webserver.hpp"
#include "apps/mmbox/apps/webserver/handler/RecordHandler.hpp"

#include <sstream>

// we want to be in:
namespace NMM {

  TvTimer::TvTimer(string id, MMBoxApplication* _app) 
    : MMBoxState(id, _app),
      recorder(NULL),
      stringPos(0),
      dvbChannel(0),
      contentTmp('a'),
      editing(false),
      timerChange(false),
      timerLocation("localhost"),
      displayRecorded(false),
      displaySleeping(true),
      displayRecording(true),
      channelConfig(NULL),
      timer_load(false),
      dvb(),
      rec_table(),
      rec_cur(rec_table.begin()),
      timertext(false),
      active(false),
      rootpath("/"),
      timerList(NULL),
      timerDetails(NULL),
      timerUpdate_listener(this, &TvTimer::timerUpdate)
  {

    winStatus = LIST;

    MMBoxConfig* config = app->getMMBoxConfig();
    if (!(config->getValue("rootpath", rootpath))) {
      rootpath="/";
    }

    //Initialize TVTimerControl
    config = app->getMMBoxConfig();
    if (!(config->getValue("tvtimerlocation", timerLocation))) {
      timerLocation="";
    }
    
    
    connectTVTimerServer(timerLocation);        

    month["January"]   = 1;
    month["February"]  = 2;
    month["March"]     = 3;
    month["April"]     = 4;
    month["May"]       = 5;
    month["June"]      = 6;
    month["July"]      = 7;
    month["August"]    = 8 ;
    month["September"] = 9;
    month["October"]   = 10;
    month["November"]  = 11;
    month["December"]  = 12;


    time_t now = time(NULL); localtime(&now); //make sure, the timezone-vaiable is initialized

    dvb = recorder->getChannelConfig();
    list<TVChannel> channels = dvb.getChannels();

    for (unsigned int i= 0; i <= channels.size(); i++) {
      try{
        channel[string(dvb.getChannelByNumber(i).getChannelName())] = i;
      }catch(const Exception &e){}
    }

    //register Timer-Handler on Webserver
    SMMBoxWebServer::getInstance().registerHandler(new RecordHandler(recorder.get()));

    winStatus = LIST;
  }


  TvTimer::~TvTimer(){
    editing = false;
    deactivateThread();
    if(channelConfig) delete channelConfig;
  }

  void TvTimer::setWinProperties( int winnr, win_properties_t* prop ) {
    memcpy( &winprops[winnr], prop, sizeof( win_properties_t ) );
  }
  
  /******************** MMBoxState interface ********************/
  Result TvTimer::doCreate(string line) {
    GLOBAL_DEBUG_STREAM("TvTimer: create: " << stateid << endl);

    //TVTimerControl reconnect to TVTimerControl, just to be sure
    MMBoxConfig *config = app->getMMBoxConfig();
    if (!(config->getValue("tvtimerlocation", timerLocation))) {
      timerLocation="localhost";
    }

    string timerDisplaySelection;
    if ((config->getValue("displaySleeping", timerDisplaySelection))) {
      if(timerDisplaySelection == "yes" || timerDisplaySelection == "1"){
        displaySleeping = true;
      }else{
        displaySleeping = false;
      }
    }
    if ((config->getValue("displayRecording", timerDisplaySelection))) {
      if(timerDisplaySelection ==  "yes" || timerDisplaySelection ==  "1"){
        displayRecording = true;
      }else{
        displayRecording = false;
      }
    }
    if ((config->getValue("displayRecorded", timerDisplaySelection))) {
      if(timerDisplaySelection == "yes" || timerDisplaySelection == "1"){
        displayRecorded = true;
      }else{
        displayRecorded = false;
      }
    }

    connectTVTimerServer(timerLocation);

    if (isBackgroundFree()) {
      if (FAILURE == app->createDefaultGraph(background)) {
        return FAILURE;
      }

      if (FAILURE == app->startDefaultGraph()) {
        return FAILURE;
      }
    }

    createWindow(LIST, "D  M Start End  Channel  Period/Status");
    createWindow(DETAILS, "Timer Information");

    showTimerList(atoi(line.c_str()));
    updateColorKeyBar();

    if(recorder.get()){
      recorder->getParentObject()->registerEventListener(ITVTimerEvents::timerUpdate_event, &timerUpdate_listener);
    }

    return SUCCESS;
  }

  Result TvTimer::doKill() {
    GLOBAL_DEBUG_STREAM("TvTimer: kill: " << stateid << endl);

    editing = false;
    deactivateThread();

    app->getOsdManager()->hideColorBar();
    app->getOsdManager()->update();

    deleteWindow();

    if (isBackgroundFree()) {
      app->disconnectVideoGraph();
    }

    active = false;

    if(recorder.get()){
      recorder->getParentObject()->removeEventListener(&timerUpdate_listener);
    }

    return SUCCESS;
  }

  Result TvTimer::doInitialize(string) {
    GLOBAL_DEBUG_STREAM("TvTimer: initialize: " << stateid << endl);

    if (isBackgroundFree()) {
      if (FAILURE == app->createDefaultGraph(background)) {
        return FAILURE;
      }

      if (FAILURE == app->startDefaultGraph()) {
        return FAILURE;
      }
    }

    if(recorder.get()){
      recorder->getParentObject()->registerEventListener(ITVTimerEvents::timerUpdate_event, &timerUpdate_listener);
    }

    showOSD();
    return SUCCESS;
  }

  Result TvTimer::doUninitialize() {
    GLOBAL_DEBUG_STREAM("TvTimer: uninitialize: " << stateid << endl);

    hideOSD();
    editing = false;
    deactivateThread();

    if(recorder.get()){
      recorder->getParentObject()->removeEventListener(&timerUpdate_listener);
    }

    return SUCCESS;
  }

  Result TvTimer::Down() {
    GLOBAL_DEBUG_STREAM("TvTimer: Down" << endl);

    if(winStatus == LIST || winStatus == CONFLICT){
      timerList->selectDown();
      timerList->update();
      nextTimer();
      showDetails();
    }else if(winStatus == TIMER){
      //if(stringPos >=17){
        unsigned int sel = timerList->getSelectedLine();
        string t = timerList->getLineSubject(sel);

        if(dvbChannel > 0 && dvbChannel <= (int)channel.size()){
          dvbChannel--;
          string c = dvb.getChannelByNumber(dvbChannel).getChannelName();
          t.replace(18, t.length(), c);
          timerList->changeSelectedLineSubject(t);
          timerList->update();
        }

        // }


    }else if(winStatus == DETAILS || winStatus == CONFLICTDETAILS){
      timerDetails->selectDown();
      timerDetails->update();
    }
    updateColorKeyBar();
    app->getOsdManager()->update();

    return SUCCESS;
  }

  Result TvTimer::Up() {
    GLOBAL_DEBUG_STREAM("TvTimer: Up" << endl);

    if(winStatus == LIST || winStatus == CONFLICT){
      timerList->selectUp();
      previousTimer();
      timerList->update();
      showDetails();
    }else if(winStatus == TIMER){
      // if(stringPos >=17){
        unsigned int sel = timerList->getSelectedLine();
        string t = timerList->getLineSubject(sel);

        if(dvbChannel >= 0 && dvbChannel < (int)channel.size()){
          dvbChannel++;
          string c = dvb.getChannelByNumber(dvbChannel).getChannelName();
          t.replace(18, t.length(), c);
          timerList->changeSelectedLineSubject(t);
          timerList->update();
        }

        // }

    }else if(winStatus == DETAILS || winStatus == CONFLICTDETAILS){
      timerDetails->selectUp();
      timerDetails->update();
    }

    updateColorKeyBar();
    app->getOsdManager()->update();

    return SUCCESS;
  }

  Result TvTimer::Left() {
    GLOBAL_DEBUG_STREAM("TvTimer: Left" << endl);

    if(winStatus == LIST ){
      winStatus = DETAILS;
      timerList->deSelectWindow();
      timerList->update();
      timerDetails->selectWindow();
      timerDetails->update();
    }else if(winStatus == CONFLICT){
      winStatus = CONFLICTDETAILS;
      timerList->deSelectWindow();
      timerList->update();
      timerDetails->selectWindow();
      timerDetails->update();
    }else if(winStatus == TIMER){
      if(stringPos > 0){
        string content = timerList->getLineSubject(timerList->getSelectedLine());
        timerList->changeSelectedLineSubject(content.replace(stringPos, 1, string(1,contentTmp)));
        timerList->update();        
        stringPos--;
        
        if(stringPos == 2 || stringPos == 5 || stringPos == 8 || stringPos == 11 
           || stringPos == 14 || stringPos == 17)
          stringPos--;

        contentTmp = content[stringPos];

      }
    }else if(winStatus == DETAILS){
      winStatus = LIST;
      timerList->selectWindow();
      timerList->update();
      timerDetails->deSelectWindow();
      timerDetails->deSelectLine();
      timerDetails->update();
    }else if(winStatus == CONFLICTDETAILS){
      winStatus = CONFLICT;
      timerList->selectWindow();
      timerList->update();
      timerDetails->deSelectWindow();
      timerDetails->deSelectLine();
      timerDetails->update();
    }
    updateColorKeyBar();
    app->getOsdManager()->update();

    return SUCCESS;
  }

  Result TvTimer::Right() {
    GLOBAL_DEBUG_STREAM("TvTimer: Right" << endl);

    if(winStatus == LIST){
      winStatus = DETAILS;
      timerList->deSelectWindow();
      timerList->update();
      timerDetails->selectWindow();
      timerDetails->update();
    }else if(winStatus == CONFLICT){
      winStatus = CONFLICTDETAILS;
      timerList->deSelectWindow();
      timerList->update();
      timerDetails->selectWindow();
      timerDetails->update();
    }else if(winStatus == TIMER){
      if(stringPos < 17){
        string content = timerList->getLineSubject(timerList->getSelectedLine());
        timerList->changeSelectedLineSubject(content.replace(stringPos, 1, string(1,contentTmp)));
        timerList->update();        
        stringPos++;

        if(stringPos == 2 || stringPos == 5 || stringPos == 8 || stringPos == 11 
           || stringPos == 14 )// || stringPos == 17 )
          stringPos++;

        contentTmp = content[stringPos];

      }
    }else if(winStatus == DETAILS){
      winStatus = LIST;
      timerList->selectWindow();
      timerList->update();
      timerDetails->deSelectWindow();
      timerDetails->deSelectLine();
      timerDetails->update();
    }else if(winStatus == CONFLICTDETAILS){
      winStatus = CONFLICT;
      timerList->selectWindow();
      timerList->update();
      timerDetails->deSelectWindow();
      timerDetails->deSelectLine();
      timerDetails->update();
    }
    updateColorKeyBar();
    app->getOsdManager()->update();

    return SUCCESS;
  }

  Result TvTimer::Back() {
    GLOBAL_DEBUG_STREAM("TvTimer: Back" << endl);

    deactivate();
    
    return goToLastState();
  }

  Result TvTimer::Record() {
    GLOBAL_DEBUG_STREAM("TvTimer: Record" << endl);

    return SUCCESS;
  }


  Result TvTimer::specialKey01() {
    GLOBAL_DEBUG_STREAM("TvTimer: specialKey01" << endl);
    return SUCCESS;
  }

  Result TvTimer::specialKey02() {
    GLOBAL_DEBUG_STREAM("TvTimer: specialKey02" << endl);

    if(winStatus == TIMER){
        
    }
    return SUCCESS;
  }

  Result TvTimer::colorKeyRed(){
    if(winStatus == LIST){
      if(deleteTimer() != SUCCESS){
        app->getOsdManager()->displayMessage("No permisson" , "You are not allowed to delete the timer", "OK" );
        app->getOsdManager()->update();
        changeState("OSDInputState");
      }

      int line = timerList->getSelectedLine();
      showTimerList(line - 1);
      
    }else if(winStatus == TIMER){
      editing = false;
      deactivateThread();
      timerChange = false;
      unsigned int line = timerList->getSelectedLine();
      if(line > rec_table.size()){
        showTimerList(line - 1);
      }else{
        showTimerList(line);
      }
      winStatus = LIST;

    }else if(winStatus == DETAILS){
      if(deleteTimer() != SUCCESS){
        app->getOsdManager()->displayMessage("No permisson" , "You are not allowed to delete the timer", "OK" );
        app->getOsdManager()->update();
        changeState("OSDInputState");
      }
      timerList->selectWindow();
      timerList->update();
      timerDetails->deSelectWindow();
      timerDetails->update();
      int line = timerList->getSelectedLine();
      showTimerList(line -1);
    }else if (winStatus == CONFLICT || winStatus == CONFLICTDETAILS){
      if(deleteTimer() != SUCCESS){
        app->getOsdManager()->displayMessage("No permisson" , "You are not allowed to delete the timer", "OK" );
        app->getOsdManager()->update();
        changeState("OSDInputState");
      }else{
        if(rec_cur == rec_table.begin()){
          //we delted the first timer in the list
          winStatus = LIST;
          showTimerList(0);
          return SUCCESS;
        }
      }
      int line = timerList->getSelectedLine();
      showTimerList(line - 1);
    }
    updateColorKeyBar();
    return SUCCESS;
  }
  Result TvTimer::colorKeyGreen(){
    if(winStatus == LIST || winStatus == DETAILS || winStatus == CONFLICT || winStatus == CONFLICTDETAILS){
      if(rec_table.empty()) return FAILURE;
      if(rec_cur->getStatus() == TVTimerEntry::RECORDING) return FAILURE;
      winStatus = TIMER;
      stringPos = 0;
      timerList->selectWindow();
      timerList->update();
      timerDetails->deSelectWindow();
      timerDetails->update();

      int pos = timerList->getSelectedLine();
      timerList->selectLine2(pos);

      timerChange = true;
      editing = true;
      activateThread();

    }else if(winStatus == TIMER){
      editing = false;
      deactivateThread();

      try{
        Result res  = programCurrentSelection(timerChange);
        
        if(res == FAILURE){
          app->getOsdManager()->displayMessage("Timer Not programmed!", "Timer has conflicts and no new TVCard could be found", "OK" );
          app->getOsdManager()->update();
          changeState("OSDInputState");
        }
      }catch(const ActionNotAllowedException &e){
        GLOBAL_ERROR_STREAM(e << endl);
        app->getOsdManager()->displayMessage("No permisson" , "You are not allowed to change the timer", "OK" );
        app->getOsdManager()->update();
        changeState("OSDInputState");
      }catch(const Exception &e){
        GLOBAL_ERROR_STREAM(e << endl);
      }catch(...){
      }
     
      int line = timerList->getSelectedLine();
      showTimerList(line);
      winStatus = LIST;

    }else{
    }

    updateColorKeyBar();
    return SUCCESS;
  }
  Result TvTimer::colorKeyYellow(){
    if(winStatus == LIST || winStatus == DETAILS){

      if(rec_cur->getStatus() == TVTimerEntry::RECORDING){
        DvbState* dvbstate = dynamic_cast<DvbState*>(app->getState("DVB-TV"));
        if(dvbstate){
          dvbstate->setTVTimerPlayback(*rec_cur);
        }else{
          return FAILURE;
        }
        changeState("DVB-TV");
      }else if(rec_cur->getStatus() == TVTimerEntry::RECORDED
               || (rec_cur->getStatus() == TVTimerEntry::SLEEPING && rec_cur->getPeriod() != TVTimerEntry::ONCE)){
        if(rec_cur != rec_table.end()){
          if(!rec_cur->getFileURL().empty()){
            changeState("Playlist", rec_cur->getFileURL());
          }
        }
      }else if(rec_cur->getStatus() == TVTimerEntry::SLEEPING){
      }else if(rec_cur->getStatus() == TVTimerEntry::CONFLICT){
        conflictEntry = *rec_cur;
        if(winStatus == DETAILS){
          Left();
        }
        winStatus = CONFLICT;
        int line = timerList->getSelectedLine();
        showTimerList(line -1);
      }

    }else if(winStatus == CONFLICT || winStatus == CONFLICTDETAILS){
      if(winStatus == CONFLICTDETAILS){
          Left();
        }
      winStatus = LIST;
      showTimerList(0);
    }else if(winStatus == TIMER){
      int sel = timerList->getSelectedLine();
      string cont = timerList->getLineContent(sel);

      if(cont == "Daily"){
        timerList->chooseLineContent(timerList->getSelectedLine(),"Weekly");
      }else if(cont == "Weekly"){
        timerList->chooseLineContent(timerList->getSelectedLine(),"Monthly");
      }else if(cont == "Monthly"){
        timerList->chooseLineContent(timerList->getSelectedLine(), "Weekdays");
      }else if(cont == "Weekdays"){
        timerList->chooseLineContent(timerList->getSelectedLine(), "Weekends");
      }else if(cont == "Weekends"){
        timerList->chooseLineContent(timerList->getSelectedLine(), "*Channel");
      }else if(cont == "*Channel"){
        timerList->chooseLineContent(timerList->getSelectedLine(), "*ChannelTime");
      }else if(cont == "*ChannelTime"){
        timerList->chooseLineContent(timerList->getSelectedLine(), "*");
      }else if(cont == "*"){
        timerList->chooseLineContent(timerList->getSelectedLine(), " ");
      }else{
        timerList->chooseLineContent(timerList->getSelectedLine(),"Daily");
      }

      timerList->update();
      app->getOsdManager()->update();
    }else{
    }
    updateColorKeyBar();
    return SUCCESS;
  }

  Result TvTimer::colorKeyBlue(){
    if(winStatus == LIST || winStatus == DETAILS){
      winStatus = TIMER;
      stringPos = 0;
      dvbChannel = 0;

      //fill new timer with default-date
      time_t ct = time(NULL)+60;
      struct tm* currenttime  = localtime(&ct);
      currenttime->tm_min++;
      
      char currenttimetmp[30];
      strftime(currenttimetmp, sizeof(currenttimetmp), "%d %m %H:%M", currenttime);
      
      string newline(currenttimetmp);
      
      currenttime->tm_min++;
      strftime(currenttimetmp, sizeof(currenttimetmp), "%H:%M", currenttime);
      newline += " " + string(currenttimetmp) + " ";
      newline += dvb.getChannelByNumber(0).getChannelName();

      DoubleListUnit::line_t newLineList;
      newLineList.push_back(" ");
      newLineList.push_back("Weekly");
      newLineList.push_back("Daily");
      newLineList.push_back("Monthly");
      newLineList.push_back("Weekdays");
      newLineList.push_back("Weekends");
      newLineList.push_back("*Channel");
      newLineList.push_back("*ChannelTime");
      newLineList.push_back("*");
      
      timerList->addLine(newline, newLineList);
      timerList->selectLine2(timerList->getNumLines());

      timerDetails->deleteAllLines();
      timerDetails->addLine("Create New Timer");
      timerDetails->addLine(string(""));
      timerDetails->addLine("Pattern:");
      timerDetails->addLine("DD MM hh:mm hh:mm");

      timerList->selectWindow();
      timerList->update();
      timerDetails->deSelectWindow();
      timerDetails->update();
      app->getOsdManager()->update();

      timerChange = false;
      editing = true;
      activateThread();

    }else if(winStatus == TIMER){
    }else if(winStatus == DETAILS){
    }else if(winStatus == CONFLICT){
    }
    updateColorKeyBar();
    return SUCCESS;
  }

  Result TvTimer::numKey0() {
    numKey("0");

    return SUCCESS;
  }

  Result TvTimer::numKey1() {
    numKey("1");

    return SUCCESS;
  }

  Result TvTimer::numKey2() {
    numKey("2");

    return SUCCESS;
  }

  Result TvTimer::numKey3() {
    numKey("3");

    return SUCCESS;
  }

  Result TvTimer::numKey4() {
    numKey("4");

    return SUCCESS;
  }

  Result TvTimer::numKey5() {
    numKey("5");

    return SUCCESS;
  }

  Result TvTimer::numKey6() {
    numKey("6");

    return SUCCESS;
  }

  Result TvTimer::numKey7() {
    numKey("7");

    return SUCCESS;
  }

  Result TvTimer::numKey8() {
    numKey("8");

    return SUCCESS;
  }

  Result TvTimer::numKey9() {
    numKey("9");

    return SUCCESS;
  }

  Result TvTimer::Return() {
    GLOBAL_DEBUG_STREAM("TvTimer: Return" << endl);

    if(winStatus == LIST){
    

    }else if(winStatus == TIMER){

    }

    app->getOsdManager()->update();

    return SUCCESS;
  }

  Result TvTimer::receiveMessage(string) {
    app->getOsdManager()->deleteMessage();
    app->getOsdManager()->update();
    return SUCCESS;
  }
  /******************** MMBoxState interface ********************/

  Result TvTimer::timerUpdate(){
    //cerr << "TIMER-UPDATE" <<endl;
    if(winStatus != TIMER){
      showTimerList();
    }
    return SUCCESS;
  }


  /******************** private helpers ********************/
  void TvTimer::updateColorKeyBar(){
    if(hasFocus()){
      if(winStatus == LIST || winStatus == DETAILS){
        if(rec_table.empty()){
          app->getOsdManager()->setColorBarText("", "", "", "New Timer");
        }else{
          switch(rec_cur->getStatus()){
          case TVTimerEntry::RECORDED:
            app->getOsdManager()->setColorBarText("Delete", "Change", "Play", "New Timer");
            break;
          case TVTimerEntry::RECORDING:
            app->getOsdManager()->setColorBarText("Delete", "", "View", "New Timer");
            break;
          case TVTimerEntry::RECORDED_WITH_ERROR:
            app->getOsdManager()->setColorBarText("Delete", "Change", "", "New Timer");
            break;
          case TVTimerEntry::CONFLICT:
            app->getOsdManager()->setColorBarText("Delete", "Change", "Solve", "New Timer");
            break;
          case TVTimerEntry::SLEEPING:
            app->getOsdManager()->setColorBarText("Delete", "Change", "", "New Timer");
            break;
          default:
            app->getOsdManager()->setColorBarText("Delete", "", "", "New Timer");
            break;
          }
        }
      }else if(winStatus == TIMER){
        
        int sel = timerList->getSelectedLine();
        string cont = timerList->getLineContent(sel);
        
        if(cont == "Daily"){
          app->getOsdManager()->setColorBarText("Discard", "Program", "Weekly", "");
        }else if(cont == "Weekly"){
          app->getOsdManager()->setColorBarText("Discard", "Program", "Monthly", "");
        }else if(cont == "Monthly"){
          app->getOsdManager()->setColorBarText("Discard", "Program", "Weekdays", "");
        }else if(cont == "Weekdays"){
          app->getOsdManager()->setColorBarText("Discard", "Program", "Weekends", "");
        }else if(cont == "Weekends"){
          app->getOsdManager()->setColorBarText("Discard", "Program", "*Channel", "");
        }else if(cont == "*Channel"){
          app->getOsdManager()->setColorBarText("Discard", "Program", "*ChannelTime", "");
        }else if(cont == "*ChannelTime"){
          app->getOsdManager()->setColorBarText("Discard", "Program", "*", "");
        }else if(cont == "*"){
          app->getOsdManager()->setColorBarText("Discard", "Program", "Once", "");
        }else{
          app->getOsdManager()->setColorBarText("Discard", "Program", "Daily", "");
        }
        
        
      }else if(winStatus == CONFLICT || winStatus == CONFLICTDETAILS){
          app->getOsdManager()->setColorBarText("Delete", "Change", "Show All", "New Timer");
      }else{
        app->getOsdManager()->setColorBarText("Delete", "Change", "", "New Timer");
      }
      app->getOsdManager()->displayColorBar();
      app->getOsdManager()->update();
    }
  }

  void TvTimer::showOSD(){
    if(timerList){
      timerList->show();
      timerList->update();
    }
    if(timerDetails){
      timerDetails->show();
      timerDetails->update();
    }
    updateColorKeyBar();
    app->getOsdManager()->update();

  }

  void TvTimer::hideOSD(){
    if(timerList){
      timerList->hide();
      timerList->update();
    }
    if(timerDetails){
      timerDetails->hide();
      timerDetails->update();
    }

    app->getOsdManager()->hideColorBar();
    app->getOsdManager()->update();

  }


  time_t TvTimer::stringsToTime(string s_time, string s_day, string s_month, string s_year) {
    struct tm abs_time;
    unsigned int pos;
    int min, h, day, mon, year;

    pos = s_time.find_first_of(':');
    if ( pos != string::npos ) {
      string temp;
      temp = s_time.substr( 0, pos );
      h = atoi( temp.c_str() );

      temp = s_time.substr( pos+1, s_time.length() );
      min = atoi( temp.c_str() );
    }

    day  = atoi( s_day.c_str() );
    if ( s_month.length() > 2 ) mon  = month[s_month];
    else mon = atoi( s_month.c_str() );
    year = atoi( s_year.c_str() );

    abs_time.tm_sec  = 0;
    abs_time.tm_min  = min;
    abs_time.tm_hour = h;
    abs_time.tm_mday = day;
    abs_time.tm_mon  = mon-1;
    abs_time.tm_year = year-1900;

    return( mktime( &abs_time ) );
  }

  time_t TvTimer::stringToTime(string s) {
    string s_time, s_year, s_month, s_day;
    unsigned int pos;

    pos = s.find_first_of(' ');
    if ( pos != string::npos ) {
      string temp;
      s_time  = s.substr( pos+1, s.length() );
      temp = s.substr( 0, pos );

      pos = temp.find_first_of('-');
      if ( pos != string::npos ) {
        s_year = temp.substr( 0, pos );
        temp   = temp.substr( pos+1, temp.length() );
        pos = temp.find_first_of('-');
        if ( pos != string::npos ) {
          s_month = temp.substr( 0, pos );
          s_day   = temp.substr( pos+1, temp.length() );
        }
      }

    }

    return( stringsToTime(s_time, s_day, s_month, s_year) );
  }

  string TvTimer::TimeToString(time_t ltime) {
    if(ltime == 0){
      return string("unlimited");
    }
    char temp[64];
    strftime(temp, sizeof(temp), "%d-%m-%Y %H:%M",  localtime(&ltime));
    return( string(temp) );
  }

 Result TvTimer::deleteTimer() {
   try{
     if (rec_cur != rec_table.end()) {
       recorder->deleteTimer(rec_cur->getID(), UserInfo().getUsername());
     }
   }catch(const ActionNotAllowedException &e){
     GLOBAL_ERROR_STREAM(e << endl);
     return FAILURE;
   }catch(const Exception &e){
     GLOBAL_ERROR_STREAM(e << endl);
     return FAILURE;
   }

   return SUCCESS;
 }

  Result TvTimer::programCurrentSelection(bool change) throw (const ActionNotAllowedException&){

    if(!recorder.get()) return FAILURE;

    time_t now = time(NULL);
    struct tm * startendtime = localtime(&now); // initialize with current time
    mktime(startendtime);

    string t = timerList->getLineSubject(timerList->getSelectedLine());
    
    TVTimerEntry *newEntry = new TVTimerEntry(UserInfo().getUsername());

    if(change){
      newEntry->setID(rec_cur->getID());
      if(!rec_cur->getTimerInfo().empty()){
        newEntry->setTimerInfo(rec_cur->getTimerInfo());
      }
    }

    try{
      startendtime->tm_mon  = atoi(t.substr(3,2).c_str()) - 1 ;
      startendtime->tm_mday = atoi(t.substr(0,2).c_str());
      startendtime->tm_hour = atoi(t.substr(6,2).c_str());
      startendtime->tm_min  = atoi(t.substr(9,2).c_str());
      startendtime->tm_sec  = 0;
      newEntry->setStartTime(mktime(startendtime));

      startendtime->tm_hour = atoi(t.substr(12,2).c_str());
      startendtime->tm_min = atoi(t.substr(15,2).c_str());
      newEntry->setEndTime(mktime(startendtime));

    }catch(const Exception &e){
      GLOBAL_ERROR_STREAM("Error druring programming timer " << e << endl);
      return FAILURE;
    }catch(...){
      GLOBAL_ERROR_STREAM("Error during programming timer" << endl);
      return FAILURE;
    }
    
    try{
      newEntry->setChannelName(t.substr(18));
    }catch(const Exception &e){
    }catch(...){

      return FAILURE;
    }

    //get the Program-Info
    string channel; 
    ITVGuide_var itvgserver(NULL);
    NMMObject* o(SReferenceManager::getInstance().getService("TVGuide"));
    if(o){
      itvgserver.reset(o->getInterface<ITVGuide>());
    }else{
      GLOBAL_ERROR_STREAM("Service TVGuide not available" << endl);
    }    

    if((!change) //if we program a new timer
       || !(rec_cur->getStartTime() == newEntry->getStartTime() //or if we changed time or channel
           && rec_cur->getEndTime() == newEntry->getEndTime()
           && rec_cur->getChannelName() == newEntry->getChannelName())
       ){
      if(itvgserver.get()){

        channel = itvgserver->getChannelByName(t.substr(18));  
        char start[25]; char endt[25];
        time_t s = newEntry->getStartTime();
        time_t e = newEntry->getEndTime();

        //respect Daylight Saving Time
        time_t now = time(NULL); 
        struct tm *abs_time = localtime(&now);
        struct tm *tmp_tm = localtime(&s);
        double d = -1*timezone/60/60*100;

        if(tmp_tm->tm_isdst >=0){
          s += (abs_time->tm_isdst - tmp_tm->tm_isdst)*3600;              
          e += (abs_time->tm_isdst - tmp_tm->tm_isdst)*3600;
          d += tmp_tm->tm_isdst * 100;
        }
        char tmp[6];
        sprintf(tmp, " %0+5.0f", d);
        tz = string(tmp);

        strftime(start, sizeof(start), "%Y%m%d%H%M", localtime(&s));
        strftime(endt, sizeof(endt), "%Y%m%d%H%M", localtime(&e));   
        ProgramDescription pd = itvgserver->getEvent(start+tz,endt+tz,channel);
        if(!pd.empty()){
          pd.localizeTimes();
          newEntry->setTimerInfo(pd);
        }
      }
    }
    
    if(timerList->getLineContent(timerList->getSelectedLine()) == "Daily"){
      newEntry->setPeriod(TVTimerEntry::DAILY);
    }else if(timerList->getLineContent(timerList->getSelectedLine()) == "Weekly"){
      newEntry->setPeriod(TVTimerEntry::WEEKLY);
    }else if(timerList->getLineContent(timerList->getSelectedLine()) == "Monthly"){
      newEntry->setPeriod(TVTimerEntry::MONTHLY);
    }else if(timerList->getLineContent(timerList->getSelectedLine()) == "Weekdays"){
      newEntry->setPeriod(TVTimerEntry::WEEKDAYS);
    }else if(timerList->getLineContent(timerList->getSelectedLine()) == "Weekends"){
      newEntry->setPeriod(TVTimerEntry::WEEKENDS);
    }else if(timerList->getLineContent(timerList->getSelectedLine()) == "*Channel"){
      newEntry->setPeriod(TVTimerEntry::RANDOM_ONLY_CHANNEL);
      // look for dates in the future (search by title)
      ProgramFilter pf;
      pf.setStartTime(newEntry->getTimerInfo().getEndTime());
      pf.addChannel(newEntry->getTimerInfo().getChannelID());
      pf.setSearchPhrase(newEntry->getTimerInfo().getTitles().front());
      
      list<ProgramDescription> resList = itvgserver->search(pf);
      resList.sort();
      newEntry->setNextTimers(resList);

    }else if(timerList->getLineContent(timerList->getSelectedLine()) == "*ChannelTime"){
      newEntry->setPeriod(TVTimerEntry::RANDOM_ONLY_CHANNEL_AND_TIME);
      // look for dates in the future (search by title)
      ProgramFilter pf;
      pf.setStartTime(newEntry->getTimerInfo().getEndTime());
      pf.addChannel(newEntry->getTimerInfo().getChannelID());
      pf.setSearchPhrase(newEntry->getTimerInfo().getTitles().front());
      
      list<ProgramDescription> resList = itvgserver->search(pf);
      resList.sort();
      list<ProgramDescription> resList2;

      for(list<ProgramDescription>::iterator i = resList.begin(); i != resList.end(); i++){
        i->localizeTimes();
        if(atoi(i->getStartTime().substr(8,4).c_str()) >= atoi(newEntry->getTimerInfo().getStartTime().substr(8,4).c_str())
           && atoi(i->getStartTime().substr(8,4).c_str()) <= atoi(newEntry->getTimerInfo().getEndTime().substr(8,4).c_str())){
          resList2.push_back(*i);
        }
        
      }

      newEntry->setNextTimers(resList2);
 
    }else if(timerList->getLineContent(timerList->getSelectedLine()) == "*"){
      newEntry->setPeriod(TVTimerEntry::RANDOM);
      // look for dates in the future (search by title)
      ProgramFilter pf;
      pf.setStartTime(newEntry->getTimerInfo().getEndTime());
      pf.setSearchPhrase(newEntry->getTimerInfo().getTitles().front());
      
      list<ProgramDescription> resList = itvgserver->search(pf);
      resList.sort();
      newEntry->setNextTimers(resList);

    }else{
      newEntry->setPeriod(TVTimerEntry::ONCE);
    }

    try{
      if (!change) {
        Result res = recorder->programTimer(*newEntry);
        return res; 
      } 
      else {
        try{
          return recorder->changeTimer(*newEntry);
          
        }catch(const ActionNotAllowedException& e){
          GLOBAL_ERROR_STREAM(e);
          return FAILURE;
        }catch (const Exception &e){
          return FAILURE;
        }catch(...){
          return FAILURE;
        }        
      }
    }catch(const TimerAlreadyProgrammedException &e){
      app->getOsdManager()->displayMessage("Timer Alredy programmed!", "There is already a timer programmed to record this show", "OK" );
      app->getOsdManager()->update();
      changeState("OSDInputState");

    }catch(const Exception &e){
      cerr << "ERROR " << e << endl;
    }catch(...){
      //cerr << " aehm" << endl;
    }
    return SUCCESS;
  }

  Result TvTimer::nextTimer() {
    if(rec_cur != rec_table.end()){
      rec_table_t::iterator rec_cur2 = rec_cur;
      rec_cur2++;
      if(rec_cur2 != rec_table.end()){
        rec_cur++;
      }
    }
    return SUCCESS;
  }

  Result TvTimer::previousTimer() {
    if(rec_cur != rec_table.begin()){
      rec_cur--;
    }
    return SUCCESS;
  }


  void TvTimer::numKey(string num) {


    //if not in Timer-Modification-mode -> ignore num-keys
    if(winStatus != TIMER)
      return;
    
    MutexGuard mg(mutex);

    unsigned int sel = timerList->getSelectedLine();
    string t = timerList->getLineSubject(sel);

    t.replace(stringPos, 1, num);
    timerList->changeSelectedLineSubject(t);
    stringPos++;


    if(stringPos == 2 || stringPos == 5 || stringPos == 11 || stringPos == 17){
      t.replace(stringPos, 1, " ");
      stringPos++;
    }

    if(stringPos == 8 || stringPos == 14){
      t.replace(stringPos, 1, ":");
      stringPos++;
    }
    if(stringPos >= 18) return ;

    timerList->update();
    app->getOsdManager()->update();


  }


  void TvTimer::createWindow( winType_e w, string h) {

    if(w == LIST){
      if(timerList){
        delete timerList;
        timerList = NULL;
      }

      timerList = new DoubleListUnit(app->getOsdManager(), &app->getOSDWidgetFactory(), "medium");
      timerList->setDimension(winprops[w].width, winprops[w].height);
      timerList->setPosition(winprops[w].xpos, winprops[w].ypos );
      timerList->setColors( winprops[w].win_active_color,
                            winprops[w].win_inactive_color,
                            winprops[w].sel_active_color,
                            winprops[w].sel_active_color2,
                            winprops[w].sel_inactive_color );
      timerList->setHeader(h);
      timerList->selectWindow();
      timerList->show();
      timerList->update();
      
    }else if(w == DETAILS){
      if(timerDetails){
        delete timerDetails;
        timerDetails = NULL;
      }

      timerDetails = new ListUnit(app->getOsdManager(), &app->getOSDWidgetFactory(), "medium");
      timerDetails->setDimension(winprops[w].width, winprops[w].height);
      timerDetails->setPosition(winprops[w].xpos, winprops[w].ypos );
      timerDetails->setColors(winprops[w].win_active_color,
                              winprops[w].win_inactive_color,
                              winprops[w].sel_active_color,
                              winprops[w].sel_active_color2,
                              winprops[w].sel_inactive_color );
      timerDetails->deSelectWindow();
      timerDetails->setHeader(h);
      timerDetails->show();
      timerDetails->update();
    }

    app->getOsdManager()->update();
  }

  void TvTimer::deleteWindow() {
    if (timerList) {
      delete timerList;
      timerList = NULL;
    }
    if (timerDetails) {
      delete timerDetails;
      timerDetails = NULL;
    }
 
  }


  void TvTimer::showTimerList(int markLine ){
    
    timerList->deleteAllLines();
    timerList->setFirstColumnWidth(60);

    if(recorder.get()){
      if(winStatus == CONFLICT){

        rec_table = recorder->getConflictsWithTimer(conflictEntry);
        rec_table.push_front(conflictEntry);

      }else{
        winStatus = LIST;

        rec_table.clear();

        rec_table_t tmpList = recorder->getConflictTimers();
        if(!tmpList.empty()) rec_table.splice(rec_table.begin(), tmpList);

        if(displaySleeping){
          rec_table_t tmpList = recorder->getSleepingTimers();
          if(!tmpList.empty()) rec_table.splice(rec_table.begin(), tmpList);
        }
        if(displayRecording){
          rec_table_t tmpList = recorder->getRecordingTimers();
          if(!tmpList.empty()) rec_table.splice(rec_table.begin(), tmpList);
        }
        if(displayRecorded){
          rec_table_t tmpList = recorder->getRecordedTimers();
          if(!tmpList.empty()) rec_table.splice(rec_table.begin(), tmpList);
        }
        
      }

    }

    for(rec_table_t::iterator i = rec_table.begin(); i !=  rec_table.end();i++){
      char tmp[25];
      string linecontent("");
      time_t t = i->getStartTime();
      strftime(tmp, sizeof(tmp), "%d %m %H:%M", localtime(&t));
      linecontent += string(tmp) + " ";
      t = i->getEndTime();
      if(t != 0){
        strftime(tmp, sizeof(tmp), "%H:%M", localtime(&t));
      }else{
        sprintf(tmp, "--:--");
      }
      linecontent += string(tmp) + " ";

      linecontent += string(i->getChannelName());

      DoubleListUnit::line_t line;
      string linecontent2("");

      
      switch(i->getPeriod()){
      case TVTimerEntry::ONCE:
        break;
      case TVTimerEntry::WEEKLY:
        linecontent2 += "Weekly";
        break;
      case TVTimerEntry::DAILY:
        linecontent2 += "Daily";
        break;
      case TVTimerEntry::MONTHLY:
        linecontent2 += "Monthly";
        break;
      case TVTimerEntry::WEEKDAYS:
        linecontent2 += "Weekdays";
        break;
      case TVTimerEntry::WEEKENDS:
        linecontent2 += "Weekends";
        break;
      case TVTimerEntry::RANDOM_ONLY_CHANNEL:
        linecontent2 += "*Channel";
        break;
      case TVTimerEntry::RANDOM_ONLY_CHANNEL_AND_TIME:
        linecontent2 += "*ChannelTime";
        break;
      case TVTimerEntry::RANDOM:
        linecontent2 += "*";
        break;
      }


      if(i->hasConflicts()){
        linecontent2 += "Conflict";
      }else{
      }

      if(i->getStatus() == TVTimerEntry::RECORDING){ 
        linecontent2 = "Recording ";
      }

      if(i->getStatus() == TVTimerEntry::RECORDED){ 
        line.clear();
        line.push_back(string("Recorded"));
        linecontent2 += "Recorded";
      }

      if(i->getStatus() == TVTimerEntry::RECORDED_WITH_ERROR){ 
        line.clear();
        line.push_back(string("Error"));
        linecontent2 += "Error";
      }

      if(i->getStatus() == TVTimerEntry::CONFLICT){ 
        line.clear();
        line.push_back(string("Conflict"));
        linecontent2 += "Conflict";
      }

      line.push_back(linecontent2);
      line.push_back(" ");
      line.push_back("Weekly");
      line.push_back("Daily");
      line.push_back("Monthly");
      line.push_back("Weekdays");
      line.push_back("Weekends");
      line.push_back("*Channel");
      line.push_back("*ChannelTime");
      line.push_back("*");
      timerList->addLine(linecontent, line);

    }

    //select the current timer
    int tmpsize = rec_table.size();
    if(markLine <= 0 || markLine > tmpsize) //check for value
      markLine = 1;
    timerList->selectLine(markLine);
    rec_cur = rec_table.begin();
    advance(rec_cur, markLine-1); //first timer has no. 1 but first item in list no. 0

    timerDetails->show();
    timerDetails->update();

    timerList->show();
    timerList->update();

    showDetails();
    updateColorKeyBar();
    app->getOsdManager()->update();
    
  }

  void TvTimer::showDetails(){
    //if(!hasFocus()) return;

    timerDetails->deleteAllLines();

    if(rec_table.empty()) {
      timerDetails->update();
      app->getOsdManager()->update();
      return;
    }

    string content = timerList->getLineSubject(timerList->getSelectedLine());    
    
    //get the Program-Info
    string channel; 
    ITVGuide_var itvgserver(NULL);
    NMMObject* o(SReferenceManager::getInstance().getService("TVGuide"));
    if(o){
      itvgserver.reset(o->getInterface<ITVGuide>());
    }else{
      GLOBAL_ERROR_STREAM("Service TVGuide not available" << endl);
    }

    ProgramDescription pd = rec_cur->getTimerInfo();
    
    if(!pd.empty()){
      try{
        timerDetails->addLine("From : " + transformTime(pd.getStartTime()));
        timerDetails->addLine("Until  : " + transformTime(pd.getEndTime()));
        timerDetails->addLine("On   : " + content.substr(18)); 
        if(!rec_cur->getUserName().empty()){
          timerDetails->addLine("User: "+ rec_cur->getUserName());
        }

        timerDetails->addLine(string(""));
        if(!pd.getTitles().empty()){
          list<string> titleList = pd.getTitles();
          for(list<string>::iterator i = titleList.begin(); i != titleList.end(); i++){
            timerDetails->addLine(*i);
          } 
        }
        timerDetails->addLine(string(""));
        timerDetails->addMultiLine(pd.getDescription());
      }catch(const Exception& e){
      }
    }else{
        timerDetails->addLine("From : "   + TimeToString(rec_cur->getStartTime()));
        timerDetails->addLine("Until  : " + TimeToString(rec_cur->getEndTime()));
        timerDetails->addLine("On   : "   + rec_cur->getChannelName()); 
        if(!rec_cur->getUserName().empty()){
          timerDetails->addLine("User: "+ rec_cur->getUserName());
        }
    }

    if(!rec_cur->getNextTimers().empty() ){
      timerDetails->addLine(string(""));
      list<ProgramDescription> l = rec_cur->getNextTimers();
      if(!l.empty()){
        timerDetails->addLine("Available dates:");
        for(list<ProgramDescription>::iterator i = l.begin(); i != l.end(); i++){
          ProgramDescription pd = *i;
          pd.localizeTimes();
          timerDetails->addLine(transformTime(pd.getStartTime()));
        }
      }
    }
      
    timerDetails->update();
    app->getOsdManager()->update();
  }

  string TvTimer::transformTime(string s){
    
    if(s.empty()) return s;
    try{
      string ret = s.substr(6,2)
        + "-" + s.substr(4,2)
        + "-" + s.substr(0,4)
        + " " + s.substr(8,2)
        + ":" + s.substr(10,2);

      return ret;
    }catch(const Exception &e){
      return string("");
    }
  }

  /******************** private helpers ********************/


  /******************* Run-Method for threadedObject ********/
  void TvTimer::run(){
    
    int sel = timerList->getSelectedLine();
    string content = timerList->getLineSubject(sel);

    while(editing){
      if(timerList){
        sel = timerList->getSelectedLine();
        content = timerList->getLineSubject(sel);
        
        MutexGuard mg(mutex);
                
        if(content[stringPos] != '_'){
          contentTmp = content[stringPos];
          
          content.replace(stringPos, 1, string("_"));
        }else{
          content.replace(stringPos, 1, string(1,contentTmp));
        }
        
      }

      timerList->changeSelectedLineSubject(content);
      timerList->update();
      app->getOsdManager()->update();

      mutex.unlock();

      usleep(500000);

    }//end while(true)

    content.replace(stringPos, 1, string(1,contentTmp));
    timerList->changeSelectedLineSubject(content);
    timerList->update();
    app->getOsdManager()->update();
    
  }

  TCPAddress TvTimer::parseHostString(const string s){
    string host = "localhost";
    int port = ClientRegistry::default_port;

      
    try{
      unsigned int pos = s.find_first_of(":");
      host = s.substr(0,pos);
      if(pos != std::string::npos){
        port = atoi(s.substr(++pos).c_str());
      }
    }catch(const Exception &e){
      GLOBAL_ERROR_STREAM("host-String in wrong format" << endl);
    }

    return TCPAddress(host, port);

  }

  void TvTimer::connectTVTimerServer(string s){
    if((s != "localhost") && (s != "")){ 
      if(recorder.get()){
        SReferenceManager::getInstance().removeService("TVTimer");
      }
      ServiceProxy service_proxy;
      TCPAddress tcp_address = parseHostString(s);
      
      try{
        service_proxy.connectTo(tcp_address, "TVTimer");
        recorder.reset(service_proxy.getInterface<ITVTimer>());
        SReferenceManager::getInstance().registerService("TVTimer", recorder->getParentObject());
        return;
      } catch (const BindingException& e) {
        GLOBAL_ERROR_STREAM("Could not connect to session location: " << s << " because of " << e << endl);
        GLOBAL_WARNING_STREAM("Using local Recorder instead of remote" << endl);
      }
    }

    if(!recorder.get()){
      recorder.reset(STVTimerControl::getInstance().getInterface<ITVTimer>());
      if(recorder.get()){
        STVTimerControl::getInstance().setApplication(app->getNMMApplication());
        SReferenceManager::getInstance().registerService("TVTimer", recorder->getParentObject());
        string dest_dir;
        if (!(app->getMMBoxConfig()->getValue("tvdestinationdirectory", dest_dir))) {
          dest_dir = rootpath + "/TV-Recordings";
        }else{
          dest_dir = rootpath + "/" + dest_dir;
        }

        DirReader dr;
        dr.setRootDir("/");
        if(!dr.changeDir(dest_dir)){
          dr.makeDir(dest_dir);
        }

        recorder->setDirectory(dest_dir);

        string channelFile;
        if ((app->getMMBoxConfig()->getValue("recorderChannelFile", channelFile))) {
          channelConfig = new TVChannelList(channelFile);
          STVTimerControl::getInstance().setChannelConfig(channelConfig);
        }
        
        ITVGuide* itvgserver = NULL;
        NMMObject* o(SReferenceManager::getInstance().getService("TVGuide"));
        if(o){
          itvgserver = o->getInterface<ITVGuide>();
          recorder->setTVGuide(itvgserver);
        }else{
          GLOBAL_ERROR_STREAM("Service TVGuide not available" << endl);
        }
        
        recorder->loadTimer(rootpath + "/tvtimers");
        
      }else{
        return;
      }   
      //register TVTimer as service internally so other states can access the same object
      SReferenceManager::getInstance().registerService("TVTimer", recorder->getParentObject());
    }
  }

  /******************* Run-Method for threadedObject ****end*/

  /**** Public function to program timers *****/
  int TvTimer::programTimer(ProgramDescription p){
    GLOBAL_DEBUG_STREAM("TvTimer::programTimer(ProgramDescripottion)" << endl);
    p.localizeTimes();

    try{
      time_t t = time(NULL);
      struct tm *abs_time = localtime(&t);
      abs_time->tm_sec  = 0;//atoi(p.getStartTime().substr(12,2).c_str());
      abs_time->tm_min  = atoi(p.getStartTime().substr(10,2).c_str());
      abs_time->tm_hour = atoi(p.getStartTime().substr(8,2).c_str());
      abs_time->tm_mday = atoi(p.getStartTime().substr(6,2).c_str());
      abs_time->tm_mon  = atoi(p.getStartTime().substr(4,2).c_str())-1;
      abs_time->tm_year = atoi(p.getStartTime().substr(0,4).c_str())-1900;
      time_t startTime = mktime( abs_time );
      abs_time->tm_sec  = 0;//atoi(p.getEndTime().substr(12,2).c_str());
      abs_time->tm_min  = atoi(p.getEndTime().substr(10,2).c_str());
      abs_time->tm_hour = atoi(p.getEndTime().substr(8,2).c_str());
      abs_time->tm_mday = atoi(p.getEndTime().substr(6,2).c_str());
      abs_time->tm_mon  = atoi(p.getEndTime().substr(4,2).c_str())-1;
      abs_time->tm_year = atoi(p.getEndTime().substr(0,4).c_str())-1900;
      time_t stopTime =  mktime( abs_time );
      
      ITVGuide* itvgserver = NULL;
      NMMObject* o(SReferenceManager::getInstance().getService("TVGuide"));
      if(o){
        itvgserver = o->getInterface<ITVGuide>();
      }else{
        GLOBAL_ERROR_STREAM("Service TVGuide not available" << endl);
      }
      
      string ch = itvgserver->getChannelById(p.getChannelID());
      
      GLOBAL_DEBUG_STREAM("From TVTIMER: ID " << p.getChannelID() << " Name " << ch << endl);
            
      char tmp[20];
      strftime(tmp, sizeof(tmp), " %Y%m%d%H%M", localtime(&startTime));
      strftime(tmp, sizeof(tmp), " %Y%m%d%H%M", localtime(&stopTime));

      TVTimerEntry *newEntry = new TVTimerEntry(UserInfo().getUsername());
      newEntry->setStartTime(startTime - EPGTIMEOFFSET);
      newEntry->setEndTime(stopTime + EPGTIMEOFFSET);
      newEntry->setChannelName(ch);
      newEntry->setTimerInfo(p);
      
      if(recorder.get()){
        Result res = recorder->programTimer(*newEntry);
        
        if(res){
          return 0;
        }
      }else{
        return -1;
      }
      
    }catch(const TimerAlreadyProgrammedException &e){
      app->getOsdManager()->displayMessage("Timer Alredy programmed!", "There is already a timer programmed to record this show", "OK" );
      app->getOsdManager()->update();
      changeState("OSDInputState");

      return e.getTimerID();
    }catch(const Exception &e){
      //we catch the TimerAlreadyProgrammedException this is the only exception throw by programTimer
      return -1;
    }catch(...){
      return -1;
    }

    return -1;
  }

  int TvTimer::programTimer(const TVTimerEntry &e){
    GLOBAL_DEBUG_STREAM("TvTimer::programTimer(TVTimerEntry)" << endl);

    try{
      
      if(recorder.get()){
        Result res = recorder->programTimer(e);
        
        if(res){
          return 0;
        }
      }else{
        return -1;
      }
    }catch(TimerAlreadyProgrammedException &e){

      app->getOsdManager()->displayMessage("Timer Alredy programmed!", "There is already a timer programmed to record this show", "OK" );
      app->getOsdManager()->update();
      changeState("OSDInputState");

      return e.getTimerID();

    }catch(const Exception &e){
      return -1;
    }catch(...){
      return -1;
    }
    return -1;
  }


  list<TVTimerEntry> TvTimer::getTimers(string day, string channel){

    list<TVTimerEntry> result;
    if(recorder.get()){
      result = recorder->getTimers(day, channel);
    }

    return result;
  }

} // end namespace NMM  
