/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        David Maass <dama@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "WindowSelectorPlaylist.hpp"
#include <fstream>
#include <sstream>
#include "nmm/base/utils/GraphURL.hpp"

#include "nmm/interfaces/device/cdda/ICDDA.hpp"
#include "nmm/interfaces/device/cdda/ICDDB.hpp"
#include "nmm/interfaces/base/registry/IPluginManager.hpp"
#include "nmm/base/ProxyObject.hpp"
#include "nmm/base/NMMApplication.hpp"
#include "nmm/utils/utils.hpp"
#include <iostream>

namespace NMM {

  WindowSelectorPlaylist::WindowSelectorPlaylist(IOSDManager* _manager, IWidgetFactory* _factory, PlaylistTemplateFunctions* _state):
    playlist(NULL),
    track(NULL),
    selection(NULL),
    state(_state),
    app(NULL),
    rootpath(""),
    cdrompath("/dev/cdrom"),
    actual_file(""),
    playlistfile("mmbox.pls"),
    playlist_mode(NORMAL)
  {
    manager.reset(dynamic_cast<IOSDManager*>(_manager->copy()));
    factory.reset(dynamic_cast<IWidgetFactory*>(_factory->copy()));
    useprefix = true;
    dirreader.setNiceDir( true );
    focus = 0;
  
    elements.clear();
    updateColorBar();
  } 
 
  WindowSelectorPlaylist::~WindowSelectorPlaylist() {
    deleteWindows();
  }

  void WindowSelectorPlaylist::setApplication(MMBoxApplication* _app) {
    app=_app;
  }

  void WindowSelectorPlaylist::setWinProperties(int nr, win_properties_t* prop) {
    memcpy( &winprops[nr], prop, sizeof( win_properties_t ) );
  }

  void WindowSelectorPlaylist::deleteWindows() {
    savePlayList( rootpath+"/"+ playlistfile);

    if (selection) {
      delete selection;
      selection = NULL;
    }

    if (track) {
      delete track;
      track = NULL;
    }

    if (playlist) {
      delete playlist;
      playlist = NULL;
    }

    elements.clear();

    while (!dir_position.empty()) {
      dir_position.pop();
    }
    if (mounted) {
      mounted=false;
      system(("umount "+cdrompath).c_str());
    }
  }

  void WindowSelectorPlaylist::createWindows() {
    manager->displaySymbol("PLAY");
    manager->hideSymbol();
    focus=0;
    selection = new ListUnit(manager.get(),factory.get(),"medium");
    elements.push_back(selection);
    
    track = new ListUnit(manager.get(),factory.get(),"medium");
    elements.push_back(track);
    
    playlist = new ListUnit(manager.get(),factory.get(),"medium");
    elements.push_back(playlist);
    
    for (uint i=0;i<3;i++) {
      elements[i]->setHeader(winprops[i].name);

      /* set colors according to properties */
      elements[i]->setColors( winprops[i].win_active_color,
                              winprops[i].win_inactive_color,
                              winprops[i].sel_active_color,
                              winprops[i].sel_active_color2,
                              winprops[i].sel_inactive_color );
      
      /* set width and height of window */
      elements[i]->setDimension( winprops[i].width, winprops[i].height );
      
      /* set position of the window */
      elements[i]->setPosition( winprops[i].xpos, winprops[i].ypos );
      
    }
    
    selection->addLine("Disk");
    selection->addLine("DataCD");
    selection->addLine("AudioCD");
    selection->selectLine(1);
    selection->selectWindow();
    track->deSelectWindow();
    playlist->deSelectWindow();
    loadPlayList( rootpath+"/"+playlistfile);

    //show();
    //update();
  }

  Result WindowSelectorPlaylist::colorKeyRed() {
    switch( focus ) {
    case 0: //selection window
      if (mounted) {
	system(("umount "+cdrompath).c_str());
	selection->getTextWidget()->deSelectLine2();
	mounted=false;
        track->setHeader("");
        track->deleteAllLines();
        track->update();
      } else {
	system(("mount "+cdrompath).c_str());
        mounted=true;
	selection->getTextWidget()->selectLine2(selection->getTextWidget()->getSelectedLine());
	selection->update();
      }
      break;
    case 2: //PLAYLIST
      {
        /* if we are in playlist window delete selected entry from playlist */
        unsigned int nr = playlist->getTextWidget()->getSelectedLine();
        /* delete line only if it is not the playing titel */
        if ( nr != playlist->getTextWidget()->getSelectedLine2() ) {
          playlist->getTextWidget()->deleteLine( nr );
          playlist->update();
          manager->update();
        }
      }
      break;
    case 1: //track window
      /* if we are in track window and selecte entry is a filename, the file can be deleted
         (after confirmation) */
      if (  prefix == "file://" ) {
        unsigned int nr = track->getTextWidget()->getSelectedLine();
        if ( nr ) {
          string name = dir[nr-1].name;
          if ( (name!=".") && (name!="..") ) {
            string temp = "DEL:" + dir[nr-1].path + name;
            manager->displayMessage( "Delete", "Delete file/whole dir?" , temp, "NO", "", 1 );
            manager->update();
           
            /* change to OSDINputState to get confirmation if file should be deleted */
            state->doChangeState(string("OSDInputState"));
          }
        }
      }
      break;
    default:
      break;
    }

    updateColorBar();
     
    return SUCCESS;
  }

  Result WindowSelectorPlaylist::colorKeyGreen() {
    switch (focus) {
    case 2: //playlist window
      {
        if (playlist_mode==NORMAL) {
	  if (playlist->getTextWidget()->getSelectedLine()) {
	    playlist->selectWindow(EDIT);
	    playlist_mode = EDIT;
	  }
        } else {
          playlist->selectWindow(NORMAL);
          playlist_mode = NORMAL;
        }
	updateColorBar();
        break;
      }

    case 1: //trackwindow
      {
        unsigned int nr = track->getTextWidget()->getSelectedLine();
        if ( nr ) {
          /* if we are in track window, add entry with correct prefix to playlist */
          if ( prefix == "file://" ) {
            if ( !dir[nr-1].isdir ) {
              /* if we have a m3u-File we don't want to add the file but the playlist inside */
              if ( dirreader.getSuffix( dir[nr-1].name ) == "m3u" ) {
                ifstream playlist_file;
                  playlist_file.open((dirreader.getCurDir()+dir[nr-1].name).c_str() );
                  if (!playlist_file.fail() ) {
                  
                    string tmp,url;
                    while( getline(playlist_file, tmp) ) {
                      url = GraphURL::urlencode((dir[nr-1].path + tmp));
                      while (url.find("%2F") != url.npos) {
                        url.replace(url.find("%2F"), 3, string("/"));
                      }
                      if (useprefix)
                        url = prefix + url;
                      playlist->getTextWidget()->insertLine( playlist->getTextWidget()->getSelectedLine(), dirreader.getNiceFileName(tmp,40), url );
                      playlist->getTextWidget()->selectDown();
                    }
                  }
              } else {
                string url(GraphURL::urlencode(dir[nr-1].path + dir[nr-1].name));
                while (url.find("%2F") != url.npos) {
                  url.replace(url.find("%2F"), 3, string("/"));
                }
                if (useprefix)
                  url = prefix + url;
                playlist->getTextWidget()->insertLine( playlist->getTextWidget()->getSelectedLine(), dirreader.getFileName( dir[nr-1].name, 40 ), url );
                playlist->getTextWidget()->selectDown();
                //                Down();
              }
            } else {
              add_all_files(dir[nr-1].name);
            }
          } else
            if ( prefix == "audiocd://" ) {
              string url(prefix + string("?track=") + ((int) nr));
              playlist->getTextWidget()->insertLine( playlist->getTextWidget()->getSelectedLine(), track->getTextWidget()->getLineContent(nr), url );
              playlist->getTextWidget()->selectDown();
            }
            else {
              // unknown prefix is possible if called from WindowSelectorTranscode
              return FAILURE;
            }
          
          unsigned int playing_nr  = playlist->getTextWidget()->getSelectedLine2();
          unsigned int selected_nr = playlist->getTextWidget()->getSelectedLine();
          if ( playing_nr != 0 ) {
            if ( playing_nr >= selected_nr ) playlist->getTextWidget()->selectLine2( playing_nr+1 );
          }
          playlist->update();
        }
      }
      break;

    default:
      break;
    }

    elements[focus]->update();
    manager->update();
    return SUCCESS;
  }

  Result WindowSelectorPlaylist::colorKeyYellow() {
    if (focus==2) {
      playlist->deleteAllLines();
      updateColorBar();
      playlist->update();
      manager->update();
    }
    return SUCCESS;
  }

  Result WindowSelectorPlaylist::colorKeyBlue() {
    return SUCCESS;
  }
          

  Result WindowSelectorPlaylist::selectDown() {
    switch (focus) {
    case 2: //playlist window
      {
        // if we move down in EDIT mode, the entry must be moved down as well
        if ( playlist_mode == EDIT ) {
          unsigned int line = playlist->getTextWidget()->getSelectedLine();
          if ( line < playlist->getTextWidget()->getNumLines() ) {
            string content = playlist->getTextWidget()->getLineContent(line);
            string url     = playlist->getTextWidget()->getLineURL(line);
            playlist->getTextWidget()->deleteLine(line);
            playlist->getTextWidget()->insertLine(line, content, url);
            
            // if new position is the position of the playing titel, select playing titel correctly on new position
            if ( line == playlist->getTextWidget()->getSelectedLine2()-1 ) {
              playlist->getTextWidget()->selectLine2( line );
            } else {
              if ( line == playlist->getTextWidget()->getSelectedLine2() ) {
                playlist->getTextWidget()->selectLine2( line+1 );
              }
            }
          }
        }
        break;
      }
      
    default:
      break;
    }
    elements[focus]->selectDown();
    elements[focus]->update();

    updateColorBar();
    manager->update();

    return(SUCCESS);
  }

  Result  WindowSelectorPlaylist::selectUp() {
    switch( focus ) {
    case 2: //playlist window
      /* if we move up in EDIT mode, the entry must be moved up as well */
      if ( playlist_mode == EDIT ) {
        unsigned int line = playlist->getTextWidget()->getSelectedLine();
        if ( line > 1 ) {
          string content = playlist->getTextWidget()->getLineContent(line);
          string url     = playlist->getTextWidget()->getLineURL(line);
          playlist->getTextWidget()->deleteLine(line);
          playlist->getTextWidget()->insertLine(line-2, content, url);

          if ( line == playlist->getTextWidget()->getNumLines() ) {
            playlist->getTextWidget()->selectDown();
          }
        }
           
        /* if new position is the position of the playing titel, select playing titel correctly on new position */
        if ( line == playlist->getTextWidget()->getSelectedLine2()+1 ) {
          playlist->getTextWidget()->selectLine2( line );
        } else {
          if ( line == playlist->getTextWidget()->getSelectedLine2() ) {
            playlist->getTextWidget()->selectLine2( line-1 );
          }
        }
      }
      break;
      
    default:
      break;
    }

    elements[focus]->selectUp();    
    elements[focus]->update();

    updateColorBar();
    manager->update();

    return(SUCCESS);
  }

  Result WindowSelectorPlaylist::selectLeft() {
    if (focus==0)
      selectWindow(2);
    else selectWindow(focus-1);

    updateColorBar();

    return SUCCESS;
  }

  Result WindowSelectorPlaylist::selectRight() {
    if (focus==2)
      selectWindow(0);
    else selectWindow(focus+1);

    updateColorBar();

    return SUCCESS;
  }

  Result WindowSelectorPlaylist::select() {
    switch (focus) {
    case 0: //selection window
      {
        string sel = selection->getTextWidget()->getLineContent(selection->getTextWidget()->getSelectedLine() );
        /* if we are in selection window and the "Disk" entry is activated:
           read directoy and show it in track window */
        if ( sel == "Disk" ) {
          prefix = "file://";
          dirreader.setRootDir( rootpath );
          dirreader.changeDir( "." );
          track->setHeader( dirreader.getNiceDirName( dirreader.getCurDirWithoutRoot(), 30) );
          dirToText();
        } else if (sel =="DataCD") {
          if (!mounted) {
            system(("mount "+cdrompath).c_str());
            mounted=true;
          }
          prefix = "file://";
          dirreader.setRootDir(cdrompath);
          dirreader.changeDir(".");
          track->setHeader( dirreader.getNiceDirName( dirreader.getCurDirWithoutRoot(), 30) );
	  selection->getTextWidget()->selectLine2(selection->getTextWidget()->getSelectedLine());
	  selection->update();
          dirToText();
	  mounted=true;
        }
	/* if we are in selection window and the "CD" entry is activated:
	   read content of the audio cd and show it in track window */
        else if ( sel == "AudioCD" ) {
          prefix = "audiocd://";
          track->getTextWidget()->deleteAllLines();
          
          if (app) {
            try {
            ClientRegistry& registry = app->getNMMApplication()->getRegistry();
              INode_var cdda;
              {
                RegistryLock lock(registry);
                list<Response> cdda_response = registry.initRequest(NodeDescription("CDDANode"));
                
                if (cdda_response.empty()) {
                  GLOBAL_ERROR_STREAM("Size of cdda_response " << cdda_response.size() << endl);
                   throw Exception(string("A required node is not available\n"));
                }
                 
                cdda.reset(registry.requestNode(cdda_response.front()));
              }
              
              cdda->reachActivated();
              ICDDA_var icd(cdda->getParentObject()->getInterface<ICDDA>());
              unsigned int total_tracks(icd->getTotalTracks());
              
              ICDDB_var cddb(cdda->getParentObject()->getInterface<ICDDB>());
              cddb->setCDDBServer(string("freedb.freedb.org"), 8880);
              cddb->setCDDB(true);
              
              for (unsigned int i=0; i < total_tracks; i++) {
                track->getTextWidget()->addLine(cddb->getTrackName(i + 1));
              }
            }catch (const Exception& e) {
              manager->displayMessage("Error","Could not find a valid AudioCD","OK");
              state->doChangeState("OSDInputState");
              manager->update();
              return FAILURE;}
          }
        }
        else {
          //unknown selection: might be possible if called by WindowSelectorTranscode
          return FAILURE;
        }

        track->getTextWidget()->selectLine( 1 );
        selectWindow(1); //select Track-Window
        return SUCCESS;
        break;
      }
    case 1: //track window
      {
        unsigned int nr = track->getTextWidget()->getSelectedLine();
        if ( nr ) {
          string url;

          if ( prefix == "file://") {
            /* if directoy entry is selected, switch to directoy and
               show it in track list window */
            if ( dir[nr-1].isdir ) {
              string dirname=dir[nr-1].name;
              dirreader.changeDir( dirname );
              track->setHeader( dirreader.getNiceDirName( dirreader.getCurDirWithoutRoot(), 30) );
              dirToText();
              
              track->getTextWidget()->reset();
              if (dirname=="Go Up") {
                track->getTextWidget()->selectLine(dir_position.top());
                dir_position.pop();
              } else {
                track->getTextWidget()->selectLine( 1 );
                dir_position.push(nr);
              }
              track->update();
              manager->update();

              return SUCCESS;
            } else {
              /* build URL and hand it back to state */
              url = GraphURL::urlencode(dir[nr-1].path + dir[nr-1].name);
              while (url.find("%2F") != url.npos) {
                url.replace(url.find("%2F"), 3, string("/"));
              }
              actual_file = track->getTextWidget()->getLineContent(nr);
	      if (useprefix)
		return state->doSelectTrackWindow("file://" + url);
	      else return state->doSelectTrackWindow(url);
	    }
	  } else if (prefix == "audiocd://") {
            actual_file = track->getTextWidget()->getLineContent(nr);
	    return state->doSelectTrackWindow(prefix + string("?track=") + ((int) nr));
          } else {
            //unknown prefix is possible if called from WindowSelectorTranscode
            return FAILURE;
          }
          
        }
      }
    case 2: //playlistwindow
      {
        int line = playlist->getTextWidget()->getSelectedLine();
	actual_file = playlist->getTextWidget()->getLineContent(line);
	if (state->doSelectPlaylistWindow(playlist->getTextWidget()->getLineURL(line))) {
	  playlist->getTextWidget()->selectLine2(line);
	  playlist->update();
	  return SUCCESS;
	} else
	  return FAILURE;
      }
    default:
      break;
    }
    return FAILURE;
  }

  void WindowSelectorPlaylist::selectWindow(int winid) {
    ListUnit* temp;
    //we should test if dynamic cast fails - but it couldn't fail in this WindowTemplate -> no test
    temp = dynamic_cast<ListUnit*>(elements[focus]);
    temp->deSelectWindow();
    temp->update();

    temp = dynamic_cast<ListUnit*>(elements[winid]);
    temp->selectWindow();
    if (!temp->getSelectedLine())
      temp->selectLine(1);
    temp->update();

    manager->update();
    focus = winid;

    updateColorBar();
  }

  void WindowSelectorPlaylist::setRootpath(string path) {
    rootpath = path;
  }

  string WindowSelectorPlaylist::getSimpleName( string name ) {
    unsigned int pos = name.find_last_of('/');
    if ( pos != string::npos ) {
      string temp;
      temp = name.substr( pos+1, name.length() );
      return( temp );
    } else {
      return( name );
    }
  }

  void WindowSelectorPlaylist::savePlayList( string filename ) {
    unsigned int lines;
    if ( !playlist ) return;

    lines = playlist ->getTextWidget()->getNumLines();

    ofstream playlistfile;
    playlistfile.open( filename.c_str(), ios::trunc | ios::out );

    for( unsigned int i=1; i<=lines; i++ ) {
      string url = playlist->getTextWidget()->getLineURL( i );
      playlistfile << url << endl;
    }

    playlistfile.close();
  }

  bool WindowSelectorPlaylist::loadPlayList( string filename ) {
    ifstream playlistfile;
    
    playlistfile.open( filename.c_str() );
    if ( playlistfile.fail() ) return false;
    
    playlist->getTextWidget()->deleteAllLines();
    
    string temp;
    while( getline(playlistfile, temp) ) {
      string shortname = getSimpleName( GraphURL::urldecode(temp) );
      playlist->getTextWidget()->addLine( dirreader.getFileName( shortname, 40 ), temp );
    }
    
    playlistfile.close();
    
    return( true );
  }

 void WindowSelectorPlaylist::dirToText() {
   dir = dirreader.getDirEntries();
    
    track->deleteAllLines();
    for ( unsigned int i=0; i < dir.size(); i++ ) {
      if ( dir[i].isdir ) {
        string name = dir[i].name;
        track->addLine( "DIR: " + dirreader.getFileName( dir[i].name, 30 ) );
      }
      else track->addLine( dirreader.getFileName( dir[i].name, 30 ) );
    }
  }

  void WindowSelectorPlaylist::add_all_files(string dirname) {
    if (dirname!="Go Up") {
      dirreader.changeDir(dirname);
      vector<DirReader::directory_t> files = dirreader.getRegFiles();
      for (uint i=0;i<files.size();i++) {
        string url(GraphURL::urlencode(files[i].path + files[i].name));
        while (url.find("%2F") != url.npos) {
          url.replace(url.find("%2F"), 3, string("/"));
        }
        if (useprefix)
          url = prefix + url;
        
        playlist->getTextWidget()->insertLine( playlist->getTextWidget()->getSelectedLine(), dirreader.getFileName( files[i].name, 40 ), url );
        playlist->getTextWidget()->selectDown();
      }
      
      vector<DirReader::directory_t> dirs = dirreader.getDirsOnlyEntries();
      for (uint i=0;i<dirs.size();i++)
        if (dirs[i].name!="Go Up")
          add_all_files(dirs[i].name);
      
      dirreader.changeDir("Go Up");
    }
  }

  void WindowSelectorPlaylist::refreshFiles() {
    dirreader.changeDir( "." );
    dirToText();
    track->update();
    manager->update();
  }

  string WindowSelectorPlaylist::getPreviousPlaylistfile() {
    unsigned int act_line = playlist->getTextWidget()->getSelectedLine2();
    if (act_line-1 > 0) {
      playlist->getTextWidget()->selectLine2(act_line-1);
      playlist->update();
      manager->update();
      actual_file = playlist->getTextWidget()->getLineContent(act_line-1);
      return playlist->getTextWidget()->getLineURL(act_line-1);
    } 
    playlist->getTextWidget()->deSelectLine2();
    playlist->update();
    return "";
  }

  string WindowSelectorPlaylist::getPreviousPlaylistName() {
    unsigned int act_line = playlist->getTextWidget()->getSelectedLine2();
    if (act_line-1 > 0) {
      playlist->getTextWidget()->selectLine2(act_line-1);
      playlist->update();
      manager->update();
      actual_file = playlist->getTextWidget()->getLineContent(act_line-1);
      return actual_file;
    } 
    playlist->getTextWidget()->deSelectLine2();
    playlist->update();
    return "";
  }

  string WindowSelectorPlaylist::getNextPlaylistfile() {
    unsigned int lines = playlist->getTextWidget()->getNumLines();
    unsigned int act_line = playlist->getTextWidget()->getSelectedLine2();
    if (act_line+1 <= lines) {
      playlist->getTextWidget()->selectLine2(act_line+1);
      playlist->update();
      manager->update();
      actual_file = playlist->getTextWidget()->getLineContent(act_line+1);
      return playlist->getTextWidget()->getLineURL(act_line+1);
    } 
    playlist->getTextWidget()->deSelectLine2();
    playlist->update();
    return "";
  }

  string WindowSelectorPlaylist::getNextPlaylistName() {
    unsigned int lines = playlist->getTextWidget()->getNumLines();
    unsigned int act_line = playlist->getTextWidget()->getSelectedLine2();
    if (act_line+1 <= lines) {
      playlist->getTextWidget()->selectLine2(act_line+1);
      playlist->update();
      manager->update();
      actual_file = playlist->getTextWidget()->getLineContent(act_line+1);
      return actual_file;
    } 
    playlist->getTextWidget()->deSelectLine2();
    playlist->update();
    return "";
  }

  string WindowSelectorPlaylist::getActualFile() {
    return actual_file;
  }

  void WindowSelectorPlaylist::getAllPlaylistFiles(vector<GraphURL>& content) {
    playlist->getTextWidget()->selectLine2(0);
    string temp = playlist->getTextWidget()->getLineURL(playlist->getTextWidget()->getSelectedLine2());
    while (temp!="") {
      content.push_back(GraphURL(temp));
      temp = getNextPlaylistfile();
    }
  }

  void WindowSelectorPlaylist::getAllPlaylistNames(vector<string>& content) {
    playlist->getTextWidget()->selectLine2(0);
    string temp = playlist->getTextWidget()->getLineContent(playlist->getTextWidget()->getSelectedLine2());
    while (temp!="") {
      content.push_back(temp);
      temp = getNextPlaylistName();
    }
  }

  void WindowSelectorPlaylist::displaySymbol(string sym) {
    manager->displaySymbol(sym);
    manager->update();
  }
  
  void WindowSelectorPlaylist::addprefix(bool op) {
    useprefix=op;
  }

  void WindowSelectorPlaylist::updateColorBar() {
    switch (focus) {
    case 0: {
      if (selection) {
        string sel = selection->getTextWidget()->getLineContent(selection->getTextWidget()->getSelectedLine() );
        if (sel == "DataCD")
          manager->setColorBarText(mounted ? "Unmount" : "Mount", "", "", "");
        else 
          manager->setColorBarText("", "", "", "");
      }
      else 
        manager->setColorBarText("", "", "", "");
      break;
    }
    case 1:
      if (prefix == "file://" && track->getTextWidget()->getSelectedLine()) {
	manager->setColorBarText("Delete", "Add", "", "");
      }
      else {
	manager->setColorBarText("", "", "", "");
      }
      break;
    case 2:
      if (playlist->getTextWidget()->getSelectedLine()) {
	if (playlist->getTextWidget()->getSelectedLine() == playlist->getTextWidget()->getSelectedLine2()) {
	  manager->setColorBarText("", playlist_mode == NORMAL ? "Drag" : "Drop", "Del. all", "");
	}
	else {
	  manager->setColorBarText("Delete", playlist_mode == NORMAL ? "Drag" : "Drop", "Del. all", "");
	}
      }
      else {
	manager->setColorBarText("", "", "", "");
      }
      break;
    default:
      manager->setColorBarText("", "", "", "");
    }
    manager->update();
  }
}
