/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        NMM work group,
 *                    http://www.networkmultimedia.org
 *                     
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "nmm/base/connect.hpp"
#include "nmm/base/ProxyObject.hpp"

// include used interfaces
#include "nmm/interfaces/base/INode.hpp"
#include "nmm/interfaces/file/IFileHandler.hpp"
#include "nmm/interfaces/device/audio/IAudioDevice.hpp"

// the plugin manager
#include "nmm/base/registry/ClientRegistry.hpp"
#include "nmm/base/ProxyApplication.hpp"

#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <sys/time.h>
#include <unistd.h>
#include <getopt.h>

/* Global options */
bool verbose           = false;
bool ac3_passthrough   = false;
string filename;

using namespace NMM;

void usage( char *name ) {
  cerr << endl
       << "Usage:" << endl
       << name << " [options] [filename]" << endl
       << endl
       << "    filename:            ac3 file" << endl
       << "    -a --ac3passthrough  send ac3 data directly to soundcard" << endl
       << "    -v --verbose:        Verbose mode (lots of messages!)" << endl
       << "    -h --help:           Shows this help and exits." << endl;
  exit(0);
}

void turnOffAllMessages() {
  NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
}

int main(int argc, char ** argv) {
  // create the NMM application
  NMMApplication* app = ProxyApplication::getApplication(argc, argv);

  // get the registry
  ClientRegistry& registry = app->getRegistry();
  
  try {
    const char *optstring = "avh";
    struct option longopts[] = {
      {"ac3passthrough", 0, NULL, 'a'},
      {"verbose",        0, NULL, 'v'},
      {"help",           0, NULL, 'h'},
      {NULL,             0, NULL, 0  } /* last row must be zero */
    };

    int ch, optionindex;

    while(EOF != (ch=getopt_long(argc,argv,optstring,longopts,&optionindex))){
      // which option did we get?
      switch (ch) {
      case 'a':
        ac3_passthrough = true;
        break;
      case 'v':
        verbose = true;
        break;
      default: //includes 'h'
        usage(argv[0]);
      }
    }

    if (argc <= optind) { // no file specified
      usage(argv[0]);
    } else {
      filename = string(argv[optind]);
    }

    // maybe turn off debug messages
    if (!verbose) {
      turnOffAllMessages();
    }
    
    // create requests for nodes
    vector<NodeDescription> request(3);
    request[0].setNodeName("GenericReadNode");
    request[1].setNodeName("AC3DecodeNode");
    request[2].setNodeName("PlaybackNode");
    
    // lock the IRegistry and request nodes
    registry.initTransaction();
    vector< list<Response> > response(request.size());
    for (unsigned int i=0; i<request.size(); ++i) {
      response[i] = registry.initRequest(request[i]);
      if (response[i].empty()) {
        cerr << "Failed to request required node " << request[i].getNodeName() << endl;
        exit(1);
      }
    }
    
    // request the nodes
    INode_var readfile(registry.requestNode(response[0]));
    INode_var ac3dec(registry.requestNode(response[1]));
    INode_var audio_play(registry.requestNode(response[2]));
    
    // unlock transaction
    registry.releaseTransaction();
    
    // initialize file reader
    readfile->init();
    IFileHandler_var file_source(readfile->getParentObject()->getCheckedInterface<IFileHandler>());
    file_source->setFilename(filename);
    
    // initialize AC3 decoder
    if ( !ac3_passthrough ) {
      ac3dec->init();
    }
    
    // initialize playback node
    IAudioDevice_var audio_device(audio_play->getParentObject()->getCheckedInterface<IAudioDevice>());
    audio_device->setDevice("/dev/dsp");
    audio_play->init();
    
    // setup source file format
    // (needed because we are reading raw files)
    Format* file_format = new Format("audio/ac3");
    
    // connect and activate the nodes
    readfile->initOutput();
    
    if ( !ac3_passthrough ) {
      CommunicationChannel* readfile_to_ac3dec = c_connect(readfile.get(), ac3dec.get(), file_format);
      readfile_to_ac3dec->setTransportStrategy( readfile_to_ac3dec->getStrategyList().front() );
      
      readfile->activate();
      ac3dec->initOutput();
      
      CommunicationChannel* ac3dec_to_audio_play = c_connect(ac3dec.get(), audio_play.get());
      ac3dec_to_audio_play->setTransportStrategy( ac3dec_to_audio_play->getStrategyList().front() );
      
      ac3dec->activate();
    } else {
      CommunicationChannel* readfile_to_audio_play = c_connect(readfile.get(), audio_play.get(), file_format); 
      readfile_to_audio_play->setTransportStrategy( readfile_to_audio_play->getStrategyList().front() );
      
      readfile->activate();
    }
    
    audio_play->initOutput();
    audio_play->activate();
    
    // start nodes
    audio_play->start();
    if ( !ac3_passthrough ) {
      ac3dec->start();
    }
    readfile->start();
    
    getchar();
    
    // stop nodes
    readfile->stop();
    ac3dec->stop();
    audio_play->stop();
    
    // deactivate nodes
    readfile->deactivate();
    ac3dec->deactivate();
    audio_play->deactivate();
    
    // deinitOutput nodes
    readfile->deinitOutput();
    ac3dec->deinitOutput();
    audio_play->deinitOutput();
    
    // deinit nodes
    readfile->deinit();
    ac3dec->deinit();
    audio_play->deinit();
    
    // release nodes
    registry.releaseNode(*readfile);
    registry.releaseNode(*ac3dec);
    registry.releaseNode(*audio_play);
  } catch(Exception e) {
    cerr << e << endl;
    return 1;
  }
  
  registry.releaseTransaction();
  delete app;
  return 0;
}
