/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

// distributed mp3 player using GraphDescription with error handling and access to 'serverregistry'
// playback on local and remote host
// notice: you need a running 'serverregistry' on the remote host 

#include <iostream>

// include NMM 
#include "nmm/base/connect.hpp"
#include "nmm/base/ProxyObject.hpp"
#include "nmm/base/ProxyApplication.hpp"
#include "nmm/base/proxy/tcp/TCPStrategy.hpp"

// include used interfaces
#include "nmm/interfaces/file/IFileHandler.hpp"

// import namespace
using namespace NMM;

int main(int argc, char ** argv)
{
  if(argc != 3) {
    cerr << "Usage:" << endl
	 << "  " << argv[0] << " <mp3file> <remote host where serverregistry is running>" << endl
	 << endl;
    return 1;
  }
  cerr << argv[0] << " running ... " << endl << endl;

  // disable all debug output (error, warning, message, debug)
  // see http://graphics.cs.uni-sb.de/NMM/current/Docs/messages/index.html
  NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);
  
  // the graph description for the example
  GraphDescription graph;

  // the NMM application for the example
  NMMApplication* app = 0;

  // NMM-operations may throw exception, in this simple example, all exceptions
  // are just caught at the end, and the application is terminated
  try {
    // create application object for example
    app = ProxyApplication::getApplication(argc, argv);

    // create the descriptions for the GenericReadNode, MPEGAudioDecodeNode, and PlaybackNode
    NodeDescription readfile_descr("GenericReadNode");
    NodeDescription decoder_descr("MPEGAudioDecodeNode");
    NodeDescription audioplay_descr("PlaybackNode");
    // create the descriptions for the remote MPEGAudioDecodeNode, and remote PlaybackNode
    NodeDescription decoder_descr2("MPEGAudioDecodeNode");
    decoder_descr2.setLocation(argv[2]);
    NodeDescription audioplay_descr2("PlaybackNode");
    audioplay_descr2.setLocation(argv[2]);
    
    // create a flow graph: readfile -> remote decoder -> remote audioplay
    graph.addEdges(&readfile_descr, &decoder_descr2, &audioplay_descr2);
    // add edges: readfile -> local decoder -> local audioplay
    graph.addEdges(&readfile_descr, &decoder_descr, &audioplay_descr);
    
    // request complete graph
    cerr << "request GraphDescription from registry" << endl;
    ClientRegistry& registry = app->getRegistry();
    registry.requestGraph(graph);

    // set the filename by requesting appropriate interface
    cerr << "set filename to " << argv[1] << endl;
    INode* readfile = graph.getINode(readfile_descr);
    IFileHandler_var filehandler (readfile->getParentObject()->getCheckedInterface<IFileHandler>());
    filehandler->setFilename(argv[1]);

    // setup distributed inter-stream synchronization
    cerr << "setup synchronization" << endl;
    graph.connectSinkSynchronizer();

    // realize and start graph
    cerr << "realizeGraph()" << endl;
    graph.realizeGraph();
    cerr << "startGraph()" << endl;
    graph.startGraph();
    
    // wait ...
    cerr << "Playing audio ... enter 'q <enter>' to exit." << endl;
    string s;
    cin >> s;
    
    // stop and release graph
    cerr << "stopGraph()" << endl;
    graph.stopGraph();

    cerr << "releaseGraph()" << endl;
    registry.releaseGraph(graph);
    return 0;
  }
  catch(const Exception& e) { // catch NMM-exception and print it
    cerr << e << endl;

    // release graph
    cerr << "releaseGraph()" << endl;
    ClientRegistry& registry = app->getRegistry();
    registry.releaseGraph(graph);

    return 1;
  }
  catch(...) { // any other exception
    cerr << "something went wrong!" << endl;

    // release graph
    cerr << "releaseGraph()" << endl;
    ClientRegistry& registry = app->getRegistry();
    registry.releaseGraph(graph);

    return 1;
  }
}
