/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


/*
 * Navigieren in Verzeichnissen mit MP3-Dateien
 */
#include <iostream>
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <dirent.h>
#include <unistd.h>
#include <curses.h>
#include <getopt.h>
#include <sys/stat.h>
#include <vector>
#include <string>
#include <algorithm>

#include "nmm/base/connect.hpp"
#include "nmm/base/ProxyObject.hpp"
#include "nmm/plugins/video/display/X/XDisplayNode.hpp"
#include "nmm/plugins/video/osd/OSDOverlayNode.hpp"
#include "nmm/plugins/video/converter/RGBtoYV12ConverterNode.hpp"
#include "nmm/plugins/audio/mpegdec/MPEGAudioDecodeNode.hpp"
#include "nmm/plugins/audio/PlaybackNode.hpp"
#include "nmm/plugins/file/GenericReadNode.hpp"
#include "nmm/plugins/file/png/PNGReadNode.hpp"
#include "nmm/misc/lcd/LCDWriter.hpp"
#include "nmm/misc/lirc/LircEvents.hpp"

#include "nmm/utils/utils.hpp"

#define USECURSES //comment out for debugging

using namespace NMM;

string binname;
const string mp3suffix = ".mp3";
unsigned int cols = 0;
unsigned int lines = 0;
// number of entries (songs) to be seen
#ifndef USECURSES
unsigned int dispentrynum = 10;
#else
unsigned int dispentrynum = 0;
#endif
const unsigned int disptextlen = 70;

class LircReceiver :
  public IEventReceiver {

protected:
  EventDispatcher* dispatcher;

  Result receiveEventInternal(const CEvent& ce, const IEventSender* sender) const {
    Result result = SUCCESS;
    for (unsigned int i = 0; i< ce.size(); ++i) {
      Result r = receiveEvent(ce[i], sender);
      if (SUCCESS != r) {
        result = r;
      }
    }

    return result;
  }

  Result receiveEventInternal(const Event& e, const IEventSender* sender) const {
    if (!dispatcher) {
      return FAILURE;
    }
    
    return dispatcher->dispatch(e, sender);
  }

public:
  explicit LircReceiver(EventDispatcher* ed = NULL) :
    dispatcher(ed) {
  }

  void setEventDispatcher(EventDispatcher* ed) {
    dispatcher = ed;
  }

  virtual Result receiveEvent(const CEvent& ce, const IEventSender* sender = 0 ) {
    return receiveEventInternal(ce, sender);
  }

  virtual Result receiveEvent(const Event& e, const IEventSender* sender = 0 ) {
    return receiveEventInternal(e, sender);
  }

  virtual Result receiveEvent(const CEvent& ce, const IEventSender* sender = 0 ) const {
    return receiveEventInternal(ce, sender);
  }

  virtual Result receiveEvent(const Event& e, const IEventSender* sender = 0 )  const {
    return receiveEventInternal(e, sender);
  }
};

class MP3Console
{
private:
  // Variables
  enum MP3Status { PLAYING, STOPPED };
  MP3Status state;
  DIR* indir;
  struct dirent* indirEntry;
  string curSelPath;
  string mp3rootpath;
  string filePath;
  vector<struct dirent> curDirEntries;
  vector<string> playList;
  unsigned int countitems;
  unsigned int viewoffset;
  unsigned int selection;
  const int width;
  const int height;
  const int framerate;
  bool useGraphics;
  bool useLirc;
  bool useFullScreen;
  bool first_file;

  // audio parts
  GenericReadNode *readfile;
  MPEGAudioDecodeNode *mp3dec;
  PlaybackNode *audio_play;

  INode_var ireadfile;
  INode_var imp3dec;
  INode_var iaudio_play;
  
  IFileHandler_var ifile_source;

  // video parts
  PNGReadNode             *pngread;
  RGBtoYV12ConverterNode  *converter; /* convert overlay output to YUV */
  OSDOverlayNode          *osd;       /* On-Screen-Display */
  XDisplayNode            *display;

  INode_var ipngread;
  INode_var iconverter;
  INode_var iosd;
  INode_var idisplay;

  IOSDOverlay_var iosd_osd;
  IVideoDisplay_var idis_display;

  LircProducer lirc_producer;
  LircReceiver lirc_receiver;
  EventDispatcher *dispatcher;
  LCDWriter *lcdWriter;
  
  // Functions
  bool myIsDir(struct dirent* entry);
  bool myIsReg(struct dirent* entry);
  void initTerm();
  void resetTerm();
  void initMP3Console();
  void initGUI();
  void initSound();
  void initProducers();
  void playmp3(string filename);
  void stopmp3();
  void filterDir(string curSelPath_);
  void printAllTerminal(string curPath_, int selection_);
  void printAllOSD(string curPath_, int selection_);
  void getNewPath(string* oldPath_, int selection_);
  void turnOffAllMessages();
  void showNewPath(string newPath_);
  void addToPlayList();
  
  // Sub-Classes
  class cmpdirents
  {
    // to compare the direntries, needed by sort.
  public:
    int operator() (const struct dirent& dir1, const struct dirent& dir2) const
    {
      string s1 = dir1.d_name;
      string s2 = dir2.d_name;
      return (s1<s2);
    }
  };

public:
  MP3Console();
  ~MP3Console();

  void start(char *, bool, bool, bool);
  // all the Callback-Functions
  Result KeyUpPressed();
  Result KeyDownPressed();
  Result KeyLeftPressed();
  Result KeyRightPressed();
  Result KeyPageUpPressed();
  Result KeyPageDownPressed();
  Result selectEntry(string &);
  Result selectEntry();
  Result Quit();
  Result switchFullScreen();

  Result  receiveEndTrack();
};

MP3Console::MP3Console():width(720), height(576),framerate(25)
{
  turnOffAllMessages();
  state = STOPPED;
  countitems=0;
  viewoffset=0;
  selection =1;
  curDirEntries = vector<struct dirent>(dispentrynum);
  playList = vector<string>(0);
  readfile = new GenericReadNode("MP3READ");
  mp3dec = new MPEGAudioDecodeNode("MP3DECODE", StreamQueue::MODE_SUSPEND, 1);
  audio_play = new PlaybackNode("MP3PLAYBACK", StreamQueue::MODE_SUSPEND, 1);
  pngread = new PNGReadNode("PNGReadNode");
  converter = new RGBtoYV12ConverterNode("RGBtoYUV Converter");
  osd = new OSDOverlayNode();
  display = new XDisplayNode("MP3-File-Chooser");
  dispatcher = new EventDispatcher("mp3console-ev-dispatcher");
  lcdWriter = new LCDWriter();
  lcdWriter->init();

  ireadfile.reset(readfile->getInterface<INode>());
  imp3dec.reset(mp3dec->getInterface<INode>());
  iaudio_play.reset(audio_play->getInterface<INode>());
  ipngread.reset(pngread->getInterface<INode>());
  iconverter.reset(converter->getInterface<INode>());
  iosd.reset(osd->getInterface<INode>());
  idisplay.reset(display->getInterface<INode>());
  
  ifile_source.reset(readfile->getInterface<IFileHandler>());

  audio_play->getEventDispatcher().registerEventListener(
	      ITrack::endTrack_event, new TEDObject0<MP3Console>(this, &MP3Console::receiveEndTrack));
  audio_play->getEventDispatcher().registerEventListener(
	      ITrack::startTrack_event, new TEDObject1<LCDWriter, string>(lcdWriter, &LCDWriter::receiveStartTrack));

  indir = 0;
  indirEntry = 0;
  curSelPath = "";
  filePath = "";
  mp3rootpath="";
  useGraphics = false;
  useLirc = false;
  useFullScreen = false;
  first_file = true;
}

MP3Console::~MP3Console()
{
  delete readfile;
  delete mp3dec;
  delete audio_play;
  delete pngread;
  delete converter;
  delete osd;
  delete display;
  delete dispatcher;
  delete lcdWriter;
}

bool MP3Console::myIsDir(struct dirent* entry)
{
  struct stat statbuf;
  lstat(entry->d_name, &statbuf);
  return(S_ISDIR(statbuf.st_mode));
}

bool MP3Console::myIsReg(struct dirent* entry)
{
  struct stat statbuf;
  lstat(entry->d_name, &statbuf);
  return(S_ISREG(statbuf.st_mode));
}

Result MP3Console::receiveEndTrack()
{
  // Just select the next track (BUGGY YET)

  cerr << "EndTrack received" << endl;
  cerr << "EndTrack received" << endl;
  cerr << "EndTrack received" << endl;
  cerr << "EndTrack received" << endl;
  cerr << "EndTrack received" << endl;
  cerr << "EndTrack received" << endl;
  cerr << "EndTrack received" << endl;
  cerr << "EndTrack received" << endl;
  cerr << "EndTrack received" << endl;
  cerr << "EndTrack received" << endl;
  cerr << "EndTrack received" << endl;
  cerr << "EndTrack received" << endl;
  cerr << "EndTrack received" << endl;

  selection++;
  if (selection==dispentrynum)
  {
    selection = 1;
  }
  selectEntry();
  return SUCCESS;
}

void MP3Console::initTerm()
{
  // curses-functions
  initscr();
  keypad(stdscr, TRUE);
  cbreak();
  noecho();
  
  // get size of the terminal
  cols = COLS;
  lines = LINES;
  // determine number of entries to be displayed
  // -...: leave space for directory and other information
  dispentrynum = lines - 12;
}

void MP3Console::resetTerm()
{
  keypad(stdscr, FALSE);
  nocbreak();
  echo();
}

void MP3Console::initMP3Console()
{
  if (useGraphics)
    initGUI();
  initSound();
  initProducers();
  state = STOPPED;
}

void MP3Console::initGUI()
{
  IMultipleSourceFileHandler_var ipngread_sourcefile(pngread->getInterface<IMultipleSourceFileHandler>());
  ipngread_sourcefile->addFilename(getNMMResource("timber_back.png"));
  ipngread->init();
  iconverter->init();
  iosd->init();
  idis_display.reset(display->getInterface<IVideoDisplay>());
  idis_display->setFullscreen(useFullScreen);
  display->init();
  
  ipngread->initOutput();

  iosd_osd.reset(osd->getInterface<IOSDOverlay>());
  iosd_osd->setFont(getNMMResource("font/font.desc"), 0.75 );
  
  cerr << "connect png,con" << endl;
  connect(ipngread, iconverter);
  cerr << "png,con connected" << endl;
  ipngread->activate();
  iconverter->initOutput();
  cerr << "connect con,osd" << endl;
  connect(iconverter, iosd);
  cerr << "con,osd connected" << endl;
  iconverter->activate();
  iosd->initOutput();
  cerr << "connect png,con" << endl;
  connect(iosd, idisplay);
  iosd->activate();
  idisplay->initOutput();
  idisplay->activate();
  
  ipngread->start();
  iconverter->start();
  iosd->start();
  
  // Stop the readnode, because the osd will cache the image,
  // and we don't need to get a Buffer from pngread 25 times/second.
  iosd_osd->doPause(true);
  ipngread->stop();
  
  idisplay->start();
}

void MP3Console::initSound()
{
  ireadfile->init();
  
  imp3dec->init();
  
  IDevice_var iaudio_dev(audio_play->getInterface<IDevice>());
  iaudio_dev->setDevice(string("/dev/dsp"));
  iaudio_play->init();
}

void MP3Console::initProducers()
{
  if (useLirc || useGraphics)
  {
    if (useLirc) {
      lirc_producer.readConfig( getNMMResource("mp3console.lirc") );
      lirc_producer.setSendExtendedEvents(true);
      lirc_producer.connectTo( &lirc_receiver );
      lirc_producer.start();
      lirc_receiver.setEventDispatcher( dispatcher );
    }

    if (useGraphics) {
      // Enable display events
      IDisplayEventSender_var display_event_sender(idisplay->getParentObject()->getInterface<IDisplayEventSender>());
      if (idisplay->isActivated() && display_event_sender.get()) {
	display_event_sender->enableDisplayEvents(getNMMResource("mp3console.xml"));
      }
      else cerr << "COULD NOT ENABLE DISPLAY EVENTS!" << endl;
    }
    
    dispatcher->registerEvent( "KEY_q", new TEDObject0<MP3Console>(this, &MP3Console::Quit) );
    dispatcher->registerEvent( "KEY_f", new TEDObject0<MP3Console>(this, &MP3Console::switchFullScreen) );
    dispatcher->registerEvent( "KEY_Up", new TEDObject0<MP3Console>(this, &MP3Console::KeyUpPressed) );
    dispatcher->registerEvent( "KEY_Down", new TEDObject0<MP3Console>(this, &MP3Console::KeyDownPressed) );
    dispatcher->registerEvent( "KEY_Pageup", new TEDObject0<MP3Console>(this, &MP3Console::KeyPageUpPressed) );
    dispatcher->registerEvent( "KEY_Pagedown", new TEDObject0<MP3Console>(this, &MP3Console::KeyPageDownPressed) );
    dispatcher->registerEvent( "KEY_Left", new TEDObject0<MP3Console>(this, &MP3Console::KeyLeftPressed) );
    dispatcher->registerEvent( "KEY_Right", new TEDObject0<MP3Console>(this, &MP3Console::KeyRightPressed) );
    dispatcher->registerEvent( "KEY_Return", new TEDObject0<MP3Console>(this, &MP3Console::selectEntry) );
    dispatcher->registerEvent( "KEY_0", new TEDObject1<MP3Console, string>(this, &MP3Console::selectEntry) );
    dispatcher->registerEvent( "KEY_1", new TEDObject1<MP3Console, string>(this, &MP3Console::selectEntry) );
    dispatcher->registerEvent( "KEY_2", new TEDObject1<MP3Console, string>(this, &MP3Console::selectEntry) );
    dispatcher->registerEvent( "KEY_3", new TEDObject1<MP3Console, string>(this, &MP3Console::selectEntry) );
    dispatcher->registerEvent( "KEY_4", new TEDObject1<MP3Console, string>(this, &MP3Console::selectEntry) );
    dispatcher->registerEvent( "KEY_5", new TEDObject1<MP3Console, string>(this, &MP3Console::selectEntry) );
    dispatcher->registerEvent( "KEY_6", new TEDObject1<MP3Console, string>(this, &MP3Console::selectEntry) );
    dispatcher->registerEvent( "KEY_7", new TEDObject1<MP3Console, string>(this, &MP3Console::selectEntry) );
    dispatcher->registerEvent( "KEY_8", new TEDObject1<MP3Console, string>(this, &MP3Console::selectEntry) );
    dispatcher->registerEvent( "KEY_9", new TEDObject1<MP3Console, string>(this, &MP3Console::selectEntry) );
  }
}

void MP3Console::playmp3(string filename)
{
  GLOBAL_DEBUG_STREAM("MSADEBUG: In method playmp3..." << endl);
  if (state == STOPPED)
  {
    ifile_source->setFilename(filename);
    
    if (first_file) {
      first_file = false;
      
      ireadfile->initOutput();
      connect(ireadfile, imp3dec);
      ireadfile->activate();
      imp3dec->initOutput();
      connect(imp3dec, iaudio_play);
      imp3dec->activate();
      iaudio_play->initOutput();
      iaudio_play->activate();
    }
    
    if (ireadfile->isActivated() && !ireadfile->isStarted()) ireadfile->start();
    if (imp3dec->isActivated() && !imp3dec->isStarted()) imp3dec->start();
    if (iaudio_play->isActivated() && !iaudio_play->isStarted()) iaudio_play->start();
    
    GLOBAL_DEBUG_STREAM("readfile is in state " << ireadfile->getState() << endl);
    GLOBAL_DEBUG_STREAM("mp3dec is in state " << imp3dec->getState() << endl);
    GLOBAL_DEBUG_STREAM("audio_play is in state " << iaudio_play->getState() << endl);
    
    state = PLAYING;
  }
}

void MP3Console::stopmp3()
{
  if (state == PLAYING)
  {
    try
    {
      if (ireadfile->isStarted())
	ireadfile->stop();
      if (imp3dec->isStarted())
	imp3dec->stop();
      imp3dec->flush();
      //usleep(200);
      if (iaudio_play->isStarted()) {
	iaudio_play->stop();
      }
      state = STOPPED;
    }
    catch (NMM::Exception e)
    {
      cerr << e << endl;
    }
  }
}

void MP3Console::filterDir(string curSelPath_)
{
  countitems = 0;
  curDirEntries.clear();

  vector<struct dirent> dirs(dispentrynum);
  vector<struct dirent> regFiles(dispentrynum);

  dirs.clear();
  regFiles.clear();
  // First read the Names of any Sub-Directories ...
  if ((indir = opendir(curSelPath_.c_str()))!=0)
  {
    chdir(curSelPath_.c_str());
    while((indirEntry = readdir(indir))!=0)
    {
      if(myIsDir(indirEntry) && strcmp(indirEntry->d_name,"."))
      {
	if (countitems==dirs.size())
        {
	  dirs.resize(dirs.size() + dispentrynum);
	}
	dirs[countitems++] = *indirEntry;
      }
    }
    closedir(indir);
  }
  else
  {
    cerr << "Fehler beim Oeffnen des Verzeichnisses" << endl;
  }

  dirs.resize(countitems);
  sort(dirs.begin(), dirs.end(), cmpdirents());

  unsigned int regfilecount = 0;

  // ... and now the mp3-files
  if ((indir = opendir(curSelPath_.c_str()))!=0)
  {
    chdir(curSelPath_.c_str());
    while((indirEntry = readdir(indir))!=0)
    {
      if(myIsReg(indirEntry))
      {
	// filter for mp3-files by calculating the last position of
	// a substring ".mp3", add 4 and compare to len.
	string name = indirEntry->d_name;
	uint lpos = name.rfind(mp3suffix);
	uint len = name.length();
	if ((lpos != string::npos) && (len == lpos + mp3suffix.length()))
	{
	  if (regfilecount==regFiles.size())
	  {
	    regFiles.resize(regFiles.size() + dispentrynum);
	  }
	  regFiles[regfilecount++] = *indirEntry;
	}
      }
    }
    closedir(indir);
  }
  else
  {
    cerr << "Fehler beim Oeffnen des Verzeichnisses" << endl;
  }

  // resize to fit exactly (because sort would bring
  // empty entries to be shown at top)
  regFiles.resize(regfilecount);
  // sort list
  if (regfilecount > 0)
    sort(regFiles.begin(), regFiles.end(), cmpdirents());

  // copy dir entries in front of regular files
  curDirEntries.insert(curDirEntries.begin(), regFiles.begin(), regFiles.end());
  curDirEntries.insert(curDirEntries.begin(), dirs.begin(), dirs.end());

  cerr << endl << "curDirEntries.size(): " << curDirEntries.size() << endl;

  // reset countitems to the last entry
  countitems += regfilecount - 1;
}

void MP3Console::printAllOSD(string curPath_, int selection_)
{
  char textElemName[10];
  char textElemText[disptextlen];
  int  menu_offset_x = 20;
  int  list_offset_x = 50;
  int  row_height = 40;
  int  border_offset_y = 20;

  snprintf(textElemText, disptextlen, "Current dir: %s", curPath_.c_str());
  osd->addText("header", textElemText, menu_offset_x, border_offset_y);

  if (viewoffset >= dispentrynum)
    iosd_osd->addText("prevpage", "p: Previous Page", menu_offset_x, 1*row_height + border_offset_y);
  else
    iosd_osd->addText("prevpage", "", menu_offset_x, 1*row_height + border_offset_y);

  for (unsigned int i=viewoffset; i < viewoffset + dispentrynum; i++)
  {
    if (i <= countitems)
    {
      if (i == viewoffset + selection_)
      {
	// print "01 trackname" to string and add this text to osd
	snprintf(textElemName, menu_offset_x, "text%d", i-viewoffset);
	snprintf(textElemText, disptextlen, "-- %d: %s", i - viewoffset, curDirEntries[i].d_name);
	iosd_osd->addText(textElemName, textElemText, list_offset_x - 20, (i - viewoffset + 2)*row_height + border_offset_y);
      }
      else
      {
	// print "09          " to string and add this text to osd
	snprintf(textElemName, menu_offset_x, "text%d", i-viewoffset);
	snprintf(textElemText, disptextlen, "%d: %s", i - viewoffset, curDirEntries[i].d_name);
	iosd_osd->addText(textElemName, textElemText, list_offset_x, (i - viewoffset + 2)*row_height + border_offset_y);
      }
    }
    else
    {
      snprintf(textElemName, 10, "text%d", i-viewoffset);
      snprintf(textElemText, disptextlen, " ");
      iosd_osd->addText(textElemName, textElemText, list_offset_x, (i - viewoffset + 2)*row_height + border_offset_y);
    }
  }

  if (countitems > viewoffset + dispentrynum)
    iosd_osd->addText("nextpage", "n: Next Page", menu_offset_x, 12*row_height + border_offset_y);
  else
    iosd_osd->addText("nextpage", "", menu_offset_x, 12*row_height + border_offset_y);

  if (state == PLAYING)
  {
    snprintf(textElemText, disptextlen, "Playing file: %s", filePath.c_str());
    iosd_osd->addText("fileinfo", textElemText, menu_offset_x, 13*row_height + border_offset_y);
  }
  else
    iosd_osd->addText("fileinfo", "", menu_offset_x, 13*row_height + border_offset_y);

  // update OSD
  iosd_osd->update();
}

void MP3Console::printAllTerminal(string curPath_, int selection_)
{
  /* Ausgabe: */
  clear();
  move(0,0);

  printw("Current dir:\n%s\n\n", curPath_.c_str());

  if (viewoffset >= dispentrynum)
  {
    printw("p: Previous Page\n\n");
  }
  else printw("\n\n");

  for (unsigned int i=viewoffset; i < viewoffset + dispentrynum; i++)
  {
    if (i <= countitems)
    {
      // invert printout of selected Entry
      if (i == viewoffset + selection_)
	attron(A_REVERSE);
      if (myIsDir(&curDirEntries[i]))
	attron(A_BOLD);
      
      printw("%d: %s\n", i - viewoffset, curDirEntries[i].d_name);
      
      if (i == viewoffset + selection_)
	attroff(A_REVERSE);
      attroff(A_BOLD);
    }
    else
      printw("%d: \n", i - viewoffset);
  }

  if (countitems > viewoffset + dispentrynum)
  {
    printw("\nn: Next Page\n");
  }
  else printw("\n\n");

  if (state == PLAYING)
  {
    printw("\nPlaying file:\n%s\n", filePath.c_str());
  }

  refresh();
}

void MP3Console::getNewPath(string* oldPath_, int selection_)
{
  // Path must not end with "/"!
  if (!strcmp(curDirEntries[selection_].d_name,".."))
  {
    // ".." selected in mp3rootpath shouldn't do anything.
    if (strcmp(oldPath_->c_str(),mp3rootpath.c_str()))
    {
      int lastPos = oldPath_->find_last_of('/');
      *oldPath_ = oldPath_->substr(0,lastPos);
    }
  }
  else
  {
    *oldPath_ += "/";
    *oldPath_ += curDirEntries[selection_].d_name;
  }
}

void MP3Console::turnOffAllMessages()
{
  NamedObject::getGlobalInstance().setErrorStream(new ofstream("logfile.errors"), NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setDebugStream(new ofstream("logfile.debug"), NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setWarningStream(new ofstream("logfile.warnings"), NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setMessageStream(new ofstream("logfile.messages"), NamedObject::ALL_LEVELS);
//    NamedObject::getGlobalInstance().setErrorStream(NULL, NamedObject::ALL_LEVELS);
//    NamedObject::getGlobalInstance().setDebugStream(NULL, NamedObject::ALL_LEVELS);
//    NamedObject::getGlobalInstance().setWarningStream(NULL, NamedObject::ALL_LEVELS);
//    NamedObject::getGlobalInstance().setMessageStream(NULL, NamedObject::ALL_LEVELS);
}

void MP3Console::showNewPath(string newPath_)
{
  filterDir(newPath_);
  selection = 1;
#ifdef USECURSES
  printAllTerminal(newPath_, selection);
#endif
  if (useGraphics)
    printAllOSD(newPath_, selection);
} 

void MP3Console::addToPlayList()
{
  int entrynum = 0;

  if (selection <= countitems)
  {
    entrynum = viewoffset + selection;
    // in case of a directory
    if(myIsDir(&curDirEntries[entrynum]))
    {
      cerr << "Selection of whole directories is not implemented yet!" << endl;
    }
    // in case of a regular file
    else if(myIsReg(&curDirEntries[entrynum]))
    {
      // Remember complete Path to selected file
      string addFilePath = curSelPath;
      addFilePath += "/";
      addFilePath += curDirEntries[entrynum].d_name;
      playList.insert(playList.end(), addFilePath);
      cerr << "Adding " << addFilePath << " to PlayList..." << endl;
    }
  }
}

Result MP3Console::KeyUpPressed()
{
  if (selection == 0)
  // KEY_UP on first visible entry -> page up
  {
    if (viewoffset >= dispentrynum)
    {
      viewoffset -= dispentrynum;
      selection = dispentrynum - 1;
    }
  }
  else
    selection--;
#ifdef USECURSES
  printAllTerminal(curSelPath, selection);
#endif
  if (useGraphics)
    printAllOSD(curSelPath, selection);
  return SUCCESS;
}

Result MP3Console::KeyDownPressed()
{
  if (selection == dispentrynum - 1)
  // KEY_DOWN on last visible entry -> page down
  {
    if (viewoffset + dispentrynum < countitems)
    {
      viewoffset += dispentrynum;
      selection = 0;
    }
  }
  else if (selection + viewoffset < countitems)
    selection++;
#ifdef USECURSES
  printAllTerminal(curSelPath, selection);
#endif
  if (useGraphics)
    printAllOSD(curSelPath, selection);
  return SUCCESS;
}

Result MP3Console::KeyPageUpPressed()
{
  if (viewoffset >= dispentrynum)
  {
    viewoffset -= dispentrynum;
    selection = dispentrynum - 1;
#ifdef USECURSES
    printAllTerminal(curSelPath, selection);
#endif
    if (useGraphics)
      printAllOSD(curSelPath, selection);
  }
  return SUCCESS;
}

Result MP3Console::KeyPageDownPressed()
{
  if (viewoffset + dispentrynum < countitems)
  {
    viewoffset += dispentrynum;
    selection = 0;
#ifdef USECURSES
    printAllTerminal(curSelPath, selection);
#endif
    if (useGraphics)
      printAllOSD(curSelPath, selection);
  }
  return SUCCESS;
}

Result MP3Console::KeyLeftPressed()
{
  viewoffset = 0;
  selection = 0;
  selectEntry();
  return SUCCESS;
}

Result MP3Console::KeyRightPressed()
{
  selectEntry();
  return SUCCESS;
}

Result MP3Console::selectEntry()
{
  char c[4];
  snprintf(c, 4, "%d", selection);
  string keyStr(c);
  selectEntry(keyStr);
  return SUCCESS;
}

Result MP3Console::selectEntry(string &eventvalue)
{
  // if ENTER is pressed, the selection stays the same.
  // otherwise, the number to be selected is expected in eventvalue
  if (eventvalue != "ENTER") selection = atoi(eventvalue.c_str());
  int entrynum = 0;

  if (selection <= countitems)
  {
    entrynum = viewoffset + selection;
    // in case of a directory
    if(myIsDir(&curDirEntries[entrynum]))
    {
      // change Directory, get Contents, display and reset viewoffset
      viewoffset = 0;
      getNewPath(&curSelPath, entrynum);
      showNewPath(curSelPath);
    }
    // in case of a regular file
    else if(myIsReg(&curDirEntries[entrynum]))
    {
      // Remember complete Path to selected file
      filePath = curSelPath;
      filePath += "/";
      filePath += curDirEntries[entrynum].d_name;
      // TODO:
      // problem: file permissions:
      // if a file is not readable for the user, it
      // will be displayed, but can't be played.
      stopmp3();
      playmp3(filePath);
      // to get a screen update:
#ifdef USECURSES
      printAllTerminal(curSelPath, selection);
#endif
      if (useGraphics)
	printAllOSD(curSelPath, selection);
    }
  }
  return SUCCESS;
}

Result MP3Console::Quit()
{
#ifdef USECURSES
  resetTerm();
#endif
  exit(0);
}

Result MP3Console::switchFullScreen()
{
  useFullScreen = 1 - useFullScreen;
  if (useGraphics)
    idis_display->setFullscreen(useFullScreen);
  return SUCCESS;
}

void MP3Console::start(char *mp3rootpath_, bool useLirc_, bool useGraphics_, bool useFullScreen_)
{
  useLirc = useLirc_;
  useGraphics = useGraphics_;
  useFullScreen = useFullScreen_;
  mp3rootpath = mp3rootpath_;
  
  // free mp3rootpath from last "/"
  unsigned int lastPos = mp3rootpath.find_last_of('/');
  if (lastPos + 1 == mp3rootpath.length())
  {
    string tmpPath(mp3rootpath, 0, lastPos);
    mp3rootpath = tmpPath;
  }

  curSelPath = mp3rootpath;

#ifdef USECURSES
  initTerm();
#endif
  
  initMP3Console();

  showNewPath(curSelPath);

  int key = 0;

  while(key != 'q')
  {
    switch (key)
    {
    case '0': // fall through
    case '1': // fall through
    case '2': // fall through
    case '3': // fall through
    case '4': // fall through
    case '5': // fall through
    case '6': // fall through
    case '7': // fall through
    case '8': // fall through
    case '9': // fall through
      selection = key-'0'; //fall through
    case KEY_ENTER: // fall through
    case '\n':
      selectEntry();
      break;
    case 'n': // fall through
    case KEY_NPAGE:
      KeyPageDownPressed();
      break;
    case 'p': // fall through
    case KEY_PPAGE:
      KeyPageUpPressed();
      break;
    case 'c': // TODO: Pause and Continue
      /*if (0!=filePath.length())
	playmp3(filePath);*/
      break;
    case 'a':
      addToPlayList();
      break;
    case KEY_UP:
      KeyUpPressed();
      break;
    case KEY_DOWN:
      KeyDownPressed();
      break;
    case KEY_LEFT:
      KeyLeftPressed();
      break;
    case KEY_RIGHT:
      KeyRightPressed();
      break;
    default:
      break;
    }
    key = getch();
  }
  Quit();
}

void usage()
{
  cerr << "USAGE:" << endl;
  cerr << binname << " [-f|-g] [-h] [-l] mp3rootpath" << endl;
  cerr << " -f --fullscreen : enable fullscreen, implies '-g'" << endl;
  cerr << " -g --graphics   : enable graphical User Interface" << endl;
  cerr << " -h --help       : print this help screen" << endl;
  cerr << " -l --lirc       : enable Lirc" << endl;
  cerr << " mp3rootpath     : The path to all your mp3-files" << endl;
  exit(0);
}

int main (int argc, char* argv[])
{
  bool lirc = false; bool graphics = false;
  bool fullscreen = false;
  const char *optstring = "fghl";
  struct option longopts[] = {
    {"fullscreen", 0, NULL, 'f'},
    {"graphics"  , 0, NULL, 'g'},
    {"help"      , 0, NULL, 'h'},
    {"lirc"      , 0, NULL, 'l'},
    {NULL        , 0, NULL, 0  } /* last row must be zero */
  };

  int ch, optionindex;

  binname = argv[0];
  
  while(EOF != (ch=getopt_long(argc,argv,optstring,longopts,&optionindex))){
    // which option did we get?
    switch (ch) {
    case 'f':
      fullscreen = true;
      graphics = true;
      break;
    case 'g':
      graphics = true;
      break;
    case 'l':
      lirc = true;
      break;
    default:
      usage();
    }
  }  

  if (optind >= argc) { // assure that mp3rootpath is given
    usage();
  }
  
  MP3Console* mp3console = new MP3Console();
  mp3console->start(argv[optind], lirc, graphics, fullscreen);
  delete mp3console;
}
