/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef NMM_COMPOSITESTRATEGY_HPP
#define NMM_COMPOSITESTRATEGY_HPP

#include "TransportStrategy.hpp"

#include "ObjectAcceptor.hpp"

#include "nmm/interfaces/base/ICompositeStrategyImpl.hpp"
#include "nmm/base/sync/TimedElement.hpp"
#include "nmm/utils/Average.hpp"

namespace NMM {

  class CompositeStrategy;
  
  class IConnectionManager;
  class ConnectorAddress;
  
  /**
   *  @brief Composite address
   *
   *  This is the address type used by the CompositeStrategy. The address consists of
   *  two child addresses for the child strategies of the composite strategy and a master address
   *  which is used to connect the CompositeStrategy itself.
   *
   *  A composite address is valid iff both child addresses and the master address are non-NULL,
   *  the master address is a ConnectorAddress with strategy type "CompositeAddress" and neither
   *  of the child addresses is a CompositeAddress.
   */
  class CompositeAddress
    : public TSAddress
  {
  public:
    /**
     *  @brief Default constructor
     *  @param ref NMM reference
     *
     *  Construct a composite address with no child addresses.
     */
    CompositeAddress(const NMMReference& ref = 0);
    
    /** 
     *  @brief Constructor
     *  @param ref NMM reference
     *  @param master Master address
     *  @param buffer_address Buffer child address
     *  @param event_address Event child address
     *
     *  Construct a composite address with the given master and child addresses.
     */
    CompositeAddress(const NMMReference& ref, const TSAddress& master, const TSAddress& buffer_address, const TSAddress& event_address);
    
    /** 
     *  @brief Copy constructor
     */
    CompositeAddress(const CompositeAddress&);
    
    /** 
     *  @brief Virtual copy constructor
     */
    virtual CompositeAddress* copy() const;
    //@}
    
    /**
     *  @name Accessors
     */
    //@{
    /** 
     *  @brief Get master address
     *  @return The master address of this composite address
     */
    const TSAddress* getMasterAddress() const;
    
    /** 
     *  @brief Get buffer child address
     *  @return The buffer child address of this composite address
     */
    const TSAddress* getBufferAddress() const;
    
    /** 
     *  @brief Get event child address
     *  @return The event child address of this composite address
     */
    const TSAddress* getEventAddress() const;
    
    /**
     *  @brief Set master address
     *  @param address The new master address of this composite address
     */
    void setMasterAddress(const TSAddress& address);
    
    /**
     *  @brief Set buffer child address
     *  @param address The new buffer child address of this composite address
     */
    void setBufferAddress(const TSAddress& address);
    
    /**
     *  @brief Set event child address
     *  @param address The new event child address of this composite address
     */
    void setEventAddress(const TSAddress& address);
    //@}
    
  private:
    /** The master address.
     */
    TValue<TSAddress*> __master_address;
    
    /** The buffer child address.
     */
    TValue<TSAddress*> __buffer_address;
    
    /** The event child address.
     */
    TValue<TSAddress*> __event_address;
  };
  
  inline const TSAddress* CompositeAddress::getMasterAddress()  const {
    return __master_address.getValue();
  }
  
  inline const TSAddress* CompositeAddress::getBufferAddress()  const {
    return __buffer_address.getValue();
  }
  
  inline const TSAddress* CompositeAddress::getEventAddress() const {
    return __event_address.getValue();
  }
  
  inline void CompositeAddress::setMasterAddress(const TSAddress& address) {
    __master_address.setValue(address.copy());
  }
  
  inline void CompositeAddress::setBufferAddress(const TSAddress& address)  {
    __buffer_address.setValue(address.copy());
  }
  
  inline void CompositeAddress::setEventAddress(const TSAddress& address) {
    __event_address.setValue(address.copy());
  }
  
  
  /**
   *  @brief Instream message and signal handler
   *
   *  This class performs the multiplexing of instream messages from both child strategies
   *  of a CompositeStrategy.
   *
   *  The handler receives all instream messages and signals from one of the child strategies
   *  and forwards them to the right handling methods of the parent strategy.
   */
  class InstreamHandler
    : public IMessageReceiver,
      public ISignalReceiver
  {
  public:
    /**
     *  @brief Construct an %InstreamHandler with default properties.
     */
    InstreamHandler() throw();
    
    /**
     *  @brief Connect handler
     *  @brief parent Parent strategy
     *  @brief child Child strategy, must be a child strategy of @c parent
     *
     *  This method sets the parent and the child strategy of this handler. It also sets
     *  the message and signal receiver of the child strategy to itself.
     */
    virtual void connect(CompositeStrategy* parent, TransportStrategy* child) throw();
    
    /**
     *  @brief Disconnect handler
     *
     *  This method resets the parent and child strategy of this handler to 0. If a child strategy
     *  was set, then its message and signal receiver are both set to 0.
     */
    virtual void disconnect();
    
    /**
     *  @brief Get parent strategy
     *  @return The parent strategy of this handler
     */
    CompositeStrategy& getParent() const throw();
    
    /**
     *  @brief Get child strategy
     *  @return The child strategy of this handler
     */
    TransportStrategy& getTransportStrategy() const throw();

    /**
     *  @brief Receive signal
     *
     *  This method is the signal handler for the child strategy of this handler.
     *  It forwards the signal to the method CompositeStrategy::receiveChildSignal of the parent.
     */
    virtual void receiveSignal(Signal);
    
    /**
     *  @brief Receive instream message
     *
     *  This method is the signal handler for the child strategy of this handler.
     *  It forwards the signal to the method CompositeStrategy::receiveChildMessage of the parent.
     */
    virtual Result receiveMessage(Message*, const IMessageSender* = 0);
    
    /**
     *  @brief Forward message to sub binding
     */
    Result forwardMessage(Message* message, const IMessageSender* sender = 0);
    
    /**
     *  @brief Check if sub binding is blocked
     *  @return @c true if the sub binding is blocked
     */
    bool isBlocked() const;
    
    /**
     *  @brief Get pending message
     *  @return The pending message
     */
    Message* getMessage() const;
    
    /**
     *  @brief Remove pending message
     *
     *  This method sets the pending message of this binding to @c NULL. A @c FREE_QUEUE_SLOT signal is sent to the
     *  sub binding if it is not currently processing a receiveMessage call.
     */
    void unblock();

    /**
     *  @brief Get current receiver timestamp
     *  @return The timestamp of the most recently received message
     */
    const Timestamp& getRecvTimestamp() const;

    /**
     *  @brief Get current sender timestamp
     *  @return The timestamp of the most recently sent message
     */
    const Timestamp& getSendTimestamp() const;
    
    /**
     *  @brief Get average transmission delay
     *  @return The average transmission delay for this sub binding
     */
    const Interval getAvgTransmissionDelay() const;
    
  private:
    CompositeStrategy* __parent;                 ///< Parent CompositeStrategy
    TransportStrategy* __child;                  ///< Child strategy
    
    /* BEGIN> Sender state */
    Timestamp __send_timestamp;                  ///< Timestamp of most recently sent message
    bool __blocked;                              ///< Sub-binding is blocked
    ThreadMutex __send_mutex;                    ///< Sender mutex
    /* END< */

    /* BEGIN> Receiver state */
    Message* __pending;                          ///< Most recently received message, if any
    TAverage<Interval> __transmission_delay;     ///< Average transmission delay
    Timestamp __recv_timestamp;                  ///< Timestamp of most recently received message
    ThreadMutex __recv_mutex;                    ///< Receiver mutex
    bool __is_current;                           ///< Currently in a receiveMessage call
    /* END< */
  };
  
  inline CompositeStrategy& InstreamHandler::getParent() const throw() {
    return *__parent;
  }
  
  inline TransportStrategy& InstreamHandler::getTransportStrategy() const throw() {
    return *__child;
  }
  
  inline bool InstreamHandler::isBlocked() const {
    return __blocked;
  }
  
  inline Message* InstreamHandler::getMessage() const {
    return __pending;
  }  
  
  inline const Timestamp& InstreamHandler::getRecvTimestamp() const {
    return __recv_timestamp;
  }
  
  inline const Timestamp& InstreamHandler::getSendTimestamp() const {
    return __send_timestamp;
  }
  
  inline const Interval InstreamHandler::getAvgTransmissionDelay() const {
    return __transmission_delay.get() /*+ __transmission_delay.getVariance()*/;
  }
  
  
  /** 
   *  @brief Composite transport strategy
   *
   *  A composite transport strategy uses two transport strategies (called child strategies) for sending instream messages.
   *  Buffers and events are transmitted separately using either of the two child strategies.
   */
  class CompositeStrategy
    : public TransportStrategy,
      public ICompositeStrategyImpl,
      public TimedElement,
      public TSAddressOwner
  {
  public:
    /**
     *  @brief Default constructor
     */
    CompositeStrategy();
    
    /**
     *  @brief Virtual destructor
     */
    virtual ~CompositeStrategy();
    
    /**
     *  @brief Schedule destruction
     */
    virtual void scheduleDestruction();
    //@}
    
    /**
     *  @name Child Strategies
     */
    //@{
    /**
     *  @brief Set buffer child strategy type
     *  @name strategy_type Type of transport strategy, must be the name of a registered TransportStrategy class
     *  @exception BindingException if the transport strategy can not be created
     *
     *  This method creates a transport strategy that will be used as the buffer child strategy of this
     *  composite transport strategy. The child strategy is of type @c strategy_type.
     */
    void setBufferStrategyType(const string& strategy_type) throw (const BindingException&);
    
    /**
     *  @brief Set event child strategy type
     *  @name strategy_type Type of transport strategy, must be the name of a registered TransportStrategy class
     *  @exception BindingException if the transport strategy can not be created
     *
     *  This method creates a transport strategy that will be used as the event child strategy of this
     *  composite transport strategy. The child strategy is of type @c strategy_type.
     */
    void setEventStrategyType(const string& strategy_type) throw (const BindingException&);
    
    /**
     *  @brief Get buffer child strategy
     *  @return The buffer child strategy of this composite strategy
     *
     *  If no buffer child strategy exists, this method returns NULL.
     */
    TransportStrategy* getBufferStrategy() const;
    
    /**
     *  @brief Get event child strategy
     *  @return The event child strategy of this composite strategy
     *
     *  If no event child strategy exists, this method returns NULL.
     */
    TransportStrategy* getEventStrategy() const;
    
    /**
     *  @brief Get buffer child strategy
     */
    virtual ITransportStrategy* getBufferChildStrategy() const;
    
    /**
     *  @brief Get event child strategy
     */
    virtual ITransportStrategy* getEventChildStrategy() const;
    
    /**
     *  @brief Set buffer child strategy
     *  @param strategy The transport strategy
     *
     *  This method sets the buffer child strategy of this composite strategy. This composite strategy
     *  becomes responsible for destroying the child strategy. If a buffer child strategy already exists,
     *  then it is destroyed.
     */
    void setBufferStrategy(TransportStrategy* strategy);
    
    /**
     *  @brief Set event child strategy
     *  @param strategy The transport strategy
     *
     *  This method sets the event child strategy of this composite strategy. This composite strategy
     *  becomes responsible for destroying the child strategy. If a event child strategy already exists,
     *  then it is destroyed.
     */
    void setEventStrategy(TransportStrategy* strategy);
    //@}
    
    /**
     *  @name Connect & Configure
     */
    //@{
    /**
     *  @brief Set address
     *  @param address Address to connect to, must be a valid CompositeAddress
     *  @exception BindingException if the connection setup fails
     *
     *  This method sets the destination address and connects the transport strategy.
     *  The @c address must be a valid CompositeAddress.
     *
     *  Connection setup is done in the following steps: Buffer, a partner strategy of class
     *  CompositeStrategy is created on the remote host. This is done using the ConnectionManager
     *  and the master address of @c address. The partner strategy is connected to the partner
     *  object. Then the two child strategies are created and connected with the partner strategy.
     *  If any of these steps fail, then this method also fails.
     */
    virtual void setAddress(const TSAddress& address) throw (const BindingException&);
    
    /**
     *  @brief Get local address
     *  @return A dummy address
     */
    virtual const TSAddress& getLocalAddress() const throw();
    
    /**
     *  @brief Set ValueStream
     *
     *  This method does nothing because this strategy doesn't use value streams.
     */
    virtual void setValueStream(const IValueStream&, const OValueStream&) {}
    
    /**
     *  @brief Get strategy type
     *  @return "CompositeStrategy"
     *
     *  This method returns the type name of this transport strategy, which is "CompositeStrategy".
     */
    virtual string getStrategyType() const { return getStaticStrategyType(); }
    
    /**
     *  @brief Configure for input jack
     *  @param format Stream format
     *
     *  This method configures the transport strategy for an input jack with the given @c format.
     *  The call is forwarded to both child strategies, which both must exist and be connected.
     */
    virtual void configureForInputJack(const Format& format);
    
    /**
     *  @brief Configure for output jack
     *  @param format Stream format
     *
     *  This method configures the transport strategy for an output jack with the given @c format.
     *  The call is forwarded to both child strategies, which both must exist and be connected.
     */
    virtual void configureForOutputJack(const Format& format);    
    //@}
    
    /**
     *  @name ICompositeStrategy interface
     */
    //@{
    /**
     *  @brief Set synchronization mode
     *
     *  Internal method for setting the synchronization mode. TEMPORARY!
     */
    virtual void setMode(const ICompositeStrategy::Mode& mode);

    /**
     *  @brief Set data loss interval
     *
     *  This method sets the interval for simulating data loss in unreliable mode.
     *  It is only meant for benchmarks and should not be used. An interval of 0 disables data loss simulation.
     */
    virtual void setLossInterval(const unsigned int& loss_interval);
    
    /**
     *  @brief Connect buffer child strategy
     *  @exception BindingException if this event was not sent through a TransportStrategy
     *
     *  This event is sent by the CompositeStrategy that initiates a connection to connect
     *  the partner strategy of its buffer child strategy to its partner strategy.
     *  The receiver assumes that the sender of the event is the partner strategy of the child strategy
     *  and sets its child strategy accordingly.
     */
    void connectToCompositeBuffer() throw(const BindingException&);
    
    /**
     *  @brief Connect event child strategy
     *  @exception BindingException if this event was not sent through a TransportStrategy
     *
     *  This event is sent by the CompositeStrategy that initiates a connection to connect
     *  the partner strategy of its buffer child strategy to its partner strategy.
     *  The receiver assumes that the sender of the event is the partner strategy of the child strategy
     *  and sets its child strategy accordingly.
     */
    void connectToCompositeEvent() throw(const BindingException&);
    
    /**
     *  @brief Notify the receiver that the last buffer before an event has been sent
     *  @param stream_counter Sequence number of event
     *  @param time Maximum time to wait for the buffer
     *
     *  This notification is sent from the sender to the receiver when the last buffer before an event has been sent.
     *  The receiver must wait at most @c time for the last buffer before the event. Then the buffer must be treated as lost.
     */
    void notifyEvent(const int& stream_counter, const Interval& time);
    //@}
    
    /**
     *  @name Sender & Receiver Interface
     */
    //@{
    /**
     *  @brief Receive instream message
     *
     *  This method forwards the message to the correct child strategy.
     */
    virtual Result receiveMessage(Message*, const IMessageSender* = 0);
    
    /**
     *  @brief Receive CEvent
     *
     *  This method forwards the event to the event child strategy.
     */
    virtual Result receiveEvent(const CEvent&, const IEventSender* = 0);
    
    /**
     *  @brief Receive event
     *
     *  This method forwards the event to the event child strategy.
     */
    virtual Result receiveEvent(const Event&, const IEventSender* = 0);
    
    /**
     *  @brief Receive CEvent
     *
     *  This method forwards the event to the event child strategy.
     */
    virtual Result receiveEvent(const CEvent&, const IEventSender* = 0) const;
    
    /**
     *  @brief Receive event
     *
     *  This method forwards the event to the event child strategy.
     */
    virtual Result receiveEvent(const Event&, const IEventSender* = 0) const;
    
    /**
     *  @brief Send CEvent
     *
     *  This method forwards the event to the event child strategy.
     */
    virtual Result sendEvent(const CEvent&);
    
    /**
     *  @brief Send event
     *
     *  This method forwards the event to the event child strategy.
     */
    virtual Result sendEvent(const Event&);
    
    /**
     *  @brief Send CEvent
     *
     *  This method forwards the event to the event child strategy.
     */
    virtual Result sendEvent(const CEvent&) const;
    
    /**
     *  @brief Send event
     *
     *  This method forwards the event to the event child strategy.
     */
    virtual Result sendEvent(const Event&) const;
    
    /**
     *  @brief Send instream message
     *
     *  This method forwards the message to the correct child strategy.
     */
    virtual Result sendMessage(Message*);
    
    /**
     *  @brief Send reliable event
     *
     *  Send a reliable event to the partner object using the ConnectionManager
     */
    virtual Result sendReliableEvent(Event&) const;
    
    /**
     *  @brief Receive reliable event
     *
     *  Send a reliable event to the partner object using the ConnectionManager.
     *  Handle an incoming reliable event.
     */
    virtual Result receiveReliableEvent(Event&, const IEventSender*) const;
    
    /**
     *  @brief Receive signal
     *
     *  This is the signal handler for the owner of this transport strategy.
     *  If the signal is @c FREE_QUEUE_SLOT, then this method unblocks the message handler.
     */
    virtual void receiveSignal(Signal);
    
    /**
     *  @brief Receive signal from child
     *  @param signal The signal which was received
     *  @param child The child which sent the signal
     *
     *  This method handles signals from child strategies.
     *
     *  If the signal is @c FREE_DOWNSTR_QUEUE_SLOT or @c FREE_UPSTR_QUEUE_SLOT, then this method
     *  checks whether the signal should be forwarded to the signal receiver of this strategy.
     */
    virtual void receiveChildSignal(Signal signal, TransportStrategy* child);
    
    /**
     *  @brief Child received instream message
     *  @param message The instream message which was received
     *  @param source The sub binding where the message was received from
     *
     *  This method handles instream messages received by child strategies.
     *  It is responsible for synchronizing the streams from the two child strategies.
     */
    virtual Result receiveChildMessage(Message* message, InstreamHandler& source);
    
    /**
     *  @brief Process messages
     *
     *  This method is triggered by the instream handlers on the receiver side.
     *  It processes messages until no more messages are available, messages are missing or
     *  the partner returned @c BLOCKED.
     */
    void run();
    //@}
    
    /**
     *  @name Static type information
     */
    //@{    
    /**
     *  @brief Get strategy type
     *  @return "CompositeStrategy"
     *
     *  This method returns the type name of strategy, which is "CompositeStrategy".
     */
    static string getStaticStrategyType();
    //@}
    
  protected:
    /**
     *  @brief Forward message to partner
     *  @param message The instream message which was received
     *  @param source The sub binding where the message was received from
     */
    virtual Result forwardMessage(Message* message, InstreamHandler& source);
    
    /**
     *  @brief Set send value stream
     *
     *  Does nothing.
     */
    virtual void setSendValueStream(const string&);
    
    /**
     *  @brief Set receive value stream
     *
     *  Does nothing.
     */
    virtual void setReceiveValueStream(const string&);    

  private:
    /**
     *  @brief Receive event ACK
     */
    void eventReceived(const int& stream_counter);
    
    /**
     *  @brief Send event ACK
     */
    void sendEventReceived(const int& stream_counter);
    
    /**
     *  @brief Wait for message timeout
     */
    void waitForTimeout();
    
    /**
     *  @brief Update receiver timestamp
     *  @param timestamp New receiver timestamp
     *  @param type Message type (only for debug output and statistics)
     *  @param lost Message was lost (only for debug output and statistics)
     */
    void setRecvTimestamp(const Timestamp& timestamp, Message::Type type, bool lost);
    
  private:
    ConnectorAddress* __remote_address;       ///< Remote address
    ConnectorAddress* __local_address;        ///< Local address
    ThreadMutex __send_mutex;                 ///< Mutex for message sending.
    ThreadMutex __recv_mutex;                 ///< Mutex for message sending.
    ThreadCondition __timeout_cond;           ///< Condition for waiting for timeout.
    ICompositeStrategy::Mode __mode;          ///< Mode of operation.
    unsigned int __loss_interval;             ///< Interval for data loss simulation.
    InstreamHandler __sub_binding[2];         ///< Sub bindings.
    int __send_seq_no;                        ///< Sequence number.
    
    /* BEGIN> Message multiplexer */
    bool __wait_event_ack;                    ///< Waiting for event ACK.
    bool __wait_for_timeout;                  ///< Waiting for timeout.
    int __timeout_stream_counter;             ///< Stream counter of buffer we are waiting for.
    Timestamp __recv_timestamp;               ///< Receiver timestamp.
    bool __blocked;                           ///< Partner is blocked.
    
    Time __last_buffer_recv_time;             ///< Time when last buffer was forwarded.
    Time __last_timeout;                      ///< Last timeout
    int  __last_timeout_stream_counter;       ///< Stream counter of last timeout
    /* END< Message multiplexer */
    
  private:
    /** "CompositeStrategy".
     */
    static const char __strategy_type[];
  };
  
  inline TransportStrategy* CompositeStrategy::getBufferStrategy() const {
    return &__sub_binding[Message::BUFFER].getTransportStrategy();
  }
  
  inline TransportStrategy* CompositeStrategy::getEventStrategy() const {
    return &__sub_binding[Message::EVENT].getTransportStrategy();
  }
  
  inline string CompositeStrategy::getStaticStrategyType() {
    return __strategy_type;
  }
  
} // namespace NMM

#endif // NMM_COMPOSITESTRATEGY_HPP
