/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "ConnectorAcceptor.hpp"

#include "TransportStrategy.hpp"
#include "TSFactory.hpp"
#include "ConnectionManager.hpp"
#include "ConnectorAddress.hpp"

#include "nmm/interfaces/base/IConnectionManager.hpp"

namespace NMM {
  
  ConnectorAcceptor::ConnectorAcceptor(const std::string& strategy_type)
    : __strategy_type(strategy_type)
  {}
  
  ConnectorAcceptor::~ConnectorAcceptor()
  {}
  
  void ConnectorAcceptor::handleConnect(const ConnectorAddress* remote_addr, ConnectorAddress*& local_addr)
    throw (const BindingException&)
  {
    TransportStrategy* strategy = NULL;
    
    try {
      GLOBAL_DEBUG_STREAM("ConnectorAcceptor: handle connection request" << std::endl);
      
      /* create a partner strategy */
      GLOBAL_DEBUG_STREAM("ConnectorAcceptor: create partner strategy" << std::endl);
      strategy = STSFactory::getInstance().createStrategy(getStrategyType());
      if(!strategy) {
        GLOBAL_ERROR_STREAM("ConnectorAcceptor: Failed to create partner TransportStrategy" << endl);
        throw BindingException("Failed to create partner TransportStrategy");
      }
      
      /* get local reference */
      NMMReference local_ref = strategy->getReference();
      
      /* set event receiver so that other side can call connectToObject */
      strategy->setEventReceiver(this);
      TSAddressOwner* address_owner = dynamic_cast<TSAddressOwner*>(strategy);
      if(!address_owner) {
        GLOBAL_ERROR_STREAM("ConnectorAcceptor: Strategy has no TSAddressOwner interface" << std::endl);
        throw BindingException("Strategy has no TSAddressOwner interface");
      }
      
      /* connect local strategy to remote address */
      GLOBAL_DEBUG_STREAM("ConnectorAcceptor: Set remote address" << std::endl);
      strategy->setAddress(*remote_addr);
      
      /* get address of local strategy */
      GLOBAL_DEBUG_STREAM("ConnectorAcceptor: get local address" << std::endl);
      const TSAddress& ts_address = address_owner->getLocalAddress();
      local_addr = dynamic_cast<ConnectorAddress*>(ts_address.copy());
      local_addr->setTSReference(local_ref);
      
      GLOBAL_DEBUG_STREAM("ConnectorAcceptor: connection accepted" << std::endl);
    } catch(...) {
      /* clean up */
      if(strategy) {
        strategy->scheduleDestruction();
      }
      
      /* rethrow exception */
      throw;
    }
  }
  
  void ConnectorAcceptor::connect(const ConnectorAddress* local_addr, ConnectorAddress*& remote_addr, IConnectionManager* remote_icm)
    throw(const BindingException&)
  {
    GLOBAL_DEBUG_STREAM("ConnectorAcceptor: request connection" << std::endl);
    
    /* we need a remote connection manager */
    if(!remote_icm) {
      GLOBAL_ERROR_STREAM("ConnectorAcceptor: No remote connection manager interface given" << endl);
      throw BindingException("No remote connection manager interface given");
    }
    
    /* forward request to remote connection manager */
    // XXX: need a dynamic cast here because we can't pass ConnectorAddress to the interface
    TSAddress* remote_address = NULL;
    remote_icm->handleConnect(local_addr, remote_address);
    remote_addr = dynamic_cast<ConnectorAddress*>(remote_address);
    
    GLOBAL_DEBUG_STREAM("ConnectorAcceptor: connection successful" << std::endl);
  }
  
}
