/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef _NMM_CONNECTOR_ACCEPTOR_HPP_
#define _NMM_CONNECTOR_ACCEPTOR_HPP_

#include "ObjectAcceptor.hpp"

namespace NMM {
  
  class IConnectionManager;
  class ConnectorAddress;
  
  /**
   *  @brief Virtual acceptor base class
   *
   *  A connector acceptor is an %ObjectAcceptor which is triggered by the %ConnectionManager in response to a
   *  conenction request. Its purpose it to simulate reliable connection setup for transport strategies that
   *  don't provide reliable connection setup of their own.
   *
   *  @see ConnectorAddress
   *  @see ConnectionManager
   */
  class ConnectorAcceptor :
    public ObjectAcceptor
  {
  public:
    /**
     *  @brief Construct a %ConnectorAcceptor
     *  @param strategy_type Type of transport strategy
     */
    ConnectorAcceptor(const std::string& strategy_type);
    
    /**
     *  @beief Virtual destructor
     */
    virtual ~ConnectorAcceptor();
    
    /** 
     *  @name Remote Interface
     */
    //@{
    /** 
     *  @brief Handle connection request
     *  @param remote_addr Fully specified address of remote transport strategy
     *  @param local_addr Returns the fully specified address of local transport strategy
     *  @exception BindingException if the connection request has failed
     *
     *  This method is called by the ConnectionManager when it has received a connection request.
     *  The implementation is expected to handle this connection request and create a suitable
     *  partner TransportStrategy. It should also return the local address of the partner strategy
     *  and its NMM reference.
     */
    virtual void handleConnect(const ConnectorAddress* remote_addr, ConnectorAddress*& local_addr)
      throw (const BindingException&);
    //@}
    
    /** @name Local Interface
     */
    //@{
    /**
     *  @brief Request a connection
     *  @param local_addr Fully specified address of the local transport strategy
     *  @param remote_addr Returns the fully specified address of the remote transport strategy
     *  @param remote_icm Connection manager interface on the remote side
     *  @exception BindingException if the connection could not be set up
     *
     *  This method is called by the ConnectionManager when a local transport strategy requests
     *  a connection to a remote host. The implementation is expected to forward the request to
     *  the remote host. This is typically done by calling handleConnect on the remote IConnectionManager
     *  interface, but different connection setup strategies (such as SIP) may do something else
     *  entirely.
     */
    virtual void connect(const ConnectorAddress* local_addr, ConnectorAddress*& remote_addr, IConnectionManager* remote_icm)
      throw(const BindingException&);
    
    /** 
     *  @brief Get strategy type
     *  @return The startegy type of this connector
     *
     *  This method returns the type of transport strategy that this connector can handle.
     */
    const std::string& getStrategyType() const;
    //@}
    
  private:
    std::string __strategy_type;          ///< Transport strategy type
  };
  
  
  inline const std::string& ConnectorAcceptor::getStrategyType() const {
    return __strategy_type;
  }
  
}

#endif
