/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *                    Marco Lohse        <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */



#include "Event.hpp"
#include "nmm/interfaces/base/INode.hpp"

namespace NMM {

  Event::Event(const string& k)
  : SerializedValue(type_name, "NMM"), key(k), __arg_return(1), __internal_return_vector(1)
  {
    assign();
    __arg_return[0] = 0;
    __internal_return_vector[0] = 0;
  }

  Event::Event(const string& k, unsigned  int arg_size)
  : SerializedValue(type_name, "NMM"), key(k), arg_array(arg_size), __arg_return(1), __internal_return_vector(1)
  {
    for (unsigned int i=0; i < arg_array.size(); ++i) {
      arg_array[i] = 0;
    }
    __arg_return[0] = 0;
    __internal_return_vector[0] = 0;

    assign();
  }

  Event::Event(const string& k, Value* value_0, Value* value_1, Value* value_2, Value* value_3, Value* value_4, Value* value_5, Value* value_6)
   : SerializedValue(type_name, "NMM"), key(k), __arg_return(1), __internal_return_vector(1)
  {
    assign();
    if (value_0) {
      arg_array.insert(arg_array.end(), value_0);
      if (value_1) {
        arg_array.insert(arg_array.end(), value_1);
        if (value_2) {
          arg_array.insert(arg_array.end(), value_2);
          if (value_3) {
            arg_array.insert(arg_array.end(), value_3);
            if (value_4) {
              arg_array.insert(arg_array.end(), value_4);
              if (value_5) {
                arg_array.insert(arg_array.end(), value_5);
                if (value_6) {
                  arg_array.insert(arg_array.end(), value_6);
                }
              }
            }
          }
        }
      }
    }
    __arg_return[0] = 0;
    __internal_return_vector[0] = 0;
  }

  void Event::assign() {
    insertArgument(&key);
    insertArgument(&arg_array);
    insertArgument(&__arg_return);
    insertArgument(&__exception);
  }

  Event::Event(const Event& e) : SerializedValue(e), __arg_return(1), __internal_return_vector(1)
  {
    *this           = e;
    assign();
    __arg_return[0] = 0;
    __internal_return_vector[0] = 0;
  }

  Event::~Event(){
    erase();
  }

  void Event::erase() {
    for (unsigned int i = 0; i< arg_array.size(); ++i)
      if (arg_array[i])
        arg_array[i]->releaseValue();
    arg_array.clear();
    __arg_return[0] = 0;
    delete getException();
    __exception.setValue(0);
  }

  void Event::setValue(unsigned int position, Value* value) {
    if (position >= arg_array.size()) {
      GLOBAL_ERROR_STREAM2("Value could not be stored. Inavlid position");
      return;
    }

    if (value == arg_array[position])
      return;

    if (arg_array[position])
      arg_array[position]->releaseValue();
    arg_array[position] = value;
  }

  const Event& Event::operator=(const Event& e) {
    if (this == &e)
      return *this;

    key = e.getKey();

    erase();

    for (unsigned int i = 0; i< e.getArgSize(); ++i)
      if (e.getValue(i))
        arg_array.insert(arg_array.end(), e.getValue(i)->copy());

    if (e.getException()) {
      __exception.setValue(e.getException()->copy());
    }

    return *this;
  }

  const Value*  Event::getValue(unsigned int position)   const {
    if (position < arg_array.size())
      return arg_array[position];
    return 0;
  }

  Value* Event::getValue(unsigned int index) {
    if (index < arg_array.size())
      return arg_array[index];

    return 0;
  }

  void Event::disableMemoryType(Value::MemoryMgm type) {
    if (arg_array.size() > 0) {
      /* resize the internal_arg_vector to the size of the value*/
      __internal_arg_vector.resize(arg_array.size());

      /* Begin > Compare each Value's memory type with the given type */
      for (unsigned int i = 0; i < arg_array.size(); ++i) {
        /* Begin> If the memory types are the samen, store the Value in the internal-array and set
         *        the arg-array to NULL */
        if (arg_array[i]->getMemoryStrategy() == type) {
          __internal_arg_vector[i] = arg_array[i];
          arg_array[i] = 0;
        } else {
          /* Set the position in teh internal-array to NULL */
          __internal_arg_vector[i] = 0;
        }
      }
      /* End< */
    }

    /* Begin> handle the return values if OUTVAULES are disabled. */
    if (type == Value::OUTVALUE) {
      __internal_return_vector.resize(1);
      __internal_return_vector[0] = __arg_return[0];
      __arg_return[0] = 0;
    }
    /* End< */
  }

  void Event::resetMemoryType() {
    /* Begin> reset the old configuration and copy the pointer from the internal-arg vector to the arg-array */
    for (unsigned int i = 0; i < __internal_arg_vector.size(); ++i)
      if ( __internal_arg_vector[i] ) {
        arg_array[i] = __internal_arg_vector[i];
        __internal_arg_vector[i] = 0;
      }
    /* End< */
    if (__internal_return_vector.size() == 1)
      if (__internal_return_vector[0] != 0)
        __arg_return[0] = __internal_return_vector[0];
  }


  void  Event::setException(const Exception& exception) {
    delete __exception.getRValue();           //delete old exception
    __exception.setValue(exception.copy());   //store a copy
  }

  const string Event::type_name = "Event";


  CEvent::CEvent(bool b)
    : Message(Message::EVENT), b_mgr(b)
  {
    setMandatoryBit(true); // event should not be discarded during flush
    event_array.useTValue(false);
    insertArgument(&event_array);
    setTypeName(type_name);
    setNamespace("NMM");
    increaseCount();
  }


  CEvent::CEvent(unsigned int size, bool b)
    : Message(Message::EVENT), event_array(size), b_mgr(b)
  {
    setMandatoryBit(true); // event should not be discarded during flush
    event_array.useTValue(false);
    insertArgument(&event_array);
    for (unsigned int i = 0; i< size; ++i)
      event_array[i] = 0;

    setTypeName(type_name);
    setNamespace("NMM");

    increaseCount();
  }

  CEvent::CEvent(const Event& event, bool b)
    : Message(Message::EVENT), b_mgr(b)
  {
    setMandatoryBit(true); // event should not be discarded during flush
    event_array.useTValue(false);
    insertArgument(&event_array);

    setTypeName(type_name);
    setNamespace("NMM");

    event_array.insert(event_array.end(), new Event(event));
    increaseCount();
  }

  CEvent::CEvent(Event* event, bool b)
    : Message(Message::EVENT), b_mgr(b)
  {
    setMandatoryBit(true); // event should not be discarded during flush
    event_array.useTValue(false);
    insertArgument(&event_array);

    setTypeName(type_name);
    setNamespace("NMM");

    event_array.insert(event_array.end(), event);
    increaseCount();
  }

  CEvent::CEvent(const Event& e1, const Event& e2, bool b)
    : Message(Message::EVENT), b_mgr(b)
  {
    setMandatoryBit(true); // event should not be discarded during flush
    event_array.useTValue(false);
    insertArgument(&event_array);

    setTypeName(type_name);
    setNamespace("NMM");

    event_array.insert(event_array.end(), new Event(e1));
    event_array.insert(event_array.end(), new Event(e2));
    increaseCount();
  }

  void CEvent::erase() const {
    if (b_mgr)
      delete this;
  }

  CEvent::CEvent(const CEvent& ce) 
    : Message(ce), event_array(ce.size()) 
  {
    setMandatoryBit(true); // event should not be discarded during flush
    event_array.useTValue(false);
    insertArgument(&event_array);

    for (unsigned int i = 0; i< event_array.size(); ++i )
      event_array[i] = new Event(ce[i]);

    setTypeName(type_name);
    setNamespace("NMM");
    setDirection(ce.getDirection());

    increaseCount();
  }

  CEvent::~CEvent() {
   for (unsigned int i = 0; i< event_array.size(); ++i)
     delete event_array[i];
  }

  Event& CEvent::operator[](unsigned int i) {
    return *event_array[i];
  }

  const Event& CEvent::operator[](unsigned int i) const {
    return *event_array[i];
  }

  void CEvent::insertEvent(unsigned int p, const Event& event) {
    if (p >= event_array.size()) {
      event_array.insert(event_array.end(), new Event(event));
      return;
    }

    vector<Event*>::iterator it = event_array.begin();
    advance(it, p);
    event_array.insert(it, new Event(event));
  }

  void CEvent::insertEvent(unsigned int p, Event* event) {
    if (p >= event_array.size()) {
      event_array.insert(event_array.end(), event);
      return;
    }

    vector<Event*>::iterator it = event_array.begin();
    advance(it, p);
    event_array.insert(it, event);
  }

  void CEvent::removeEvent(unsigned int p) {
    if (p >= event_array.size()) {
      GLOBAL_ERROR_STREAM2("CEvent > Invalid position. Could not remove Event" << endl);
      return;
    }
    vector<Event*>::iterator it = event_array.begin();
    for (unsigned int i = 0; i < p; ++i)
      ++it;
    delete *it;
    event_array.erase(it);
  }


  CEvent& CEvent::operator=(const CEvent& ce) {
    Message::operator=(ce); //call the assignmenet operator from base class

    for (unsigned int i = 0; i< event_array.size(); ++i) {
      delete event_array[i];
    }
    event_array.resize(ce.event_array.size());

    for (unsigned int i = 0; i< ce.event_array.size(); ++i) {
      if (ce.event_array[i])
        event_array[i] = ce.event_array[i]->copy();
      else
        event_array[i] = 0;
    }

    return *this;
  }
  const string CEvent::type_name = "CEvent";

  TConstructor<Event>                     __nmm__event_value    ("Event", "NMM");
  TConstructor<CEvent>                    __nmm__cevent_value   ("CEvent", "NMM");

  /**
   * Special instream event for flushing nodes.
   */
  FlushEvent::FlushEvent() : CEvent(Event(INode::flush_event)) {
    setType(Message::FLUSH_EVENT);
    setMandatoryBit(true); // event should not be discarded during flush
  }

  FlushEvent::~FlushEvent() {
  }

} //end namespace
