/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *                    Marco Lohse        <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */



#ifndef NMM_EVENT_HPP
#define NMM_EVENT_HPP

#include "Message.hpp"

namespace NMM {
  class Value;
  class CEvent;

  /**
  */
  class Event : public SerializedValue {
  public:
    /** Default-Constructor
     */
    Event(const string& key = "");

    /** Creates an Event with arg_size Values
     */
    Event(const string& key, unsigned int arg_size);

    /** Creates an Event with three Values
     */
    Event(const string& key, Value*, Value* = 0, Value* = 0, Value* = 0, Value* = 0, Value* = 0, Value* = 0);

    /** Copy constructor
     */
    Event(const Event&);

    /** destructor
     */
    virtual ~Event();

    /** assignement operator. The values stored inside the event desides if a deep copy is done
     *  or not.
     */
    const Event&  operator=(const Event& e);

    /** set a Value if position > arg_size, the value is not set.
     */
    void setValue(unsigned int position, Value* value);

    /** Assign a new key to this Event
     */
    inline void setKey(const string& k)                   { key.setValue(k); }

    /** returns the event-key
     */
    inline const string& getKey()                   const { return key.getRValue(); }

    /** returns the value-array's size
     */
    inline unsigned int getArgSize()                const { return arg_array.size(); }

    /** returns the i'th value. NULL if this value is not set or i > arg_size
     */
    const Value*  getValue(unsigned int position)   const ;

    /** returns the i'th value. NULL if this value is not set or i > arg_size
     */
          Value*  getValue(unsigned int position) ;

    /** This method sets the returns value. The meory management is done by the event.
      */
    inline  void setReturnValue(Value* value)             { __arg_return[0] = value; }

    /** This method returns the return value of an executed method. The memoty management
      * of the returned value is done by the event. If the caller needs a copy of the value
      * he can use the copy() method.
      */
    inline       Value* getReturnValue()                  { return __arg_return[0]; }

    /** This method returns the return value of an executed method. The memoty management
      * of the returned value is done by the event. If the caller needs a copy of the value
      * he can use the copy() method.
      */
    inline const Value* getReturnValue()            const { return __arg_return[0]; }

    /** This method returns a copy of this event.
      */
    virtual Event* copy()                           const { return new Event(*this);}

    /** This method resizes the internal value array to store more values.
      */
    inline void resize(unsigned int size)                 { arg_array.resize(size); }

    /** clears all values
      */
    void erase();

    /** This method returns an excetption which was thrown during the execution of the event.
      */
    inline       Exception* getException()       { return __exception.getValue(); }

    /** This method returns an excetption which was thrown during the execution of the event.
      */
    inline const Exception* getException() const { return __exception.getValue(); }

    /** This method sets the exception which threw the executed method.
      */
           void  setException(const Exception& exception);

    /** This method can be used to disabel certain types. This mehtod is interest for complex
      * data structures to disable certain Memory types like In or Out-Values for (de-)serialization
      */
    virtual void disableMemoryType(Value::MemoryMgm type);

    /** resets the internal datastructure wich was disabled before.
      */
    virtual void resetMemoryType();

    static const string type_name;                      ///< stores the typename of the event
  private:
    void assign();
    TValue<string> key;                                 ///< stores the event key                
    SequenceValue<> arg_array;                          ///< stores the arguments of the event

    SequenceValue<>       __arg_return;                 ///< stores the return value
    TOutValue<Exception*> __exception;                  ///< stores the return value
    vector<Value*>        __internal_arg_vector;        ///< stores the disabled argments of the event
    vector<Value*>        __internal_return_vector;     ///< stores the disabled return value.
  };

  /** This class can be used to send Events to a Node. All Events are executed
   * atomically and in the specified order. The event size is dynamic.
   */
  class CEvent : public Message {
  public:
    /** Constructor
     */
    explicit CEvent(bool b_mgr = true);

    /** Constructor
     */
    explicit CEvent(unsigned int i, bool b_mgr = true);

    /** Constructor
     */
    explicit CEvent(const Event&, bool b_mgr = true);

    /** Constructor
     */
    explicit CEvent(Event*, bool b_mgr = true);

    /** Creates a CEvent with two Events
     */
    explicit CEvent(const Event&, const Event&, bool b_mgr = true);

    /** Copy constructor
      */
    explicit CEvent(const CEvent& ce);

    /** Destructor
     */
    virtual ~CEvent();

    /** returns the count of stored events
     */
    unsigned int size()                         const {return event_array.size();}

    /** This method can be used to access the Events
     */
    Event& operator[](unsigned int i);

    /** This method can be used to access the Events
     */
    const Event& operator[](unsigned int i) const;

    /** returns the i'th element.
     */
    Event& getEvent(unsigned int i)               {return *event_array[i];}

    /** assign a new Event
     */
    void setEvent(unsigned int i, const Event& e) {  delete event_array[i]; event_array[i] = new Event(e);}

    /** assign a new Event
     */
    void setEvent(unsigned int i, Event* e) {  delete event_array[i]; event_array[i] = e;}

    /** Inserts a new Event e at position p.
      * \param p : Position of Event e.(from 0 to size-1). If p > size-1, the
      *            Event is inserted at the end.
      * \param e : The new Event.
      */
    void insertEvent(unsigned int p, const Event& e);

    /** Inserts a new Event e at the end.
      * \param e : The new Event.
      */
    inline void insertEvent(const Event& e) { event_array.insert(event_array.end(), new Event(e)); };

    /** Inserts a new Event e at the end.
      * \param e : The new Event.
      */
    inline void insertEvent(Event* e) { event_array.insert(event_array.end(), e); };

    /** Inserts a new Event e at position p.
      * \param p : Position of Event e.(from 0 to size-1). If p > size-1, the
      *            Event is inserted at the end.
      * \param e : The new Event.
      */
    void insertEvent(unsigned int p, Event* e);

    /** Removes the i'th Event.
      * \param i : Postion of Event to delete (from 0 to size-1). If i > size-1
      *            this method returns immediatly.
      */
    void removeEvent(unsigned int i);

    CEvent& operator=(const CEvent& ce);
    virtual CEvent* copy() const  {return new CEvent(*this);}
    static const string type_name;
  protected:
    /** erase implementation for pure virtual method in superclass
     */
    virtual void erase() const;

  private:
    SequenceValue<Event> event_array;
    bool b_mgr;
  };

  /**
   * Special instream event for flushing nodes.
   */
  class FlushEvent : public CEvent {

  public:
    FlushEvent();
    ~FlushEvent();
  };

}
#endif
