/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, 
 * USA.
 */

#include "GenericProducer.hpp"

#include "nmm/base/IReceiver.hpp"

#include <unistd.h>
#include <sys/poll.h>

namespace NMM {

  GenericProducer::GenericProducer()
    : receiver(NULL),
      mutex(ThreadMutex::RECURSIVE)
  {
  }
  
  GenericProducer::~GenericProducer() {}
  
  void GenericProducer::connectTo(IEventReceiver* receiver) {
    MutexGuard guard(mutex);
    this->receiver = receiver;
  }
  
  void GenericProducer::disconnect() {
    MutexGuard guard(mutex);
    receiver = NULL;
  }
  
  IEventReceiver* GenericProducer::getReceiver() const {
    MutexGuard guard(mutex);
    return receiver;
  }
  
  Result GenericProducer::sendEvent(const Event& event) {
    MutexGuard guard(mutex);
    if (receiver) {
      return receiver->receiveEvent(event, this);
    } else {
      return FAILURE;
    }
  }
  
  Result GenericProducer::sendEvent(const Event& event) const {
    MutexGuard guard(mutex);
    if (receiver) {
      return receiver->receiveEvent(event, this);
    } else {
      return FAILURE;
    }
  }
  
  Result GenericProducer::sendEvent(const CEvent& cevent) {
    MutexGuard guard(mutex);
    if (receiver) {
      return receiver->receiveEvent(cevent, this);
    } else {
      return FAILURE;
    }
  }
  
  Result GenericProducer::sendEvent(const CEvent& cevent) const {
    MutexGuard guard(mutex);
    if (receiver) {
      return receiver->receiveEvent(cevent, this);
    } else {
      return FAILURE;
    }
  }
  
  
  InputProducer::InputProducer()
    : fd(-1),
      timeout(-1)
  {
    pipefd[0] = -1;
    pipefd[1] = -1;
  }
  
  InputProducer::~InputProducer() {
    /* wake up poll and deactivate thread */
    InputProducer::stop();
  }
  
  void InputProducer::start() {
    GLOBAL_DEBUG_STREAM("InputProducer: start" << endl);
    
    /* create internal pipe for interrupting poll() */
    if ( pipe(&pipefd[0]) != 0 ) {
      GLOBAL_ERROR_STREAM("Could not create the internal pipe" << endl);
      throw Exception("Could not create the internal pipe");
    }
    
    /* start the thread */
    activateThread();
  }
  
  void InputProducer::stop() {
    GLOBAL_DEBUG_STREAM("InputProducer: stop" << endl);
    
    /* write to internal pipe to interrupt poll() */
    if (pipefd[1] >= 0) {
      write(pipefd[1], &pipefd[1], 1);
    }
    
    /* stop the thread */
    deactivateThread();
    
    /* close internal pipe */
    if (pipefd[1] >= 0) {
      close(pipefd[0]);
      pipefd[0] = -1;
      close(pipefd[1]);
      pipefd[1] = -1;
    }
  }
  
  void InputProducer::setTimeout(int timeout) {
    this->timeout = timeout;
  }
  
  void InputProducer::setFileDescriptor(int fd) {
    this->fd = fd;
  }
  
  void InputProducer::run() {
    
    /* setup array of poll FDs (fd and internal pipe fd) */
    pollfd pollfds[] = {
      { fd       , POLLIN, 0 },
      { pipefd[0], POLLIN, 0 }
    };
    
    int retval = 0;
    while(isRunning()) {
      
      /* call poll on fd */
      retval = poll(pollfds, 2, timeout);
      
      /* check for error */
      if (retval < 0) {
        GLOBAL_ERROR_STREAM("InputProducer: Something went wrong during the poll() call" << endl);
      }
      
      /* if change on a fd call handler */
      if (retval > 0 ) {
        if (pollfds[0].revents & POLLIN) {
          handleInput();
        }
        
        if (pollfds[1].revents & POLLIN) {
          char dummy;
          read(pipefd[0], &dummy, 1);
          break;
        }
      }
      
    }
  }
}
