/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, 
 * USA.
 */

#ifndef _NMM_GENERICPRODUCER_HPP_
#define _NMM_GENERICPRODUCER_HPP_ _NMM_GENERICPRODUCER_HPP_

#include "nmm/utils/thread/ThreadedObject.hpp"
#include "nmm/base/ISender.hpp"

namespace NMM {

  class IEventReceiver;
  
  
  /**
   * Generic Producer.
   * An IEventSender object that sends events to an IEventReceiver while it is started.
   */
  class GenericProducer :
    public IEventSender 
  {
  private:
    IEventReceiver* receiver;
    ThreadMutex mutex;
    
  public:
    GenericProducer();
    virtual ~GenericProducer();
    
    /** Connects this object to a receiver */
    void connectTo(IEventReceiver*);
    
    /** Disconnect this object from its receiver */
    void disconnect();
    
    /** Get event receiver this object is connected to */
    IEventReceiver* getReceiver() const;
    
    /** Starts the event-handling */
    virtual void start() = 0;
    
    /** Stops the event-handling */
    virtual void stop() = 0;
    
    
    /** Forward event to event receiver */
    virtual Result sendEvent(const CEvent& ce);
    
    /** Forward event to event receiver */
    virtual Result sendEvent(const Event& e);
    
    /** Forward event to event receiver */
    virtual Result sendEvent(const CEvent& ce) const;
    
    /** Forward event to event receiver */
    virtual Result sendEvent(const Event& e) const;
  };
  
  
  /**
   * Input producer.
   * A producer that polls a file descriptor and sends events when input is available.
   */
  class InputProducer :
    public GenericProducer,
    private ThreadedObject
  {
  private:
    int fd;
    int timeout;
    
    int pipefd[2];
    
  protected:
    /** Handler called when fd is readable */
    virtual void handleInput() = 0;
    
    /** Set the file descriptor to poll */
    void setFileDescriptor(int fd);
    
  public:
    InputProducer();
    virtual ~InputProducer();
    
    /**
     * Set timeout for poll
     * @param timeout timeout in milliseconds
     */
    void setTimeout(int timeout);
    
    /** Starts the event-handling in a new thread */
    void start();
    
    /** Stops the event-handling in a new thread */
    void stop();
    
  private:
    void run();
  };
  
}

#endif /* _NMM_GENERICPRODUCER_HPP_ */
