/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "OutOfBandChannel.hpp"

#include "nmm/base/TSFactory.hpp"
#include "nmm/base/ProxyObject.hpp"
#include "nmm/base/proxy/tcp/TCPStrategy.hpp"
#include "nmm/base/serialize/net/BufferedNetStream.hpp"

namespace NMM {
  
  OutOfBandChannel::OutOfBandChannel(NMMObject& parent)
    : CommunicationChannel(getStaticTypeInfo()), __parent(parent),
      __strategy(0), __i_strategies_valid(false), __default_address(0)
  {

  }

  void OutOfBandChannel::setTransportStrategy(TransportStrategy* strategy) {
    if (__strategy) {
      __strategy->setBindingChannel(0);
    }

    __strategy = strategy;
    if (__strategy) {
      /* Begin> Configure the TransportStrategy to the NMMObject or ProxyObject as receiver for incoming messages */
      __strategy->setBindingChannel(this);
      __strategy->setEventReceiver(&__parent);
      __strategy->setMessageReceiver(dynamic_cast<IMessageReceiver*>(&__parent));
      /* End < */

      /* Set the default address */
      for (list<TSAddress*>::iterator it = __address_list.begin(); it != __address_list.end(); ++it)
        if ((*it)->getStrategyType() == strategy->getStrategyType())
          setDefaultAddress(**it);

      /* Begin> HACK!! The configuration of the BufferedNetStream should if the NetworkStack is configured.
       * Unfortunately this is not implemented at the moment.
       */
      TCPStrategy* tcp_strategy = dynamic_cast<TCPStrategy*>(strategy);
      if (tcp_strategy) {
         BufferedNetStream* stream = dynamic_cast<BufferedNetStream*>(&tcp_strategy->getNetStream());
         if (stream) {
           stream->setSendBuffer(2048);
         }
      }
      /* End < Hack */
    } else {
      /* Clear the interfaces */
      getInterfaceList().clear();

      /* Clear the partner interfaces */
      getPartnerInterfaceList().clear();
    }
  }



  OutOfBandChannel::~OutOfBandChannel() {
    for (list<TSAddress*>::iterator it = __address_list.begin(); it != __address_list.end(); ++it)
      delete *it;

    delete __default_address;
  }

  OutOfBandChannel::OutOfBandChannel(const OutOfBandChannel& channel)
    : CommunicationChannel(channel), __parent(channel.__parent)
  {

  }

  void OutOfBandChannel::setTransportStrategy(const string& strategy_type) {
    /* First check whether the required transport strategy can be used */
    if ( find( getStrategyList().begin(), getStrategyList().end(), strategy_type ) == getStrategyList().end()
         && strategy_type != "NULL")
      if ( ! __default_address )
        throw BindingException("The TransportStrategy " + strategy_type + " can not be used to connect this ProxyObject with its NMMObject" );
      else
        if ( __default_address->getStrategyType() != strategy_type )
          throw BindingException("The TransportStrategy " + strategy_type + " can not be used to connect this ProxyObject with its NMMObject" );

    setTransportStrategy(STSFactory::getInstance().createStrategy(strategy_type));
  }

  void OutOfBandChannel::deinit() {
    getInterfaceList().clear();
    getPartnerInterfaceList().clear();
  }

  void  OutOfBandChannel::addAddress(const TSAddress& addr) {
    __address_list.push_back(addr.copy());
    getStrategyList().push_back(addr.getStrategyType());
  }

  const TSAddress*   OutOfBandChannel::getAddress(const string& strategy)  const {
    for (list<TSAddress*>::const_iterator it = __address_list.begin(); it != __address_list.end(); ++it)
      if ((*it)->getStrategyType() == strategy)
        return *it;

    return 0;
  }


  void OutOfBandChannel::setDefaultAddress(const TSAddress& address) {
    delete __default_address;
    __default_address = address.copy();
  }

  void OutOfBandChannel::setValueStream(const IValueStream& iv, const OValueStream& ov) {
    if (!__strategy ) {
      GLOBAL_WARNING_STREAM("Could not set ValueStreams. No TransportStrategy set" << endl);
      return;
    }

    if (!__i_strategies_valid)
      updateIStrategies();

    if (! __i_strategy.get() )
      throw Exception("Could not get PartnerInterface");

    __i_strategy->setSendValueStream(iv.getIValueStreamType());
    __i_partner_strategy->setSendValueStream(SValueFactory::getInstance().getOValueStreamPartner( ov.getOValueStreamType() ));
  }

  void OutOfBandChannel::updateIStrategies() {
    if (!__strategy)
      return;

    if (!__i_strategy.get() )
      __i_strategy.reset(__strategy->getCheckedInterface<ITransportStrategy>());

    if (! __i_partner_strategy.get() )
      __i_partner_strategy.reset( __strategy->getPartnerStrategy() );

    __i_strategies_valid = true;
  }

  list<Interface*>& OutOfBandChannel::getInterfaceList() {
    if (!__strategy)
      return CommunicationChannel::getInterfaceList();
    return __strategy->getInterfaceList();
  }

  list<Interface*>& OutOfBandChannel::getPartnerInterfaceList() {
    if (!__i_strategies_valid)
      updateIStrategies();

    if (!__i_partner_strategy.get())
      return CommunicationChannel::getInterfaceList();

    return __i_partner_strategy->getParentObject()->getInterfaceList();
  }

  void OutOfBandChannel::updateStrategyList(const Interface& interface) {
    getStrategyList().clear();

    /*Begin> specify, wich binding type we can use. Use the method from base class
     *       to check the object position
     */
    TSAddress::BindingType binding_type;
    if (getParent().isLocalTo(interface)) {
      binding_type = TSAddress::LOCAL;
    }

    if (getParent().isSystemTo(interface)){
      binding_type = TSAddress::SYSTEM;
    }

    if (getParent().isNetworkTo(interface)){
      binding_type = TSAddress::NETWORK;
    }
    /* End< */

    /* Begin> Set the TransportStrategies we can use */
    switch ( binding_type ) {
    case TSAddress::LOCAL :
      /* Begin> Set the TransportStrategies we can use */
      for (unsigned int i = 0;i< interface.getAddressList().size(); ++i)
        if (   interface.getAddressList()[i]->isBindingType(TSAddress::LOCAL)
               && STSFactory::getInstance().isStrategyAvailable(interface.getAddressList()[i]->getStrategyType()))
          {
            addAddress(*interface.getAddressList()[i]);
          }
      /* End< */
      /* Don't call break because System and Network strategies can be used too */

    case TSAddress::SYSTEM :
      /* Begin> Set the TransportStrategies we can use */
      for (unsigned int i = 0;i< interface.getAddressList().size(); ++i)
        if (   interface.getAddressList()[i]->isBindingType(TSAddress::SYSTEM)
               && STSFactory::getInstance().isStrategyAvailable(interface.getAddressList()[i]->getStrategyType()))
          {
            addAddress(*interface.getAddressList()[i]);
          }
      /* End< */
      /* Don't call break because network strategies can be used too */
      
    case TSAddress::NETWORK :
      /* Begin> Set the TransportStrategies we can use */
      for (unsigned int i = 0;i< interface.getAddressList().size(); ++i)
        if (   interface.getAddressList()[i]->isBindingType(TSAddress::NETWORK)
               && STSFactory::getInstance().isStrategyAvailable(interface.getAddressList()[i]->getStrategyType()))
          {
            addAddress(*interface.getAddressList()[i]);
          }
      /* End< */
    }
    /* End< */
  
    /* Strategy list is complete */
    finishStrategyList(binding_type);
  }
  
  TypeInfo OutOfBandChannel::getStaticTypeInfo() {
    return TypeInfo("OutOfBandChannel", "NMM");
  }
  
}
