/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        NMM work group,
 *                    http://www.networkmultimedia.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef _NMM_PROPERTY_HPP
#define _NMM_PROPERTY_HPP

#include "definition.hpp"
//#include "nmm/base/format/Format.hpp"
#include <memory>

namespace NMM {

class IValueStream;
class OValueStream;

    /************************************************************************
     This class represents the property of one NMM-Node.
     Depending on the node-type, a node has the following propertys:
     SOURCE        : Output-Property
     SINK          : Input-Property
     FILTER        : Input- / Output-Property
     CONVERTER     : Input- / Output-Property
     MULTIPLEXER   : Input- / Output-Property
     DEMULTIPLEXER : Input- / Output-Property
     The propertys consists of the formats, which on their part consists
     of several parameters.
  *************************************************************************/

class Property {
    
private:
    
  //! The type of the property
  std::string         m_property_type;

  //! STL-List of all the supported formats.
  std::list<Format *> m_format_list;
  
  //! The default format of this property.
//   Format *  m_default_format;
  //! The default formats of this property.
  std::map<const std::string, Format *>  m_default_formats;

private:
public:

  //! The default constructor
  Property();

  //! The default destructor
   ~Property();

  Property(const Property& );

  const Property& operator=(const Property& );
  /**
     This method creates a new format and adds it to the m_format_list.
     For every format you have to specifie the type/subtype
     \param string format_type: the type/subtype of the new format.
     \return Format *: pointer to the new format.
  **/
  Format * addNewFormat(const std::string & format_type);
  
  /**
     Add a new Format to the list of formats. The new format is already complete.
     \param Format * new_format: the new format that will be added to the list of formats.
  **/
  void addNewFormat(Format * new_format);

  /**
     This method deletes the format specified by the argument pFormat.
     \param Format *: the format which will be deleted.
     \return true in success otherwise false.
  **/
  void deleteFormat(Format *pFormat);
  
  /**
     This function returns a list of Formats in success, specified by the desired Type/subType.
     If no format is machting, the returned format_list will be empty.
     \param string pFormatType: the Type/SubType of the desired Format.
     \return FormatList: all the formats matching the desired Type/SubType.
  **/
  format_list getFormat(const std::string & pFormatType) const;
  
  /**
     This functions returns a list of formats in success, matching several defined parameters,
     specified by the desired argument of function 'pFormat'.
     By filling out the parameter-list of the 'pFormat', you can specify the desired parameters.
     \param pFormat *Format: the template format.
     \return format_list: all formats that fulfil the requirements of the template.
  **/
  format_list getFormat(const Format *pFormat) const;
  
  /**
     This function returns the complete list of formats of the Property.
     \return FormatList: list of all formats.
  **/
  const format_list & getFormatList() const;

  /**
     Returns the desired IOPartner of the specified Format.
     If the Format has no IOPartner (it's a sink or source), the function returns NULL.
     \param Format *: the format which IOPartner is given back.
     \return Format *: the appropriate IOPartner for the format.
  **/
  const Format * getIOPartner(const Format * pFormat) const;

  /**
     Show all the Parameters of the list.
     Just a debugging function.
  **/
  void showFormats() const;

  /**
     Set the default Format for this Device.
     \param Format*: the default format.
     \param const string: the jack - specified by its jack tag, whose default format should be set.
  **/
  void setDefaultFormat(Format *pFormat, const std::string jack_tag = "default");

  /**
     Get the default format of the device.
     \param const string: the jack - specified by its jack tag, whose default format should be returned.
     \returns The default format (of the specified jack) if it is set or NULL otherwise.
  **/
  const Format * getDefaultFormat(const std::string jack_tag = "default") const;

  /** deletes all Formats stored in the property
    */
  void erase();

  /**
   * Returns true if one of the format stored in this property includes the given format,
   * false otherwise.
   */
  bool isFormatIncluded(const Format& format, const std::string& jack_tag = "default") const;

  /**
   * Returns true if one of the format stored in this property matches the given format
   * and jack tag, false otherwise.
   */
  bool isFormatMatching(const Format& format, const std::string& jack_tag = "default") const;

  /**
   * Creates list of formats that match given format and its jack tag
   * \param format format to match 
   * \param property_jack_tag specifies with which formats in this property the intersection should be done
   * \param include_def_format if true default format will also be matched and eventually included in list
   * \returns format list
   */
  const_format_list createIntersectionFormats(const Format* format, 
					      const std::string& property_jack_tag, 
					      bool include_def_format = true) const;

  /**
   * Creates list of formats that match within this and given property.
   * \param out_jack_tag jack tag within this property
   * \param property other property (should be an input property)
   * \param in_jack_tag jack tag within other property
   * \param include_def_format if true default format will also be matched and eventually included in list
   * \returns format list
   */
  const_format_list createIntersectionFormats(const std::string& out_jack_tag, 
					      const Property* property, const std::string& in_jack_tag, 
					      bool include_def_format = true) const ;

  friend IValueStream& operator>>(IValueStream& v, Property& );
  friend OValueStream& operator<<(OValueStream& v, const Property& );
};

typedef auto_ptr<Property> Property_var;

}


#endif
