/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef _NMM_BUFFERMANAGER_HPP_
#define _NMM_BUFFERMANAGER_HPP_

#include "nmm/utils/thread/Thread.hpp"

namespace NMM {

  class Buffer;

  /**
   * This is a "fully abstract" class (an "interface") for
   * managing buffers. This is the superclass for all
   * classes which allocate, dispense, release and reuse
   * buffers.
   * \internal I wanted it to be abstract, but C++ refused to
   * let a function return a pointer to it, since it's abstract!
   * Duh!
   */
  class BufferManager {

  public:
    BufferManager();

    /**
     * Request an unused buffer of a certain minimal size. 
     * If the buffer manager is able to comply with the request
     * (by reusing a buffer that is large enough, or by allocating
     * a new one), such a buffer is returned. Otherwise, null
     * is returned. This call does not block until an adequate
     * Buffer is available.
     *
     * The reperence count is already increased to one inside
     * this function. A Node should NOT increase the reference
     * count of the returned Buffer unless it is reallt passing
     * it on multiple times!
     *
     * \param size the minimal size of the buffer.
     * \returns a pointer to a Buffer object, or null if no buffer
     * is available.
     */
    virtual Buffer* requestBuffer (const size_t size) const = 0;

    /**
     * Release a buffer. This is called from the Buffer class when
     * the corresponding buffer's reference count has dropped to 0.
     * \param buffer The buffer to be released.
     */
    virtual void releaseBuffer (Buffer* buffer) const = 0;

    virtual BufferManager* request();
    virtual void release();

protected:
    /**
     * The main destructor, which is empty.
     * The destructor is protected, because a BufferManager deletes themself if the refernce counter is 0.
     */
    virtual ~BufferManager() {};

    virtual void erase() const = 0;
private:
    int ref_count;
    ThreadMutex mutex;

  };

}

#endif //_NMM_BUFFERMANAGER_HPP_
