/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *                    Marco Lohse        <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "ObjectDescription.hpp"
#include "nmm/base/Interface.hpp"
#include "ClientRegistry.hpp"
#include <algorithm>

namespace NMM {

ObjectDescription::ObjectDescription(MemoryMgm mgm)
  : SerializedValue("ObjectDescription", "NMM", mgm), __location(UserInfo().getHostname()),
    __location_flag(true), __port(ClientRegistry::default_port)
{
  registerAttributes();
}

ObjectDescription::ObjectDescription(const string& object_name, MemoryMgm mgm)
  : SerializedValue("ObjectDescription", "NMM", mgm), __object_name(object_name), 
    __location(UserInfo().getHostname()),
    __location_flag(true), __port(ClientRegistry::default_port)
{
  registerAttributes();
}


ObjectDescription::~ObjectDescription() {

}


ObjectDescription::ObjectDescription(const ObjectDescription& nd)
  : SerializedValue(nd), __location(UserInfo().getHostname()), __location_flag(true)
{
  registerAttributes();
  *this = nd;
}

void ObjectDescription::addInterface(const Interface& interface) {
  addInterface(interface.getTypeName());
}

void ObjectDescription::addInterface(const string& interface) {
  __interface_list.getRValue().push_back(interface);
}

void ObjectDescription::setLocation(const string& location)
{ 
  if(location.empty() || location == "localhost") {
    __location.setValue(UserInfo().getHostname());
  }
  else {
    __location.setValue(location); 
  }
}
 
void ObjectDescription::disableLocation() { 
  __location_flag.setValue(false);
}

void ObjectDescription::enableLocation() { 
  __location_flag.setValue(true);
}

bool ObjectDescription::isSubset(const ObjectDescription& nd) const {
  
  /*Begin> Compare the ObjectName if it is set*/
  if (!nd.getObjectName().empty()) {
    if (getObjectName() != nd.getObjectName()) {
      return false;
    }
  }
  /*End< */


  /*Begin> Check if locations of objects are identical */
  if (nd.isLocationEnabled()){
    if((getLocation() != nd.getLocation())) {
      return false;
    }
  }
  /*End<*/


  /*Begin> Compare the library name if it is set*/
  if (!nd.getLibraryName().empty()) {
    if (getLibraryName() != nd.getLibraryName())
      return false;
  }
  /*End< */

  /* Begin> check the interfaces */
  if (!isInterfaceSubset(nd)){
    return false;
  }
  /* End< */

  /* If all parameters match return true */
  return true;
}

bool ObjectDescription::isInterfaceSubset(const ObjectDescription& np) const {
  for (list<string>::const_iterator it = np.getInterfaceList().begin(); it != np.getInterfaceList().end(); ++it)
  {
    if (find(__interface_list.getRValue().begin(), __interface_list.getRValue().end(), *it) == __interface_list.getRValue().end())
      return false;
  }
  return true;
}

void ObjectDescription::setPort(int port) {
  __port.setValue(port);
}

void ObjectDescription::registerAttributes() {
  insertArgument(&__object_name);
  insertArgument(&__library_name);
  insertArgument(&__interface_list);
  insertArgument(&__service_description);
  insertArgument(&__config_event);
  insertArgument(&__location);
  insertArgument(&__location_flag);
  insertArgument(&__port);
}

TConstructor<ObjectDescription>    __nmm__objectdescription_info("ObjectDescription", "NMM");

} //end namespace
