/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include <iostream>
#include "MagicNumberValueStream.hpp"
#include "Value.hpp"
#include "nmm/base/BaseExceptions.hpp"


namespace NMM {

const TypeInfo* MagicNumberManager::getTypeInfo(int magic_number) {
  /* find the iterator of the magic number */
  HASH_MAP_STL::hash_map<int, TypeInfo*>::iterator it = __magic_number_map.find(magic_number);
  /* If the magic number is unknown return NULL */
  if (it == __magic_number_map.end())
    return 0;

  /* return the TypeInfo */
  return it->second;
}


int MagicNumberManager::getMagicNumber(const string& type_name, const string& ns) {
  /* find the iterator of the given TypeInfo */
  HASH_MAP_STL::hash_map<string, HASH_MAP_STL::hash_map<string, int, string_hash >, string_hash >::iterator it = __type_info_map.find(ns);

  /* If the namespace is unknown return an invalid magic number */
  if ( it == __type_info_map.end() )
    return -1;

  HASH_MAP_STL::hash_map<string, int, string_hash >::iterator it_2 = it->second.find(type_name);
  /* If the typename is unknown return an invalid magic number */
  if (it_2 == it->second.end())
    return -1;

  /* Return the magic number */
  return it_2->second;
}


void MagicNumberManager::addMagicNumber(const TypeInfo& type, int magic_number) {
  HASH_MAP_STL::hash_map<string, HASH_MAP_STL::hash_map<string, int, string_hash >, string_hash >::iterator it = __type_info_map.find(type.t_namespace);
  if (it == __type_info_map.end())
    it = __type_info_map.insert( pair<string, HASH_MAP_STL::hash_map<string, int, string_hash > >
          (type.t_namespace, HASH_MAP_STL::hash_map<string, int, string_hash>()) ).first;

  it->second.insert(pair<string, int>(type.t_name, magic_number));

  __magic_number_map.insert(pair< int, TypeInfo* >(magic_number, new TypeInfo(type)));
}

const TypeInfo* GenericMagicNumber::magic_number_map[] =
  { 0, 0,
    &u_int16_type_info, &int16_type_info,
    &u_int32_type_info, &int32_type_info,
    &u_int64_type_info, &int64_type_info,
    &bool_type_info,
    &unsigned_long_type_info, &long_type_info,
    &double_type_info, &float_type_info,
    &char_type_info, &unsigned_char_type_info,
    &char_ptr_type_info, &unsigned_char_ptr_type_info,
    &string_type_info
  };

const int GenericMagicNumber::magic_number[] = {
    0, 1, 2, 3, 4, 5, 6, 7, 8, 9,
    10, 11, 12, 13, 14, 15, 16, 17, 18
  };



MagicNumberIValueStream::MagicNumberIValueStream()
  : __current_type_info(0), __next_element(0), __is_empty(true),  __stream_empty(true), __magic_number_manager(&SMagicNumberManager::getInstance())
{

}


MagicNumberIValueStream::~MagicNumberIValueStream()
{

}

const TypeInfo& MagicNumberIValueStream::getCurrentTypeInfo() const {
  /* If no TypeInfo is available return a string */
  if (!__current_type_info) {
    throw Exception("Stream is empty");
  }

  return *__current_type_info;
}

const string& MagicNumberIValueStream::getCurrentType() const {
  /* If no TypeInfo is available return a string */
  if (!__current_type_info) {
    return __empty_string;
  }

  /* return the current typename */
  return __current_type_info->t_name;
}

const string& MagicNumberIValueStream::getCurrentNamespace() const {
  if (!__current_type_info) {
    const_cast<MagicNumberIValueStream*>(this)->nextElement();
  }

  /* If no TypeInfo is available return a string */
  if (!__current_type_info) {
    GLOBAL_ERROR_STREAM("Stream is empty");
    return __empty_string;
  }

  /* return the current namespace */
  return __current_type_info->t_namespace;
}

bool MagicNumberIValueStream::isEmpty() const {
  if (__stream_empty)
    return true;

  return __is_empty;
}


void MagicNumberIValueStream::goUp() {
  if (__open_stack.empty() )
    return ;
  /* remove the first type info of latest complex type */
  __open_stack.pop_back();
  __is_empty = false;
  /* read the TypeInfo of the next element */
  nextElement();
}


bool MagicNumberIValueStream::goDown() {
  if ( (__next_element != magic_number[COMPLEX_BEGIN]) && (__next_element < 20)) {
    return false;
  }
  /* push the current TypeInfo to the internal vector */
  __open_stack.push_back(__current_type_info);
  /* read the TypeInfo of the next element */
  nextElement();
  return true;
}

bool MagicNumberIValueStream::nextElement() {
  /* If no Partner stream is set, no element can be read */
  if(!getIPartnerStream()) {
    return false;
  }

  /* If the stream is empty return false */
  if (__stream_empty) {
    return false;
  }

  /* Is the next element a NULL-Pointer? */
  bool skip_null = (__next_element == 32);

  /* read the magic number of the next element */
  (*getIPartnerStream()) >> __next_element;
  switch (__next_element) {
    /* For the next element no magic number was available. We must read it. */
    case 1:
      __is_empty = false;
      (*getIPartnerStream()) >> __complex_type_info.t_name;
      (*getIPartnerStream()) >> __complex_type_info.t_namespace;
      __current_type_info = &__complex_type_info;
      break;


    /* A complex struct was closed. */
    case 0:
      __is_empty = true;
      if (__current_type_info)
        if (__current_type_info->t_name == "NULL")
          skip_null = true;
      __current_type_info = 0;
      /* If the current Element is a null pointerm, read the next TypeInfo */
      if (skip_null)
        nextElement();
      break;

    /* The Stream is empty */
    case 18:
      __stream_empty = true;
      break;

    /* Map the read magic number to a TypeInfo */
    default:
      __is_empty = false;
      /* is it in our internal map ?*/
      if ( (__next_element >= 0) && ( __next_element <= 18) )
        __current_type_info = magic_number_map[__next_element];
      else
        /* Test if its in the MagicNumberManager */
        __current_type_info = __magic_number_manager->getTypeInfo(__next_element);


      if (__current_type_info == 0) {
        cerr << "Invalid Magic Number " << __next_element << endl; cerr.flush();
        throw Exception("Invalid Magic number");
      }

      break;
  }
  return true;
}

void MagicNumberIValueStream::readBuffer(char* buffer, unsigned int size) {
  /* If stream is empty, return */
  if (isEmpty()) {
    cerr << "Stream is Empty for type id : " << "char_buffer" << endl;
    return;
  }

  /* Type check */
  if (__next_element != CHAR_PTR) {
    cerr << "Invalid type conversion from char_buffer to " << getCurrentType()  << endl;
  }

  /* read the buffer from the partner stream */
  if (getIPartnerStream()) {
    getIPartnerStream()->readBuffer(buffer, size);
  }

  /* read the TypeInfo of the next element */
  nextElement();
}

void MagicNumberIValueStream::readBuffer(unsigned char* buffer, unsigned int size) {
  /* If stream is empty, return */
  if (isEmpty()) {
    cerr << "Stream is Empty for type id : " << "unsigned_char_buffer" << endl;
    return;
  }

  /* Type check */
  if (__next_element != UNSIGNED_CHAR_PTR) {
    cerr << "Invalid type conversion from char_buffer to " << getCurrentType()  << endl;
  }

  /* read the buffer from the partner stream */
  if (getIPartnerStream()) {
    getIPartnerStream()->readBuffer(buffer, size);
  }

  /* read the TypeInfo of the next element */
  nextElement();
}

void MagicNumberIValueStream::readData() {
  if (getIPartnerStream()) {
    getIPartnerStream()->readData();
  }
  
  __is_empty = false;
  __stream_empty = false;
  nextElement();
}


MagicNumberIValueStream& MagicNumberIValueStream::operator>>(int16_t&  i) {
  if (getIPartnerStream()) {
    /* Type check */
    if (__next_element != magic_number[INT16]) {
      throw Exception("Invalid Type assignment. Excpected type INT16"  + getCurrentType());
    } else {
      /* read the value */
      (*getIPartnerStream()) >> i;
    }
    nextElement();
  }

  return *this;
}


MagicNumberIValueStream& MagicNumberIValueStream::operator>>(u_int16_t& i) {
  if (getIPartnerStream()) {
    /* Type check */
    if (__next_element != magic_number[U_INT16]) {
      throw Exception("Invalid Type assignment. Excpected type U_INT16"  + getCurrentType());
    } else {
      /* read the value */
      (*getIPartnerStream()) >> i;
    }
    nextElement();
  }

  return *this;
}


MagicNumberIValueStream& MagicNumberIValueStream::operator>>(int32_t& i) {
  if (getIPartnerStream()) {
    /* Type check */
    if (__next_element != magic_number[INT32] && __next_element != magic_number[LONG]) {
      throw Exception("Invalid Type assignment. Excpected type INT32, actual type "  + getCurrentType());
    } else {
      /* read the value */
      (*getIPartnerStream()) >> i;
    }
    nextElement();
  }
  
  return *this;
}

MagicNumberIValueStream& MagicNumberIValueStream::operator>>(u_int32_t&  i) {
  if (getIPartnerStream()) {
    /* Type check */
    if (__next_element != magic_number[U_INT32]) {
      throw Exception("Invalid Type assignment. Excpected type U_INT32, actual type "  + getCurrentType());
    } else {
      /* read the value */
      (*getIPartnerStream()) >> i;
    }
    nextElement();
  }

  return *this;
}


MagicNumberIValueStream& MagicNumberIValueStream::operator>>(int64_t& i) {
  if (getIPartnerStream()) {
    /* Type check */
    if (__next_element != magic_number[INT64]) {
      throw Exception("Invalid Type assignment. Excpected type INT64, actual type "  + getCurrentType());
    } else {
      /* read the value */
      (*getIPartnerStream()) >> i;
    }
    nextElement();
  }

  return *this;
}


MagicNumberIValueStream& MagicNumberIValueStream::operator>>(u_int64_t& i) {
  if (getIPartnerStream()) {
    /* Type check */
    if (__next_element != magic_number[U_INT64]) {
      throw Exception("Invalid Type assignment. Excpected type u_INT64, actual type "  + getCurrentType());
    } else {
      /* read the value */
      (*getIPartnerStream()) >> i;
    }
    nextElement();
  }

  return *this;
}



MagicNumberIValueStream& MagicNumberIValueStream::operator>>(unsigned long& i) {
  if (getIPartnerStream()) {
    /* Type check */
    if (__next_element != magic_number[UNSIGNED_LONG]) {
      throw Exception("Invalid Type assignment. Excpected type UNSIGNED LONG, actual type "  + getCurrentType());
    } else {
      /* read the value */
      (*getIPartnerStream()) >> i;
    }
    nextElement();
  }

  return *this;
}

 MagicNumberIValueStream& MagicNumberIValueStream::operator>>(long& i) {
   if (getIPartnerStream()) {
     /* Type check */
     if (__next_element != magic_number[LONG] && __next_element != magic_number[INT32]) {
       throw Exception("Invalid Type assignment. Excpected type LONG"  + getCurrentType());
     } else {
       /* read the value */
       (*getIPartnerStream()) >> i;
     }
     nextElement();
   }

   return *this;
 }

MagicNumberIValueStream& MagicNumberIValueStream::operator>>(double& i) {
  if (getIPartnerStream()) {
    /* Type check */
    if (__next_element != magic_number[DOUBLE]) {
      throw Exception("Invalid Type assignment. Excpected type DOUBLE"  + getCurrentType());
    } else {
      /* read the value */
      (*getIPartnerStream()) >> i;
    }
    nextElement();
  }

  return *this;
}

MagicNumberIValueStream& MagicNumberIValueStream::operator>>(float& i) {
  if (getIPartnerStream()) {
    /* Type check */
    if (__next_element != magic_number[FLOAT]) {
      throw Exception("Invalid Type assignment. Excpected type FLOAT. Current type " + getCurrentType());
    } else {
      /* read the value */
      (*getIPartnerStream()) >> i;
    }
    nextElement();
  }

  return *this;
}

MagicNumberIValueStream& MagicNumberIValueStream::operator>>(char& i) {
  if (getIPartnerStream()) {
    /* Type check */
    if (__next_element != magic_number[CHAR]) {
      throw Exception("Invalid Type assignment Excpected type char. Current type"  + getCurrentType());
    } else {
      /* read the value */
      (*getIPartnerStream()) >> i;
    }
    nextElement();
  }

  return *this;
}


MagicNumberIValueStream& MagicNumberIValueStream::operator>>(unsigned char& i) {
  if (getIPartnerStream()) {
    /* Type check */
    if (__next_element != magic_number[UNSIGNED_CHAR]) {
      throw Exception("Invalid Type assignment Excpected type unsigne char. Current type"  + getCurrentType());
    } else {
      /* read the value */
      (*getIPartnerStream()) >> i;
    }
    nextElement();
  }

  return *this;
}

MagicNumberIValueStream& MagicNumberIValueStream::operator>>(string& i) {
  if (isEmpty()) {
    return *this;
  }


  if (getIPartnerStream()) {
    /* Type check */
    if (__next_element != magic_number[STRING]) {
      throw Exception("Invalid Type assignment Excpected type string. Current type"  + getCurrentType());
    } else {
      /* read the value */
      (*getIPartnerStream()) >> i;
    }
  }

  nextElement();
  return *this;
}

MagicNumberIValueStream& MagicNumberIValueStream::operator>>(bool& i) {
  if (getIPartnerStream()) {
    /* Type check */
    if (__next_element != magic_number[BOOL]) {
      throw Exception("Invalid Type assignment Excpected type bool. Current type"  + getCurrentType());
    } else {
      /* read the value */
      (*getIPartnerStream()) >> i;
    }
    nextElement();
  }

  return *this;
}

const string MagicNumberIValueStream::__i_value_stream_type = "MagicNumberIValueStream";
const string MagicNumberIValueStream::__empty_string;

/* MagicNumberOValueStream*/




MagicNumberOValueStream::~MagicNumberOValueStream() {
}


MagicNumberOValueStream::MagicNumberOValueStream()
  : __magic_number_manager(&SMagicNumberManager::getInstance())
{

}

void MagicNumberOValueStream::beginType(const string& type, const string& ns) {
  int m_number = __magic_number_manager->getMagicNumber(type, ns);
  if (m_number == -1) {
    if (getOPartnerStream()) {
      (*getOPartnerStream()) << magic_number[COMPLEX_BEGIN] << type << ns;
    }
  } else {
    if (getOPartnerStream()) {
      (*getOPartnerStream()) << m_number;
    }
  }
}


void MagicNumberOValueStream::beginType(int m_number) {
  if (getOPartnerStream()) {
    (*getOPartnerStream()) << m_number;
  }
}

void MagicNumberOValueStream::endType() {
  if (getOPartnerStream()) {
    (*getOPartnerStream()) << magic_number[COMPLEX_END];
  }
}


void MagicNumberOValueStream::writeBuffer(const char* buffer, unsigned int size) {
  if (getOPartnerStream()) {
    (*getOPartnerStream()) << magic_number[CHAR_PTR];
    getOPartnerStream()->writeBuffer(buffer, size);
  }
}

void MagicNumberOValueStream::writeData()  {
  if (getOPartnerStream()) {
    (*getOPartnerStream()) << magic_number[FINFLAG];
    getOPartnerStream()->writeData();
  }
}

void MagicNumberOValueStream::writeBuffer(const unsigned char* buffer, unsigned int size) {
  if (getOPartnerStream()) {
    (*getOPartnerStream()) << magic_number[UNSIGNED_CHAR_PTR];
    getOPartnerStream()->writeBuffer(buffer, size);
  }
}


MagicNumberOValueStream& MagicNumberOValueStream::operator<<(const u_int16_t &i) {
  if (getOPartnerStream()) {
    (*getOPartnerStream()) << magic_number[U_INT16] << i;
  } else {
    throw Exception("No partner stream available");
  }
  return *this;
}

MagicNumberOValueStream& MagicNumberOValueStream::operator<<(const int16_t&  i) {
  if (getOPartnerStream()) {
    (*getOPartnerStream()) << magic_number[INT16] << i;
  }
  return *this;
}

MagicNumberOValueStream& MagicNumberOValueStream::operator<<(const u_int32_t& i) {
  if (getOPartnerStream()) {
    (*getOPartnerStream()) << magic_number[U_INT32] << i;
  }
  return *this;
}

MagicNumberOValueStream& MagicNumberOValueStream::operator<<(const int32_t&  i) {
  if (getOPartnerStream()) {
    (*getOPartnerStream()) << magic_number[INT32] << i;
  }
  return *this;
}


MagicNumberOValueStream& MagicNumberOValueStream::operator<<(const u_int64_t& i) {
  if (getOPartnerStream()) {
    (*getOPartnerStream()) << magic_number[U_INT64] << i;
  }
  return *this;
}

MagicNumberOValueStream& MagicNumberOValueStream::operator<<(const int64_t& i) {
  if (getOPartnerStream()) {
    (*getOPartnerStream()) << magic_number[INT64] << i;
  }
  return *this;
}

MagicNumberOValueStream& MagicNumberOValueStream::operator<<(const unsigned long& i) {
  if (getOPartnerStream()) {
    (*getOPartnerStream()) << magic_number[UNSIGNED_LONG] << i;
  }
  return *this;
}

MagicNumberOValueStream& MagicNumberOValueStream::operator<<(const long& i) {
   if (getOPartnerStream()) {
     (*getOPartnerStream()) << magic_number[LONG] << i;
   }
   return *this;
}

MagicNumberOValueStream& MagicNumberOValueStream::operator<<(const double& i) {
  if (getOPartnerStream()) {
    (*getOPartnerStream()) << magic_number[DOUBLE] << i;
  }
  return *this;
}

MagicNumberOValueStream& MagicNumberOValueStream::operator<<(const float& i) {
  if (getOPartnerStream()) {
    (*getOPartnerStream()) << magic_number[FLOAT] << i;
  }
  return *this;
}

MagicNumberOValueStream& MagicNumberOValueStream::operator<<(const char& i) {
  if (getOPartnerStream()) {
    (*getOPartnerStream()) << magic_number[CHAR] << i;
  }
  return *this;
}

MagicNumberOValueStream& MagicNumberOValueStream::operator<<(const unsigned char& i) {
  if (getOPartnerStream()) {
    (*getOPartnerStream()) << magic_number[UNSIGNED_CHAR] << i;
  }
  return *this;
}

MagicNumberOValueStream& MagicNumberOValueStream::operator<<(const string& i) {
  if (getOPartnerStream()) {
    (*getOPartnerStream()) << magic_number[STRING] << i;
  }
  return *this;
}

MagicNumberOValueStream& MagicNumberOValueStream::operator<<(const bool& i) {
  if (getOPartnerStream()) {
    (*getOPartnerStream()) << magic_number[BOOL] << i;
  }
  return *this;
}

const string MagicNumberOValueStream::__o_value_stream_type = "MagicNumberOValueStream";


TValueStreamConstructor<MagicNumberIValueStream, MagicNumberOValueStream> magic_number_value_stream(MagicNumberIValueStream::__i_value_stream_type, MagicNumberOValueStream::__o_value_stream_type);


/* register some magic number */

MagicNumberRegister __nmm_event_magic_number(TypeInfo("Event", "NMM"), 20);
MagicNumberRegister __nmm_cevent_magic_number(TypeInfo("CEvent", "NMM"), 21);
MagicNumberRegister __nmm_buffer_magic_number(TypeInfo("Buffer", "NMM"), 22);
MagicNumberRegister __nmm_nmmguid_magic_number(TypeInfo("NMMGuid", "NMM"), 23);
MagicNumberRegister __nmm_nodedescription_magic_number(TypeInfo("NodeDescription", "NMM"), 24);
MagicNumberRegister __nmm_userinfo_magic_number(TypeInfo("UserInfo", "NMM"), 25);
MagicNumberRegister __nmm__registry_id_magic_number(TypeInfo("RegistryID", "NMM"), 26);
MagicNumberRegister __nmm_list_Response_value_magic_number(TypeInfo("list_Response", "std"), 27);
MagicNumberRegister __nmm_list_RegistryID_value_magic_number(TypeInfo("list_RegistryID", "std"), 28);
MagicNumberRegister __nmm_list_ObjectID_value_magic_number(TypeInfo("list_ObjectID", "std"), 29);
MagicNumberRegister __nmm_pair_response_value_magic_number(TypeInfo("pair_RegistryID_list_ObjectID", "std"), 30);
MagicNumberRegister __nmm_sequence_value_magic_number(TypeInfo("SequenceValue", "NMM"), 31);

MagicNumberRegister __nmm_NULL_value_magic_number(TypeInfo("NULL", "std"), 32);

MagicNumberRegister __nmm_node_info_number(TypeInfo("NodeInfo", "NMM"), 33);
MagicNumberRegister __nmm_mgm_number(TypeInfo("MemoryMgm", "NMM"), 34);

MagicNumberRegister __nmm_timestamp_magic_number(TypeInfo("Timestamp", "NMM"), 40);
MagicNumberRegister __nmm_time_magic_number(TypeInfo("Time", "NMM"), 41);



MagicNumberRegister __nmm_list_string_value_magic_number(TypeInfo("list_string", "std"), 200);
MagicNumberRegister __nmm_list_float_value_magic_number (TypeInfo("list_float", "std"), 201);
MagicNumberRegister __nmm_list_double_value_magic_number(TypeInfo("list_double", "std"), 202);
MagicNumberRegister __nmm_list_u_int_value_magic_number(TypeInfo("list_unsigned_int", "std"), 202);
MagicNumberRegister __nmm_list_su_int_value_magic_number(TypeInfo("list_short_unsigned_int", "std"), 203);
MagicNumberRegister __nmm_list_int_value_magic_number(TypeInfo("list_int", "std"), 204);
MagicNumberRegister __nmm_list_s_int_value_magic_number(TypeInfo("list_short_int", "std"), 205);
MagicNumberRegister __nmm_list_long_value_magic_number(TypeInfo("list_long", "std"), 206);
MagicNumberRegister __nmm_list_u_long_value_magic_number (TypeInfo("list_unsigned_long", "std"), 207);
MagicNumberRegister __nmm_list_NodeInfo_magic_number(TypeInfo("list_NodeInfo", "std"), 208);


MagicNumberRegister __nmm_vector_nmm_int32_magic_number(TypeInfo("vector_int32_t", "std"), 300);

MagicNumberRegister __nmm_pair_int_int_value_magic_number (TypeInfo("pair_int_int", "std"), 400);
MagicNumberRegister __nmm_pair_unigned_long_long_unsigned_long_long_magic_number (TypeInfo("pair_unsigned_long_long_unsigned_long_long", "std"), 401);

MagicNumberRegister __nmm_map_string_string_magic_number (TypeInfo("map_string_string", "std"), 500);


} //end namespace NMM


