/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef _NMM_VALUE_HPP_
#define _NMM_VALUE_HPP_

#include <string>
#include <list>
#include <vector>
#include "ValueFactory.hpp"

using namespace std;

namespace NMM {

class IValueStream;
class OValueStream;

/** This class is a superclass for all "value" classes.
  */
class Value {
public:
  enum ValueType{ SIMPLE  = 0, ARRAY, SEQUENCE, COMPOUND};
  enum MemoryMgm{ INVALUE = 1, OUTVALUE = 2, INOUTVALUE = 4};

  Value(ValueType vt = SIMPLE, MemoryMgm memory_mgm = INOUTVALUE)
    :  _vt(vt), _memory_mgm(memory_mgm) {}

/** The main destructor.
  */
  virtual ~Value()                              {}

/** set the value from string
  */
  virtual void assignFromStream(IValueStream&)      {};

  virtual void createFromStream(IValueStream&)      {};

/** Write the Value to a stream
  */
  virtual void toStream(OValueStream&)        const {};

/** virtual copy method
  */
  virtual Value* copy()                       const { return 0; };

/** release method for memory management
  */
  virtual void releaseValue()                 const {};


/** returns true if the stored value is a pointer
  */
  virtual bool isPointerType()                const { return false; }

/** This method can be used to disabel certain types. This mehtod is interest for complex
  * data structures to disable certain Memory types like In or Out-Values for (de-)serialization
  */
  virtual void disableMemoryType(Value::MemoryMgm) {}

/** resets the internal datastructure wich was disabled before.
  */
  virtual void resetMemoryType() {}

/** return the value type. This metho is obsolete.
  */
  inline       ValueType getValueType()       const {return _vt;}

/** The used memory management strategy
  */
  inline       MemoryMgm getMemoryStrategy()  const {return _memory_mgm;}

/** Set the used memory management strategy
  */
  inline       void setMemoryStrategy(MemoryMgm mgm) {_memory_mgm = mgm;}

private:
  ValueType _vt;
  MemoryMgm _memory_mgm;
};


OValueStream& operator<<(OValueStream&, const Value&);
IValueStream& operator>>(IValueStream&, Value& );

OValueStream& operator<<(OValueStream&, const Value::MemoryMgm &);
IValueStream& operator>>(IValueStream&, Value::MemoryMgm& );

template<typename T>
class TValue : public Value {
public:

/** constructor
  */
  TValue(MemoryMgm mem_mgr = INOUTVALUE, bool is_mgm_enabled = true)
    : Value(Value::SIMPLE, mem_mgr), __is_mgm_enabled(is_mgm_enabled), __serialize_mgm_type(false)   {}

/** constructor
  */
  TValue(const T& value, MemoryMgm mem_mgr = INOUTVALUE, bool is_mgm_enabled = true)
    : Value(Value::SIMPLE, mem_mgr), _value(value), __is_mgm_enabled(is_mgm_enabled), __serialize_mgm_type(false)   {}

/** copy constructor
  */
  TValue(const TValue<T>& v)
    : Value(v), _value(v.getRValue()), __is_mgm_enabled(v.__is_mgm_enabled), __serialize_mgm_type(v.__serialize_mgm_type) {}

/** returns a copy of the current value
  */
  inline T  getValue()                              const   { return _value; }

/** returns a refernce to the current value
  */
  inline  const T& getRValue()                      const   { return _value; }

/** returns a refernce to the current value
  */
  inline        T& getRValue()                              { return _value; }

/** returns a pointer to the current value. You MUST NOT delete this pointer.
  */
  inline  const T* getPValue()                      const   { return &_value; }

/** returns a pointer to the current value. You MUST NOT delete this pointer.
  */
  inline        T* getPValue()                              { return &_value; }

/** returns whether the automatic memory management is enabled or not.
  */
  inline  bool isMgmEnabled()                       const   { return __is_mgm_enabled; }

/** This method enables or disables the automtic memory management is true or not
  * bool mgm: TRUE enables the automatic memory management, FALSE disables the automatic memory management.
  */
  inline void setMgmEnabled(bool mgm)                       { __is_mgm_enabled = mgm; }

/** set a new value
  */
  inline void setValue(const T& new_value)                  { _value = new_value; }

/** set the value from ValueStream
  */
  virtual void assignFromStream(IValueStream& s)            { if (s.getCurrentType() == "NULL") s.nextElement();  else { s >> _value; if (s.getCurrentType() == "MemoryMgm") {MemoryMgm mgm; s >> mgm; setMemoryStrategy(mgm);} } }

  virtual void createFromStream(IValueStream& s)            { if (s.getCurrentType() == "NULL") s.nextElement();  else { s >> _value; if (s.getCurrentType() == "MemoryMgm") {MemoryMgm mgm; s >> mgm; setMemoryStrategy(mgm);} } }
/** Write the Value to a ValueStream
  */
  virtual void toStream(OValueStream& s)             const  { s << _value; if (__serialize_mgm_type) { s << getMemoryStrategy();} }

/** operator for implicit type conversion
  */
  template <class D>
  operator TValue<D>()                                      { return TValue<D>(_value); }

/** copy method (virtual copy-constructor)
  */
  virtual TValue<T>* copy()                          const  { if (getMemoryStrategy() == INVALUE) return new TValue<T>(*this);  return const_cast<TValue<T>* >(this);}

/** assignement operator
  */
  void operator=(const TValue<T>& v)                        { Value::operator=(v);  _value = v.getRValue();}

  virtual void releaseValue()                        const  { if (getMemoryStrategy() == INVALUE && __is_mgm_enabled) delete this;}

protected:
  void serializeMgmType (bool value)                        { __serialize_mgm_type = value ;}
private:
  mutable T _value;                                  ///< The stored value
  bool    __is_mgm_enabled;                          ///< specifies whether the automatic memory management is enabled for InValues or not. Default is true. 
  bool    __serialize_mgm_type;                      ///< specifies whether the memory type is serialized as well. This is only true if subclasses of TValue are used
};


template<>
class TValue<char*> : public Value {
public:

/** constructor
  */
  TValue(MemoryMgm mem_mgr = INOUTVALUE, bool is_mgm_enabled = true)
    : Value(Value::SIMPLE, mem_mgr) , _value(0), _size(0), __is_mgm_enabled(is_mgm_enabled), __serialize_mgm_type(false) {}

/** constructor
  */
  TValue(char* value, unsigned int size, MemoryMgm mem_mgr = INOUTVALUE, bool is_mgm_enabled = true)
    : Value(Value::SIMPLE, mem_mgr), _value(value), _size(size), __is_mgm_enabled(is_mgm_enabled), __serialize_mgm_type(false)  {}

/** copy constructor
  */
  TValue(const TValue<char*>& v)
    : Value(v), _value(v.getRValue()), _size(v.getSize()), __is_mgm_enabled(v.__is_mgm_enabled), __serialize_mgm_type(v.__serialize_mgm_type)  {}

/** returns a copy of the current value
  */
  inline unsigned int getSize()                      const   { return _size; }

/** returns a copy of the current value
  */
  inline const char* getValue()                      const   { return _value; }
  inline       char* getValue()                              { return _value; }

/** returns a refernce to the current value
  */
  inline       char*& getRValue()                            { return _value; }
  inline       char*  getRValue()                    const   { return _value; }

/** returns a pointer to the current value. You MUST NOT delete this pointer.
  */
  inline       char** getPValue()                    const   { return &_value; }

/** returns whether the automatic memory management is enabled or not.
  */
  inline  bool isMgmEnabled()                        const   { return __is_mgm_enabled; }  

/** This method enables or disables the automtic memory management is true or not
  * bool mgm: TRUE enables the automatic memory management, FALSE disables the automatic memory management.
  */
  inline void setMgmEnabled(bool mgm)                        { __is_mgm_enabled = mgm; }

/** set a new value
  */
  inline void setValue(char* new_value, unsigned int size)   { _value = new_value; _size = size; }

/** set the value from ValueStream
  */
  virtual void assignFromStream(IValueStream& s)             { if (s.getCurrentType() != "NULL") { s.readBuffer(_value, _size);  if (s.getCurrentType() == "MemoryMgm") {MemoryMgm mgm; s >> mgm; setMemoryStrategy(mgm);}} else s.nextElement();}

  virtual void createFromStream(IValueStream& s)             { if (s.getCurrentType() != "NULL") { s.readBuffer(_value, _size);  if (s.getCurrentType() == "MemoryMgm") {MemoryMgm mgm; s >> mgm; setMemoryStrategy(mgm);}} else { s.nextElement(); setValue(0, 0); }}
/** Write the Value to a ValueStream
  */
  virtual void toStream(OValueStream& s)              const  { s.writeBuffer(_value, _size);  if (__serialize_mgm_type) { s << getMemoryStrategy();} }

/** operator for implicit type conversion
  */
  template <class D>
  operator TValue<D>()                                      { return TValue<D>(_value); }

/** copy method (virtual copy-constructor)
  */
  virtual TValue<char*>* copy()                      const  { if (getMemoryStrategy() == INVALUE) return new TValue<char*>(*this);  return const_cast<TValue<char*>* >(this);}

/** assignement operator
  */
  void operator=(const TValue<char*>& v)                    { Value::operator=(v);  _value = v.getRValue();}

  virtual void releaseValue()                        const  { if (getMemoryStrategy() == INVALUE && __is_mgm_enabled) delete this;}
protected:
  void serializeMgmType (bool value)                        { __serialize_mgm_type = value ;}
private:
  mutable char*  _value;
  unsigned int _size;
  bool    __is_mgm_enabled;                          ///< specifies whether the automatic memory management is enabled for InValues or not. Default is true.
  bool    __serialize_mgm_type;                      ///< specifies whether the memory type is serialized as well. This is only true if subclasses of TValue are used
};


template<class T>
class TValue<T*> : public Value {
public:

/** constructor
  */
  TValue(MemoryMgm mem_mgr = INOUTVALUE, bool is_mgm_enabled = true)
    : Value(Value::SIMPLE, mem_mgr), _value(0), __is_mgm_enabled(is_mgm_enabled), __serialize_mgm_type(false) { }

/** constructor
  */
  TValue(const T* value, MemoryMgm mem_mgr = INOUTVALUE, bool is_mgm_enabled = true)
    : Value(Value::SIMPLE, mem_mgr), _value(const_cast<T*>(value)), __is_mgm_enabled(is_mgm_enabled), __serialize_mgm_type(false){ }

/** copy constructor
  */
  TValue(const TValue<T*>& v)
    : Value(v), _value(0), __is_mgm_enabled(v.__is_mgm_enabled), __serialize_mgm_type(v.__serialize_mgm_type)  { if (v.getValue() != 0) _value = const_cast<T*>(v.getValue()); }

/**
  */
  ~TValue() { }

/** returns true if the stored value is a pointer
  */
  virtual bool isPointerType()                    const   { return true; }

/** returns a copy of the current value
  */
  inline const T*  getValue()                     const   { return _value; }

/** returns a copy of the current value
  */
  inline       T*  getValue()                             { return _value; }

/** returns a refernce to the current value
  */
  inline T* const& getRValue()                    const   { return _value; }

/** returns a refernce to the current value
  */
  inline       T*& getRValue()                            { return _value; }

/** returns a pointer to the current value. You MUST NOT delete this pointer.
  */
  inline       T** getPValue()                            { return &_value; }

/** returns a pointer to the current value. You MUST NOT delete this pointer.
  */
  inline T* const* getPValue()                    const   { return &_value; }

/** set a new value
  */
  inline void setValue(T* new_value)                      { _value = new_value; }

/** returns whether the automatic memory management is enabled or not.
  */
  inline  bool isMgmEnabled()                     const   { return __is_mgm_enabled; }  

/** This method enables or disables the automtic memory management is true or not
  * bool mgm: TRUE enables the automatic memory management, FALSE disables the automatic memory management.
  */
  inline void setMgmEnabled(bool mgm)                     { __is_mgm_enabled = mgm; } 
  
/** set the value from ValueStream
  */
  virtual void assignFromStream(IValueStream& s) {
    if (s.getCurrentType() == "NULL") {
      _value = 0;
      s.nextElement();
    } else {
      if (_value) {
        s >> *_value;   //if a value is set we use this one
        if (s.getCurrentType() == "MemoryMgm") { 
          MemoryMgm mgm; s >> mgm; setMemoryStrategy(mgm);
        } 
        return;
      }

      if ( ! SValueFactory::getInstance().isValueAvailable(s.getCurrentTypeInfo()) ) {
        /* Creates a dummy type */
        Value* dummy_value = SValueFactory::getInstance().createValue(s.getCurrentTypeInfo());
        s >> *dummy_value;
        delete dummy_value;
        if (s.getCurrentType() == "MemoryMgm") { 
          MemoryMgm mgm; s >> mgm; setMemoryStrategy(mgm);
        } 
	return;
      }

      Value* tmp_value = 0;

      /* create the instance */
      tmp_value = SValueFactory::getInstance().createValue(s.getCurrentTypeInfo());

      /* cast to the correct type */
      _value = dynamic_cast<T*>(tmp_value);

      /* if all fails, create the value using the new Operator. */
      if (!_value)
        _value = new T();

      /* Read the value */
      if (_value!= 0) {
        s >> *_value;
        if (s.getCurrentType() == "MemoryMgm") { 
          MemoryMgm mgm; s >> mgm; setMemoryStrategy(mgm);
        } 
      }
      else {
        GLOBAL_ERROR_STREAM("Runtime error in Value"<< endl);
        delete tmp_value;
      }
    }
  }

  virtual void createFromStream(IValueStream& s){
    delete _value;
    if (s.getCurrentType() == "NULL") {
      _value = 0;
      s.nextElement();
    } else {
      if ( ! SValueFactory::getInstance().isValueAvailable(s.getCurrentTypeInfo()) ) {
        /* Creates a dummy type */
        Value* dummy_value = SValueFactory::getInstance().createValue(s.getCurrentTypeInfo());
        s >> *dummy_value;
        if (s.getCurrentType() == "MemoryMgm") { 
          MemoryMgm mgm; s >> mgm; setMemoryStrategy(mgm);
        } 
        delete dummy_value;
        return;
      }

      Value* tmp_value = 0;

      /* create the instance */
      tmp_value = SValueFactory::getInstance().createValue(s.getCurrentTypeInfo());

      /* cast to the correct type */
      _value = dynamic_cast<T*> (tmp_value);

      /* Read the value */
      if (!_value) {
        _value = new T();
      }

      if (_value!= 0) {
        s >> *_value;        
        if (s.getCurrentType() == "MemoryMgm") { 
          MemoryMgm mgm; s >> mgm; setMemoryStrategy(mgm);
        } 
      }
      else {
        GLOBAL_ERROR_STREAM("Runtime error in Value"<< endl);
        delete _value;
      }
    }
  }

/** Write the Value to a ValueStream
  */
  virtual void toStream(OValueStream& s)             const  {
    if (_value == 0) {
      s.beginType("NULL");
      s.endType();
    } else {
      s << *_value;
      if (__serialize_mgm_type) { s << getMemoryStrategy(); }
    }
  }

/** operator for implicit type conversion
  */
  template <class D>
  operator TValue<D>()                                      { return TValue<D>(_value); }

  T* operator->()                                           { return _value; }

/** copy method (virtual copy-constructor)
  */
  virtual TValue<T*>* copy()                         const  { if (getMemoryStrategy() == INVALUE) return new TValue<T*>(*this);  return const_cast< TValue<T*> * >(this);}

/** assignement operator
  */
  void operator=(const T* v)                                { if (v) { Value::operator=(*v); _value = dynamic_cast<T*>(v->copy()); } }

/** assignement operator
  */
  void operator=(const TValue<T*>& v)                       { Value::operator=(v);  if (v.getValue()) _value = dynamic_cast<T*>(v.getValue()->copy()); else _value = 0; }

  virtual void releaseValue()                        const  { if (getMemoryStrategy() == INVALUE && __is_mgm_enabled) delete this;}
protected:
  void serializeMgmType (bool value)                        { __serialize_mgm_type = value ;}

private:
  mutable T* _value;
  bool    __is_mgm_enabled;                          ///< specifies whether the automatic memory management is enabled for InValues or not. Default is true. 
  bool    __serialize_mgm_type;                      ///< specifies whether the memory type is serialized as well. This is only true if subclasses of TValue are used
};


/** This class can be used for types which have Value not as base class
  */
template <class T>
class TValueConstructor : public ValueConstructor {
public:
/** @name Constructor & Destructor
  */
//@{
/** Constructor
  */
  TValueConstructor(const string& type_name, const string& ns)
    :  ValueConstructor(TypeInfo(type_name, ns))
  { }

/** Constructor, that needs a TypeInfo
  */
  TValueConstructor(const TypeInfo& ti)
    :  ValueConstructor(ti)
  { }

/** virtual copy constructor
  */
  virtual ValueConstructor* copy()  const    { return new TValueConstructor<T>(*this); }
//@}

/** @name Factory method
  */
//@{
/** Creates an instance of the value. The caller of this method is responsible to delete the object
  */
  virtual TValue<T>* createValue()      const    { return new TValue<T>();  }

/** Creates an instance of the value with an TValue-Wrapper. The caller of this method is responsible to delete the object
  */
  virtual TValue<T>* createTValue()     const    { return new TValue<T>();  }

/** Creates an TValue-Wrapper with a pointer to the object. The caller of this method is responsible to delete the object
  */
  virtual TValue<TValue<T> * >* createTPointerValue()  const {return new TValue<TValue<T>*>();}
//@}
};

/** This class can be used for types which have Value as base class
  */
template<class T>
class TConstructor : public ValueConstructor {
public:
/** @name Constructor & Destructor
  */
//@{
/** Constructor
  */
  TConstructor(const string& type_name, const string& ns)
    : ValueConstructor(TypeInfo(type_name, ns))
  { }

/** Constructor, that needs a TypeInfo
  */
  TConstructor(const TypeInfo& t_info)
    : ValueConstructor(t_info)
  { }

/** virtual copy constructor
  */
  virtual ValueConstructor* copy()  const    { return new TConstructor<T>(*this); }
//@}

/** @name Factory method
  */
//@{
/** Creates an instance of the value. The caller of this method is responsible to delete the object
  */
  virtual Value*          createValue()           const {return new T();}

/** Creates an instance of the value with an TValue-Wrapper. The caller of this method is responsible to delete the object
  */
  virtual Value*          createTValue()          const {return new TValue<T>();}

/** Creates an TValue-Wrapper with a pointer to the object. The caller of this method is responsible to delete the object
  */
  virtual Value*          createTPointerValue()   const {return new TValue<T*>();}
//@}
};


template<typename T>
class TOutValue : public TValue<T> {
public:

  TOutValue()
    : TValue<T>(Value::OUTVALUE)   { this->serializeMgmType(true); }

  TOutValue(const T& value,  bool is_mgm_enabled = true)
    : TValue<T>(value, Value::OUTVALUE, is_mgm_enabled)   {this->serializeMgmType(true);}

};

template<typename T>
class TOutValue<T*> : public TValue<T*> {
public:

  TOutValue()
    : TValue<T*>(Value::OUTVALUE)   {this->serializeMgmType(true);}

  TOutValue(const T* value,  bool is_mgm_enabled = true)
    : TValue<T*>(value, Value::OUTVALUE,is_mgm_enabled )   {this->serializeMgmType(true);}
};


template<typename T>
class TInValue : public TValue<T> {
public:

  TInValue()
    : TValue<T>(Value::INVALUE)   {this->serializeMgmType(true);}

  TInValue(const T& value, bool is_mgm_enabled = true)
    : TValue<T>(value, Value::INVALUE, is_mgm_enabled)   {this->serializeMgmType(true);}

};

template<typename T>
class TInValue<T*> : public TValue<T*> {
public:

  TInValue()
    : TValue<T*>(Value::INVALUE)   {this->serializeMgmType(true);}

  TInValue(const T* value, bool is_mgm_enabled = true)
    : TValue<T*>(value, Value::INVALUE, is_mgm_enabled)   {this->serializeMgmType(true);}

};

template<typename T>
class TInOutValue : public TValue<T> {
public:

  TInOutValue()
    : TValue<T>(Value::INOUTVALUE)   {this->serializeMgmType(true);}

  TInOutValue(const T& value, bool is_mgm_enabled = true)
    : TValue<T>(value, Value::INOUTVALUE, is_mgm_enabled)   {this->serializeMgmType(true);}

};

template<typename T>
class TInOutValue<T*> : public TValue<T*> {
public:

  TInOutValue()
    : TValue<T*>(Value::INOUTVALUE)   {this->serializeMgmType(true);}

  TInOutValue(const T* value, bool is_mgm_enabled = true)
    : TValue<T*>(value, Value::INOUTVALUE, is_mgm_enabled)   {this->serializeMgmType(true);}
};

}

#endif // _NMM_VALUE_HPP_


