/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "BufferedNetStream.hpp"

#include "nmm/utils/net/Sock.hpp"

#include <errno.h>       // for errnos
#include <sys/socket.h>  // for inet_ntoa
#include <sys/types.h>
#include <fstream>

namespace NMM {

  BufferedNetStream::BufferedNetStream(Sock& sock_stream) :
    NetStream(sock_stream),
    __send_buffer(111), __send_buffer_used(0),
    __recv_buffer(111), __recv_buffer_begin(0), __recv_buffer_end(0)
  {}  // Set the sender and receiver size to 111 (167 without magic numbers), because this is the size of a BufferHeader, serialized with a MagicNumberValueStream

  BufferedNetStream* BufferedNetStream::copy() const {
    return new BufferedNetStream(*this);
  }

  BufferedNetStream::~BufferedNetStream()
  {}

  void BufferedNetStream::flush() {
    if(__send_buffer_used > 0) {
      /* write the temp array */
      internalWriteValue(&__send_buffer[0], __send_buffer_used);
      __send_buffer_used = 0;
    }
  }

  void BufferedNetStream::writeData() {
    flush();
  }

  void BufferedNetStream::writeValue(const void* value, size_t size) {
    if(size >= __send_buffer.size()) {
      flush();
      internalWriteValue(value, size);
    } else {
      /* if not enough space is left flush the queue */
      if((__send_buffer.size() - __send_buffer_used) < size) {
        flush();
      }
      memcpy(&__send_buffer[__send_buffer_used], value, size);
      __send_buffer_used += size;
    }
  }

  void BufferedNetStream::internalWriteValue(const void* value, size_t size) {
    /* write bytes to the socket */
    ssize_t retval = getSock().send(static_cast<const char*>(value), size, MSG_NOSIGNAL);
    if(retval == 0) {
      throw ConnectionClosed(EPIPE);
    }

    /* count bytes */
    countWriteBytes(retval);

    /* if not all values are written, write the rest */
    if(static_cast<size_t>(retval) != size) {
      const char* new_offset = static_cast<const char*>(value) + retval;
      internalWriteValue(new_offset, size - retval);
    }
  }
  
  void BufferedNetStream::readValue(void* value, size_t size) {
    while (size > 0) {
      /* get number of bytes in the buffer */
      unsigned used = __recv_buffer_end - __recv_buffer_begin;
      
      if (used >= size) {
        /* copy bytes from the buffer to the value */
        memcpy(value, &__recv_buffer[__recv_buffer_begin], size);
        __recv_buffer_begin += size;
        break;
      }
      
      if (used > 0) {
        /* copy all bytes from the buffer to the value */
        memcpy(value, &__recv_buffer[__recv_buffer_begin], used);
        size -= used;
        value = static_cast<char*>(value) + used;
      }
      
      /* buffer now empty */
      __recv_buffer_begin = 0;
      __recv_buffer_end = 0;
      
      if (size > __recv_buffer.size()) {
        /* read data directly to avoid unnecessary copying */
        ssize_t retval = internalReadValue(value, size);
        value = static_cast<char*>(value) + retval;
        size -= retval;
      } else {
        /* read new data into the buffer */
        __recv_buffer_end = internalReadValue(&__recv_buffer[__recv_buffer_begin], __recv_buffer.size());
      }
    }
  }
  
  size_t BufferedNetStream::internalReadValue(void* value, size_t size) {
    /* read bytes from the socket */
    ssize_t retval = getSock().recv(static_cast<char*>(value), size, MSG_NOSIGNAL);
    if(retval == 0) {
      throw ConnectionClosed(EPIPE);
    }
    
    /* count bytes */
    countReadBytes(retval);
    
    return retval;
  }

  bool BufferedNetStream::isEmpty() const {
    if ( (__recv_buffer_end - __recv_buffer_begin) == 0 )
      return true;

    return false;
  }

  void BufferedNetStream::setSendBuffer(size_t size) {
    /* If more data is cached then the new buffer_size allows, we flush the NetStream*/
    if (__send_buffer_used > size) {
      flush();
    }

    __send_buffer.resize(size);
  }

  void BufferedNetStream::setReceiveBuffer(size_t size) {
    if ( ((__recv_buffer_begin != 0) || (__recv_buffer_end != 0))
         && __recv_buffer.size() < size )
    {
      GLOBAL_ERROR_STREAM("Could not resize receiving buffer, because it still stores data" << endl);
      return;
    }

    __recv_buffer.resize(size);
  }
}
