/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */



#include "SourceController.hpp"
#include "GenericSyncSourceNode.hpp"

namespace NMM {
  GenericSyncSourceNode::GenericSyncSourceNode(const char* name,
					   const StreamQueue::Mode queue_mode,
					   const int size)
    : GenericSourceNode(name, queue_mode, size)
  {
    setSynchronized(false);
    setSyncMaster(false);
    setConstantRateProducer(true);
    controller = new SourceController(this, IController::ONLINE);

    last_buffer = NULL;
    valid_bytes_send = 0;
    buf_count = 0;
    time_skew = Interval();
  };

  GenericSyncSourceNode::~GenericSyncSourceNode() 
  {
    if(controller != 0) {
      delete controller;
    }
  };

  Result GenericSyncSourceNode::setSynchronized(const bool& toggle) 
  { 
    synchronized = toggle; 
    return SUCCESS;
  };

  bool GenericSyncSourceNode::isSynchronized()
  {
    return synchronized;
  }

  void GenericSyncSourceNode::setSyncMaster(const bool& master)
  {
    this->sync_master = master;
  }

  bool GenericSyncSourceNode::isSyncMaster()
  {
    return sync_master;
  }

  void GenericSyncSourceNode::setConstantRateProducer(const bool& const_rate)
  {
    const_rate_producer = const_rate;
  }

  bool GenericSyncSourceNode::isConstantRateProducer()
  {
    return const_rate_producer;
  }
  
  Result GenericSyncSourceNode::syncEnable() 
  {
    setSynchronized(true);

    DEBUG_STREAM2("sync_enable completed." << endl);
    return SUCCESS;
  };

  Result GenericSyncSourceNode::syncDisable() 
  {
    setSynchronized(false);
    
    DEBUG_STREAM2("sync_disable completed." << endl);
    return SUCCESS;
  };

  Result GenericSyncSourceNode::syncReset()
  {
    resetTime();
    return SUCCESS;
  }

  ISourceController* GenericSyncSourceNode::getController() {
    return controller->getCheckedInterface<ISourceController>();
  }

  void GenericSyncSourceNode::resetTime() {
    MutexGuard mg_reset(reset_mutex);
    valid_bytes_send = 0;
  }

  Message* GenericSyncSourceNode::processBuffer(Buffer*)
  {
    // get buffer from virtual procduceBuffer,
    // expects a valid timestamp on the buffer
    Message* message = NULL;
    bool using_last_buffer = false;
    if (getWorkingFlag()) {
      setWorkingFlag(false);
      // the refcount doesn't have to be decreased again because this
      // buffer gets to the graph twice and therefore will be released
      // twice at the sink
      message = last_buffer;
      using_last_buffer = true;
    } else {
      message = produceBuffer(NULL);
      if (NULL == message) return NULL;
    }

    // If we got an event, just return it.
    if (message->getType() != Message::BUFFER) {
      return message;
    }

    Buffer* buffer = static_cast<Buffer*>(message);

    if (synchronized) {
      // If the node is synchronized AND produces its data at a constant rate,
      // check all buffers if they are valid, or too much too early or too late.
      if (const_rate_producer) {
	// getLocalTime is called with the number of bytes send BEFORE this buffer.
	// This is because the time of the first buffer shall be 0 and not 1/recording_rate
	// (e.g. 1/recording_rate = 40ms for a PAL TV grabber card)
	time_skew = getLocalTime(valid_bytes_send) - buffer->getTimestamp().sync_time;
 	if (!isSyncMaster() && buf_count % 25 == 1 ||
	    isSyncMaster() && buf_count % 2500 == 1) DEBUG_STREAM("time_skew is: " << time_skew << endl);
	controller->setParentSkew(time_skew);
		
	if(!controller->isBufferValid(buffer->getTimestamp())) {
	  DEBUG_STREAM2("Release Buffer " 
			<< buffer->getTimestamp().stream_counter  
			<< endl);
	  buffer->release();
	  return 0;
	}
	
	if (controller->isBufferLate()) {
	  setWorkingFlag(true);
	  last_buffer = buffer;
	  // Increase the refcount of the buffer because it is remembered locally and
	  // will be used a second time
	  last_buffer->increaseCount();
	}
	
	valid_bytes_send += buffer->getUsed();
	buf_count++;
      }
      else { // not a constant rate producer
	DEBUG_STREAM("correcting Timestamp" << endl);
	controller->correctTimestamp(buffer);
      }
    }
    return buffer;
  }
}
