/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include <iostream>
#include "Rational.hpp"
#include "nmm/base/serialize/Value.hpp"

namespace NMM {

  RationalDivisionByZeroException::RationalDivisionByZeroException() {
    cerr << "Rational - Division by zero." << endl;
  }

  int64_t Rational::greatestCommonDivisor(int64_t a, int64_t b) 
  {
    if(a == 0 || b == 0) return 0;  // gcd(0, x) = gcd(x, 0) = 0

    if(a < 0) {      // make sure that the two integers are non-negative
      a = -a;
    }

    if(b < 0) {
      b = -b;
    }

    int64_t t = 1;       // and perform Euklid's algorithm
    while(t != 0) {
      t = a % b;
      a = b;
      b = t;
    }

    return a;
  };

  int64_t Rational::leastCommonMultiple(int64_t a, int64_t b)
  {
    if(a == 0 || b == 0) return 0;

    int64_t lcm = a / greatestCommonDivisor(a, b) * b;
    if(lcm < 0) {
      return -lcm;
    }
    return lcm;
  };

  void Rational::reduce() 
  {
    if(numerator == 0) {
      denominator = 1;
      return;
    }

    int64_t gcd = greatestCommonDivisor(numerator, denominator);
    numerator   /= gcd;
    denominator /= gcd;
  };

  void Rational::invert()
    throw (const RationalDivisionByZeroException&)
  {
    if (numerator == 0) throw RationalDivisionByZeroException();

    int64_t a = numerator;
    numerator = denominator;
    denominator = a;
  };

  Rational Rational::getReciprocal()
    throw (const RationalDivisionByZeroException&)
  {
    if (numerator == 0) throw RationalDivisionByZeroException();

    return Rational(denominator, numerator);
  };

  Rational::Rational(const int64_t& numerator, const int64_t& denominator)
    throw (const RationalDivisionByZeroException&) 
  {
    if(denominator == 0) {
      this->numerator = 0;    // although an exception is thrown,
      this->denominator = 1;  // set the fraction to the value zero.
      throw RationalDivisionByZeroException();
    } else {
      if(denominator < 0) {                // the numerator should carry the sign 
	this->denominator = - denominator; // of the fraction, so the denominator
	this->numerator   = - numerator;   // is made positive
      } else {
	this->numerator   = numerator;
	this->denominator = denominator;
      }
    }
  };

  Rational& Rational::operator- ()
  {
    numerator = -numerator;

    return *this;
  };

  Rational& Rational::operator+=(const Rational& b)
  {
    numerator = numerator * b.denominator + b.numerator * denominator;
    denominator *= b.denominator;
    reduce();

    return *this;
  };

  Rational& Rational::operator-=(const Rational& b)
  {
    numerator = numerator * b.denominator - b.numerator * denominator;
    denominator *= b.denominator;
    reduce();

    return *this;
  };

  Rational& Rational::operator*=(const Rational& b)
  {
    numerator   *= b.numerator;
    denominator *= b.denominator;
    reduce();

    return *this;
  };

  Rational& Rational::operator/=(const Rational& b)
    throw (const RationalDivisionByZeroException&)
  {
    if(b.numerator == 0) throw RationalDivisionByZeroException();

    numerator   *= b.denominator;
    denominator *= b.numerator;
    if(denominator < 0) {
      denominator = -denominator;
      numerator   = -numerator;
    }
    reduce();

    return *this;
  };

  Rational operator+ (const Rational& a, const Rational& b)
  {
    Rational sum = a;
    return sum += b;
  };

  Rational operator- (const Rational& a, const Rational& b)
  {
    Rational diff = a;
    return diff -= b;
  };

  Rational operator* (const Rational& a, const Rational& b)
  {
    Rational prod = a;
    return prod *= b;
  };

  Rational operator/ (const Rational& a, const Rational& b)
  {
    Rational quot = a;
    return quot /= b;
  };

  ostream& operator<< (ostream& s, const Rational& a)
  {
    return s << a.numerator << " / " << a.denominator;
  }

  OValueStream& operator<<(OValueStream& s, const Rational& r) {
    s.beginType("Rational", "NMM");
    s << r.getNumerator();
    s << r.getDenominator();
    s.endType();
    return s;
  }

  IValueStream& operator>>(IValueStream& s, Rational& r) {
    /* Begin> Error Handling */
    if (s.getCurrentTypeInfo() != TypeInfo("Rational", "NMM") ){
      GLOBAL_ERROR_STREAM ("Invalid type or namespace:" << endl);
      GLOBAL_ERROR_STREAM ("Current type : " << s.getCurrentType() << "                Expected type: Rational" << endl);
      GLOBAL_ERROR_STREAM ("Current namespace : " << s.getCurrentNamespace() << "      Expected namespace: NMM " << endl);
      s.nextElement();
      return s;
    }

    /* Access the internal attributes */
    if (!s.goDown()) {
      s.nextElement();
      return s;
    }
    /* End< Error Handling */

    /* set values */
    s >> r.numerator ;
    s >> r.denominator;

    /* Leave the formats attributes */
    s.goUp();

    return s;
  }

  TValueConstructor<Rational> __rational_value("Rational", "NMM");
}
