
#ifndef NMM_ICompositeStrategy_HPP
#define NMM_ICompositeStrategy_HPP

#include "nmm/NMMTypes.hpp"
#include "nmm/base/Interface.hpp"
#include "nmm/base/Event.hpp"
#include <memory>

#include "ITransportStrategy.hpp"

namespace NMM
{
  /**
   *  @brief Interface of CompositeStrategy
   *
   *  This is the interface of the CompositeStrategy class.
   */

  class ICompositeStrategy : virtual public Interface { 
  public:    /**
     *  @brief Modes of operation
     */

  enum Mode {       /** Reliable buffer and event strategies.
       */

    MODE_RELIABLE_RELIABLE ,       /** Reliable buffer and unreliable event strategy.
       */

    MODE_RELIABLE_UNRELIABLE  };

    ICompositeStrategy() {  setTypeInfo(TypeInfo(ICompositeStrategy::ICompositeStrategy_type, "NMM")); };

    ICompositeStrategy(NMMObject& parent) : Interface(parent) {  setTypeInfo(TypeInfo(ICompositeStrategy::ICompositeStrategy_type, "NMM")); setEventReceiver(parent); };

    virtual ~ICompositeStrategy() {};

    static const char* ICompositeStrategy_type;

    virtual  string getInterfaceType() const { return ICompositeStrategy_type; }

    Interface* copy() const { return new ICompositeStrategy(*this); }
    static const char* getBufferChildStrategy_event;
    static const char* getEventChildStrategy_event;
    static const char* setMode_event;
    static const char* setLossInterval_event;
    static const char* connectToCompositeBuffer_event;
    static const char* connectToCompositeEvent_event;
    static const char* eventReceived_event;
    static Event* create_eventReceived(const int& stream_counter);
    /**
     *  @name Child strategy access
     */
//@{
    /**
     *  @brief Get buffer child strategy
     */

    virtual ITransportStrategy* getBufferChildStrategy() const
      throw(const Exception&);
    /**
     *  @brief Get event child strategy
     */

    virtual ITransportStrategy* getEventChildStrategy() const
      throw(const Exception&);
//@}
    /**
     *  @name Internal configuration events
     */
//@{
    /**
     *  @brief Set synchronization mode
     *
     *  Internal method for setting the synchronization mode. TEMPORARY!
     */

    virtual void setMode(const ICompositeStrategy::Mode& mode)
      throw(const Exception&);
    /**
     *  @brief Set data loss interval
     *
     *  This method sets the interval for simulating data loss in unreliable mode.
     *  It is only meant for benchmarks and should not be used. An interval of 0 disables data loss simulation.
     */

    virtual void setLossInterval(const unsigned int& loss_interval)
      throw(const Exception&);
    /**
     *  @brief Connect buffer sub strategy
     *  @exception BindingException if this event was not sent through a TransportStrategy
     *
     *  This event is sent by the CompositeStrategy that initiates a connection to connect
     *  the partner strategy of its buffer child strategy to its partner strategy.
     *  The receiver assumes that the sender of the event is the partner strategy of the sub strategy
     *  and sets its sub strategy accordingly.
     */

    virtual void connectToCompositeBuffer()
      throw(const Exception&, const BindingException&);
    /**
     *  @brief Connect event sub strategy
     *  @exception BindingException if this event was not sent through a TransportStrategy
     *
     *  This event is sent by the CompositeStrategy that initiates a connection to connect
     *  the partner strategy of its buffer child strategy to its partner strategy.
     *  The receiver assumes that the sender of the event is the partner strategy of the sub strategy
     *  and sets its sub strategy accordingly.
     */

    virtual void connectToCompositeEvent()
      throw(const Exception&, const BindingException&);
  /* instream (protected) and non-instream (public) */        /**
     *  @brief Event acknowledgement
     *  @param stream_counter Sequence number of the event
     *
     *  This event is sent from the receiver to the sender when a reliable event has been received.
     */

    virtual void eventReceived(const int& stream_counter) /* in-stream */ 
      throw(const Exception&);

//@}

  };

  typedef auto_ptr<ICompositeStrategy> ICompositeStrategy_var;

  OValueStream& operator<<(OValueStream& ov, const ICompositeStrategy::Mode& t);
  IValueStream& operator>>(IValueStream& iv, ICompositeStrategy::Mode& t);



}

#endif
