
#ifndef NMM_INode_HPP
#define NMM_INode_HPP

#include "nmm/NMMTypes.hpp"
#include "nmm/base/Interface.hpp"
#include "nmm/base/Event.hpp"
#include <memory>

#include "nmm/base/format/Format.hpp"
#include "IInputJack.hpp"
#include "IOutputJack.hpp"

namespace NMM
{

  class Property;

  class Format;

  class INode : virtual public Interface { 
  public:
  enum Type { 
    UNKNOWN = 0 , 
    SOURCE , 
    SINK , 
    FILTER , 
    CONVERTER , 
    MULTIPLEXER , 
    DEMULTIPLEXER , 
    MUX_DEMUX , 
    COMPOSITE  };

  enum State { 
    CONSTRUCTED = 1 , 
    INITIALIZED = 2 , 
    OUTPUT_INITIALIZED = 4 , 
    ACTIVATED = 8 , 
    STARTED = 16  };

    INode() {  setTypeInfo(TypeInfo(INode::INode_type, "NMM")); };

    INode(NMMObject& parent) : Interface(parent) {  setTypeInfo(TypeInfo(INode::INode_type, "NMM")); setEventReceiver(parent); };

    virtual ~INode() {};

    static const char* INode_type;

    virtual  string getInterfaceType() const { return INode_type; }

    Interface* copy() const { return new INode(*this); }
    static const char* reachState_event;
    static const char* reachConstructed_event;
    static const char* reachInitialized_event;
    static const char* reachOutputInitialized_event;
    static const char* reachActivated_event;
    static const char* reachStarted_event;
    static const char* setPrefetchQueueSize_event;
    static const char* getPrefetchQueueSize_event;
    static const char* waitPrefetchFinished_event;
    static const char* init_event;
    static const char* initOutput_event;
    static const char* activate_event;
    static const char* flush_event;
    static const char* start_event;
    static const char* stop_event;
    static const char* deactivate_event;
    static const char* deinitOutput_event;
    static const char* deinit_event;
    static const char* isConstructed_event;
    static const char* isInitialized_event;
    static const char* isOutputInitialized_event;
    static const char* isActivated_event;
    static const char* isStarted_event;
    static const char* getSuccessorNode_event;
    static const char* getPredecessorNode_event;
    static const char* getInputJack_event;
    static const char* getInputStream_event;
    static const char* getInputStreamTags_event;
    static const char* isInputConnected_event;
    static const char* getInputStreamFormat_event;
    static const char* getOutputJack_event;
    static const char* getUnintializedOutputJack_event;
    static const char* getOutputStream_event;
    static const char* getOutputStreamTags_event;
    static const char* isOutputConnected_event;
    static const char* cloneOutputJack_event;
    static const char* getOutputStreamFormat_event;
    static const char* getCost_event;
    static const char* getType_event;
    static const char* getState_event;
    static const char* getInputProperty_event;
    static const char* getWorkingInputProperty_event;
    static const char* getOutputProperty_event;
    static const char* getWorkingOutputProperty_event;
    static const char* getComputationConfig_event;
    static const char* getWorkingComputationConfig_event;
    static const char* setComputationConfig_event;
    static const char* setWorkingComputationConfig_event;
    static const char* getInternalCost_event;
    static const char* sendData_event;
    static Event* create_sendData();
    static const char* sendDataFailed_event;
    static Event* create_sendDataFailed();
    /**
     * This method will try to reach the speficied state by calling the
     * required state transition methods
     */

    virtual void reachState(const INode::State& state)
      throw(const Exception&, const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&);
    /**
     * This method tries to reach the state CONSTRUCTED
     */

    virtual void reachConstructed()
      throw(const Exception&, const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&);
    /**
     * This method tries to reach the state INITIALIZED
     */

    virtual void reachInitialized()
      throw(const Exception&, const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&);
    /**
     * This method tries to reach the state OUTPUT_INITIALIZED
     */

    virtual void reachOutputInitialized()
      throw(const Exception&, const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&);
    /**
     * This method tries to reach the state ACTIAVTED
     */

    virtual void reachActivated()
      throw(const Exception&, const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&);
    /**
     * This method tries to reach the state STARTED
     */

    virtual void reachStarted()
      throw(const Exception&, const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&);
    /**
     * This method enables the pre-fetch mode and at the same time sets
     * the number of messages to be received before pre-fetching can be
     * finished. This method has to be called in state CONSTRUCTED.
     */

    virtual void setPrefetchQueueSize(const unsigned int& pre_fetch_queue_size) /* limited(INode::CONSTRUCTED) */
      throw(const Exception&, const IllegalStateException&);
    /**
     * Request the number of messages to be prefetched. If zero,
     * prefetching is disabled.
     */

    virtual unsigned int getPrefetchQueueSize()
      throw(const Exception&);
    /**
     * Blocks until the process of prefetching is finished. This
     * method has to be called in state ACTIVATED.
     */

    virtual void waitPrefetchFinished() /* limited(INode::ACTIVATED) */
      throw(const Exception&, const IllegalStateException&);
    /**
     * This method initialize a node and the node enters state INITIALIZED.
     */

    virtual void init() /* limited(INode::CONSTRUCTED) */
      throw(const Exception&, const IllegalStateException&, const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&);
    /**
     * This method initialize a node and the node enters state OUPUT_INITIALIZED.
     */

    virtual void initOutput() /* limited(INode::INITIALIZED) */
      throw(const Exception&, const IllegalStateException&, const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&);
    /**
     * This method activates a node. 
     * The node enters an ACTIVATED state. This does not mean output
     * is generated from any input since the node's play method needs
     * to be called.
     */

    virtual void activate() /* limited(INode::OUTPUT_INITIALIZED) */
      throw(const Exception&, const IllegalStateException&, const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&);
    /**
     * This method causes the node to perform a flush-operation.
     */

    virtual Result flush() /* limited(INode::ACTIVATED) */ /* limited(INode::STARTED) */
      throw(const Exception&, const IllegalStateException&);
    /*
     * This method causes the node to start generating output
     * or to generate output from input.
     */

    virtual void start() /* limited(INode::ACTIVATED) */
      throw(const Exception&, const IllegalStateException&, const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&);
    /**
     * This method causes the node to stop generating output
     * or to no longer generate output from input.
     */

    virtual void stop() /* limited(INode::STARTED) */
      throw(const Exception&, const IllegalStateException&, const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&);
    /**
     * This method deactivates a node. 
     */

    virtual void deactivate() /* limited(INode::ACTIVATED) */
      throw(const Exception&, const IllegalStateException&, const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&);
    /**
     * This method deinits the output of a node. 
     */

    virtual void deinitOutput() /* limited(INode::OUTPUT_INITIALIZED) */
      throw(const Exception&, const IllegalStateException&, const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&);
    /**
     * This method deinits a node. 
     */

    virtual void deinit() /* limited(INode::INITIALIZED) */
      throw(const Exception&, const IllegalStateException&, const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&);

    virtual bool isConstructed() const
      throw(const Exception&);

    virtual bool isInitialized() const
      throw(const Exception&);

    virtual bool isOutputInitialized() const
      throw(const Exception&);

    virtual bool isActivated() const
      throw(const Exception&);

    virtual bool isStarted() const
      throw(const Exception&);
    /**
     *  This method returns the successor node of this node.
     *  \param out_jack the name of the output.
     *  \return a node or NULL.
     */

    virtual INode* getSuccessorNode(const string& out_jack)
      throw(const Exception&);
    /**
     * This method returns the predecessor node of this node.
     * \param in_jack the name of the input.
     * \return a node or NULL.
     */

    virtual INode* getPredecessorNode(const string& in_jack)
      throw(const Exception&);
    /**
     * This method returns an input jack with the given format
     * and jack tag.
     * \param format the format for the jack
     * \param jack_tag the name of the jack.
     * \return an input jack or 0 if no input jack with the given parameters could be obtained.
     */

    virtual IInputJack* getInputJack(const Format& format, const string& jack_tag) /* limited(INode::INITIALIZED) */
      throw(const Exception&, const IllegalStateException&);
    /**
     * This method returns a connected input jack with the given jack tag.
     * \param jack_tag the name of the jack.
     * \return an input jack or 0 if no input jack with the given parameter could be obtained.
     */

    virtual IInputJack* getInputStream(const string& jack_tag) /* limited(INode::INITIALIZED) */
      throw(const Exception&, const IllegalStateException&);
    /**
     * This method returns the list of available intput stream tags.
     * \return a list of input stream tags as strings.
     */

    virtual list<string >  getInputStreamTags() const
      throw(const Exception&);
    /**
     * This method can be used to test if a InputJack of a Node is connected.
     * \param jack_tag The name of the Jack to test.
     * \return true if the Jack is connected or false if not or the the Jack does not exist.
     */

    virtual bool isInputConnected(const string& jack_tag = "default") const
      throw(const Exception&);
    /**
     * Obtain the format associated with the named input stream.
     *
     * \param jack_tag The tag of the input stream.
     * \returns a pointer to the corresponding format.
     */

    virtual Format* getInputStreamFormat(const string& jack_tag = "default")
      throw(const Exception&);
    /**
     * This method returns an output jack with the given format
     * and jack tag.
     * \param format the format for the jack
     * \param jack_tag the name of the jack.
     * \return an output jack or 0 if no output jack with the given parameters could be obtained.
     */

    virtual IOutputJack* getOutputJack(const Format& format, const string& jack_tag) /* limited(INode::OUTPUT_INITIALIZED) */
      throw(const Exception&, const IllegalStateException&);
    /**
     * This method returns an output jack with the given format
     * and jack tag.
     * \param format the format for the jack
     * \param jack_tag the name of the jack.
     * \return an output jack or 0 if no output jack with the given parameters could be obtained.
     */

    virtual IOutputJack* getUnintializedOutputJack(const Format& format, const string& jack_tag) /* limited(INode::INITIALIZED) */
      throw(const Exception&, const IllegalStateException&);
    /**
     * This method returns a connected output jack with the given jack tag.
     * \param jack_tag the name of the jack.
     * \return an output jack or 0 if no output jack with the given parameter could be obtained.
     */

    virtual IOutputJack* getOutputStream(const string& jack_tag) /* limited(INode::OUTPUT_INITIALIZED) */
      throw(const Exception&, const IllegalStateException&);
    /**
     * This method returns the list of available output stream tags.
     * \return a list of output stream tags as strings.
     */

    virtual list<string >  getOutputStreamTags() const
      throw(const Exception&);
    /**
     * This method can be used to test if a OutputJack of a Node is connected.
     * \param jack_tag The name of the Jack to test.
     * \return true if the Jack is connected or false if not or the the Jack does not exist.
     */

    virtual bool isOutputConnected(const string& jack_tag = "default") const
      throw(const Exception&);
    /**
     * This method returns a cloned output jack and fills the given format with
     * the intersection format
     * \param jack_tag the name of the jack
     * \param intersec_format the format to be filled with the intersection format
     * \return the cloned output jack
     */

    virtual IOutputJack* cloneOutputJack(Format*& intersec_format, const string& jack_tag = "default") /* limited(INode::OUTPUT_INITIALIZED) */ /* limited(INode::ACTIVATED) */
      throw(const Exception&, const IllegalStateException&);
    /**
     * Obtain the format associated with the named output stream.
     *
     * \param jack_tag The tag of the output stream.
     * \returns a pointer to the corresponding format.
     */

    virtual Format* getOutputStreamFormat(const string& jack_tag = "default")
      throw(const Exception&);
    /**
     * This method returns the cost for the specified jack_tag.
     * \param jack_tag name of the jack.
     * \return the costs for this jack.
     **/

    virtual float getCost(const string& jack_tag) const
      throw(const Exception&);
    /**
     * Returns the type of the Node.
     *  \return the type of the node.
     **/

    virtual INode::Type getType() const
      throw(const Exception&);
    /** This method returns current state
     */

    virtual INode::State getState() const
      throw(const Exception&);
    /**
     * This method returns the InputProperty of the node
     * \return Property *
     */

    virtual Property* getInputProperty()
      throw(const Exception&);

    virtual Property* getWorkingInputProperty()
      throw(const Exception&);
    /**
     * This method returns the OutputProperty of the node
     * \return Property *
     */

    virtual Property* getOutputProperty()
      throw(const Exception&);

    virtual Property* getWorkingOutputProperty()
      throw(const Exception&);
     /**
      * This method returns the computation configuration of the node
      */

    virtual ComputationConfig* getComputationConfig() const
      throw(const Exception&);

    virtual ComputationConfig* getWorkingComputationConfig() const
      throw(const Exception&);
     /**
      * This method sets the computation configuration of the node
      */

    virtual void setComputationConfig(const ComputationConfig* cc)
      throw(const Exception&);

    virtual void setWorkingComputationConfig(const ComputationConfig* cc)
      throw(const Exception&);
    /**
     * Returns the internal cost for the node. These costs accrue for the basic
     * functionality of the node. The total costs for a node are the sum of the
     * internal cost and the cost of the input- and output-jacks.
     * return the internal cost of the node.
     **/

    virtual float getInternalCost() const
      throw(const Exception&);
  /* instream (protected) and non-instream (public) */        /**
     * this method signals instream, that a successing node requests
     * some data to do its "GenericNode::analyseData()".
     * \returns DELETE always.
     */

    virtual Result sendData() /* limited(INode::OUTPUT_INITIALIZED) */ /* limited(INode::ACTIVATED) */ /* limited(INode::STARTED) */ /* in-stream */ 
      throw(const Exception&, const IllegalStateException&);
        /**
     * this method signals instream, that a predecessing node could
     * not provide some data for "GenericNode::analyseData()"
     * \returns DELETE always.
     */

    virtual Result sendDataFailed() /* limited(INode::INITIALIZED) */ /* in-stream */ 
      throw(const Exception&, const IllegalStateException&);


  };

  typedef auto_ptr<INode> INode_var;

  OValueStream& operator<<(OValueStream& ov, const INode::Type& t);
  IValueStream& operator>>(IValueStream& iv, INode::Type& t);
  OValueStream& operator<<(OValueStream& ov, const INode::State& t);
  IValueStream& operator>>(IValueStream& iv, INode::State& t);



}

#endif
