
#ifndef NMM_IAudioHandoverSynchronizer_HPP
#define NMM_IAudioHandoverSynchronizer_HPP

#include "nmm/NMMTypes.hpp"
#include "nmm/base/Interface.hpp"
#include "nmm/base/Event.hpp"
#include <memory>

#include "ISinkSynchronizer.hpp"
#include "IMultiAudioSinkSynchronizer.hpp"
#include "ISwitch.hpp"

namespace NMM
{
  /**
   * Audio-audio sink synchronizer.
   * Interface for a synchronizer which performs inter-stream synchronization between
   * an audio and a seconed audio sink.
   */

  class IAudioHandoverSynchronizer : virtual public ISinkSynchronizer, virtual public IMultiAudioSinkSynchronizer { 
  public:
  enum Mode { 
    INSTREAM_SWITCH = 0 , 
    OUTOFBAND_SWITCH = 1 , 
    INSTREAM_FLUSHED_SWITCH = 2  };

  enum SinkId { 
    AUDIO_0_SINK = 0 , 
    AUDIO_1_SINK = 1  };

    IAudioHandoverSynchronizer() {  setTypeInfo(TypeInfo(IAudioHandoverSynchronizer::IAudioHandoverSynchronizer_type, "NMM")); };

    IAudioHandoverSynchronizer(NMMObject& parent) : ISinkSynchronizer(parent), IMultiAudioSinkSynchronizer(parent) {  setTypeInfo(TypeInfo(IAudioHandoverSynchronizer::IAudioHandoverSynchronizer_type, "NMM")); setEventReceiver(parent); };

    virtual ~IAudioHandoverSynchronizer() {};

    static const char* IAudioHandoverSynchronizer_type;

    virtual  string getInterfaceType() const { return IAudioHandoverSynchronizer_type; }

    Interface* copy() const { return new IAudioHandoverSynchronizer(*this); }
    static const char* setFirstSinkController_event;
    static const char* setSecondSinkController_event;
    static const char* switchControllers_event;
    static const char* setSkew_event;
    static const char* setSwitch_event;
    static const char* setMode_event;
    static const char* setOverlap_event;
    static const char* waitForSwitchControllers_event;
    /**
     * Set the audio sink controller.
     * This sets the controller for the audio sink.
     * @param controller controller for this sink
     */

    virtual Result setFirstSinkController(const ISinkController* controller)
      throw(const Exception&);
    /**
     * Set the second audio sink controller.
     * This sets the controller for the audio sink.
     * @param controller controller for this sink
     */

    virtual Result setSecondSinkController(const ISinkController* controller)
      throw(const Exception&);
    /**
     * Makes the slave (2nd) the master (1st) controller and vica versa
     */

    virtual void switchControllers()
      throw(const Exception&);
    /**
     * Sets a skew between audio0 and audio1 output.
     * The audio1 output is presented later than the audio0 output.
     */

    virtual Result setSkew(const Interval& skew)
      throw(const Exception&);
    /**
     * Sets the ISwitch interface of this synchronizer
     */

    virtual void setSwitch(const ISwitch* s)
      throw(const Exception&);
    /**
     * Sets the modus for handover: either
     * INSTREAM_SWITCH = ISwitch::switchOutputs()
     * OUTOFBAND_SWITCH = Mute/unmute method calls from synchronizer
     * INSTREAM_FLUSHED_SWITCH = ISwitch::flushOutputs(), ISwitch::switchOutputs()
     */

    virtual void setMode(const IAudioHandoverSynchronizer::Mode& mode)
      throw(const Exception&);
    /**
     * Sets the overlapping interval during which master and
     * slave perform playback during handover (only possible
     * for modus 1)
     */

    virtual void setOverlap(const Interval& overlap)
      throw(const Exception&);
    /**
     * Blocking method that waits until handover is complete, 
     * must be called after adding slave controller with addAudioSinkController
     */

    virtual void waitForSwitchControllers()
      throw(const Exception&);


  };

  typedef auto_ptr<IAudioHandoverSynchronizer> IAudioHandoverSynchronizer_var;

  OValueStream& operator<<(OValueStream& ov, const IAudioHandoverSynchronizer::Mode& t);
  IValueStream& operator>>(IValueStream& iv, IAudioHandoverSynchronizer::Mode& t);
  OValueStream& operator<<(OValueStream& ov, const IAudioHandoverSynchronizer::SinkId& t);
  IValueStream& operator>>(IValueStream& iv, IAudioHandoverSynchronizer::SinkId& t);



}

#endif
