
#ifndef NMM_ITVGuide_HPP
#define NMM_ITVGuide_HPP

#include "nmm/NMMTypes.hpp"
#include "nmm/base/Interface.hpp"
#include "nmm/base/Event.hpp"
#include <memory>

#include "ChannelDescription.hpp"
#include "ProgramDescription.hpp"
#include "ProgramFilter.hpp"
#include "nmm/interfaces/misc/tvguide/ITVGuidePlugin.hpp"
#include "nmm/interfaces/misc/tvguide/ITVGuidePeerObject.hpp"

namespace NMM
{

  class ITVGuide : virtual public Interface { 
  public:
    ITVGuide() {  setTypeInfo(TypeInfo(ITVGuide::ITVGuide_type, "NMM")); };

    ITVGuide(NMMObject& parent) : Interface(parent) {  setTypeInfo(TypeInfo(ITVGuide::ITVGuide_type, "NMM")); setEventReceiver(parent); };

    virtual ~ITVGuide() {};

    static const char* ITVGuide_type;

    virtual  string getInterfaceType() const { return ITVGuide_type; }

    Interface* copy() const { return new ITVGuide(*this); }
    static const char* insertChannel_event;
    static const char* insertChannelDescription_event;
    static const char* insertEvent_event;
    static const char* setDirectory_event;
    static const char* getDirectory_event;
    static const char* getChannelsList_event;
    static const char* getChannelById_event;
    static const char* getChannelByName_event;
    static const char* getEvent_event;
    static const char* getEventsOnDayOnChannel_event;
    static const char* search_event;
    static const char* addExtDescription_event;
    static const char* getCurrentEventOnChannel_event;
    static const char* getFollowingEventOnChannel_event;
    static const char* getCategories_event;
    static const char* start_event;
    static const char* stop_event;
    static const char* init_event;
    static const char* deinit_event;
    static const char* getGrabberInterface_event;
    static const char* addPeerHost_event;
    static const char* clearPeerHosts_event;
    static const char* cleanDatabase_event;
    /**
     * Adds a channel to the channel-list
     *\param id ID of the channel wich is used to find show of this channel
     *\param name Name of the channel
     */

    virtual Result insertChannel(const string& id = "", const string& name = "")
      throw(const Exception&);
    /**
     * Adds a channel to the channel-list
     *\param cd ChannelDescription to be added
     */

    virtual Result insertChannelDescription(const ChannelDescription& cd)
      throw(const Exception&);
   /**
     * Adds an Event to the database
     *\param pd PrgoramDescription, containing the inrformation about the event
     */

    virtual Result insertEvent(const ProgramDescription& pd)
      throw(const Exception&);
    /**
     * Sets the directory, where the XML-Files are located/stored
     *\param path path to XML-Files
     */

    virtual void setDirectory(const string& path = "")
      throw(const Exception&);
    /**
     * returns the path, where the XML-Files are stored
     *\return Path to XML-Files
     */

    virtual string getDirectory()
      throw(const Exception&);
    /**
     * Returns the Channels, for wich informaion is available
     *\return Channel-list
     */

    virtual list<ChannelDescription >  getChannelsList()
      throw(const Exception&);
    /**
     * Returns the channel with the given ID
     *\param id ID of the channel
     *\return name of the channel, if available, empty string otherwise
     */

    virtual string getChannelById(const string& id)
      throw(const Exception&);
    /**
     *Return the ID of the Channel with the given name
     *\param name name o the channel
     *\return ID of the requested channel, empty string if not found
     */

    virtual string getChannelByName(const string& name)
      throw(const Exception&);
    /**
     *Returns the Event with the given starttime, endtime and channel-id
     *\param start StartTime of the Event
     *\param end EndTime of the Event
     *\param id channel-id of the event
     *\return ProgramDescription containg the information about the event
     */

    virtual ProgramDescription getEvent(const string& start, const string& end, const string& id)
      throw(const Exception&);
    /**
     *Returns all show on the specified day and channel
     *\param day date of the day
     *\param chan channel
     *\return List of ProgramDescripytions containing the information of the shows
     */

    virtual list<ProgramDescription >  getEventsOnDayOnChannel(const string& day, const string& chan)
      throw(const Exception&);
    /**
     * Returns a List with Events containing the given search-phrase
     *\param phrase Phrase to search for
     *\param peer Interface to the application the it is notified of new events
     *\return Event-Hitlist
     */

    virtual list<ProgramDescription >  search(const ProgramFilter& pf, const ITVGuidePeerObject* peer = NULL)
      throw(const Exception&);
    /**
     * adds a ExtendedDescription to a event (needed, when source is DVB)
     */

    virtual void addExtDescription(const string& start, const string& end, const string& id, const string& desc)
      throw(const Exception&);
    /**
     * returns the current event on the specified channel
     *\param channel channel_ID
     *@return current event in the specified channel
     */

    virtual ProgramDescription getCurrentEventOnChannel(const string& channel)
      throw(const Exception&);
    /**
     * returns the next to current event on the specified channel
     *\param channel channel_ID
     *@return next event in the specified channel
     */

    virtual ProgramDescription getFollowingEventOnChannel(const string& channel)
      throw(const Exception&);
    /** 
     * returns a list with all categories, stored in the DB
     *@return strin-list with categories
     */

    virtual list<string >  getCategories()
      throw(const Exception&);
    /**
     * this method starts the object and starts all relateed plugins
     */

    virtual Result start()
      throw(const Exception&);
    /**
     * stops all related plugins
     */

    virtual Result stop()
      throw(const Exception&);
    /**
     * initialized the requested plugins. Plugins must be rewuestetd before using 'getGrabberInterface'
     */

    virtual Result init()
      throw(const Exception&);
    /**
     * deinitializes all requested plugins
     */

    virtual Result deinit()
      throw(const Exception&);
    /**
     * requests the TVGuide-Plugin-Interface specified in type
     *@return ITVGuideUpdatePlugin* interface of the requested grabber
     *@param type Type of grabber, wich is its plugin-name, eg IXMLTVPlugin
     */

    virtual ITVGuidePlugin* getGrabberInterface(const string& type)
      throw(const Exception&);
    /**
       add a Host to query, when database-Information is incomplete
     */

    virtual void addPeerHost(const string& host, const int& port)
      throw(const Exception&);
    /**
       clear the list of peer-hosts
     */

    virtual void clearPeerHosts()
      throw(const Exception&);
    /**
       clean up the database. specify the amount of days from today wich data should be kept. older data will be erased
    */

    virtual void cleanDatabase(const int& days)
      throw(const Exception&);


  };

  typedef auto_ptr<ITVGuide> ITVGuide_var;




}

#endif
