
#ifndef NMM_ITextView_HPP
#define NMM_ITextView_HPP

#include "nmm/NMMTypes.hpp"
#include "nmm/base/Interface.hpp"
#include "nmm/base/Event.hpp"
#include <memory>

#include "IWidget.hpp"

namespace NMM
{
  /**
   * Text view widget.
   * Interface for a text view widget.
   */

  class ITextView : virtual public IWidget { 
  public:
    ITextView() {  setTypeInfo(TypeInfo(ITextView::ITextView_type, "NMM")); };

    ITextView(NMMObject& parent) : IWidget(parent) {  setTypeInfo(TypeInfo(ITextView::ITextView_type, "NMM")); setEventReceiver(parent); };

    virtual ~ITextView() {};

    static const char* ITextView_type;

    virtual  string getInterfaceType() const { return ITextView_type; }

    Interface* copy() const { return new ITextView(*this); }
    static const char* setFont_event;
    static const char* selectLine_event;
    static const char* selectLine2_event;
    static const char* deSelectLine_event;
    static const char* deSelectLine2_event;
    static const char* getSelectedLine_event;
    static const char* getSelectedLine2_event;
    static const char* getLineContent_event;
    static const char* changeLineContent_event;
    static const char* getLineURL_event;
    static const char* getLineWidget_event;
    static const char* setAutoFocus_event;
    static const char* setSelectionColor_event;
    static const char* setSelectionColor2_event;
    static const char* reset_event;
    static const char* setMaxLines_event;
    static const char* getMaxLines_event;
    static const char* setStartLine_event;
    static const char* getStartLine_event;
    static const char* getNumLines_event;
    static const char* focusSelection_event;
    static const char* insertLine_event;
    static const char* addLine_event;
    static const char* addLineWidget_event;
    static const char* deleteLine_event;
    static const char* deleteAllLines_event;
    static const char* markLine_event;
    static const char* unmarkLine_event;
    static const char* getMarkedLines_event;
    /**
     * Set font.
     * Set the font all text on this widget is drawn in.
     * @param value font name
     */

    virtual void setFont(const string& value)
      throw(const Exception&);
    /**
     * Selects the given line.
     * @param line line that should be selected (0 if no line should be selected)
     * @return linenumber that was selected
     */

    virtual unsigned int selectLine(const unsigned int& line)
      throw(const Exception&);
    /**
     * Selects the given line (second selection).
     * @param line line that should be selected (0 if no line should be selected)
     * @return linenumber that was selected
     */

    virtual unsigned int selectLine2(const unsigned int& line)
      throw(const Exception&);
    /**
     * Deselect line ( set selected_line = 0).
     */

    virtual void deSelectLine()
      throw(const Exception&);
    /**
     * Deselect second line ( set selected_line2 = 0).
     */

    virtual void deSelectLine2()
      throw(const Exception&);
    /**
     * Get selected line.
     * @return selected line (0, if no line is selected)
     */

    virtual unsigned int getSelectedLine() const
      throw(const Exception&);
    /**
     * Get selected line (second selection).
     * @return second selected line (0, if no line is selected)
     */

    virtual unsigned int getSelectedLine2() const
      throw(const Exception&);
    /**
     * Get line content.
     * @param line line number
     * @returns the text displayed in that line
     */

    virtual string getLineContent(const unsigned int& line) const
      throw(const Exception&);
    /**
     * changes the content of a line
     * @param text new content of the line
     * @param line to be changed
     */

    virtual void changeLineContent(const unsigned int& line, const string& text)
      throw(const Exception&);
    /**
     * Get line url.
     * @param line line number
     * @returns the url in that line, if it contains one.
     */

    virtual string getLineURL(const unsigned int& line) const
      throw(const Exception&);
    /**
     * Get line widget.
     * @param line line number
     * @returns the widget in that line, if it contains one.
     */

    virtual IWidget* getLineWidget(const unsigned int& line) const
      throw(const Exception&);
    /**
     * Set autofocus for selectLine.
     * @param value autofocus mode. on=true or off=false.
     */

    virtual void setAutoFocus(const bool& value)
      throw(const Exception&);
    /**
     * Sets selection color.
     * @param color rgba color
     */

    virtual void setSelectionColor(const u_int32_t& color)
      throw(const Exception&);
    /**
     * Sets selection color for second highlight.
     * @param color rgba color
     */

    virtual void setSelectionColor2(const u_int32_t& color)
      throw(const Exception&);
    /**
     * Sets standard selection color, deselects line.
     */

    virtual void reset()
      throw(const Exception&);
    /** Sets maximum number of visible lines. */

    virtual void setMaxLines(const unsigned int& value)
      throw(const Exception&);
    /** Gets maximum number of visible lines. */

    virtual unsigned int getMaxLines() const
      throw(const Exception&);
    /**
     * Sets first visible line.
     * @return first visible line
     */

    virtual unsigned int setStartLine(const unsigned int& start)
      throw(const Exception&);
    /** Get first visible line. */

    virtual unsigned int getStartLine() const
      throw(const Exception&);
    /** Get number of lines. */

    virtual unsigned int getNumLines() const
      throw(const Exception&);
    /**
     * Focus selection.
     * Adjusts first visible line so that the selected line is visible
     */

    virtual void focusSelection()
      throw(const Exception&);
    /**
     * Insert a line with content.
     * @param pos position where the new line should be inserted.
     * @param text line text
     * @param url url associated with this line
     */

    virtual void insertLine(const int& pos, const string& text, const string& url = "")
      throw(const Exception&);
    /**
     * Adds a line with content.
     * @param pos position where the new line should be inserted.
     * @param text line text
     * @param url url associated with this line
     * @return line number of the new line
     */

    virtual int addLine(const string& text, const string& url = "")
      throw(const Exception&);
    /**
     * Set line widget
     * Add a widget to a line (such as another text view).
     */

    virtual void addLineWidget(const unsigned int& line, const IWidget& widget)
      throw(const Exception&);
    /** Delete the given line. */

    virtual void deleteLine(const unsigned int& line)
      throw(const Exception&);
    /** Delete all lines. */

    virtual void deleteAllLines()
      throw(const Exception&);
    /** mark line (using selection_color_2)
     *@param line line to mark
     */

    virtual void markLine(const unsigned int& line)
      throw(const Exception&);
    /** unmark line
     *@param line line to unmark
     */

    virtual void unmarkLine(const unsigned int& line)
      throw(const Exception&);
    /** get marked line-numbers
     *@return list of line-numbers
     */

    virtual list<unsigned int >  getMarkedLines()
      throw(const Exception&);


  };

  typedef auto_ptr<ITextView> ITextView_var;




}

#endif
