
#ifndef NMM_ISPUDecoder_HPP
#define NMM_ISPUDecoder_HPP

#include "nmm/NMMTypes.hpp"
#include "nmm/base/Interface.hpp"
#include "nmm/base/Event.hpp"
#include <memory>


namespace NMM
{
  /**
   * Subtitle Picture Unit (SPU) decoder.
   * Interface that defines the events that control a SPU decoder. These events
   * are sent for example by the DVD reader.
   */

  class ISPUDecoder : virtual public Interface { 
  public:
    ISPUDecoder() {  setTypeInfo(TypeInfo(ISPUDecoder::ISPUDecoder_type, "NMM")); };

    ISPUDecoder(NMMObject& parent) : Interface(parent) {  setTypeInfo(TypeInfo(ISPUDecoder::ISPUDecoder_type, "NMM")); setEventReceiver(parent); };

    virtual ~ISPUDecoder() {};

    static const char* ISPUDecoder_type;

    virtual  string getInterfaceType() const { return ISPUDecoder_type; }

    Interface* copy() const { return new ISPUDecoder(*this); }
    static const char* setPaletteItem_event;
    static Event* create_setPaletteItem(const int& index, const int& color_y, const int& color_u, const int& color_v, const int& alpha);
    static const char* initColorScheme_event;
    static Event* create_initColorScheme(const int& scheme_nr, const int& selection, const int& activation);
    static const char* setDVDSelectedButton_event;
    static Event* create_setDVDSelectedButton(const int& button);
    static const char* setDVDButtonCount_event;
    static Event* create_setDVDButtonCount(const int& buttons);
    static const char* setDVDButtonCoords_event;
    static Event* create_setDVDButtonCoords(const int& start_x, const int& start_y, const int& end_x, const int& end_y);
    static const char* setDVDButtonScheme_event;
    static Event* create_setDVDButtonScheme(const int& scheme);
  /* instream (protected) and non-instream (public) */        /**
     * Instream event to set colorpalette item.
     * Assigns an RGBA color value to a color palette index.
     * @param index color palette index
     * @param color_y Y color value
     * @param color_u U color value
     * @param color_v V color value
     * @param alpha alpha value
     */

    virtual Result setPaletteItem(const int& index, const int& color_y, const int& color_u, const int& color_v, const int& alpha) /* in-stream */ 
      throw(const Exception&);
        /**
     * Instream event to init color sheme.
     * Initialize the colors of a color scheme. Assigns the colors (palette entries) for the
     * selection and activation state of a button that has this scheme.
     * @param scheme_nr index of color scheme
     * @param selection palette index of selection color
     * @param activation color palette index of activation color
     */

    virtual Result initColorScheme(const int& scheme_nr, const int& selection, const int& activation) /* in-stream */ 
      throw(const Exception&);
        /**
     * Instream event to set selected button.
     * Sent by the DVD reader if a new menu button was selected.
     * @param button the button number
     */

    virtual Result setDVDSelectedButton(const int& button) /* in-stream */ 
      throw(const Exception&);
        /**
     * Instream event to set the number of buttons.
     * Sent by the DVD reader to inform the SPU decoder about the number of buttons.
     * @param buttons number of buttons
     */

    virtual Result setDVDButtonCount(const int& buttons) /* in-stream */ 
      throw(const Exception&);
        /**
     * Instream event to set button coordinates.
     * Sent by the DVD reader to set the coordinates of the current button.
     * @param start_x left X coordinate of button
     * @param start_y upper Y coordinate of button
     * @param end_x right X coordinate of button
     * @param end_y lower Y coordinate of button
     */

    virtual Result setDVDButtonCoords(const int& start_x, const int& start_y, const int& end_x, const int& end_y) /* in-stream */ 
      throw(const Exception&);
        /**
     * Instream event to set button color scheme.
     * Sent by the DVD reader to set the color scheme of the current button.
     * @param scheme color scheme index
     */

    virtual Result setDVDButtonScheme(const int& scheme) /* in-stream */ 
      throw(const Exception&);


  };

  typedef auto_ptr<ISPUDecoder> ISPUDecoder_var;




}

#endif
