/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Christoph Wellner <chwellner@graphics.cs.uni-sb.de>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef TVCHANNELLIST_HPP
#define TVCHANNELLIST_HPP TVCHANNELLIST_HPP

#include "nmm/NMMTypes.hpp"
#include "nmm/base/serialize/ComplexValue.hpp"

#include <iostream>
#include <list>
#include <string>

namespace NMM{

  /**
   *Exception if a channel could not be found in the list
   */
  class ChannelNotFoundException : public Exception{
    NMM::TValue<string> channelName;
    NMM::TValue<unsigned int> channelNumber;
  public:
    ChannelNotFoundException();
    ChannelNotFoundException(string channel);
    ChannelNotFoundException(unsigned int channel);

    ChannelNotFoundException* copy() const { return new ChannelNotFoundException(*this); }

    /**
     *get the requested channel-number
     */
    const unsigned int& getChannelNumber(){return channelNumber.getRValue();}
    /**
       get the requested channel-name
    */
    const string& getChannelName(){return channelName.getRValue();}
  private:
    void registerAttributes();    
  };

  class TVChannel : public SerializedValue{

  public :
    typedef map<string, int> cardList_t;

  protected:
    NMM::TValue<string> channelName;
    NMM::TValue<unsigned int> channelNumber;
    NMM::TValue<list<string> > nameAliases;
    NMM::TValue<cardList_t> cards;

  public:
    TVChannel();
    TVChannel(string name);
    TVChannel(const TVChannel& t);

    ~TVChannel();

    void setChannelName(string name){channelName = name;}
    string getChannelName() const {return channelName.getRValue();}

    void setChannelNumber(unsigned int number){channelNumber = number;}
    unsigned int getChannelNumber() const {return channelNumber.getRValue();}

    void setChannelAliases(list<string> aliases){nameAliases = aliases;}
    list<string> getChannelAliases()const {return nameAliases.getRValue();}

    bool operator<(const TVChannel &cd) const;

    void addCard(string URL, unsigned int channelNumber);
    cardList_t getCards();
    
    string getName();
    list<string> getAliases();
    void addAlias(string a);

  private:
    void registerAttributes();
  };
  
  
  
  class TVChannelList : public SerializedValue {
  private:
    NMM::TValue<list<TVChannel> > channels;
    NMM::TValue<int> channelNumber; //internal only to number the channels increasing when addChannel() is called

  public:
    TVChannelList();
    TVChannelList(string filename);
    TVChannelList(const TVChannelList& t);

    ~TVChannelList();

    void addChannel(TVChannel cd);
    TVChannel getChannelByNumber(unsigned int number);
    TVChannel getChannelByName(string name);
    list<TVChannel> getChannels();

    Result load(string filename);
    Result write(string filename);

    void clear(){channels.getRValue().clear();}
    int size() {return channels.getRValue().size();}
    bool empty() {return channels.getRValue().empty();}

  private:
    void registerAttributes();

    void transform(string &s){
      size_t sIt = string::npos;
      size_t sOld = 0;
      do{
        sIt = s.find_first_of(" ", sOld);
        if(sIt != string::npos) s.erase(sIt,1);
        sOld = sIt+1;
      }while(sIt != string::npos);

      string sout = s;
      for (unsigned int i = 0; i <= s.length(); i++){
        sout[i] = toupper(s[i]);
      }
      s=sout;
    }

  };
}
#endif
