/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef _NMM_XVIRTUALIMG_HPP_
#define _NMM_XVIRTUALIMG_HPP_

#include "nmm/NMMTypes.hpp"

#include <string>

#include <X11/Xutil.h>

namespace NMM {
  
  class Node;
  class XProducer;
  class Buffer;
  class ShmBufferManager;
  
  /**
   *  @brief X11 image base class
   *
   *  This class provides common functionality for X11 display windows.
   */
  class XVirtualImg {
  public:
    /**
     *  @name Constructor and Destructor
     */
    //@{
    /**
     *  @brief Create an %XVirtualImg with the given properties
     *  @param parent The parent %Node
     *  @param width Width of the image in pixels
     *  @param width Height of the image in pixels
     *  @param title Window title
     *
     *  Fullscreen is set to @c false
     *  The aspect ratio is @c width : @c height
     */
    XVirtualImg(Node* parent, int width, int height, const std::string& title);

    /**
     *  @brief Create an %XVirtualImg with the given properties
     *  @param parent The parent %Node
     *  @param width Width of the image in pixels
     *  @param width Height of the image in pixels
     *  @param ratio_x Numerator of X:Y aspect ratio
     *  @param ratio_y Denominator of X:Y aspect ratio
     *  @param fullscreen @c true for fullscreen, @c false for windowed mode
     *  @param title Window title
     */
    XVirtualImg(Node* parent, int width, int height, int ratio_x, int ratio_y, bool fullscreen, const std::string& title);
    
    /**
     *  @brief Virtual destructor
     */
    virtual ~XVirtualImg();
    //@}
    
    /**
     *  @name Get/set attributes
     */
    //@{
    /**
     *  @brief Get the parent %Node of this image
     */
    Node* getParent() const throw();
    
    /**
     *  @brief Get the display ID of this image
     */
    Display* getDisplay() const throw();
    
    /**
     *  @brief Get the window ID of this image
     */
    Window getWindow() const throw();
    
    /**
     *  @brief Get the width of this image in pixels
     */
    int getWidth() const throw();
    
    /**
     *  @brief Get the height of this image in pixels
     */
    int getHeight() const throw();
    
    /**
     *  @brief Get gra[hics context of this image 
     */
    GC getContext() const throw();
    
    /**
     *  @brief Get the Y blit offset
     */
    int getYBlitOffset() const throw();

    /**
     *  @brief Get the Y blit offset
     */
    int getXBlitOffset() const throw();
    
    /**
     *  @brief Get the screen width in pixels
     */
    int getScreenWidth() const throw();
    
    /**
     *  @brief Get the screen height in pixels
     */
    int getScreenHeight() const throw();
    //@}
    
    /**
     *  @name Window management
     */
    //@{
    /**
     *  @brief Initialize window
     *
     *  This method does all the initialization of this image. It also invokes initialization methods
     *  that are implemented in subclasses.
     */
    void init();
    
    /**
     *  @brief Deinitialize window
     *
     *  This method does all the deinitialization of this image. It also invokes deinitialization methods
     *  that are implemented in subclasses.
     */
    void deinit();
    
    /**
     *  @brief Set image size
     *  @param width New width of this image in pixels
     *  @param height New height of this image in pixels
     *  @return @c SUCCESS, if this method succeeds, otherwise @c FAILURE
     *
     *  This method changes the size of this image.
     */
    Result setSize(int width, int height);
    
    /**
     *  @brief Set window title
     *  @param title The new title of this window
     */
    void setTitle(const std::string& title);
    
    /**
     *  @brief Set X:Y aspect ratio
     *  @param ratio_x numerator of aspect ratio
     *  @param ratio_y denominator aspect ratio
     */
    void setRatio(int ratio_x, int ratio_y);
    
    /**
     *  @brief Set fullscreen mode
     *  @param fullscreen @c true for fullscreen, @c false for windowed mode
     */
    void setFullscreen(bool fullscreen);
    //@}

    void setAlignment(int i);
    /**
     *  @name Cursor functions
     */
    //@{
    /**
     *  @brief Use a dot-shaped cursor
     */
    void useDotCursor(bool use_dot_cursor);
    
    /**
     *  @brief Use a triangle-shaped cursor
     */
    void useTriangleCursor(bool use_triangle_cursor);
    
    /**
     *  @brief Use the standard cursor
     */
    void useStandardCursor();
    //@}
    
    /**
     *  @brief %Buffer management and display
     */
    //@{
    /**
     *  @brief Update the shared memort buffer manager
     *  @param manager The shared memory buffer manager
     *  @param shared_pages Number of shared pages
     */
    virtual void updateMemoryManager(ShmBufferManager* manager, unsigned int shared_pages);
    
    /**
     *  @brief Prepare display of a buffer
     *  @param buffer The buffer to display
     */
    virtual void prepareOutput(Buffer* buffer) = 0;
    
    /**
     *  @brief Display a buffer on the image
     *  @param buffer The buffer to display
     *
     *  The buffer must have been prepared by calling prepareOutput with the same @c buffer argument.
     */
    virtual void update(Buffer*) = 0;
    
    /**
     *  @brief Get pointer to image data
     */
    virtual char* getImage() = 0;
    //@}
    
    /**
     *  @name Event production
     */
    //@{
    /**
     *  @brief Enable display events
     *  @param config_file Filename of display event config file
     *
     *  This method enables the %XProducer to send display events. The events sent in response to
     *  keys pressed are defined in the XML configuration file given by @c config_file
     */
    void enableDisplayEvents(const std::string& config_file);
    
    /**
     *  @brief Disable display events
     *
     *  This method disables sending of display events
     */
    void disableDisplayEvents();
    
    /**
     *  @brief Set whether to send key events directly without using the %XProducer
     *  @param direct_keyboard @c true to send key events directly, @c false to use the %XProducer
     */
    void setUsingKeySymKeyEvents(const bool& direct_keyboard);
    
    /**
     *  @brief Check whether key events are sent directly
     *  @return @c true if key events are sent directly, otherwise @c false
     */
    bool isUsingKeySymKeyEvents();
    
    /**
     *  @brief Set whether to send mouse events directly
     *  @param direct_keyboard @c true to send key events directly, @c false for normal mode
     */
    void setSendingMouseEventsDirect(const bool& direct_mouse);

    /**
     *  @brief Check whether mouse events are sent directly
     *  @return @c true if mouse events are sent directly, otherwise @c false
     */
    bool isSendingMouseEventsDirect();
    
    /**
     *  @brief Set X auto-repeat mode for keys
     *  @param auto_repeat @c true to enable, @c false to disable auto-repeat
     *
     *  If auto-repeat mode is active (which is the default), key events are repeated if a key is being held.
     *  Otherwise key events are sent only when the key is pressed and released.
     */
    void setXAutoRepeat(const bool& on);
    
    /**
     *  @brief Check for X auto-repeat mode
     *  @return @c true if auto-repeat mode is active, otherwise @c false
     */
    bool isXAutoRepeat();
    
    /**
     *  @brief Get default event mask
     */
    static long getEventMask();
    //@}
    
  protected:
    /**
     *  @brief Initialize visual
     *
     *  This method sets the visual information by calling the setVisualInfo method. The default implementation
     *  uses @c XMatchVisualInfo to find a matching visual. Subclasses can override this behavious by reimplementing
     *  this method.
     */
    virtual void initVisual();
    
    /**
     *  @brief Deinitialize visual
     *
     *  This method is called when the visual information set by @c initVisual is no longer needed.
     *  Subclasses can implement this method to free the @c XVisualInfo structure and any related resources.
     */
    virtual void deinitVisual();

    /**
     *  @brief Create the graphics context
     *
     *  This method sets the graphics context by calling the setContext method. The default implementation
     *  creates a standard graphics context. If any other kind of graphics context should be used in a subclass,
     *  then it must reimplement this method.
     */
    virtual void initContext();
    
    /**
     *  @brief Free the graphics context
     *
     *  This method frees the graphics context created by @c initContext
     */
    virtual void deinitContext();
    
    /**
     *  @brief Allocate image
     *
     *  This method allocates the image used to display buffers
     */
    virtual bool initImage() = 0;
    
    /**
     *  @brief This method frees the image allocated by @c allocateImage
     */
    virtual void deinitImage() = 0;
    
    /**
     *  @brief Set the visual information
     *  @param gc The new visual information
     *
     *  This method is used by implementations of initVisual to set the visual information.
     */
    void setVisualInfo(XVisualInfo* visual_info) throw();

    /**
     *  @brief Set the graphics context
     *  @param gc The new graphics context
     *
     *  This method is used by implementations of initContext to set the graphics context.
     */
    void setContext(GC gc) throw();
    
  private:    
    /**
     *  @brief Common initialization method called by all constructors
     */
    void initCommon();
    
    /**
     *  @brief Wait for X @c UnmapEvent event
     */
    void waitForXUnmapEvent();
    
    /**
     *  @brief Wait for X @c Expose event
     */
    void waitForXExposureEvent();
    
    /**
     *  @brief Wait for X @c ConfigureNotifyEvent
     */
    void waitForXConfigureNotifyEvent();
    
    /**
     *  @brief Calculate the screen size
     */
    void calcScreenWidthHeight();
    
    /**
     *  @brief Set the size hints for the window manager
     */
    void setSizeHints();
    
    /**
     *  @brief Set the cursor and window decoration for this window
     *
     *  This method turns both the cursor and the window decoration off if the window is in fullscreen mode.
     */
    void setCursorAndDecoration();
    
    /**
     *  @brief Create an X cursor
     *  @param src The cursor image source
     *  @param mask The cursor transparency mask
     *  @param width The cursor width in pixels
     *  @param height The cursor height in pixels
     *  @return The created X11 cursor
     */
    Cursor createCursor(char* src, char* mask, int width, int height);
    
    /**
     *  @brief Save window position before switching to fullscreen mode
     *  @bug Saving the window position before switching to fullscreen is currently broken
     */
    void saveXYPosition();
    
    /**
     *  @brief Set the window position
     */
    void setPosition();

    /** 
     *  @brief Initialize window
     */
    void initWindow();

    /** 
     *  @brief Deinitialize window
     */
    void deinitWindow();

    /**
     *  @brief Change window
     *
     *  This method is called when the window size or fullscreen flag has changed.
     */
    void changeWindow();

    /**
     *  @brief Turn cursor on/off
     *  @param show @c true to show, @c false to hide cursor
     */
    void showCursor(bool show);
    
    /**
     *  @brief Turn window decoration on/off
     *  @param show @c true to show, @c false to hide window decoration
     */
    void showDecorations(bool show);
    
    /**
     *  @brief Set EWMH fullscreen state
     *  @param fullscreen @c true for fullscreen, @c false for windowed mode
     *
     *  This method adds or removes the EWMH fullscreen state.
     */
    void setEWMHFullscreen(bool fullscreen);
    
    /**
     *  @brief Fill rectangle with background color
     *  @param x X position of upper-left corner
     *  @param y Y position of upper-left corner
     *  @param w Width in pixels
     *  @param h Height in pixels
     *
     *  This method is a wrapper for XFillRectangle which checks if
     *  the coordinates are valid and otherwise draws no rectangle.
     */
    void fillRectangle(int x, int y, int w, int h);
    
  private:
    Node*        __parent;                 ///< The parent node
    XProducer*   __xproducer;              ///< The X producer that converts X events into NMM events
    bool         __direct_mouse;           ///< Send mouse events directly
    bool         __direct_keyboard;        ///< Send keyboard events directly
    
    int          __xpos;                   ///< X position of the window on the screen
    int          __ypos;                   ///< Y position of the window on the screen
    int          __width;                  ///< Width of the image in pixels
    int          __height;                 ///< Height of the image in pixels
    int          __ratio_x;                ///< Numerator of the X:Y aspect ratio
    int          __ratio_y;                ///< Denominator of the X:Y aspect ratio
    int          __scaled_width;           ///< Scaled width in pixels
    int          __scaled_height;          ///< Scaled height in pixels
    int          __full_screen_width;      ///< Full screen width in pixels
    int          __full_screen_height;     ///< Full screen height in pixels
    int          __screen_width;           ///< Screen width in pixels
    int          __screen_height;          ///< Screen height in pixels
    int          __x_blit_offset;          ///< x offset in pixels where buffers should be displayed on the window
    int          __y_blit_offset;          ///< Y offset in pixels where buffers should be displayed on the window
    std::string  __title;                  ///< Window title
    bool         __fullscreen;             ///< Fullscreen mode
    int          __screen;                 ///< X Screen
    int          __depth;                  ///< Color depth
    int          __align;                  ///< Left / Center / Right alignment
    bool         __have_ewmh;              ///< Have support for EWMH
    
    Display*     __display;                ///< X display ID
    Window       __win;                    ///< X window ID
    XSizeHints   __size_hints;             ///< X size hints
    GC           __gc;                     ///< X graphics context
    XVisualInfo  __own_visual_info;        ///< Default visual information
    XVisualInfo* __visual_info;            ///< Current visual information

    /** Default event mask.
     */
    static const long WINDOW_MASK = ButtonMotionMask | ButtonPressMask | ButtonReleaseMask | KeyPressMask | KeyReleaseMask | FocusChangeMask | ExposureMask;
  };
  
  inline Node* XVirtualImg::getParent() const throw() {
    return __parent;
  }
  
  inline Display* XVirtualImg::getDisplay() const throw() {
    return __display;
  }
  
  inline Window XVirtualImg::getWindow() const throw() {
    return __win;
  }
  
  inline int XVirtualImg::getWidth() const throw() {
    return __width;
  }

  inline int XVirtualImg::getHeight() const throw() {
    return __height;
  }
  
  inline GC XVirtualImg::getContext() const throw() {
    return __gc;
  }
  
  inline int XVirtualImg::getYBlitOffset() const throw() {
    return __y_blit_offset;
  }

  inline int XVirtualImg::getXBlitOffset() const throw() {
    return __x_blit_offset;
  }
  
  inline int XVirtualImg::getScreenWidth() const throw() {
    return __screen_width;
  }
  
  inline int XVirtualImg::getScreenHeight() const throw() {
    return __screen_height;
  }
  
  inline long XVirtualImg::getEventMask() {
    return WINDOW_MASK;
  }
}

#endif
