/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Dostert Roger <dostert@studcs.uni-sb.de>
 *                    
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */
	
#include "JoystickEvents.hpp"
#include <iostream>
#include <sys/ioctl.h>
#include <sys/types.h> // for getpid()
#include <unistd.h> // for close(), read(), getpid(), usleep()

namespace NMM {

  JoystickEvents::JoystickEvents() { 
    /* set the member variables to defined values */
    dev_name    = "/dev/js0";
    thread_pid  = 0;
    num_buttons = 0;
    num_axis    = 0;
    axisscale   = NULL;
  }


  JoystickEvents::~JoystickEvents() { 
    /* close the joystick device */
    close(dev_handle);
    //    stop();
  }


  Result JoystickEvents::set_devicename(string &name) {
    dev_name = name;
    return SUCCESS;
  }  /* JoystickEvents::set_devicename(string name) */


  string JoystickEvents::get_devicename(void) {
    return dev_name;
  }  /* JoystickEvents::set_devicename(string name) */


  void JoystickEvents::connectTo( EventDispatcher *ed) {
    GLOBAL_ERROR_STREAM( "JoystickEvents connected" << endl);
    dispatcher = ed;
  } /* JoystickEvents::connectTo(...) */
 

  bool JoystickEvents::start(JoystickEventsInterface *jevint) {
    /* joystick initialization */
    int      version;
    char     name[128];
    bool     done   = false;

    axisscale = jevint;

    memset(&jevent,0,sizeof(js_event));
    
    /* open the joystick device in NON_BLOCKING mode */
    GLOBAL_DEBUG_STREAM("JoystickEvents::start() : open device " << dev_name << endl);
    dev_handle = open(dev_name.c_str(),O_RDONLY | O_NONBLOCK);
    if (-1 != dev_handle) {
      /* read out all initial events */
      GLOBAL_DEBUG_STREAM("JoystickEvents::start() : read initial events" << endl);
      while (!done) {
        if (sizeof(js_event) == read(dev_handle,&jevent,sizeof(js_event))) {

          if (JS_EVENT_INIT == (jevent.type & JS_EVENT_INIT)){
            switch(jevent.type & ~JS_EVENT_INIT) {
            case JS_EVENT_AXIS :
              GLOBAL_DEBUG_STREAM("   got JS_EVENT_INIT for AXIS " << int(jevent.number) << endl);
              break;
            case JS_EVENT_BUTTON :
              GLOBAL_DEBUG_STREAM("   got JS_EVENT_INIT for BUTTON " << int(jevent.number) << endl);
              break;
            default:
              GLOBAL_DEBUG_STREAM("   got JS_EVENT_INIT for UNKNOWN" << endl);
            }
            
          } else {
            done = true;
          } /* if(JS_EVENT_INIT ... */
        } else {
          done   = true;
        } /* if(sizeof(... */
      } /* while (!done) */
      
      /* get the capabilities of the joystick */
      GLOBAL_DEBUG_STREAM("JoystickEvents::start() : get capabilities" << endl);
      
      /* get driver version */
      ioctl(dev_handle,JSIOCGVERSION,&version);
      GLOBAL_DEBUG_STREAM("   driver version is " << 
                          ((version & 0xff000000) >> 24) << "." <<
                          ((version & 0x00ff0000) >> 16) << "." << 
                          ((version & 0x0000ff00) >>  8) << "." <<
                          ((version & 0x000000ff)) << endl);
      if (0 > ioctl(dev_handle,JSIOCGNAME(sizeof(name)),name)) {
        strncpy(name,"unknown",sizeof(name));
      }
    
      /* get name of identified joystick */
      GLOBAL_DEBUG_STREAM("   joystick name is <" << name << ">" << endl);

      /* get number of axis */
      ioctl(dev_handle,JSIOCGAXES,&num_axis);
      GLOBAL_DEBUG_STREAM("   found " << num_axis << " axis" << endl);
      
      /* get number of buttons */
      ioctl(dev_handle,JSIOCGBUTTONS,&num_buttons);
      GLOBAL_DEBUG_STREAM("   found " << num_buttons << " buttons" << endl);
      
      /* allocate memory for button buffer */
      buttons = new bool[num_buttons];
      
      /* allocate memory for axis buffer */
      axis = new signed short int[num_axis];

      /* send the number of axes and buttons to the application */
      if ( dispatcher ) { 
        dispatcher->dispatch( Event("joystick_capabilities", 
                                    new TInValue<int>(num_buttons),
                                    new TInValue<int>(num_axis)
                                    ));         
      }

      /* create thread */
      running=true;
      thread.start(getEvent_static, (void *)this);
      thread_pid = getpid();
      GLOBAL_ERROR_STREAM( "JoystickEvents started with PID " << thread_pid << endl);
      cerr << "JoystickEvents started with PID " << thread_pid << endl;
    } else {
      /* if there is no joystick we return false and the application must handle this */
      GLOBAL_ERROR_STREAM("ERROR in JoystickEvents::start() : can\'t open device " << dev_name << endl);
      return false;
    } // if(-1 != dev_handle)
    
    return true;
  } /* JoystickEvents::start() */


  void JoystickEvents::stop() {
    /* reset the scaling function */
    axisscale = NULL;
    /* the thread's mainloop should stop now */
    running   = false;

    /* release the axes and buttons buffer */
    if (axis)    { delete axis; }
    if (buttons) { delete buttons; }


    /* kick the thread */
    cerr << "\tProz kill " << thread_pid << endl;

    if ( thread_pid ) {
      kill( thread_pid, SIGIO );        /* send SIGIO to getEvent thread -> force <select> to terminate */
      thread.join();
      thread_pid = 0;
    }

    return;
  } /* JoystickEvents::stop() */


  void* JoystickEvents::getEvent_static(void *me) {
    (static_cast<JoystickEvents*>(me))->getEvent();

    /* to satisfy void* requirements, only checked with pthread_join */
    return NULL;
  }/* JoystickEvents::getEvent_static(...) */
    

  void JoystickEvents::getEvent() {
    int value;

    //    cerr << "Running is " << running << endl;

    /* while the thread is running update the joystick state */
    while(running) {
      if (sizeof(js_event) == read(dev_handle,&jevent,sizeof(js_event))) {
        /* it does no matter if there is an init event */
        jevent.type = jevent.type & ~JS_EVENT_INIT;
        switch (jevent.type) {
        case JS_EVENT_AXIS: /* the position of one axis has changed */

          /* scale the positon down with the scale function */
          if (NULL != axisscale) {
            axisscale->JoystickAxisScale(jevent.number, jevent.value, value);
              
            /* send the scaled value if it is different from the last one */
            /* and save this value for the next time                      */
            if ((axis[jevent.number] != value) && dispatcher)  { 
              cerr << "send  JOYSTICK_AXIS" << endl; 
              dispatcher->dispatch( Event("joystick_axis", 
                                          new TInValue<int>(jevent.number),
                                          new TInValue<int>(value)
                                          ));   
              axis[jevent.number] = value;
            }
          }

          break;
        case JS_EVENT_BUTTON: /* a button was pressed or released */
          buttons[jevent.number] = bool(jevent.value);
          
          /* send the button's new value to the application */
          if ( dispatcher ) {
            cerr << "send  JOYSTICK_BUTTON" << endl; 
            dispatcher->dispatch( Event("joystick_button", 
                                        new TInValue<int>(jevent.number),
                                        new TInValue<bool>(jevent.value)
                                        ));   
          }
          break;
        }
      }
      /* wait 10 ms until  the next call */
      usleep(10000);
    } /* while (running) */

  } /* JoystickEvents::getEvent() */

}; /* namespace NMM */
