/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Christoph Wellner <chwellner@graphics.cs.uni-sb.de>
 *                     
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "TVGLibXML2Database.hpp"

#include <fstream>
#include <unistd.h>
#include <time.h>
#include <regex.h>

#include "nmm/base/sync/LocalTime.hpp"
#include "nmm/base/sync/Clock.hpp"
#include "nmm/base/ProxyObject.hpp"
 
#include "nmm/utils/utils.hpp"
#include "nmm/utils/SecurityCheck.hpp"

#include "nmm/interfaces/misc/tvguide/ITVGuideUpdateFinished.hpp"

namespace NMM{

  TVGLibXML2Database::TVGLibXML2Database(ITVGuide* itvg)
    : workDir("./"),
      directory(),
      xml(NULL),
      root(NULL),
      termUpdate(false)
  {

    itvguide.reset(itvg);
    xml = new XMLTree();

    epsilon = 0;
    directory.setRootDir(workDir);
    readChannels();   
    readCategories();

  }//end Constructor


  TVGLibXML2Database::~TVGLibXML2Database(){
    saveChannels();
  }

  void TVGLibXML2Database::insertChannel(const string &id, const string &name){
    GLOBAL_DEBUG_STREAM("TVGLibXML2Database : insertChannel( " << id << " , " << name << ")"<< endl);

    ChannelDescription ins;
    ins.setID(id);
    ins.addChannelName(name);

    //check for aliases and add them to ChannelDescription
    XMLTree* channelAliases = new XMLTree(getNMMResourceDir() + "/tvguide/channelAliases.xml");
    XMLNode* rootnode = NULL;
    if(channelAliases){
      rootnode = channelAliases->root();
      if(rootnode){
        XMLNodeList channels = rootnode->children("channel");
        for(XMLNodeIterator i = channels.begin(); i != channels.end(); i++){
          XMLNode* channelNode = *i;
          XMLNodeList aliases = channelNode->children("alias");
          for(XMLNodeIterator j = aliases.begin(); j != aliases.end(); j++){
            XMLNode* aliasNode = *j;
            string tmpName = name; transform(tmpName);
            string tmpAlias = aliasNode->getData(); transform(tmpAlias);
            if(tmpAlias == tmpName){
              for(XMLNodeIterator k = aliases.begin(); k != aliases.end(); k++){
                XMLNode* n = *k;
                ins.addChannelName(n->getData());
              }
              break;
            }
          }//end(for(aliases)
        }//end for(channels)
      }//end if(rootnode)
    }//end if(xhannelAliases)
  

    if(channels.empty()){
      channels.push_back(ins);
      directory.makeDir(directory.getCurDir() +"/" + id);
    }else{
      list<ChannelDescription>::iterator it = channels.begin();
      list<ChannelDescription>::iterator insertPos = channels.end();

      while(it != channels.end()){
        ChannelDescription cur = *it;

        //don't insert a channel twice
        if(cur.getID() == id){ //same id?
          return;
        }else{ //other ID but same name? happens, if having >= 2 sources
          list<string> names = cur.getChannelNames();
          list<string>::iterator iter = names.begin();
          while(iter != names.end()){
            string s = *iter;
            iter++;
            if(s == name){
              return;
            }
          }
        }//end if(cur.getID())
        
        //insert channel in appropriate place
        if(cur.getID() > id){
          insertPos = it;
        }
         
        it++;
        
      }//end while

      //if we have come so far, there is no duplicate channel, and we can insert the new one
      channels.insert(insertPos, ins);
      directory.makeDir(directory.getCurDir() +"/" + id);
      //and send an event, that we have a new channel
      itvguide->getParentObject()->receiveEvent(Event(ITVGuideUpdateFinished::newChannel_event));

    }
    saveChannels();
  }

  void TVGLibXML2Database::setDirectory(const string &path){
    workDir = path;
    directory.setRootDir(workDir);
    readChannels();
    readCategories();
  }

  list<string> TVGLibXML2Database::getCategories() throw (const NMM::Exception&){
    return categories;
  }

  string TVGLibXML2Database::getDirectory()throw (const NMM::Exception&){
    return workDir;
  }

  list<ChannelDescription> TVGLibXML2Database::getChannelsList()throw (const NMM::Exception&){
    return channels;
  }  

  string TVGLibXML2Database::getChannelById(const string &id){
    GLOBAL_DEBUG_STREAM("TVGLibXML2Database: getChannelById( " << id << ")" << endl);

    list<ChannelDescription>::iterator it = channels.begin();
    while(it != channels.end()){
      ChannelDescription cur = *it;
      
      if(cur.getID() == id){
        return cur.getChannelNames().front();
      }else{
        it++;
      }
    }
    return string("");
  }  

  string TVGLibXML2Database::getChannelByName(const string &name){
    GLOBAL_DEBUG_STREAM("TVGLibXML2Database: getChannelByName( " << name << ")" << endl);
    
    for( list<ChannelDescription>::iterator it = channels.begin(); it != channels.end(); it++){
      ChannelDescription cur = *it;  
      list<string> channelNames = cur.getChannelNames();
      for(list<string>::iterator nameIt = channelNames.begin(); nameIt != channelNames.end(); nameIt++){
        string n = *nameIt;
        string tmp = name;
        transform(n);
        transform(tmp);
        if(n == tmp){
          return cur.getID();
        }else{
        }
      }//end while nameIt
    }//end while it
    
    return string("");
  }  

  void TVGLibXML2Database::insertEvent(const ProgramDescription &pd){
    GLOBAL_DEBUG_STREAM("TVGLibXML2Database::insertEvent(...)" << endl);

    if(getChannelById(pd.getChannelID()) == ""){
      return;
    }

    MutexGuard mg(tvmutex);

    string start = pd.getStartTime().substr(0,8);
    string chan = pd.getChannelID();
    //write xml-info to file
    string filename = workDir + "/" + chan + "/" + start + ".xml" ;

    //Security-Check
    if(!SSecurityCheck::getInstance().checkFilenameForWriting(filename)) {
      GLOBAL_ERROR_STREAM("Security check not passed for file " << filename << ", check .nmmrc for options" << endl);
      return;
    }

    //Check, if directory exists. if not, create it
    if(!directory.changeDir(chan)){
      directory.makeDir(workDir +"/" + chan);
      directory.setRootDir(workDir);
    }else{
      directory.setRootDir(workDir);
    }

    if(fileExists(filename)){


      xml->read(filename);
      if(xml)
         root = xml->root();
      
      //if root is found, file exists
      if(root){

        GLOBAL_DEBUG_STREAM("File Exists: " << filename << endl);
        
        XMLNodeList programList = root->children("programme");

        XMLNodeIterator programIter;
        programIter = programList.begin();

        //get the time 
        time_t startTime= stringToTime(pd.getStartTime());
        time_t endTime  = stringToTime(pd.getEndTime());

        //sorting of programme according to start-/endtime
        int position = 0;

        while(programIter != programList.end()){
          XMLNode* currProgram = *programIter;
          programIter++;
          position++;
          XMLNode* nextProgram = *programIter;
          
          
          time_t curProgramStop= 0;
          try{
            XMLProperty* property = currProgram->property("stop");
            if(property){
              curProgramStop = stringToTime(property->value());
            }
          }catch(const exception& e){
            return;
          }catch(...){
            return;
          }
          
          
          time_t nextProgramStart = 0;
          try{
            if(programIter != programList.end()){
              XMLProperty* property = nextProgram->property("start");
              if(property){
                nextProgramStart =  stringToTime(property->value());
              }
            }else{
              //if List has only 1 entry, there is no "next Program" and the program has to be added
              nextProgramStart = endTime;
            }
          }catch(const exception& e){
            return;
          }catch(...){return;}
          
          time_t curProgramStart = 0;
          try{
            XMLProperty* p = currProgram->property("start");
            if(p){
              curProgramStart = stringToTime(p->value());
            }
          }catch(const exception& e){
            return;
          }catch(...){
            return;
          }

          if(startTime == curProgramStart
             && endTime == curProgramStop){
            //test, if the start-/endtimes are the same (+- epsilon)
            //if yes, test the title-string of aproximate equality
            //if equal, and extended Description is lager, replace or add the description and additional value not already available.
            
            list<string> lt = pd.getTitles();
            string s1("");
            for(list<string>::iterator i = lt.begin(); i != lt.end(); i++){
              s1 += *i;
            }
            XMLNodeList nl = currProgram->children("title");
            string s2("");
            for(XMLNodeIterator i = nl.begin(); i != nl.end(); i++){
              XMLNode* node = *i;
              s2 += node->getData();
            }

            if(compareStrings(s1, s2)){
              XMLNodeList curDesc = currProgram->children("desc");
              string description("");
              for(XMLNodeIterator i = curDesc.begin(); i != curDesc.end(); i++){
                XMLNode* n = *i;
                description += n->getData() + " ";
              }
              if(description.size() < pd.getDescription().size()){
                //replace Description
                for(XMLNodeIterator i = curDesc.begin(); i != curDesc.end(); i++){
                  XMLNode* n = *i;
                  currProgram->remove_child(n);
                }
                XMLNode* newNode = currProgram->add_child("desc");
                if(newNode) newNode->add_content(transformSpecialCharsIn(pd.getDescription()));
              }

              //XMLNodeList curDuration = curProgram->children("duration");              
              //XMLNodeList curCategory = curProgram->children("category");
              xml->write();
              return;    

            }

            /*
              }else if(curProgramStop <= startTime //curProgram end before newProgram begins       
              && nextProgramStart >= endTime){ //newProgram end before nextProgram starts
            */
          }else if(curProgramStart <= startTime && startTime < nextProgramStart){

            XMLNode* newProgram = root->add_child("programme", position);
            //add start-time
            newProgram->add_property("start", pd.getStartTime());
            //add end-time
            if(pd.getEndTime() == ""){
              time_t tmpTime = stringToTime(pd.getStartTime()) + (pd.getDuration()*60);
              char t[20];
              strftime(t, sizeof(t), "%Y%m%d%H%M",localtime(&tmpTime));
              sprintf(t, "%s %s", t, pd.getStartTime().substr(pd.getStartTime().find_first_of(" ")+1, 5).c_str());
              newProgram->add_property("stop",t);
            }else{
              newProgram->add_property("stop", pd.getEndTime());
            }
            //add channel-id
            newProgram->add_property("channel", pd.getChannelID());
            
            //add Duration
            char tmp[5];
            sprintf(tmp, "%i", pd.getDuration());
            XMLNode* tmpNode = newProgram->add_child("duration");
            tmpNode->add_property("units", "min");
            tmpNode->add_content(tmp);
            
            //add description
            if(pd.getDescription() != ""){
              tmpNode = newProgram->add_child("desc");
              tmpNode->add_content(transformSpecialCharsIn(pd.getDescription()) + " ");
            }
            
            list<string> titles = pd.getTitles();
            list<string>::iterator titlesIter = titles.begin();
            while(titlesIter != titles.end()){
              string title = *titlesIter;
              titlesIter++;
              
              tmpNode = newProgram->add_child("title");
              tmpNode->add_content(transformSpecialCharsIn(title) + " ");
              
            }//end while(titlesTiter)
            
            //add Category
            if(!pd.getCategory().empty()){
              list<string> cl = pd.getCategory();
              for(list<string>::iterator i = cl.begin(); i != cl.end(); i++){
                tmpNode = newProgram->add_child("category");
                tmpNode->add_content(transformSpecialCharsIn(*i));
                categories.push_back(*i);
                categories.sort();
                categories.unique();
		saveCategories();
              }
            }
                        
            //write the new xml-tree to file
            xml->write();
            //and send an event for a new Event 
            //            itvguide->getParentObject()->receiveEvent(Event(ITVGuideUpdateFinished::newShow_event));
            return;    
          }//end if(curProgramStop <= startTime && nextProgramStart >= endTime)
          
        }//end while(programIter)
      }

    }else{

      GLOBAL_DEBUG_STREAM("Creating File: " << filename << endl);
      //writing the XML-Header and root-tree to xml-file
      ofstream fout(filename.c_str(), ios::out);
      if(fout){
        fout << "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?> \n" << endl;
        fout << "<!DOCTYPE tv SYSTEM \"xmltv.dtd\"> \n" << endl;
        fout << "<tv generator-info-name=\"NMM TVGuide\">\n" << endl;; 
        fout << "\n </tv> \n" << endl;
        fout.close();
                
        xml->read(filename);
        if(xml)
          root = xml->root();
        //add new Programme-Node to XMLTree
        if(root){
          XMLNode* newProgram = root->add_child("programme");;
          
          //add start-time
          newProgram->add_property("start", pd.getStartTime());
          //add end-time
          if(pd.getEndTime() == ""){
            time_t tmpTime = stringToTime(pd.getStartTime()) + (pd.getDuration()*60);
            char t[20];
            strftime(t, sizeof(t), "%Y%m%d%H%M",localtime(&tmpTime));
            sprintf(t, "%s %s", t, pd.getStartTime().substr(pd.getStartTime().find_first_of(" ")+1, 5).c_str());
            newProgram->add_property("stop",t);
          }else{
            newProgram->add_property("stop", pd.getEndTime());
          }

          //add channel-id
          newProgram->add_property("channel", pd.getChannelID());
          
          //add Duration
          char tmp[5];
          sprintf(tmp, "%i", pd.getDuration());
          XMLNode* tmpNode = newProgram->add_child("duration");
          tmpNode->add_property("units", "min");
          tmpNode->add_content(tmp);
          //add description
          if(pd.getDescription() != ""){
            tmpNode = newProgram->add_child("desc");
            tmpNode->add_content(transformSpecialCharsIn(pd.getDescription()) + " ");
          }

          list<string> titles = pd.getTitles();
          list<string>::iterator titlesIter = titles.begin();
          
          while(titlesIter != titles.end()){
            
            string title = *titlesIter;
            titlesIter++;
            
            tmpNode = newProgram->add_child("title");
            tmpNode->add_content(transformSpecialCharsIn(title) + " ");
            
            
          }//end while(titlesTiter)

	  //add Category
	  if(!pd.getCategory().empty()){
            list<string> cl = pd.getCategory();
            for(list<string>::iterator i = cl.begin(); i != cl.end(); i++){
              tmpNode = newProgram->add_child("category");
              tmpNode->add_content(transformSpecialCharsIn(*i));	  
              categories.push_back(*i);
              categories.sort();
              categories.unique();
	      saveCategories();
            }
          }

          //write the new xml-tree to file
          xml->write();
          //          itvguide->getParentObject()->receiveEvent(Event(ITVGuideUpdateFinished::newShow_event));
        }
      }else{
        GLOBAL_ERROR_STREAM("FILE: " << filename << " NOT CREATED!!!" << endl);
      }
    }
  
  }

  ProgramDescription TVGLibXML2Database::getEvent(const string &start, const string &end, const string &id){
    GLOBAL_DEBUG_STREAM("TVGLibXML2Database: getEvent( " << start << " , " << end << " , " << id << ")" << endl);
    MutexGuard mg(tvmutex);

    string filename = workDir + "./" + id + "/" + start.substr(0,8) + ".xml" ;
    ProgramDescription p;
    
    //if the xml-file does not exist, exit funktion
    if(!fileExists(filename))
      return p;

    //Security-Check
    if(!SSecurityCheck::getInstance().checkFilenameForReading(filename)) {
      GLOBAL_ERROR_STREAM("Security check not passed for file " << filename << ", check .nmmrc for options" << endl);
      return p;
    }

    time_t startTime = 0;
    time_t  stopTime  = 0;

    try{
      startTime = stringToTime(start);
      stopTime  = stringToTime(end);
    }catch(const Exception &e){
      GLOBAL_ERROR_STREAM("start- or end-time are in wrong format " << endl);
      return p;
    }

    xml->read(filename);
    if(xml)
      root = xml->root();

    if(root){
      XMLNodeList xmllist = root->children("programme");
      XMLNodeIterator iter = xmllist.begin();
      
      while(iter != xmllist.end()){
        XMLNode* cur = *iter;
        iter++;
        
        XMLProperty* sprop = cur->property("start");
        XMLProperty* eprop = cur->property("stop");
        XMLProperty* idprop = cur->property("channel");

        time_t curProgramStart = stringToTime(sprop->value());
        time_t curProgramStop  = 0;
        try{
          curProgramStop = stringToTime(eprop->value());
        }catch(const Exception &e){
        }catch (...){}

        if(sprop && eprop && idprop){
          if(   (curProgramStart <= startTime  && curProgramStop >= stopTime) 
                || (curProgramStart >= startTime ) && (curProgramStop <= stopTime )
             && (idprop->value() == id)){

            p.setStartTime(start);
            p.setEndTime(end);
            p.setChannelID(id);
            
            XMLProperty* property = cur->property("duration");
            if(property)
              p.setDuration(atoi((property->value()).c_str()));
            
            //read titles
            XMLNodeList titleList = cur->children("title");
            XMLNodeIterator t_iter, t_end;
            t_iter = titleList.begin();
            t_end = titleList.end();
            
            while(t_iter != t_end){
              XMLNode* title = *t_iter;
              t_iter++;
              p.addTitle(transformSpecialCharsOut(title->getData()));
              
            }//end while(title)
            
            //read description
            XMLNodeList descList = cur->children("desc");
            XMLNodeIterator d_iter, d_end;
            d_iter = descList.begin();
            d_end = descList.end();
            string d_tmp("");

            while(d_iter != d_end){
              XMLNode* desc = *d_iter;
              d_iter++;
              d_tmp += desc->getData();
            }//end while(desc)
            p.setDescription(transformSpecialCharsOut(d_tmp));
            return p;
          }
        }
        
      }//end while(xmllist)
    }
    return p;
  }

  list<ProgramDescription> TVGLibXML2Database::getEventsOnDayOnChannel(const string &day, const string &chan){
    GLOBAL_DEBUG_STREAM("TVGLibXML2Databaseserver::getEventsOnDayOnChannel( " << day << " , " << chan << ")" << endl);


    list<ProgramDescription> l;
    string filename = workDir + "/" + chan + "/" + day + ".xml" ;

    //if the xml-file does not exist, exit funktion
    if(!fileExists(filename)){
      GLOBAL_ERROR_STREAM("File " << filename << "does not exist" << endl); 
      return l;
    }
    //Security-Check
    if(!SSecurityCheck::getInstance().checkFilenameForReading(filename)) {
      GLOBAL_ERROR_STREAM("Security check not passed for file " << filename << ", check .nmmrc for options" << endl);
      return l;
    }

    MutexGuard mg(tvmutex);
    
    xml->read(filename);
    if(xml){
      root = xml->root();
    }else{
      GLOBAL_ERROR_STREAM("No SourceFIle for day " + day + " and channel " + chan + " found !!" << endl);
      return l;
    }

    if(root){
      XMLNodeList xmllist = root->children("programme");
      ProgramDescription pd;

      XMLNodeIterator iter, end;
      iter = xmllist.begin();
      end = xmllist.end();

      while(iter != end){
        ProgramDescription pd;

        XMLNode* node = *iter;
        iter++;

        //set start-time
        XMLProperty* property = node->property("start");
        if(property)
          pd.setStartTime(property->value());

        //set end-time
        property = node->property("stop");
        if(property)
          pd.setEndTime(property->value());
        
        //set channel
        property = node->property("channel");
        if(property)
          pd.setChannelID(property->value());
        
        //set duration
        property = node->property("duration");
        if(property){
          pd.setDuration(atoi((property->value()).c_str()));
        }else{
          XMLNodeList durList = node->children("duration");
          for(XMLNodeIterator i = durList.begin(); i != durList.end(); i++){
            XMLNode* n = *i;
            pd.setDuration(atoi(n->getData().c_str()));
          }
        }
        //read titles
        XMLNodeList titleList = node->children("title");
        XMLNodeIterator t_iter, t_end;
        t_iter = titleList.begin();
        t_end = titleList.end();
        
        while(t_iter != t_end){
          XMLNode* title = *t_iter;
          t_iter++;

          pd.addTitle(transformSpecialCharsOut(title->getData()));
          
        }//end while(title)

        //read description
        XMLNodeList descList = node->children("desc");
        XMLNodeIterator d_iter, d_end;
        d_iter = descList.begin();
        d_end = descList.end();
        string d_tmp("");

        while(d_iter != d_end){
          XMLNode* desc = *d_iter;
          d_iter++;
          d_tmp += desc->getData();
        }//end while(desc)
        pd.setDescription(transformSpecialCharsOut(d_tmp));
        
        //read category
        XMLNodeList catList = node->children("category");
        XMLNodeIterator c_iter, c_end;
        c_iter = catList.begin();
        c_end = catList.end();
        string c_tmp("");

        while(c_iter != c_end){
          XMLNode* cat = *c_iter;
          c_iter++;
          pd.addCategory(transformSpecialCharsOut(cat->getData()));
        }//end while(cat)
       
        l.push_back(pd);
      }//end while(programme)
    }else{
    }//end if(rootnode)

    return l;
  }

  list<ProgramDescription> TVGLibXML2Database::search(const ProgramFilter &pf){
    GLOBAL_DEBUG_STREAM("TVGLibXML2Database: search(...)" << endl);


    Clock clock;

    //read all specified subdirs (= channels)
    list<string> cl;
    if(pf.getChannels().empty()){
      for(list<ChannelDescription>::iterator it = channels.begin(); it != channels.end(); it++){
        cl.push_back(it->getID());
      }
    }else{
      cl = pf.getChannels();
    }
    
    list<ProgramDescription> l; //returnlist


    if(pf.getStartTime() == "Now" || pf.getEndTime() == "Now"){
      for(list<string>::iterator subdirIt = cl.begin(); subdirIt != cl.end(); subdirIt++){
        ProgramDescription p = getCurrentEventOnChannel(*subdirIt);
        if(!p.empty()) {
          regex_t exp;
          bool noreturn = false;
          //Check Searchphrase
          if(!(regcomp(&exp, pf.getSearchPhrase().c_str(), REG_EXTENDED | REG_ICASE))){
            if(pf.getSearchPhrase() == ""){
            }else{
              if(regexec(&exp, p.getDescription().c_str() ,0,NULL,0) == REG_NOMATCH){
                list<string> titles = p.getTitles();
                for(list<string>::iterator i = titles.begin(); i != titles.end(); i++){
                  string s = *i;
                  if(regexec(&exp, s.c_str() ,0,NULL,0) == REG_NOMATCH){
                    noreturn = true;
                  }
                }
              }
            }
          }
          //Filter by categories if necessary
          if(!pf.getCategories().empty()){
            list<string> catl = pf.getCategories();
            list<string> pcatl = p.getCategory();
            noreturn = true;
            for(list<string>::iterator i = catl.begin(); i != catl.end(); i++){
              for(list<string>::iterator j = pcatl.begin(); j != pcatl.end(); j++){
                string s1 = *i;
                string s2 = *j;
                if(s1 == s2){
                  noreturn = false;
                }
              }
            }
          }
          if(!noreturn)l.push_back(p);
        }
      }
      return l;
    }else if(pf.getStartTime() == "Next" || pf.getEndTime() == "Next"){
      for(list<string>::iterator subdirIt = cl.begin(); subdirIt != cl.end(); subdirIt++){
        ProgramDescription p = getFollowingEventOnChannel(*subdirIt);
        if(!p.empty()) {
          regex_t exp;
          bool noreturn = false;
          //Check Searchphrase
          if(!(regcomp(&exp, pf.getSearchPhrase().c_str(), REG_EXTENDED | REG_ICASE))){
            if(pf.getSearchPhrase() == ""){
            }else{
              if(regexec(&exp, p.getDescription().c_str() ,0,NULL,0) == REG_NOMATCH){
                list<string> titles = p.getTitles();
                for(list<string>::iterator i = titles.begin(); i != titles.end(); i++){
                  string s = *i;
                  if(regexec(&exp, s.c_str() ,0,NULL,0) == REG_NOMATCH){
                    noreturn = true;
                  }
                }
              }
            }
          }
          //Filter by categories if necessary
          if(!pf.getCategories().empty()){
            list<string> catl = pf.getCategories();
            list<string> pcatl= p.getCategory();
            noreturn = true;
            for(list<string>::iterator i = catl.begin(); i != catl.end(); i++){
              for(list<string>::iterator j = pcatl.begin(); j != pcatl.end(); j++){
                string s1 = *i;
                string s2 = *j;
                if(s1 == s2){
                  noreturn = false;
                }
              }
            }
          }
          if(!noreturn)l.push_back(p);
        }
      }
      return l;
    }


    //Time time0 = clock.getTime();

    //iteratore over all given subdirs
    for(list<string>::iterator subdirIt = cl.begin(); subdirIt != cl.end(); subdirIt++){
      //Time time1 = clock.getTime();
        
      if(directory.changeDir(*subdirIt)){

        //compute file to read
        vector<DirReader::directory_t> files = directory.getRegFiles();
        for(vector<DirReader::directory_t>::iterator i = files.begin(); i != files.end(); i++){
          string dirFile = i->name.substr(0,8);
          string startFile = pf.getStartTime().substr(0,8);
          string endFile   = pf.getEndTime().substr(0,8);

          string file;
          if(pf.getStartTime() == "" && pf.getEndTime() == ""){
            //neither start- nor end-time given
            file = workDir + "/" + *subdirIt + "/" + dirFile + ".xml";
          }else if(pf.getStartTime() != "" && pf.getEndTime() != ""){ 
            //start- and end-time given
            startFile = pf.getStartTime().substr(0,8);
            endFile   = pf.getEndTime().substr(0,8);
            if(atoi(startFile.c_str()) <= atoi(dirFile.c_str()) && atoi(dirFile.c_str()) <= atoi(endFile.c_str()))
              file = workDir + "/" + *subdirIt + "/" + dirFile + ".xml";
          }else if(pf.getStartTime() != ""){
            //only start-time given
            startFile = pf.getStartTime().substr(0,8);
            if(atoi(startFile.c_str()) <= atoi(dirFile.c_str()))
              file = workDir + "/" + *subdirIt + "/" + dirFile + ".xml";
          }else if(pf.getEndTime() != ""){
            //only end-time given
            endFile   = pf.getEndTime().substr(0,8);
            if(atoi(dirFile.c_str()) <= atoi(endFile.c_str()))
              file = workDir + "/" + *subdirIt + "/" + dirFile + ".xml";
          }else{
            continue;
          }

          if(file.empty()) continue;
          if(!fileExists(file)) continue;
          

          XMLTree* searchXML = new XMLTree(file);

          if(searchXML){
            XMLNode* searchRoot = searchXML->root();
            if(searchRoot){
              XMLNodeList searchProgs = searchRoot->children("programme");
              for(XMLNodeIterator searchIt = searchProgs.begin(); searchIt != searchProgs.end(); searchIt++){
                XMLNode* currentNode = *searchIt ;
                ProgramDescription p;
                //add Channel
                p.setChannelID(*subdirIt);

                XMLProperty* startProp = NULL;
                XMLProperty* endProp = NULL;

                //check startTime
                startProp = currentNode->property("start");
                if(pf.getStartTime() != ""){
                  if(startProp){
                    if(stringToTime(startProp->value()) < stringToTime(pf.getStartTime())){
                      continue;
                    }else{
                    }
                  }
                }
                
                //check endTime
                endProp = currentNode->property("stop");
                if(pf.getEndTime() != ""){
                  if(endProp){
                    if(stringToTime(endProp->value()) > stringToTime(pf.getEndTime())){
                      continue;
                    }else{
                    }
                  }
                }
                
                //check Phrase
                regex_t exp;
                int title_res = -1;
                int desc_res = -1;
                if(!(regcomp(&exp, pf.getSearchPhrase().c_str(), REG_EXTENDED | REG_ICASE))){
                  XMLNodeList liste;
                  //in titles
                  if(pf.getSearchPhrase() == ""){
                    title_res = 0;
                    liste = currentNode->children("title");
                    //no search string is given -> add every title
                    for(XMLNodeIterator i = liste.begin(); i != liste.end(); i++){
                      XMLNode* t = *i;
                      p.addTitle(transformSpecialCharsOut(t->getData()));
                    }
                  }else{
                    liste = currentNode->children("title");
                    for(XMLNodeIterator i = liste.begin(); i != liste.end(); i++){
                      XMLNode* t = *i;
                      if(regexec(&exp, transformSpecialCharsOut(t->getData()).c_str() ,0,NULL,0) != REG_NOMATCH){
                        //if we find a match, we set a flag
                        title_res = 0;
                      }                    
                      p.addTitle(transformSpecialCharsOut(t->getData()));
                    }
                  }
                    
                  //in extended description
                  liste = currentNode->children("desc");
                  string desc;
                  if(pf.getSearchPhrase() == ""){
                    desc_res = 0;
                    for(XMLNodeIterator i = liste.begin(); i != liste.end(); i++){
                      XMLNode* t = *i;
                      desc += " " + t->getData();
                    }
                  }else{
                    for(XMLNodeIterator i = liste.begin(); i != liste.end(); i++){
                      XMLNode* t = *i;
                      if(regexec(&exp, transformSpecialCharsOut(t->getData()).c_str() ,0,NULL,0) != REG_NOMATCH){
                        //if we find a match in description, we set a flag
                        desc_res = 0;
                      }
                      desc += " " + t->getData();
                    }
                  }
                  p.setDescription(transformSpecialCharsOut(desc));
                }
                //if flags are -1, then no match was found at all, and we can skip the rest
                if(title_res == -1 && desc_res == -1) continue;
                
              
                //check for Category
                if(!pf.getCategories().empty()){
                  int cat_res = -1;
                  XMLNodeList liste = currentNode->children("category");
                  for(XMLNodeIterator i = liste.begin(); i != liste.end(); i++){
                    XMLNode* t = *i;
                    list<string> catl = pf.getCategories();
                    for(list<string>::iterator j = catl.begin(); j != catl.end(); j++){
                      if(*j == t->getData()){
                        cat_res = 0;
                        p.addCategory(transformSpecialCharsOut(t->getData()));
                      }
                    }
                  }
                  if(cat_res == -1) continue;
                }
                
                //fill resulting PD
                if(startProp) p.setStartTime(startProp->value());
                if(endProp)   p.setEndTime(endProp->value());
                
                
                l.push_back(p);
              }//end for(searchIt)
            }//end if(searchRoot)
          }//end if(searchXML)

        }//end if(changeDir)
      }//end for(files)

      //Time time3 = clock.getTime();
      //cerr << "Time for Reading 1 of  " << cl.size()  << " channels (" << *subdirIt << ") : " << time3 - time1 << endl;
      
      directory.setRootDir(workDir);

    }//end for(channels)

    //Time time4 = clock.getTime();
    //cerr << "Time for Search: " << time4 - time0 << endl; 
   return l;
  }

  void TVGLibXML2Database::addExtDescription(const string& start, const string& end, const string& id, const string& desc){
    GLOBAL_DEBUG_STREAM(" TVGLibXML2Database: addExtendedDescription(...)" << endl);

    string filename = workDir + "./" + id + "/" + start.substr(0,8) + ".xml" ;

    //if the xml-file does not exist, exit funktion
    if(!fileExists(filename))
      return;

    //Security-Check
    if(!SSecurityCheck::getInstance().checkFilenameForReading(filename)) {
      GLOBAL_ERROR_STREAM("Security check not passed for file " << filename << ", check .nmmrc for options" << endl);
      return;
    }

    MutexGuard mg(tvmutex);

    string startTime = start.substr(0, 12);
    string stopTime = end.substr(0,12);

    xml->read(filename);
    if(xml)
      root = xml->root();

    if(root){
      XMLNodeList xmllist = root->children("programme");
      XMLNodeIterator iter = xmllist.begin();
      
      while(iter != xmllist.end()){
        XMLNode* cur = *iter;
        iter++;
        
        try{
          XMLProperty* sprop = cur->property("start");
          XMLProperty* eprop = cur->property("stop");
          XMLProperty* idprop = cur->property("channel");
          
          if(sprop && eprop && idprop){
            if(((sprop->value()).substr(0,12) == startTime )
               && ((eprop->value()).substr(0,12) == stopTime )
               && (idprop->value()== id)){
              
              XMLNode* description = cur->add_child("desc");
              description->add_content(transformSpecialCharsIn(desc));
              
              xml->write();
            }
          }
        }catch(const Exception &e){
        }catch (...) {}
      }//end while(xmllist)
    }
  }

  ProgramDescription TVGLibXML2Database::getCurrentEventOnChannel(const string& channel){
    GLOBAL_DEBUG_STREAM("TVGLibXML2Database: getCurrentEventOnChannel( " << channel << ")" << endl);

    ProgramDescription p;
          
    try{

      time_t currenttime;
      currenttime = time(NULL);
      
      static char tmp[25];
      strftime(tmp, sizeof(tmp), "%Y%m%d%H%M", localtime(&currenttime));     
      string day = string(tmp).substr(0,8);
      strftime(tmp, sizeof(tmp), "%H%M", localtime(&currenttime));     
      string startTime = string(tmp);
      
      string filename = workDir + "./" + channel + "/" + day + ".xml" ;
      
      //if the xml-file does not exist, exit funktion
      if(!fileExists(filename))
        return p;
      
      //Security-Check
      if(!SSecurityCheck::getInstance().checkFilenameForReading(filename)) {
        GLOBAL_ERROR_STREAM("Security check not passed for file " << filename << ", check .nmmrc for options" << endl);
        return p;
      }
      
      MutexGuard mg(tvmutex);
      
      xml->read(filename);
      if(xml){
        root = xml->root();
      }else{
        GLOBAL_ERROR_STREAM("No SourceFIle for day " + day + " and channel " + channel + " found !!" << endl);
        return p;
      }
      if(root){
        XMLNodeList xmllist = root->children("programme");
        XMLNodeIterator iter = xmllist.begin();
        
        while(iter != xmllist.end()){
          XMLNode* cur = *iter;
          iter++;
          
          XMLProperty* sprop = cur->property("start");
          XMLProperty* eprop = cur->property("stop");
          XMLProperty* idprop = cur->property("channel");
          
          if(sprop && eprop && idprop){
            if(sprop->value() == "" || eprop->value() == "" || idprop->value() == "") return p;
            if(   atoi(sprop->value().substr(8,4).c_str()) <= atoi(startTime.c_str())
                  && atoi(eprop->value().substr(8,4).c_str()) >= atoi(startTime.c_str()) 
                  && idprop->value() == channel){
              
              
              p.setStartTime(sprop->value());
              p.setEndTime(eprop->value());
              p.setChannelID(idprop->value());
                            
              XMLProperty* property = cur->property("duration");
              if(property)
                p.setDuration(atoi((property->value()).c_str()));
                            
              //read titles
              XMLNodeList titleList = cur->children("title");
              XMLNodeIterator t_iter, t_end;
              t_iter = titleList.begin();
              t_end = titleList.end();
              
              while(t_iter != t_end){
                XMLNode* title = *t_iter;
                t_iter++;
                
                p.addTitle(transformSpecialCharsOut(title->getData()));
                
              }//end while(title)
              
              //read description
              XMLNodeList descList = cur->children("desc");
              XMLNodeIterator d_iter, d_end;
              d_iter = descList.begin();
              d_end = descList.end();
              string d_tmp("");
              
              while(d_iter != d_end){
                XMLNode* desc = *d_iter;
                d_iter++;
                d_tmp += desc->getData();
              }//end while(desc)
              p.setDescription(transformSpecialCharsOut(d_tmp));
              
              //read categories
              XMLNodeList catList = cur->children("category");
              XMLNodeIterator cat_iter, cat_end;
              cat_iter = catList.begin();
              cat_end = catList.end();
              
              while(cat_iter != cat_end){
                XMLNode* cat = *cat_iter;
                cat_iter++;
                
                p.addCategory(transformSpecialCharsOut(cat->getData()));
                
              }//end while(cat)
            }
          }
          
        }//end while(xmllist)
        
      }//end if(rootnode)
      return p;
    }catch(const Exception &e){
      GLOBAL_ERROR_STREAM(e << endl);
      return p;
    }
  }

  ProgramDescription TVGLibXML2Database::getFollowingEventOnChannel(const string& channel){
    GLOBAL_DEBUG_STREAM("TVGLibXML2Database: getCurrentFollowingOnChannel( " << channel << ")" << endl);

    ProgramDescription p;
    try{
      time_t currenttime;
      currenttime = time(NULL);
      
      static char tmp[25];
      strftime(tmp, sizeof(tmp), "%Y%m%d%H%M", localtime(&currenttime));     
      string day = string(tmp).substr(0,8);
      strftime(tmp, sizeof(tmp), "%H%M", localtime(&currenttime));     
      string startTime = string(tmp);
      
      string filename = workDir + "./" + channel + "/" + day + ".xml" ;
      
      //if the xml-file does not exist, exit funktion
      if(!fileExists(filename))
        return p;
      
      //Security-Check
      if(!SSecurityCheck::getInstance().checkFilenameForReading(filename)) {
        GLOBAL_ERROR_STREAM("Security check not passed for file " << filename << ", check .nmmrc for options" << endl);
        return p;
      }
      
      MutexGuard mg(tvmutex);
      
      xml->read(filename);
      if(xml){
        root = xml->root();
      }else{
        GLOBAL_ERROR_STREAM("No SourceFIle for day " + day + " and channel " + channel + " found !!" << endl);
        return p;
      }
      
      if(root){
        XMLNodeList xmllist = root->children("programme");
        XMLNodeIterator iter = xmllist.begin();
        
        while(iter != xmllist.end()){
          XMLNode* cur = *iter;
          iter++;
          
          XMLProperty* sprop = cur->property("start");
          XMLProperty* eprop = cur->property("stop");
          XMLProperty* idprop = cur->property("channel");
          
          if(sprop && eprop && idprop){
            if(sprop->value() == "" || eprop->value() == "" || idprop->value() == "") return p;
            if(   atoi(sprop->value().substr(8,4).c_str()) <= atoi(startTime.c_str())
                  && atoi(eprop->value().substr(8,4).c_str()) >= atoi(startTime.c_str()) 
                  && idprop->value() == channel){
              
              XMLNode* cur = *iter;
              
              if(iter != xmllist.end()){
                
                /* get the values of the following show */
                XMLProperty* sprop = cur->property("start");
                XMLProperty* eprop = cur->property("stop");
                XMLProperty* idprop = cur->property("channel");
                
                p.setStartTime(sprop->value());
                p.setEndTime(eprop->value());
                p.setChannelID(idprop->value());
                
                XMLProperty* property = cur->property("duration");
                if(property)
                  p.setDuration(atoi((property->value()).c_str()));
                
                //read titles
                XMLNodeList titleList = cur->children("title");
                XMLNodeIterator t_iter, t_end;
                t_iter = titleList.begin();
                t_end = titleList.end();
                
                while(t_iter != t_end){
                  XMLNode* title = *t_iter;
                  t_iter++;
                  
                  p.addTitle(transformSpecialCharsOut(title->getData()));
                  
                }//end while(title)
                
                //read description
                XMLNodeList descList = cur->children("desc");
                XMLNodeIterator d_iter, d_end;
                d_iter = descList.begin();
                d_end = descList.end();
                string d_tmp("");
                
                while(d_iter != d_end){
                  XMLNode* desc = *d_iter;
                  d_iter++;
                  d_tmp += desc->getData();
                }//end while(desc)
                p.setDescription(transformSpecialCharsOut(d_tmp));
                
                //read categories
                XMLNodeList catList = cur->children("category");
                XMLNodeIterator cat_iter, cat_end;
                cat_iter = catList.begin();
                cat_end = catList.end();
                
                while(cat_iter != cat_end){
                  XMLNode* cat = *cat_iter;
                  cat_iter++;
                  
                  p.addCategory(transformSpecialCharsOut(cat->getData()));
                  
                }//end while(cat)

              }
            }
          }          
        }//end while(xmllist)
        
      }//end if(rootnode)
      return p;
    }catch(const Exception &e){
      GLOBAL_ERROR_STREAM(e << endl);
      return p;
    }
  }
  
  /*Begin***************** Private Helpers *******************************/
    void TVGLibXML2Database::readChannels(){
    GLOBAL_DEBUG_STREAM("TVGLibXML2Database::readChannels()" << endl);

    string filename = workDir + "/channels.xml";

    //if the xml-file does not exist, exit funktion
    if(!fileExists(filename))
      return ;

    //Security-Check
    if(!SSecurityCheck::getInstance().checkFilenameForReading(filename)) {
      GLOBAL_ERROR_STREAM("Security check not passed for file " << filename << ", check .nmmrc for options" << endl);
      return;
    }

    MutexGuard mg(tvmutex);

    xml->read(filename);

    if(xml){
      root = xml->root();

      if(root){
        channels.clear();
        //read channels out of channels-XML-file
        XMLNodeList channs;
        channs = root->children("channel");
        
        XMLNodeIterator iter, stop;
        iter = channs.begin();
        stop = channs.end();
        
        while(iter != stop){
          XMLNode* node = *iter;
          iter++;
          
          //set the id of the channel
          ChannelDescription c;
          c.setID(node->property("id")->value());

          //add the display-names, specified in the XML-File, to the channels-list
          XMLNodeList names = node->children("display-name");
          XMLNodeIterator i, end;
          i = names.begin();
          end = names.end();
          list<string> namesToAdd;
          
          while (i != end){
            XMLNode* s = *i;
            i++;
            namesToAdd.push_back(s->getData());
          }
          c.setChannelNames(namesToAdd);

          //add channel-description to list
          channels.push_back(c);
        }//end while

        GLOBAL_DEBUG_STREAM("# of channels: " << channels.size() << endl);
      }//end if(rootnode)
    }//end if(xml)
  }//end readChannels()

  void TVGLibXML2Database::saveChannels(){
    GLOBAL_DEBUG_STREAM("TVGLibXML2Database: saveChannels()" << endl);

    string filename = workDir + "/channels.xml";

    //Security-Check
    if(!SSecurityCheck::getInstance().checkFilenameForWriting(filename)) {
      GLOBAL_ERROR_STREAM("Security check not passed for file " << filename << ", check .nmmrc for options" << endl);
      return;
    }

    ofstream fout(filename.c_str(), ios::out);
    if(fout){
      fout << "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?> \n" << endl;
      fout << "<!DOCTYPE tv SYSTEM \"xmltv.dtd\"> \n" << endl;
      fout << "<tv generator-info-name=\"NMM TVGuide\">\n" << endl;; 
 
      list<ChannelDescription>::iterator it = channels.begin();

      while(it != channels.end()){
        ChannelDescription cd = *it;
        it++;
        fout << "<channel id=\"" << cd.getID() << "\">" << endl;
        list<string> names = cd.getChannelNames();
        list<string>::iterator namesIt = names.begin();
        while(namesIt != names.end()){
          string s = *namesIt;
          namesIt++;
          fout << "<display-name>" << s << "</display-name>" << endl;
        }
        fout << "</channel>" << endl;

      }
      fout << "\n </tv> \n" << endl;
      fout.close();      
    }
  }//end saveChannels()
  
  void TVGLibXML2Database::readCategories(){
    GLOBAL_DEBUG_STREAM("TVGLibXML2Database: readCategories()" << endl);

    string filename = workDir + "/categories.xml";

    //if the xml-file does not exist, exit funktion
    if(!fileExists(filename))
      return ;

    //Security-Check
    if(!SSecurityCheck::getInstance().checkFilenameForReading(filename)) {
      GLOBAL_ERROR_STREAM("Security check not passed for file " << filename << ", check .nmmrc for options" << endl);
      return;
    }

    MutexGuard mg(tvmutex);
    xml->read(filename);
    
    if(xml) root = xml->root();
    if(root){
      categories.clear();
      XMLNodeList xmlCatList = root->children("category");
      XMLNodeIterator xmlCatIt = xmlCatList.begin();
      while(xmlCatIt != xmlCatList.end()){
	XMLNode* cat = *xmlCatIt;
	xmlCatIt++;
	
	XMLProperty* catName = cat->property("name");
	if(catName){
          categories.push_back(transformSpecialCharsOut(catName->value()));
        }else{
          categories.push_back(transformSpecialCharsOut(cat->getData()));
        }
      }
    }
    categories.sort();
    categories.unique();
  }
  void TVGLibXML2Database::saveCategories(){
    GLOBAL_DEBUG_STREAM("TVGLibXML2Database: saveCategories()" << endl);

    string filename = workDir + "/categories.xml";

    //Security-Check
    if(!SSecurityCheck::getInstance().checkFilenameForWriting(filename)) {
      GLOBAL_ERROR_STREAM("Security check not passed for file " << filename << ", check .nmmrc for options" << endl);
      return;
    }
    //MutexGuard mg(tvmutex);

    ofstream fout(filename.c_str(), ios::out);
    if(fout){

      fout << "<?xml version=\"1.0\" encoding=\"UTF-8\"?> \n" << endl;
      fout << "<!DOCTYPE tv SYSTEM \"xmltv.dtd\"> \n" << endl;
      fout << "<tv generator-info-name=\"NMM TVGuide\">\n" << endl;; 

      list<string>::iterator i = categories.begin();
      while(i != categories.end()){
	string s = *i;
	i++;
	fout << "<category>" << transformSpecialCharsIn(s) << "</category>" <<  endl;
      }
      fout << "\n </tv> \n" << endl;

      fout.close();
    }
  }

  /**
   *Function to transform umlauts and other language-dependent chars to UTF8 when needed
   */
  string TVGLibXML2Database::transformSpecialCharsIn(string s){
    XMLEncoder e("ISO-8859-1");
    string ret = e.getUTF8(s);
    return ret;
  }//end transformSpecialCharsIn

  /**
   *Function to transform umlauts and other language-dependent chars from UTF8 when needed
   */
  string TVGLibXML2Database::transformSpecialCharsOut(string s){
    XMLEncoder e("ISO-8859-1");
    string ret = e.getEncodedString(s);
    return ret;
  }//end transformSpecialCharsOut

  bool TVGLibXML2Database::fileExists(string file){
    string command = "test -f " + file;
    if(!system(command.c_str())){
      GLOBAL_DEBUG_STREAM("File " << file << " exists" << endl);
      return true;
    }else{
      GLOBAL_DEBUG_STREAM("File " << file << " not exists" << endl);
      return false;
    }
  }


  bool TVGLibXML2Database::compareStrings(const string s1, const string s2){
    
    string regexString("");
    string testString("");
    //use the smaller string as pattern
    if(s1.size() < s2.size()){
      regexString = s1;
      testString = s2;
    }else{
      regexString = s2;
      testString = s1;
    }

    //TODO: get rid of special signs like !;*&%

    regex_t exp;
    if(!(regcomp(&exp, regexString.c_str(), REG_EXTENDED | REG_ICASE))){
      if(regexec(&exp, testString.c_str() ,0,NULL,0) == REG_NOMATCH){
        return false;
      }else{
        return true;
      }
    }
    return false;
  }
  /*End******************** Private Helpers ******************************/


}//end namespace NMM

