/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Christoph Wellner <chwellner@graphics.cs.uni-sb.de>
 *                     
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef NMM_TVGXERCESCDATABASE_HPP
#define NMM_TVGXERCESCDATABASE_HPP NMM_TVGXERCESCDATABASE_HPP

#include "nmm/utils/DirReader.hpp"
#include "nmm/utils/thread/Thread.hpp"

#include "nmm/interfaces/misc/tvguide/ChannelDescription.hpp"
#include "nmm/interfaces/misc/tvguide/ProgramDescription.hpp"
#include "nmm/interfaces/misc/tvguide/ProgramFilter.hpp"

#include "nmm/interfaces/misc/tvguide/ITVGuide.hpp"

#include <xercesc/util/TransService.hpp>
#include <xercesc/dom/DOMImplementation.hpp>
#include <xercesc/dom/DOMBuilder.hpp>
#include <xercesc/dom/DOMWriter.hpp>

#include <string>
#include <list>
#include <time.h>

//Dummy-classes

XERCES_CPP_NAMESPACE_USE

namespace NMM{
  class XercesTranscoder{
    
  private:
    XMLTranscoder* transcoder;
    XMLTransService::Codes res;
    char* encoding;
    
    XMLCh* fgUnicodeForm;
  public:
    XercesTranscoder();
    XercesTranscoder(string encoding);
    ~XercesTranscoder();
    
    string transcodeTo(const XMLCh* s);
    XMLCh* transcodeFrom(string s);
  };

  class WrongTimeFormatException : public Exception{
  private:
  public:
    WrongTimeFormatException() : Exception("Can't convert Time. Probably the time has the wrong format"){
      setTypeName("WrongTimeFormatException");
    }
    WrongTimeFormatException *copy(){ return new WrongTimeFormatException(*this);}
  };

  class TVGXercescDatabase{

  private:
    list<ChannelDescription> channels; /** List with all available channels */
    string workDir; /** Directory, where the information is stored */
    list<string> categories; /** List with all available categories */
    
    int epsilon;
    
    DirReader directory;
    DirReader insertdirectory;

    ThreadMutex getMutex;
    ThreadMutex searchMutex;
    ThreadMutex insertMutex;
    ThreadMutex fileAccessMutex;

    bool termUpdate;

    void transform(string &s){
      size_t sIt = string::npos;
      size_t sOld = 0;
      do{
        sIt = s.find_first_of(" ", sOld);
        if(sIt != string::npos) s.erase(sIt,1);
        sOld = sIt+1;
      }while(sIt != string::npos);

      string sout = s;
      for (unsigned int i = 0; i <= s.length(); i++){
        sout[i] = toupper(s[i]);
      }
      s=sout;
    }
   
    ITVGuide_var itvguide; //to send events to TVGuide-Object when eg new channels arrive

    map<ProgramDescription::sourceType_t, unsigned int> priorityMap;

    //DOMBuilder-Parameters
    bool doNamespaces;
    bool doSchema;
    bool schemaFullChecking;

    //DOMWriter-Parameters
    XMLCh* gOutputEncoding;
    bool   gSplitCdataSections;
    bool   gDiscardDefaultContent;
    bool   gUseFilter;
    bool   gFormatPrettyPrint;
    bool   gWriteBOM;

    DOMImplementation *impl;
    DOMBuilder *parser;
    DOMWriter *writer;

    XercesTranscoder *transcoder;

  public:

    TVGXercescDatabase(ITVGuide* itvg);
    ~TVGXercescDatabase();

    /** Insert Channel into DB. Indicates, that information for this channel is available
     *@param id ID of the channel
     *@param name Name of the channel
     */
    Result insertChannel(const string &id, const string &name);
    /** Set the directory, where information is stored
     *@param path Complete path of the directory
     */
    void setDirectory(const string &path);

    Result insertChannelDescription(const ChannelDescription &cd);

    /** Get the directory, where the informations are stored
     *@return complete path of the directory
     */
    string getDirectory() throw (const NMM::Exception&) ;

    /** Get a list with all available categories
     *@return List of categories
     */
    list<string> getCategories() throw (const NMM::Exception&);
    
    /** Get a list with all available channels. 
     *@return list of channels
     */
    list<ChannelDescription> getChannelsList()throw (const NMM::Exception&);
    /** Look for name of channel with given ID
     *@param id ID of the channel
     *@return string name of channel, if the ID is known
     *@return empty-string if ID is unknown
     */
    string  getChannelById(const string &id);
    /** Look for ID of channel with given name
     *@param name name of the channel
     *@return string ID of channel, if the name is known
     *@return empty-string if name is unknown
     */
    string getChannelByName(const string &name);

    /** Insert an event to the database
     *@param pd ProgramDescription containing the information of the event
     */
    Result insertEvent(const ProgramDescription &progDesc);
    /** look for a specific event in the time-inteval [start,end] on the channel with the given ID
     *@param start Start-Time of the interval. Time-representation: YYYYMMDDhhmm
     *@param end End-Time of the interval. Time-representation: YYYYMMDDhhmm
     *@param id ID of the Channel
     *@return ProgramDescription of the Event
     */
    ProgramDescription getEvent(const string &start, const string &end, const string &id);
    /** Get a list of all Events of a given channel on a specific day
     *@param day Representation: YYYYMMDD
     *@param id id of the channel
     *@return List of ProgramDescriptions containing the events found
     */
    list<ProgramDescription> getEventsOnDayOnChannel(const string &day, const string &chan);

    /** Search for Events matching the given filter
     *@param pf ProgramFilter to apply on the search
     *@return List with ProgramDescriptions of events found
     */
    list<ProgramDescription> search(const ProgramFilter &pf);

    /** Add an Extended Description to an event. if the event is not available in the database, it is added.
     *@param start Start-time of the event
     *@param end Stop-Time of the event
     *@param id ID of the channel
     *@param desc description to be added
     */
    void addExtDescription(const string& start, const string& end, const string& id, const string& desc);

    /** get the current event of the channel specified. Make sure, your hardware-clock is set up correctly
     *@param channel either name or id
     *@return ProgramDescription with event found
     */
    ProgramDescription getCurrentEventOnChannel(const string& channel);
    /** get the next event of the channel specified. Make sure, your hardware-clock is set up correctly
     *@param channel either name or id
     *@return ProgramDescription with event found
     */
    ProgramDescription getFollowingEventOnChannel(const string& channel);


    map<ProgramDescription::sourceType_t, unsigned int> getPriorityMap(){ return priorityMap;}
    void setPriorityMap( map<ProgramDescription::sourceType_t, unsigned int> m){ priorityMap = m;}
    
  private:
    void readChannels(); //read channels from file
    void saveChannels(); //save channels to file

    void readCategories(); //read categories from file
    void saveCategories(); //save categories to file

    bool compareStrings(const string s1, const string s2); //test if 2 Strings are approximately equal

    int compareTimes(string t1, string t2); //test, if times given as string are differnt, or not

    string transformToDatabaseTime(string s); //transforms the given time to the databases internal time, wich is localtime

    void mergeInformation(DOMNode* child, ProgramDescription pd, DOMDocument *doc);

    void updateNextStartTime(DOMDocument *doc, DOMNode* child, string pdEndTime, const ProgramDescription &pd);
    void updatePreviousEndTime(DOMNode* child, string pdStartTime, const ProgramDescription &pd);

    bool fileExists(string file);

  public:
    DOMNode* toDOMNode(ProgramDescription pd, DOMDocument *doc);
    DOMNode* toDOMNode(ChannelDescription cd, DOMDocument *doc);
    ProgramDescription toProgramDescription(DOMNode *node);
    ChannelDescription toChannelDescription(DOMNode *node);

  };
}
#endif
