/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Christoph Wellner <chwellner@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "TVCardSearch.hpp"
#include "nmm/base/proxy/tcp/TCPStrategy.hpp"
#include "nmm/interfaces/video/tv/ITVCard.hpp"
#include "nmm/base/ProxyObject.hpp"
#include "nmm/interfaces/video/tv/IChannelInfo.hpp"

namespace NMM{

  TVCardSearch::TVCardSearch()
    : app(NULL)
  {
  }

  TVCardSearch::~TVCardSearch(){}

  TVCardDescription TVCardSearch::searchNewTVCard(string channelName, int &channelNumber) throw (NoCardFoundException){
    GLOBAL_DEBUG_STREAM("TVCardSearch::searchNewTVCard(" << channelName << ")" << endl);

    if(app){
      
      for(list<TVCardDescription>::iterator i = tvcardhosts.begin(); i != tvcardhosts.end(); i++){
       
        TVCardDescription tvcardDesc = *i;

        //Test, if we already know this TV-Card for the given channel
        channelMap_t adrlist = tvcards[channelName];
        bool cardAlreadyKnown = false;
        for(channelMap_t::iterator j = adrlist.begin(); j != adrlist.end(); j++){
          if(j->second == tvcardDesc){
            //If we already know this card, then it is already tested for the requested channel
            cardAlreadyKnown = true;
          }
        }
        if(cardAlreadyKnown) continue;
        
        //If it is a new Card, request the ITVCard-Interface for querying the channel
        //NodeDescription tvcard_nd(tvcardDesc.getType());
        NodeDescription tvcard_nd = tvcardDesc;
        tvcard_nd.setLocation(tvcardDesc.getAddress().getHostname());
        tvcard_nd.setSharingType(NodeDescription::SHARED_OR_EXCLUSIVE_THEN_SHARED);
        
        NodeDescription devnull_nd("DevNullNode");
        devnull_nd.setSharingType(NodeDescription::EXCLUSIVE);
        GraphDescription gd;

        try{
          gd.addEdge(tvcard_nd, devnull_nd);
          app->getRegistry().requestGraph(gd);
        }catch(const Exception &e){
          GLOBAL_ERROR_STREAM(e << endl);
          continue;  
        }

        try{
          gd.realizeGraph();
          INode *tvcard = gd.getINode(tvcard_nd);
          if(tvcard){
            IChannelInfo_var itvcard(tvcard->getParentObject()->getCheckedInterface<IChannelInfo>());
            
            if(itvcard->hasChannel(channelNumber, channelName) == SUCCESS){
              cerr << "Have channel " << channelName << " on " << tvcard_nd.getNodeName() << endl;
              tvcardDesc.setType(tvcard_nd.getNodeName());
              gd.unrealizeGraph();
              app->getRegistry().releaseGraph(gd);
            }else{
              cerr << "Have no channel " << channelName << " on " << tvcard_nd.getNodeName() << endl;
              gd.unrealizeGraph();
              app->getRegistry().releaseGraph(gd);
              continue;
            }
          }else{
            GLOBAL_ERROR_STREAM("can't request tvcard from " << tvcardDesc.getAddress().getHostname() << endl);
            continue;
          }
          
        }catch(const Exception &e){
          GLOBAL_ERROR_STREAM("TVCardSearch: error requesting Interface " << e << endl);
          app->getRegistry().releaseGraph(gd);
          continue;
        }catch(...){
          GLOBAL_ERROR_STREAM("TVcardSearch: unknown exception" << endl);
          app->getRegistry().releaseGraph(gd);
          continue;
        }


        //OK, we have a new TVCard ....
        
        //add new TVcard to list
        if(tvcards.find(channelName) == tvcards.end()){
          channelMap_t l;
          l.push_back(pair<int, TVCardDescription>(channelNumber, tvcardDesc));
          tvcards[channelName] = l;
        }else{
          tvcards[channelName].push_back(pair<int, TVCardDescription>(channelNumber, tvcardDesc));
        }

        return tvcardDesc;
      }
    }else{
      GLOBAL_ERROR_STREAM("TVCardSearch: no NMMApplication set" << endl);
    }

    //if we haven't found a new TVCard, throw Exception
    throw NoCardFoundException(channelName);

  }

  channelMap_t TVCardSearch::getCards(string channelName){
    //Test, if we have cards for this channel. If not, throw Exception
    if(tvcards.find(channelName) == tvcards.end())
      throw NoCardFoundException(channelName);

    return tvcards[channelName];
  }

  void TVCardSearch::requestHosts(list<TVCardDescription> &l){
    l = tvcardhosts;
  }

  Result TVCardSearch::validateHost(TVCardDescription card){

    for(map<string , channelMap_t >::iterator i = tvcards.begin(); i != tvcards.end(); i++){
      channelMap_t &cmap = i->second;
      for(channelMap_t::iterator j = cmap.begin(); j != cmap.end(); j++){
        if(j->second == card){
          tvcards[i->first].erase(j);
          return SUCCESS;
        }
      }
    }
    return SUCCESS;
  } 

  void TVCardSearch::addHost(const NodeDescription &nd){
    TVCardDescription tmp(nd);
    tvcardhosts.push_back(tmp);
  }
}
