/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 */

#include "LADSPAData2IntNode.hpp"
#include "nmm/base/format/Property.hpp"
#include "nmm/base/registry/Plugin.hpp"

NMM_REGISTER_PLUGIN(NMM::LADSPAData2IntNode, "LADSPAData2IntNode")

namespace NMM {

    LADSPAData2IntNode::LADSPAData2IntNode(const char* name /*= "LADSPAData2IntNode"*/,
					   const StreamQueue::Mode qmode /*= StreamQueue::DEFAULT_MODE*/,
					   const int qsize /*= StreamQueue::DEFAULT_SIZE*/) 
    : GenericProcessorNode(name,qmode,qsize) {

	channels = 0;
    }


    LADSPAData2IntNode::~LADSPAData2IntNode(void) {
    }


    Result LADSPAData2IntNode::doInit(void) {
	/* specify the default input format */
	Format* in_format  = getOwnInputProperty()->addNewFormat("audio/raw-float");
	in_format->addStringParamValue(Format::endian_param,"little");
	in_format->addStringParamValue(Format::encoding_param,"planar");
	in_format->addStringParamValue(Format::sign_param,"signed");
	in_format->addWildcard(Format::sample_rate_param);
	in_format->addWildcard(Format::channels_param);
	in_format->addIntParamValue(Format::bitpersample_param,16);
	getOwnInputProperty()->setDefaultFormat(in_format);

	/* default output format */
	Format* out_format = getOwnOutputProperty()->addNewFormat("audio/raw");
	out_format->addStringParamValue(Format::endian_param,"little");
	out_format->addStringParamValue(Format::encoding_param,"linear");
	out_format->addStringParamValue(Format::sign_param,"signed");
	out_format->addWildcard(Format::sample_rate_param);
	out_format->addIntParamValue   (Format::channels_param,1);
	out_format->addIntParamValue   (Format::channels_param,2);
	out_format->addIntParamValue(Format::bitpersample_param,16);
	getOwnOutputProperty()->setDefaultFormat(out_format);
	
	in_format->setIOPartner(out_format);
	out_format->setIOPartner(in_format);

	return SUCCESS;
    } // LADSPAData2IntNode::doInit


    Result LADSPAData2IntNode::doDeinit(void) {
	getOwnInputProperty()->erase();
	getOwnOutputProperty()->erase();

	return SUCCESS;
    } // LADSPAData2IntNode::doDeinit


    Result LADSPAData2IntNode::doInitOutput(void) {
	const Format* in_format = getInputFormat();
	if (in_format == NULL) { return FAILURE; }

	Format* i_format = new Format(*in_format);
	getOwnWorkingInputProperty()->addNewFormat(i_format);
	getOwnWorkingInputProperty()->setDefaultFormat(i_format);

	channels = in_format->getIntValue(Format::channels_param);
	
	Format* o_format = new Format(*getOwnOutputProperty()->getDefaultFormat());
	o_format->resolveWildcards(in_format);
	o_format->addIntParamValue(Format::channels_param,channels);
	getOwnWorkingOutputProperty()->addNewFormat(o_format);
	getOwnWorkingOutputProperty()->setDefaultFormat(o_format);
	
	i_format->setIOPartner(o_format);
	o_format->setIOPartner(i_format);

	return SUCCESS;
    } // LADSPAData2IntNode::doInitOutput 


    Result LADSPAData2IntNode::doDeinitOutput(void) {
	getOwnWorkingInputProperty()->erase();
	getOwnWorkingOutputProperty()->erase();

	return SUCCESS;
    } // LADSPAData2IntNode::doDeinitOutput


    Result LADSPAData2IntNode::doActivate(void) {
	const Format* in_f = getInputFormat();
	if (!in_f) {
	    ERROR_STREAM("doActivate without input format!" << endl);
	    return FAILURE;
	}
	channels   = in_f->getIntValue(Format::channels_param);

	return SUCCESS;
    } // LADSPAData2IntNode::doActivate


    Result LADSPAData2IntNode::doDeactivate(void) {
	channels = 0;

	return SUCCESS;
    } // LADSPAData2IntNode::doDeactivate


    Message* LADSPAData2IntNode::processBuffer(Buffer* in_buffer) {
	LADSPA_Data* ld_buf = (LADSPA_Data*)(in_buffer->getData());
	int samples = in_buffer->getUsed()/sizeof(LADSPA_Data);
	int spc = samples / channels;

	Buffer* out = getNewBuffer(samples*sizeof(short));
	short* sh_buf = (short*)(out->getData());

	for (int i=0;i<spc;++i) {
	    for (int c=0;c<channels;++c) {
		(*sh_buf) = static_cast<short>(ld_buf[c*spc+i]);
		++sh_buf;
	    }
	}

	in_buffer->release();

	return out;
    } // LADSPAData2IntNode::processBuffer
    
}; // namespace NMM
