/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 */

#include "LADSPANode.hpp"
#include "nmm/base/format/Property.hpp"
#include "nmm/utils/xml/xml++.hpp"
#include "nmm/utils/utils.hpp"
#include "nmm/utils/SecurityCheck.hpp"
#include "nmm/base/registry/Plugin.hpp"

#include <stdio.h>
#include <sys/dir.h>
#include <dirent.h>
#include <iostream>

NMM_REGISTER_PLUGIN(NMM::LADSPANode, "LADSPANode")

namespace NMM {

  //
  // construction/destruction
  //
  
  LADSPANode::LADSPANode(const char* name,const StreamQueue::Mode qmode /* = StreamQueue::DEFAULT_MODE */,
                         const int qsize /* = StreamQueue::DEFAULT_SIZE */)
    : GenericProcessorNode(name,qmode,qsize) {
    
    /* initialize member variables */
    configFilename = "";
    samplerate     = 0;
    numInChannels  = 0;
    numOutChannels = 0;
  } // LADSPANode::LADSPANode
  

  LADSPANode::~LADSPANode(void) {
  } // LADSPANode::~LADSPANode
  

  //
  // ILADSPA interface
  //
  
  bool LADSPANode::addLADSPAPlugin(const string& value,string& name,unsigned int& execpos) {
    LADSPA_Plugin_t *plugin = loadLADSPAPlugin(value);
    if (plugin) {
      // by default use the Label of the LADSPA plugin
      if ("" == name) { name = std::string(plugin->descriptor->Label); }

      // check if "name" exists and make it unique if necessary
      while (LADSPA_NameMap.find(name) != LADSPA_NameMap.end()) {
        name += "*";
      }

      // clamp execpos to [0..LADPSA_Plugins.size()]
      if (execpos > LADSPA_Plugins.size()) { execpos = LADSPA_Plugins.size(); }

      std::vector<LADSPA_Plugin_t*>::iterator it2 = LADSPA_Plugins.begin();
      unsigned int i=0;
      while (it2 != LADSPA_Plugins.end() && i < execpos) {
        it2++;
        i++;
      }
      if (execpos > i) { execpos = i; }
      LADSPA_Plugins.insert(it2,plugin);


      std::map<std::string,unsigned int>::iterator it;
      for (it=LADSPA_NameMap.begin();it!=LADSPA_NameMap.end();it++) {
        if (it->second >= execpos) { it->second++; }
      }

      LADSPA_NameMap[name] = execpos;
      DEBUG_STREAM("added LADSPA plugin " << value << " as " << name << std::endl);
      return true;
    }
    return false;
  } // LADSPANode::addLADSPAPlugin


  void LADSPANode::removeLADSPAPluginByName(const string& name) {
    std::map<std::string,unsigned int>::iterator it = LADSPA_NameMap.find(name);
    if (it == LADSPA_NameMap.end()) {
      ERROR_STREAM("LADSPA plugin identified by " << name << " cannot be removed (not found)" << std::endl);
      return;
    }

    unsigned int execpos = it->second;
    LADSPA_Plugin_t *plugin = LADSPA_Plugins[execpos];
    LADSPA_Plugins[execpos] = 0;
    LADSPA_Plugins.erase(LADSPA_Plugins.begin()+execpos);
    
    LADSPA_NameMap.erase(it);
    for (it=LADSPA_NameMap.begin();it != LADSPA_NameMap.end(); it++) {
      if (it->second > execpos) { it->second--; }
    }
    
    unloadLADSPAPlugin(plugin);
    DEBUG_STREAM("LADSPA plugin identified by " << name << " removed" << std::endl);
  } // LADSPANode::removeLADSPAPluginByName


  void LADSPANode::removeLADSPAPluginByIndex(const unsigned int& index) {
    if (index >= LADSPA_Plugins.size()) {
      ERROR_STREAM("LADSPA plugin with index " << index << " cannot be removed (out of range)" << std::endl);
      return;
    }

    unsigned int execpos = index;
    LADSPA_Plugin_t *plugin = LADSPA_Plugins[execpos];
    LADSPA_Plugins[execpos] = 0;
    LADSPA_Plugins.erase(LADSPA_Plugins.begin()+execpos);

    std::map<std::string,unsigned int>::iterator it;
    std::map<std::string,unsigned int>::iterator it2 = LADSPA_NameMap.begin();
    for (it = LADSPA_NameMap.begin(); it != LADSPA_NameMap.end();it++) {
      if (it->second == execpos) { 
        it2 = it;
        continue;
      } else if (it->second > execpos) { 
        it->second--; 
      }
    }

    LADSPA_NameMap.erase(it2);
    unloadLADSPAPlugin(plugin);
    DEBUG_STREAM("LADSPA plugin with index " << execpos << " removed" << std::endl);
  } // LADSPANode::removeLADSPAPluginByIndex


  void LADSPANode::setConfigFilename(const string& value) {
    configFilename = getNMMResource(value);
  } // LADSPANode::setConfigFilename


  void LADSPANode::setConfigProfileByName(const string& name,const string& profile) {
    LADSPA_Plugin_t *plugin = getLADSPAPlugin(name);
    loadPresets(plugin,profile);
  } // LADSPANode::setConfilProfileByName


  void LADSPANode::setConfigProfileByIndex(const unsigned int& index, const string& profile) {
    LADSPA_Plugin_t *plugin = getLADSPAPlugin(index);
    loadPresets(plugin,profile);
  } // LADSPANode::setConfigProfileByIndex


  void LADSPANode::resetToBaseValuesByName(const string& name) {
    LADSPA_Plugin_t *plugin = getLADSPAPlugin(name);
    
    for (unsigned int i=0;i<plugin->numControlIn;++i) {
      plugin->bufControlIn[i] = plugin->baseValues[i];
    }
  } // LADSPANode::resetToBaseValueByName


  void LADSPANode::resetToBaseValuesByIndex(const unsigned int& index) {
    LADSPA_Plugin_t *plugin = getLADSPAPlugin(index);
    
    for (unsigned int i=0;i<plugin->numControlIn;++i) {
      plugin->bufControlIn[i] = plugin->baseValues[i];
    }
  } // LADSPANode::resetToBaseValueByIndex


  void LADSPANode::setControlInValueByName(const string& name, const unsigned int& port, const float& value) {
    DEBUG_STREAM("setControlInValueByName" << std::endl);
    DEBUG_STREAM("name = " << name << std::endl);
    DEBUG_STREAM("port = " << port << std::endl);
    DEBUG_STREAM("value = " << value << std::endl);

    LADSPA_Plugin_t *plugin = getLADSPAPlugin(name);
    int p = clampCIPort(plugin,port);
    LADSPA_Data v = clampCIValue(plugin,p,value);
    plugin->bufControlIn[p] = v;
  } // LADSPANode::setControlInValueByName


  void LADSPANode::setControlInValueByIndex(const unsigned int& index,const unsigned int& port, const float& value) {
    DEBUG_STREAM("setControlInValueByIndex" << std::endl);
    DEBUG_STREAM("index = " << index << std::endl);
    DEBUG_STREAM("port = " << port << std::endl);
    DEBUG_STREAM("value = " << value << std::endl);

    LADSPA_Plugin_t *plugin = getLADSPAPlugin(index);
    int p = clampCIPort(plugin,port);
    LADSPA_Data v = clampCIValue(plugin,port,value);
    plugin->bufControlIn[p] = v;
  } // LADSPANode::setControlInValueByIndex


  float LADSPANode::getControlInValueByName(const string& name,const unsigned int& port) {
    LADSPA_Plugin_t *plugin = getLADSPAPlugin(name);
    int p = clampCIPort(plugin,port);
    return plugin->bufControlIn[p];
  } // LADSPANode::getControlInValueByName


  float LADSPANode::getControlInValueByIndex(const unsigned int& index,const unsigned int& port) {
    LADSPA_Plugin_t *plugin = getLADSPAPlugin(index);
    int p = clampCIPort(plugin,port);
    return plugin->bufControlIn[p];
  } // LADSPANode::getControlInValueByIndex


  string LADSPANode::getControlInNameByName(const string& name,const unsigned int& port) {
    LADSPA_Plugin_t *plugin = getLADSPAPlugin(name);
    int p = clampCIPort(plugin,port);

    int hit=0;
    unsigned int i;
    for (i=0;i < plugin->descriptor->PortCount;i++) {
      if (LADSPA_IS_PORT_INPUT(plugin->descriptor->PortDescriptors[i]) && 
	  LADSPA_IS_PORT_CONTROL(plugin->descriptor->PortDescriptors[i])) {
	++hit;
	if (hit == p) { break; }
      }
    }
    return plugin->descriptor->PortNames[i];
  } // LADSPANode::getControlInNameByName


  string LADSPANode::getControlInNameByIndex(const unsigned int& index,const unsigned int& port) {
    LADSPA_Plugin_t *plugin = getLADSPAPlugin(index);
    int p = clampCIPort(plugin,port);

    int hit=0;
    unsigned int i;
    for (i=0;i < plugin->descriptor->PortCount;i++) {
      if (LADSPA_IS_PORT_INPUT(plugin->descriptor->PortDescriptors[i]) && 
	  LADSPA_IS_PORT_CONTROL(plugin->descriptor->PortDescriptors[i])) {
	++hit;
	if (hit == p) { break; }
      }
    }
    return plugin->descriptor->PortNames[i];
  } // LADSPANode::getControlInNameByIndex


  void LADSPANode::setControlInBaseByName(const string& name,const unsigned int& port, const float& value) {
    setControlInBase(getLADSPAPlugin(name),port,value);
  } // LADSPANode::setControlInBaseByName


  void LADSPANode::setControlInBaseByIndex(const unsigned int& index,const unsigned int& port, const float& value) {
    setControlInBase(getLADSPAPlugin(index),port,value);
  } // LADSPANode::setcontrolInBaseByIndex


  float LADSPANode::getControlInBaseByName(const string& name,const unsigned int& port) {
    LADSPA_Plugin_t *plugin = getLADSPAPlugin(name);
    int p = clampCIPort(plugin,port);
    return plugin->baseValues[p];
  } // LADSPANode::getControlInBaseByName


  float LADSPANode::getControlInBaseByIndex(const unsigned int& index,const unsigned int& port) {
    LADSPA_Plugin_t *plugin = getLADSPAPlugin(index);
    int p = clampCIPort(plugin,port);
    return plugin->baseValues[p];
  } // LADSPANode::getControlInBaseByIndex


  unsigned int LADSPANode::getLADSPAPluginCount() {
    return LADSPA_Plugins.size();
  } // LADSPANode::getLADSPAPluginCount


  unsigned int LADSPANode::getLADSPAPluginIndex(const string& name) {
    if (LADSPA_NameMap.find(name) == LADSPA_NameMap.end()) {
      ERROR_STREAM("no LADSPA plugin loaded for identifier " << name << std::endl);
      throw Exception("unknown LADSPA plugin identifier");
      return 0;
    }
    return LADSPA_NameMap[name];
  } // LADSPANode::getLADSPAPluginIndex


  string LADSPANode::getLADSPAPluginName(const unsigned int& index) {
    int i=index;
    if (0 > i) { i=0; }
    if ((unsigned int)i >= LADSPA_NameMap.size()) { i = LADSPA_NameMap.size()-1; }

    std::map<std::string,unsigned int>::iterator it;
    for (it = LADSPA_NameMap.begin();it != LADSPA_NameMap.end();it++) {
      if (it->second == (unsigned int)i) {
	break;
      }
    }
    return it->first;
  } // LADSPANode::getLADSPAPluginName


  unsigned int LADSPANode::getControlInCountByName(const string& name) {
    LADSPA_Plugin_t *plugin = getLADSPAPlugin(name);
    return plugin->numControlIn;
  } // LADSPANode::getcontrolInCountByName


  unsigned int LADSPANode::getControlInCountByIndex(const unsigned int& index) {
    LADSPA_Plugin_t *plugin = getLADSPAPlugin(index);
    return plugin->numControlIn;
  } // LADSPANode::getcontrolInCountByIndex


  bool LADSPANode::hasControlInUpperBoundByName(const string& name,const unsigned int& port) {
    LADSPA_Plugin_t *plugin = getLADSPAPlugin(name);
    unsigned int p = clampCIPort(plugin,port);

    LADSPA_PortRangeHint rangehint = plugin->descriptor->PortRangeHints[p];
    return (LADSPA_IS_HINT_BOUNDED_ABOVE(rangehint.HintDescriptor));
  } // LADSPANode::hasControlInUpperBoundByName  
  

  bool LADSPANode::hasControlInUpperBoundByIndex(const unsigned int& index,const unsigned int& port) {
    LADSPA_Plugin_t *plugin = getLADSPAPlugin(index);
    unsigned int p = clampCIPort(plugin,port);

    LADSPA_PortRangeHint rangehint = plugin->descriptor->PortRangeHints[p];
    return (LADSPA_IS_HINT_BOUNDED_ABOVE(rangehint.HintDescriptor));
  }// LADSPANode::hasControlInUpperBoundByIndex
  

  bool LADSPANode::hasControlInLowerBoundByName(const string& name,const unsigned int& port) {
    LADSPA_Plugin_t *plugin = getLADSPAPlugin(name);
    unsigned int p = clampCIPort(plugin,port);

    LADSPA_PortRangeHint rangehint = plugin->descriptor->PortRangeHints[p];
    return (LADSPA_IS_HINT_BOUNDED_BELOW(rangehint.HintDescriptor));
  } // LADSPANode::hasControlInLowerBoundByName  
  

  bool LADSPANode::hasControlInLowerBoundByIndex(const unsigned int& index,const unsigned int& port) {
    LADSPA_Plugin_t *plugin = getLADSPAPlugin(index);
    unsigned int p = clampCIPort(plugin,port);

    LADSPA_PortRangeHint rangehint = plugin->descriptor->PortRangeHints[p];
    return (LADSPA_IS_HINT_BOUNDED_BELOW(rangehint.HintDescriptor));
  } // LADSPANode::hasControlInLowerBoundByIndex 


  void LADSPANode::getControlInBoundsByName(const string& name,const unsigned int& port, float& lower, float& upper) {
    LADSPA_Plugin_t *plugin = getLADSPAPlugin(name);
    unsigned int p = clampCIPort(plugin,port);
    LADSPA_PortRangeHint rangehint = plugin->descriptor->PortRangeHints[p];
    lower   = rangehint.LowerBound;
    upper   = rangehint.UpperBound;
  } // LADSPANode::getControlInBoundsByName 


  void LADSPANode::getControlInBoundsByIndex(const unsigned int& index,const unsigned int& port, float& lower, float& upper) {
    LADSPA_Plugin_t *plugin = getLADSPAPlugin(index);
    unsigned int p = clampCIPort(plugin,port);
    LADSPA_PortRangeHint rangehint = plugin->descriptor->PortRangeHints[p];
    lower   = rangehint.LowerBound;
    upper   = rangehint.UpperBound;
  } // LADSPANode::getControlInBoundsByIndex  


  string LADSPANode::getLADSPAPluginRealNameByName(const string& name) {
    LADSPA_Plugin_t *plugin = getLADSPAPlugin(name);
    return std::string(plugin->descriptor->Name);
  } // LADSPANode::getLADSAPluginRealNameByName  


  string LADSPANode::getLADSPAPluginRealNameByIndex(const unsigned int& index) {
    LADSPA_Plugin_t *plugin = getLADSPAPlugin(index);
    return std::string(plugin->descriptor->Name);
  } // LADSPANode::getLADSAPluginRealNameByIndex 


  string LADSPANode::getLADSPAPluginLabelByName(const string& name){
    LADSPA_Plugin_t *plugin = getLADSPAPlugin(name);
    return std::string(plugin->descriptor->Label);
  } // LADSPANode::getLADSAPluginRealLabelByName


  string LADSPANode::getLADSPAPluginLabelByIndex(const unsigned int& index){
    LADSPA_Plugin_t *plugin = getLADSPAPlugin(index);
    return std::string(plugin->descriptor->Label);
  } // LADSPANode::getLADSAPluginRealLabelByIndex
  

  string LADSPANode::getLADSPAPluginLibraryNameByName(const string& name){
    LADSPA_Plugin_t *plugin = getLADSPAPlugin(name);
    return plugin->libname;
  } // LADSPANode::getLADSAPluginLibraryNameByName


  string LADSPANode::getLADSPAPluginLibraryNameByIndex(const unsigned int& index){
    LADSPA_Plugin_t *plugin = getLADSPAPlugin(index);
    return plugin->libname;
  } // LADSPANode::getLADSAPluginLibraryNameByIndex



  //
  // INode interface
  //

  Result LADSPANode::doInit(void) {
    /* define supported input and output formats */
    Format* in_format  = getOwnInputProperty()->addNewFormat("audio/raw-float");
    in_format->addWildcard(Format::endian_param);
    in_format->addStringParamValue(Format::encoding_param,"planar");
    in_format->addWildcard(Format::sign_param);
    in_format->addWildcard(Format::sample_rate_param);
    in_format->addWildcard(Format::channels_param);
    in_format->addWildcard(Format::bitpersample_param);
    getOwnInputProperty()->setDefaultFormat(in_format);
    
    Format* out_format = getOwnOutputProperty()->addNewFormat("audio/raw-float");
    out_format->addWildcard(Format::endian_param);
    out_format->addStringParamValue(Format::encoding_param,"planar");
    out_format->addWildcard(Format::sign_param);
    out_format->addWildcard(Format::sample_rate_param);
    out_format->addWildcard(Format::channels_param);
    out_format->addWildcard(Format::bitpersample_param);
    getOwnOutputProperty()->setDefaultFormat(out_format);
    
    in_format->setIOPartner(out_format);
    out_format->setIOPartner(in_format);
    
    return SUCCESS;
  } // LADSPANode::doInit
    

  Result LADSPANode::doDeinit(void) {
    getOwnInputProperty()->erase();
    getOwnOutputProperty()->erase();
    
    return SUCCESS;
  } // LADSPANode::doDeinit
    

  Result LADSPANode::doActivate(void) {
    /* activating all loaded LADSPA plugins */
    DEBUG_STREAM("activating LADSPA-plugins" << std::endl);
    for (unsigned int i=0;i<LADSPA_Plugins.size();++i) {
      if (LADSPA_Plugins[i]->descriptor->activate) { 
        LADSPA_Plugins[i]->descriptor->activate(LADSPA_Plugins[i]->plugin); 
      }
    }
    return SUCCESS;
  } /* LADSPANode::doActivate */
  
  
  Result LADSPANode::doDeactivate(void) {
    /* deactivating all LADSPA-plugins */
    DEBUG_STREAM("deactivating LADSPA-plugins" << std::endl);
    for (unsigned int i=0;i<LADSPA_Plugins.size();++i) {
      if (LADSPA_Plugins[i]->descriptor->deactivate) { 
        LADSPA_Plugins[i]->descriptor->deactivate(LADSPA_Plugins[i]->plugin); 
      }
    }
    
    DEBUG_STREAM("LADSPA-plugins deactivated" << std::endl);
    return SUCCESS;
  } /* LADSPANode::doDeactivate */


  Result LADSPANode::doInitOutput(void) {
    const Format* in_format = getInputFormat();
    if (in_format == NULL) { return FAILURE; }
    
    Format* i_format = new Format(*in_format);
    getOwnWorkingInputProperty()->addNewFormat(i_format);
    getOwnWorkingInputProperty()->setDefaultFormat(i_format);
    
    /* keep the format parameters we need in mind */
    samplerate = in_format->getIntValue(Format::sample_rate_param);
    numInChannels  = in_format->getIntValue(Format::channels_param);
    
    /* prepare LADSPA plugins */
    
    if (!LADSPA_Plugins.size()) {
      ERROR_STREAM("no LADSPA plugin(s) loaded before initOutputing" << std::endl);
      return FAILURE;
    }

    for (unsigned int idx=0;idx<LADSPA_Plugins.size();++idx) {
      initLADSPAPlugin(LADSPA_Plugins[idx]);
    } // for (unsigned int idx ...
    
   
    LADSPA_Plugin_t *p = LADSPA_Plugins[LADSPA_Plugins.size()-1];
    numOutChannels = numInChannels/p->numAudioIn * p->numAudioOut;
    DEBUG_STREAM(numOutChannels << " output audio channels" << std::endl);
    
    
    Format* o_format = new Format(*getOwnOutputProperty()->getDefaultFormat());
    o_format->resolveWildcards(in_format);
    o_format->addIntParamValue(Format::channels_param,numOutChannels);
    getOwnWorkingOutputProperty()->addNewFormat(o_format);
    getOwnWorkingOutputProperty()->setDefaultFormat(o_format);
    
    
    if (numInChannels != o_format->getIntValue(Format::channels_param)) {
      ERROR_STREAM("input channel number is different from output channel number" << std::endl);
      return FAILURE;
    }
    
    i_format->setIOPartner(o_format);
    o_format->setIOPartner(i_format);
    
    return SUCCESS;
  } // LADSPANode::doInitOutput
  
  
  Result LADSPANode::doDeinitOutput(void) {
    getOwnWorkingInputProperty()->erase();
    getOwnWorkingOutputProperty()->erase();
    
    /* release the memory used for port-data */
    DEBUG_STREAM("releasing portbuffers" << std::endl);

    for (unsigned int i=0;i<LADSPA_Plugins.size();++i) {
      deinitLADSPAPlugin(LADSPA_Plugins[i]);
    }
    
    return SUCCESS;
  } // LADSPANode::doDeinitOutput


  Message* LADSPANode::processBuffer(Buffer* in) {
    LADSPA_Data *ld_out = (LADSPA_Data*)(in->getData());
    int samples = in->getUsed()/sizeof(LADSPA_Data);
    
    int spc = samples / numInChannels;
    int outsize = numOutChannels*spc * sizeof(LADSPA_Data);
    
    Buffer* out = getNewBuffer(outsize);
    LADSPA_Data* ld_in = (LADSPA_Data*)(out->getData());
    
    LADSPA_Data *tmp;
    for (unsigned int p=0;p<LADSPA_Plugins.size();++p) {
      tmp = ld_in;
      ld_in = ld_out;
      ld_out = tmp;
        
      LADSPA_Plugin_t *plugin = LADSPA_Plugins[p];
      DEBUG_STREAM("plugin " << p << " : " << plugin->descriptor->Label << std::endl);

      int rounds(numInChannels/plugin->numAudioIn);

      for (int loop=0;loop < rounds;++loop) {
        int ofs_i = loop * plugin->numAudioIn * spc;
        for (unsigned int c=0;c<plugin->numAudioIn;++c) {
          unsigned int pos = ofs_i + c*spc;
          plugin->bufAudioIn[c]->data = ld_in+pos;
          plugin->descriptor->connect_port(plugin->plugin,plugin->bufAudioIn[c]->port,plugin->bufAudioIn[c]->data);
        }
        
	int ofs_o = loop * plugin->numAudioOut * spc;
	for (unsigned int c=0;c<plugin->numAudioOut;++c) {
          unsigned int pos = ofs_o + c*spc;
          plugin->bufAudioOut[c]->data = ld_out+pos;
          plugin->descriptor->connect_port(plugin->plugin,plugin->bufAudioOut[c]->port,plugin->bufAudioOut[c]->data);
	}

        if (plugin->descriptor->run_adding) {
          plugin->descriptor->run_adding(plugin->plugin,spc);
        } else {
          plugin->descriptor->run(plugin->plugin,spc);
        }
      } // for (loop ...
    }
    
    in->release();
    out->setUsed(outsize);	
    
    return out;
  } // LADSPANode::processBuffer
  
  
  //
  // internaly used utility functions
  //

  unsigned int LADSPANode::getPortCountByType(const LADSPA_Descriptor *desc, unsigned int porttype) {
    unsigned int counter = 0;
    for (unsigned int i=0;i<desc->PortCount;i++) {
      if (porttype == (desc->PortDescriptors[i] & porttype)) {
            counter++;
          }
    }
    return counter;
  } // LADSPANode::getPortCountByType


  LADSPANode::LADSPA_Plugin_t* LADSPANode::getLADSPAPlugin(const string& name) throw (const Exception&) {
    if (LADSPA_NameMap.find(name) == LADSPA_NameMap.end()) {
      ERROR_STREAM("no LADSPA plugin loaded for identifier " << name << std::endl);
      throw Exception("unknown LADSPA plugin identifier");
      return 0;
    }

    return LADSPA_Plugins[LADSPA_NameMap[name]];
  } // LADSPANode::getLADSPAPlugin(name)
  

  LADSPANode::LADSPA_Plugin_t* LADSPANode::getLADSPAPlugin(const unsigned int index) throw (const Exception&) {
    if (index >= LADSPA_Plugins.size()) {
      ERROR_STREAM("invalid index to LADSPA plugin" << std::endl);
      throw Exception("invalid LADSPA plugin index");
      return 0;
    }

    return LADSPA_Plugins[index];
  } // LADSPANode::getLADSPAPlugin(index)
  

  LADSPA_Data LADSPANode::clampCIValue(const LADSPA_Plugin_t *plugin,const unsigned int port,LADSPA_Data value) {
    LADSPA_PortRangeHint prhints = plugin->descriptor->PortRangeHints[port];
    LADSPA_PortRangeHintDescriptor prdesc = prhints.HintDescriptor;
    
    if (LADSPA_IS_HINT_BOUNDED_BELOW(prdesc) && (value < prhints.LowerBound)) {
      // lower bound
      value = prhints.LowerBound;
    }

    if (LADSPA_IS_HINT_BOUNDED_ABOVE(prdesc) && (value > prhints.UpperBound)) {
      // upper bound
      value = prhints.UpperBound;
    }
    
    if (LADSPA_IS_HINT_INTEGER(prdesc)) {
      // value must be an "integer" 
      value = (int)(value + 0.5f);
    }

    return value;
  } // LADSPANode::clampCIValue


  unsigned int LADSPANode::clampCIPort(const LADSPA_Plugin_t *plugin,unsigned int port) {
    if (port >= plugin->numControlIn) { port = plugin->numControlIn-1; }
    return port;
  } // LADSPANode::clampCIPort

  
  LADSPANode::LADSPA_Plugin_t* LADSPANode::loadLADSPAPlugin(const string& value) throw (const Exception&) {
    /* +++ search for a LADSPA plugin with name or label equal to value +++
       +++ terminate with the first hit                                 +++ */

    LADSPA_Plugin_t *plugin = 0;
    std::string ladspa_path;

    // read out LADSPA_PATH
    char *env = getenv("LADSPA_PATH");
    if (env) { ladspa_path = env; } else { ladspa_path = ""; }

    // search in all directories specified in LADSPA_PATH
    
    if ("" == ladspa_path) {
      ERROR_STREAM("LADSPA_PATH not set. LADSPA plugin " << value << " cannot be loaded" << std::endl);
      return 0;
    }

    bool found = false;
    bool done = false;
    while (!done) {
      // extract a path from LADSPA_PATH

      unsigned int pos;
      std::string path;
      if ((pos = ladspa_path.find(":")) != ladspa_path.npos) {
        path = ladspa_path.substr(0,pos) + "/";
        ladspa_path.erase(0,pos+1);
      } else { // last entry in LADSPA_PATH
        path = ladspa_path + "/";
        done = true;
      }
      DEBUG_STREAM("searching in " << path << " for LADSPA plugin " << value << std::endl);

      // search through the directory path (but only if we pass the security check)
      if (!SSecurityCheck::getInstance().checkFilenameForReading(path)) {
        ERROR_STREAM("security checknot passed for directory " 
                     << path 
                     << ". check .nmmrc for options" 
                     << std::endl);
        continue;
      }

      DIR *directory = opendir(path.c_str());
      if (!directory) { continue; }
      
      dirent *dir_entry;
      while (!found && (dir_entry = readdir(directory))) {
        std::string filename = path + dir_entry->d_name;

        if (!SSecurityCheck::getInstance().checkFilenameForReading(filename)) { 
          ERROR_STREAM("Security check not passed for file " << filename << ", check .nmmrc for options" << endl);
          continue; 
        }

        DEBUG_STREAM("searching in file " << filename << " for LADSPA plugin " << value << std::endl);

        // check if filename is a shared library containing LADSPA-Plugins
        lt_dlhandle libhandle = lt_dlopen(filename.c_str()); //,RTLD_NOW);
        if (!libhandle) { continue; } // not a shared object file

        LADSPA_Descriptor_Function descriptor_func = (LADSPA_Descriptor_Function)lt_dlsym(libhandle,"ladspa_descriptor");
        if (!descriptor_func) { continue; } // no LADSPA plugins

        // check if there is a LADSPA plugin with name or with label
        
        const LADSPA_Descriptor *descriptor;
        int loop = 0;
        while (true) {
          descriptor = descriptor_func(loop);
          if (!descriptor) { break; }

          if ((value == std::string(descriptor->Name)) ||
              (value == std::string(descriptor->Label))) {
            // hit -> search was succesful

            found = true;
            done  = true;
            plugin = new LADSPA_Plugin_t;
            plugin->libhandle  = libhandle;
            plugin->libname    = filename;
            plugin->descriptor = descriptor;
            break;
          }

          ++loop;
        } // while(true)

        if (!found) {
          lt_dlclose(libhandle);
        } else {
          DEBUG_STREAM("LADSPA plugin " << value << " found in " << plugin->libname << std::endl);
        }

      } // while ((dir_entry = ...
      
      closedir(directory);

    }; // while (!done)

    return plugin;
  } // LADSPANode::loadLADSPAPlugin


  void LADSPANode::unloadLADSPAPlugin(LADSPA_Plugin_t *plugin) {
    if (!plugin) {
      return;
    }

    lt_dlclose(plugin->libhandle);
    delete plugin;
    plugin = 0;
  } // LADSPANode::unloadLADSPAPLugin


  void LADSPANode::initLADSPAPlugin(LADSPA_Plugin_t *plugin) throw (const Exception&) {
    /* instantiate the plugins and get a handle to it */
    DEBUG_STREAM("instantiating LADSPA-plugin" << plugin->descriptor->Label << std::endl);

    plugin->plugin = plugin->descriptor->instantiate(plugin->descriptor,samplerate);

    if (!plugin->plugin) {
      ERROR_STREAM("LADSPA-Plugin " << plugin->descriptor->Label << " could not be instantiated" << std::endl);
      throw Exception("LADSPA-Plugin could not be instantiated");
    }
    
    plugin->numAudioIn    = getPortCountByType(plugin->descriptor,LADSPA_PORT_INPUT  | LADSPA_PORT_AUDIO);
    plugin->numAudioOut   = getPortCountByType(plugin->descriptor,LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO);
    plugin->numControlIn  = getPortCountByType(plugin->descriptor,LADSPA_PORT_INPUT  | LADSPA_PORT_CONTROL);
    plugin->numControlOut = getPortCountByType(plugin->descriptor,LADSPA_PORT_OUTPUT | LADSPA_PORT_CONTROL);
    
    DEBUG_STREAM("LADSPA Plugin " << plugin->descriptor->Label << ": " << std::endl);
    DEBUG_STREAM("\t" << plugin->numAudioIn << " Audio-in ports" << std::endl);
    DEBUG_STREAM("\t" << plugin->numAudioOut << " Audio-out ports" << std::endl);
    DEBUG_STREAM("\t" << plugin->numControlIn << " Control-in ports" << std::endl);
    DEBUG_STREAM("\t" << plugin->numControlOut << " Control-out ports" << std::endl);
    
    for (unsigned int i=0;i<plugin->numAudioIn;++i)   { plugin->bufAudioIn.push_back(new LADSPA_Buffer_t); }
    for (unsigned int i=0;i<plugin->numAudioOut;++i)  { plugin->bufAudioOut.push_back(new LADSPA_Buffer_t); }
    for (unsigned int i=0;i<plugin->numControlIn;++i) { plugin->baseValues.push_back(0.0f); }
    
    loadPresets(plugin,"default");

    plugin->bufControlIn.resize(plugin->numControlIn);
    for (unsigned int i=0;i<plugin->numControlIn;++i) { plugin->bufControlIn[i] = plugin->baseValues[i]; }

    // assign the buffers to the ports
    unsigned int cauin  = 0;
    unsigned int cauout = 0;
    unsigned int cconin = 0;
    for (unsigned int i=0;i<plugin->descriptor->PortCount;++i) {
      int portdesc = plugin->descriptor->PortDescriptors[i];
      
      if (LADSPA_IS_PORT_INPUT(portdesc)) {
        if (LADSPA_IS_PORT_AUDIO(portdesc)) { // audio-in port
          plugin->bufAudioIn[cauin]->data = 0;
          plugin->bufAudioIn[cauin]->port = i;
          ++cauin;
        }
        
        if (LADSPA_IS_PORT_CONTROL(portdesc)) { // control-in port
            plugin->descriptor->connect_port(plugin->plugin,i,&plugin->bufControlIn[cconin]);
            cconin++;
        }
      } // ...PORT_INPUT
      
      if (LADSPA_IS_PORT_OUTPUT(portdesc)) {
        if (LADSPA_IS_PORT_AUDIO(portdesc)) { // audio-out port
          plugin->bufAudioOut[cauout]->data = 0;
          plugin->bufAudioOut[cauout]->port = i;
          ++cauout;
        }
        
        if (LADSPA_IS_PORT_CONTROL(portdesc)) { // control-out port (they aren't supported)
          plugin->descriptor->connect_port(plugin->plugin,i,&(plugin->bufControlOut));
        }
      } // ...PORT_OUTPUT
    }
  } // LADSPANode::initLADSPAPlugin
  

  void LADSPANode::deinitLADSPAPlugin(LADSPA_Plugin_t *plugin) {
    for (unsigned int j=0;j<plugin->bufAudioIn.size();++j) {
      delete plugin->bufAudioIn[j];
    }
    
    for (unsigned int j=0;j<plugin->bufAudioOut.size();++j) {
      delete plugin->bufAudioOut[j];
    }
    
    plugin->bufAudioIn.clear();
    plugin->bufAudioOut.clear();
    plugin->bufControlIn.clear();

    plugin->descriptor->cleanup(plugin->plugin);
  } // LADSPANode::deinitLADSPAPlugin

  
  bool LADSPANode::loadPresets(LADSPA_Plugin_t *plugin,const string &profile) {
    /* getting the "pure" filename of the LADSPA plugin */
    std::string libfilename = plugin->libname;
    unsigned int pos;
    if ((pos = libfilename.rfind("/")) != std::string::npos) {
      libfilename.erase(0,pos+1);
    }


    /* searching the entry for the loaded LADSPA plugin */

    if (!SSecurityCheck::getInstance().checkFilenameForReading(configFilename)) {
      ERROR_STREAM("Security check not passed for file " << configFilename << ", check .nmmrc for options" << endl);
      return false;
    }

    XMLTree *xml = new XMLTree(configFilename);
    XMLNode *rootnode = xml->root();
    if (!rootnode) { return false; }

    if (rootnode->name() != "configuration") { return false; }

    XMLNodeList nlist = rootnode->children("LADSPA");
    XMLNode *node;
    XMLNodeIterator iter = nlist.begin();

    if (iter == nlist.end()) {
      ERROR_STREAM("readConfig() : " << configFilename << " no <LADSPA> tag found" << std::endl);
      return false;
    }

    XMLNode *hit = 0;
    XMLProperty *prop;

    while (!hit && iter != nlist.end()) {
      node = *iter;
      XMLNodeList nlist2 = node->children("plugin");
      XMLNodeIterator iter2 = nlist2.begin();
      if (iter2 == nlist2.end()) {
        ERROR_STREAM("readConfig() : " << configFilename << " no <plugin> tag found" << std::endl); 
        return false;
      }

      XMLNode *node2;
      while (!hit && iter2 != nlist2.end()) {
        node2 = *iter2;
        
        prop = node2->property("library");
        std::string lib = prop->value();
        prop = node2->property("label");
        std::string label = prop->value();
        prop = node2->property("profile");
        std::string prof = prop->value();

        if (lib == libfilename && label == std::string(plugin->descriptor->Label) && prof == profile) {
          hit = *iter2;
          break;
        }

        iter2++;
      }
      
      iter++;
    }


    if (!hit) {
      ERROR_STREAM("readConfig() : preset profile " << profile << " not found for LADSPA plugin " << plugin->descriptor->Label << std::endl);
      return false;
    }

    nlist = hit->children("ControlIn");
    iter = nlist.begin();
    if (iter == nlist.end()) {
      ERROR_STREAM("readConfig() : empty preset list for LADSPA plugin " << plugin->descriptor->Label  << std::endl);
      return false;
    }
    
    for (iter=nlist.begin();iter!=nlist.end();iter++) {
      node = *iter;

      int port;
      LADSPA_Data preset;
      prop = node->property("port");
      sscanf(prop->value().c_str(),"%d",&port);
      sscanf(node->getData().c_str(),"%f",&preset);

      DEBUG_STREAM("set input control base for port " << port << " to value " << preset << std::endl);
      setControlInBase(plugin,port,preset);
    }

    return true;
  } // LADSPANode::loadPresets


  void LADSPANode::setControlInBase(LADSPA_Plugin_t *plugin, const unsigned int& port, const float& value) {
    int p = clampCIPort(plugin,port);
    LADSPA_Data v = clampCIValue(plugin,p,value);
    plugin->baseValues[p] = v;
  } // LADSPANode::setControlInBase


} /* namespace NMM */
