/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        NMM work group,
 *                    http://www.networkmultimedia.org
 *                     
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include <stdio.h>

#include "GenericReadNode.hpp"
#include "nmm/base/format/Property.hpp"
#include "nmm/interfaces/file/ITrack.hpp"
#include "nmm/interfaces/general/ITrackSize.hpp"
#include "nmm/interfaces/general/progress/IProgressListener.hpp"
#include "nmm/interfaces/base/sync/ISyncReset.hpp"
#include "nmm/utils/utils.hpp"
#include "nmm/utils/SecurityCheck.hpp"
#include "nmm/base/utils/GraphURL.hpp"

#include "nmm/base/registry/Plugin.hpp"

NMM_REGISTER_PLUGIN(NMM::GenericReadNode, "GenericReadNode")

namespace NMM {
  GenericReadNode::GenericReadNode(const char *name /* = "GenericReadNode" */,
				   StreamQueue::Mode mode /* = StreamQueue::MODE_SUSPEND */,
				   int queue_size /* = StreamQueue::DEFAULT_SIZE */)
    : GenericSourceNode(name, mode, queue_size) {
    sendEvent = false;
    send_format = false;

    filesize = 0;
    filepos = 0;
    seekto = 0;
    
    buffer_size = 10000;
    buffers_before_progress = 5;
    buffers_progress        = 0;
    send_progress = false;
  }
  
  void GenericReadNode::sendProgressInformation(const bool& in) {
    send_progress = in;
  }

  void GenericReadNode::setProgressInterval(const int& buffers) {
    buffers_before_progress = buffers;
  }
  
  void GenericReadNode::setFilename(const string& name) {
    filename = name;
  }  
  
  void GenericReadNode::setBufferSize(const int& size) {
    buffer_size = size;
  }
  
  int GenericReadNode::getBufferSize() {
    return buffer_size;
  }
  
  Result GenericReadNode::doInit() {
    Format* out_format = getOwnOutputProperty()->addNewFormat("*/*");
    getOwnOutputProperty()->setDefaultFormat(out_format);
    
    return SUCCESS;
  }
  
  Result GenericReadNode::doInitOutput() {
    if(filename.substr(0,1) != "/") {
      // get current working directory
      char* ret = 0;
      size_t size = 128;
      do {
	char* buf = new char[size];
	ret = getcwd(buf, size);
	if(ret) {
	  filename = buf + string("/") + filename;
	}
	else {
	  size += 128;
	}
	delete[] buf;
      } while(ret == 0);

      WARNING_STREAM("No absolute path given, using file in current working directory: " << filename << endl);
    }

    if(!SSecurityCheck::getInstance().checkFilenameForReading(filename)) {
      ERROR_STREAM("Security check not passed for file " << filename << ", check .nmmrc for options" << endl);
      return FAILURE;
    }
    
    in_fstream.clear();

    GraphURL url(string() + "file://" + filename);
    Format* out_format = url.getFormat();
    DEBUG_STREAM("Open File: " << filename << endl);
    
    in_fstream.open(filename.c_str(), ios::in);
    if(in_fstream.fail()) {
      ERROR_STREAM( "Error while opening file " << filename << endl);
      return FAILURE;
    }
    
    // get filesize and seek to beginning again
    in_fstream.seekg(0, ios::end);
    filesize = in_fstream.tellg();
    in_fstream.seekg(0, ios::beg);
    
    sendEvent = true;
    if(isOutputConnected("default")) {
      send_format = true;
    }
    
    getOwnWorkingOutputProperty()->addNewFormat(out_format);
    getOwnWorkingOutputProperty()->setDefaultFormat(out_format);

    setProducingFlag(true);
    
    return SUCCESS;
  }

  Result GenericReadNode::doDeinitOutput() {
    in_fstream.close();
    return SUCCESS;
  }
  
  Message *GenericReadNode::processBuffer(Buffer *) {
    Buffer *out;
    int bytes_read;

    if ( send_progress ) {
      Event e;
      if ( buffers_progress > buffers_before_progress ) {
        buffers_progress = 0;
        CEvent* cevent(new CEvent(IProgressListener::create_setProgress(filepos, filesize)));
        cevent->setMandatoryBit(false);
        return cevent;
      }
      buffers_progress++;
    }

    if (sendEvent) {
      CEvent* ce = new CEvent();

      sendEvent = false;
      DEBUG_STREAM("Sending startTrack..." << endl);
      ce->insertEvent(ITrack::create_startTrack(filename)); 
      DEBUG_STREAM("Sending syncReset" << endl);
      ce->insertEvent(ISyncReset::create_syncReset());
      DEBUG_STREAM("Sending setFileSize" << endl);
      ce->insertEvent(ITrackSize::create_trackSize(filesize));
      

      if(send_format) {
	send_format = false;

        //change outputformat
        setOutputFormat(*getOwnWorkingOutputProperty()->getDefaultFormat());

	DEBUG_STREAM("Sending setFormat" << endl);
	ce->insertEvent(IFormatChange::create_setFormat(*getOutputFormat()));
      }

      return ce;
    }
    else {
      if ((out = getNewBuffer( buffer_size )) != 0) {
        if (seekto != 0) {
	  // test if we would get negative
	  if (seekto < 0 && labs(seekto) > static_cast<int64_t>(filepos)) {
	    // if seek result would be negative go to beginning of file
            in_fstream.seekg(0, ios::beg);
            filepos = 0;
	    DEBUG_STREAM("Seeked to beginning " << seekto << " from " << filepos << ", filesize " << filesize << endl);
	  }
	  else if (filepos + seekto > filesize) {
	    // set back to short-before-file-end
            in_fstream.seekg(0, ios::end);
	    DEBUG_STREAM("Seeked to end " << seekto << " from " << filepos << ", filesize " << filesize << endl);
	  }
	  else {
	    // seek normally
            in_fstream.seekg(filepos + seekto);
            filepos = filepos+seekto;
	    DEBUG_STREAM("Seeked normally " << seekto << " from " << filepos << ", filesize " << filesize << endl);
	  }
          seekto = 0;

          CEvent* cevent = new FlushEvent();
          cevent->insertEvent(cevent->size(), ISyncReset::create_syncReset());
          return cevent;
        }
        
        in_fstream.read(out->getData(), out->getSize());
        bytes_read = in_fstream.gcount();
        if (!in_fstream.eof()) {
	  out->setUsed(bytes_read);
          filepos+=bytes_read;
	  return(out);
	}
	else // assume EOF
	{
	  DEBUG_STREAM("Sending end_track..." << endl);
	  setProducingFlag(false); // nothing left to do, so stop producing
	  return new CEvent(Event(ITrack::endTrack_event));
	}
      }
      else {
	return 0;
      }
    }
  }

  GenericReadNode::~GenericReadNode() {
    in_fstream.close();
  }

  /******************** ISeekable interface ********************/
  bool GenericReadNode::hasPercentSeek() const {
    return true;
  }

  void GenericReadNode::seekPercent(const Rational& percent) {
    Rational internal_percent(percent);
    if (internal_percent > Rational(1)) {
      ERROR_STREAM("Seeking more than 100% not possible, defaulting to end!" << endl);
      internal_percent = Rational(1);
    }
    else if (internal_percent < Rational(-1)) {
      ERROR_STREAM("Seeking less than -100% not possible, defaulting to beginning!" << endl);
      internal_percent = Rational(-1);
    }

    Rational rat_seekto = Rational(filesize) * internal_percent;
    seekto = rat_seekto.getNumerator() / rat_seekto.getDenominator();
    DEBUG_STREAM(seekto << " = " << internal_percent << " * " << filesize << endl);
  }
   
  void GenericReadNode::seekPercentTo(const Rational& percent) {
    seekPercent(percent - Rational(filepos, filesize));
  }

  bool GenericReadNode::hasBytesSeek() const {
    return true;
  }

  void GenericReadNode::seekBytes(const int64_t& offset) {
    if ((offset < 0) && (labs(offset) > static_cast<int64_t>(filepos))) {
      ERROR_STREAM("Seeking to negativ byte-offset not possible, defaulting to beginning of file!" << endl);
      seekBytesTo(0);
    } else {
      seekBytesTo(filepos + offset);
    }
  }

  void GenericReadNode::seekBytesTo(const u_int64_t& bytes) {
    u_int64_t internal_bytes(bytes);

    if (internal_bytes > filesize) {
      ERROR_STREAM("Seeking over EOF not possible, defaulting to EOF!" << endl);
      internal_bytes = filesize;
    }

    seekto = internal_bytes - filepos;

    DEBUG_STREAM(seekto << " = " << internal_bytes << " - " << filepos << endl);
  }

  bool GenericReadNode::hasTimeSeek() const {
    return false;
  }

  void GenericReadNode::seekTime(const Interval&) {
  }

  void GenericReadNode::seekTimeTo(const Time&) {
  }
  /******************** ISeekable interface ********************/
}
