/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        NMM work group,
 *                    http://www.networkmultimedia.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "MP3ReadNode.hpp"

#include <fcntl.h>

#include "nmm/base/registry/NodeDescription.hpp"

#include "nmm/interfaces/file/ITrack.hpp"
#include "nmm/interfaces/file/ISeekReceiver.hpp"
#include "nmm/interfaces/base/IFormatChange.hpp"
#include "nmm/interfaces/general/progress/IProgressListener.hpp"
#include "nmm/interfaces/general/ITrackDuration.hpp"
#include "nmm/interfaces/base/sync/ISyncReset.hpp"
#include "nmm/interfaces/audio/icecast/IIcecast.hpp"
#include "nmm/interfaces/audio/ITagInfo.hpp"
#include "nmm/utils/SecurityCheck.hpp"
#include "nmm/base/registry/Plugin.hpp"

#include "genre.h"


#define ID3_HEADER_SIZE 10
#define ID3_TAG_SIZE 128


NMM_REGISTER_PLUGIN(NMM::MP3ReadNode, "MP3ReadNode")

namespace NMM {

  MP3ReadNode::MP3ReadNode(const char *name,
			   const StreamQueue::Mode queue_mode,
			   const int size)
    : GenericSourceNode(name, queue_mode, size),
      track_duration(0, 0)
  {
    filename      = "";
    buffer_size   = 10000;
    
    fmpeg       = 0;
    flayer      = 0;    
    fprotbit    = 0;
    fbitrate    = 0;
    fsample_rate = 0;
    fpadding    = 0;
    fnumchannels = 0;

    filepos          = 0;

    fd = -1;
    event_state = EVENT_SET_TITLE;
    seekto = 0;
    send_progress = false;
    found_tag = false;

    buffers_before_progress = 5;
    buffers_progress        = 0;
    
  }
  
  void MP3ReadNode::setFilename(const string& name) {
    filename = name;
    
    // if already playing one file, immediately open the new file.
    if (fd != -1) {
      if (FAILURE == openFile()) {
        // XXX: in this case, setFilename can fail!
        setProducingFlag(false);
        return;
      } else {
        setProducingFlag(true);
      }

      // as an "end_track" causes the application to react (e.g. in setting the *next* filename)
      // we do *not* send an "end_track" after setting a filename
//       if (event_state == EVENT_END_TRACK) {
        event_state = EVENT_START_TRACK;
//       } else {
//         event_state = EVENT_NEXT_TRACK;
//       }
    }
  }
  
  void MP3ReadNode::setBufferSize(const int& size) {
    buffer_size = size;
  }
  
  int MP3ReadNode::getBufferSize() {
    return buffer_size;
  }
  
  void MP3ReadNode::sendProgressInformation(const bool& f ) {
    send_progress = f;
  }

  void MP3ReadNode::setProgressInterval(const int& buffers) {
    buffers_before_progress = buffers;
  }
  
  Result MP3ReadNode::doInit()
  {
    fmpeg       = 0;
    flayer      = 0;
    fprotbit    = 0;
    fbitrate    = 0;
    fsample_rate = 0;
    fpadding    = 0;
    fnumchannels = 0;
    filepos          = 0;
    
    timestamp.sync_time = Time(0);
    timestamp.stream_counter = 0;
    timestamp.is_valid = false;
    
    // general output format
    Format* out_format = getOwnOutputProperty()->addNewFormat("audio/mpeg");
    out_format->addWildcard(Format::channels_param); //
    out_format->addWildcard(Format::layer_param);
    out_format->addIntParamValue(Format::audio_codec_param, 0x55);
    out_format->addIntParamValue(Format::sample_rate_param, 48000);
    out_format->addIntParamValue(Format::sample_rate_param, 44100);
    out_format->addIntParamValue(Format::sample_rate_param, 32000);
    out_format->addIntParamValue(Format::sample_rate_param, 24000);
    out_format->addIntParamValue(Format::sample_rate_param, 22050);
    out_format->addIntParamValue(Format::sample_rate_param, 16000);
    out_format->addIntParamValue(Format::sample_rate_param, 12000);
    out_format->addIntParamValue(Format::sample_rate_param, 11025);
    out_format->addIntParamValue(Format::sample_rate_param, 8000);
    out_format->addIntParamValue(Format::bitpersample_param, 16);
    
    out_format = getOwnOutputProperty()->addNewFormat("audio/mpeg");
    out_format->addWildcard(Format::layer_param);
    out_format->addIntParamValue(Format::audio_codec_param, 0x55);
    getOwnOutputProperty()->setDefaultFormat(out_format);
    
    return SUCCESS;
  }
  
  Result MP3ReadNode::openFile() 
  {
    char test[4];
    int bytes_read = 0;
    int mp3pos	 = 0;
    filepos          = 0;
    
    if(!SSecurityCheck::getInstance().checkFilenameForReading(filename)) {
      ERROR_STREAM("Security check not passed for file " << filename << ", check .nmmrc for options" << endl);
      return FAILURE;
    }
    
    // close old file first
    close(fd);
    
    DEBUG_STREAM("Open File: " << filename << endl);
    if ((fd = open(filename.c_str(), O_RDONLY)) < 0) {
      ERROR_STREAM("Could not open file!"<< endl);
      return FAILURE;
    }
    
    // get filesize and seek to beginning
    filesize = lseek(fd, 0, SEEK_END);
    lseek(fd, 0, SEEK_SET);
    DEBUG_STREAM("File size is " << filesize << endl);
    
    mp3pos = isWav();
    
    // the ID3v2 header is at the begining of the file to avoid error
    // you must read just after the IDV3 Header if it exists
    isID3v2();    	
   
    // Find first header an check for valid mp3 file
    bytes_read = read(fd, test, 4 );
    while ((bytes_read == 4) && (!(parseHeader(test)))) {    
      bytes_read = read(fd, test, 4);
    }

    char tag[4];
    
    lseek(fd, -128, SEEK_END);
    read(fd, tag, 3);
    if (!strncmp(tag, "TAG", 3)) {
      // found id3 v1 tag
      int count;
      count = read(fd, title, 30);
      title[count]='\0';

      count = read(fd,interpret,30);
      interpret[count]='\0';

      count = read(fd,album,30);
      album[count] = '\0';

      count = read(fd,year,4);
      year[count] = '\0';

      count = read(fd,comment,30);
      comment[count]='\0';

      char nr[1];
      count = read(fd,nr,1);
      nr[count]='\0';
      genre=*nr;

      found_tag = true;
    }

    
    // go to begin of mp3 data
    lseek(fd, mp3pos, SEEK_SET);
    
    // clear working output property
    getOwnWorkingOutputProperty()->erase();
    
    // initialize working output formats
    Format* out_format = getOwnWorkingOutputProperty()->addNewFormat("audio/mpeg");
    out_format->addIntParamValue(Format::layer_param, flayer);
    out_format->addIntParamValue(Format::audio_codec_param, 0x55);
    // getOwnWorkingOutputProperty()->setDefaultFormat(out_format);
    new_simple_format = *out_format;

    // complex format. uses values obtained in parseHeader
    out_format = getOwnWorkingOutputProperty()->addNewFormat("audio/mpeg");
    out_format->addIntParamValue(Format::channels_param, fnumchannels);
    out_format->addIntParamValue(Format::layer_param, flayer);
    out_format->addIntParamValue(Format::audio_codec_param, 0x55);
    out_format->addIntParamValue(Format::sample_rate_param, fsample_rate);
    out_format->addIntParamValue(Format::bitpersample_param, 16);
    getOwnWorkingOutputProperty()->setDefaultFormat(out_format);
    
    u_int64_t track_length = (filesize - mp3pos) * 8 / fbitrate;
    track_duration = Interval(track_length / 1000, (track_length - track_length / 1000 * 1000) * 1000);

    return SUCCESS;
  }
  
  Result MP3ReadNode::doInitOutput() {
    // open first file.
    // we must have a file name before we can initialize the output 
    if (FAILURE == openFile()) {
      return FAILURE;
    }

    // store complex format as "old" format for testing when having to send format changes!
    old_format = *getOwnWorkingOutputProperty()->getDefaultFormat();

    //event_state = EVENT_START_TRACK;
    event_state = EVENT_SET_TITLE;
    
    return SUCCESS;
  }

  Result MP3ReadNode::doActivate() {
    // get the connection format
    const Format* out_format(getOutputFormat());

    if (!out_format) {
      ERROR_STREAM("Output format is NULL!" << endl);
      return FAILURE;
    }

    // store connection format as the "old" one
    old_connection_format = *out_format;

    return SUCCESS;
  }

  Message *MP3ReadNode::processBuffer(Buffer *) {
    int bytes_read;
    
    // progress events can be sent at any time.
    // it is not an event state.
    if(send_progress) {
      Event e;
      if ( buffers_progress > buffers_before_progress ) {
        buffers_progress = 0;
        CEvent* cevent = new CEvent(IProgressListener::create_setProgress(filepos, filesize));
        cevent->setMandatoryBit(false);
        cevent->setTimestamp(timestamp);
        ++timestamp.stream_counter;
        return cevent;
      }
      buffers_progress++;
    }

    /* check internal state */
    Message* out = NULL;
    switch(event_state) {
    case EVENT_SET_TITLE:
      {
        if (found_tag) {
	  DEBUG_STREAM("Sending title: " << title << " ..." << endl);
	  Event e(IIcecast::setTitle_event, new TInValue<string>(title));
          out = new CEvent(e);
        }
        if (found_tag) {
          event_state = EVENT_TAGINFO;
        } else {
          event_state = EVENT_START_TRACK;
        }
      }
      break;
    case EVENT_TRACK_DURATION:
      {
	DEBUG_STREAM("Sending trackDuration: " << track_duration << endl);
        Event* e = ITrackDuration::create_trackDuration(track_duration);
        event_state = EVENT_SET_FORMAT;
        out = new CEvent(e);
      }
      break;
    case EVENT_TAGINFO:
      {
        DEBUG_STREAM("Sending Taginfo..."<<endl);
        string _genre;
        if (genre_count<genre || genre==-1)
          _genre="Unknown";
        else
          _genre=genre_table[genre];
        Event* e = ITagInfo::create_TagInfo(title,album,interpret,year,_genre);
        event_state = EVENT_START_TRACK;
        out = new CEvent(e);
      }
      break;
    case EVENT_START_TRACK:
      {
        DEBUG_STREAM("Sending startTrack..." << endl);
	Event* e = ITrack::create_startTrack(filename); //Event e (ITrack::startTrack_event, new TInValue<string>(filename));
        event_state = EVENT_TRACK_DURATION;
        out = new CEvent(e);
      }
      break;
    case EVENT_NEXT_TRACK:
      {
        DEBUG_STREAM("Sending endTrack..." << endl);
        Event e (ITrack::endTrack_event);
        event_state = EVENT_START_TRACK;
        out = new CEvent(e);
      }
      break;
    case EVENT_SET_FORMAT:
      {
        bool have_to_send_format(false);
        bool simple_connection_format(false);
        
        if (! (old_format == *getOwnWorkingOutputProperty()->getDefaultFormat())) {
          // file format has changed
          have_to_send_format = true;
          simple_connection_format = !getOutputFormat()->isParameterIncluded(Format::sample_rate_param);
          if (simple_connection_format && // connection format is the simple one
              (old_connection_format == *getOutputFormat())) { // and does not differ from new simple output format
            // no changes in output connection format!
            have_to_send_format = false;
          }
        }
        
        // store new formats as old ones
        old_format = *getOwnWorkingOutputProperty()->getDefaultFormat();
        old_connection_format = new_simple_format;
        // update the connection format:
        if (simple_connection_format) {
          setOutputFormat(new_simple_format);
        } else {
          // old format was updated above to new complex format
          setOutputFormat(old_format);
        }
        // set the next state
        event_state = EVENT_NONE;

        // send Format if it has to
        if (have_to_send_format) {
          DEBUG_STREAM("Sending setFormat..." << endl);
          if(getOutputFormat()->isParameterIncluded(Format::sample_rate_param) && 
             getOutputFormat()->isParameterIncluded(Format::channels_param)) {
            DEBUG_STREAM("Format contains: sample_rate: "
                         << getOutputFormat()->getIntValue(Format::sample_rate_param)
                         << ", channels: "
                         << getOutputFormat()->getIntValue(Format::channels_param) << endl);
          }
          // use the new connection format for sending
          out = new CEvent(IFormatChange::create_setFormat(*getOutputFormat()));
        } else {
          // nothing to do
          out = NULL;
        }
      }
      break;
    case EVENT_NONE:
      {
        if (seekto != 0) {
          // get current position in file
          filepos = lseek(fd, 0, SEEK_CUR);
          // test if we would get negative
          if ((seekto < 0) && (filepos < static_cast<u_int64_t>(-seekto))) {
            // if seek result would be negative go to beginning of file
            lseek(fd, 0, SEEK_SET);
            DEBUG_STREAM("Seeked to beginning " << seekto << " from " << filepos << ", filesize " << filesize << endl);
          }
          else if ((filepos + seekto) >= filesize) {
            // set back to short-before-file-end
            lseek(fd, filesize - 1, SEEK_SET);
            DEBUG_STREAM("Seeked to end " << filesize-1 << " from " << filepos << ", filesize " << filesize << endl);
            // as there is still some data to read (one Byte :-) )
            // do not stop producing, instead read & send the byte and send a "end_track" afterwards
            //stop producind data, if we exceed file end
            //setProducingFlag(false);
          }
          else {
            // seek normally
            lseek(fd, seekto, SEEK_CUR);
            DEBUG_STREAM("Seeked normally " << seekto << " from " << filepos << ", filesize " << filesize << endl);
          }
          //seekto = 0;
          // get current position in file for progress event
          filepos = lseek(fd, 0, SEEK_CUR);
          // make sure, progress is send next
          buffers_progress = buffers_before_progress +1;
          CEvent* cevent = new FlushEvent();
          cevent->insertEvent(cevent->size(), ISeekReciever::create_seeked(filepos,seekto));
          cevent->insertEvent(cevent->size(), ISyncReset::create_syncReset());
          
          seekto=0;
          out = cevent;
        } else {
          Buffer* buffer = NULL;
          if ((buffer = getNewBuffer( buffer_size )) != 0) {
            bytes_read = read( fd, buffer->getData(), buffer_size );
            filepos += bytes_read;
            if ( bytes_read > 0 ) {
              buffer->setUsed( bytes_read );
              out = buffer;
            } else {
              /* assume EOF */
              setProducingFlag(false);
              event_state = EVENT_END_TRACK;
              Event e (ITrack::endTrack_event);
              buffer->release();
              out = new CEvent(e);
            }
          } else {
            out = NULL;
          }
        }
      }
      break;
    case EVENT_END_TRACK:
      ERROR_STREAM("Case EVENT_END_TRACK in processBuffer, THIS MUST BE A BUG!" << endl);
      setProducingFlag(false);
      out = NULL;
      break;
    default:
      ERROR_STREAM("Invalid internal state" << endl);
      event_state = EVENT_NONE;
      out = NULL;
    }
    
    /* set correct timestamp */
    if(out) {
      out->setTimestamp(timestamp);
      ++timestamp.stream_counter;
    }
    
    return out;
  }

  MP3ReadNode::~MP3ReadNode() {
    close( fd );
  }

   /**
  * Read the ID3v2 header at the begining of the file if it exists
  * it avoids an possible error, if you try to found the mp3 header 
  * on the IDV3v2 .  
  */
  int MP3ReadNode::isID3v2() {
    int   bytes_read;
    bool  data_found;
    char  buffer[ID3_HEADER_SIZE];
    int   len;
    
    bytes_read = read( fd, buffer, ID3_HEADER_SIZE );
    if ( bytes_read != ID3_HEADER_SIZE ) return(0);
    if ( strncmp( buffer, "ID3", 3 ) == 0 ) {
        if ( buffer[3] != 0xff && buffer[4] != 0xff && 
	   (buffer[6] & 0x80 ) == 0 &&
	   (buffer[7] & 0x80 ) == 0 && 
	   (buffer[8] & 0x80 ) == 0 && 
	   (buffer[9] & 0x80 ) == 0 ) {
		
	    	len = ((buffer[6] & 0x7f ) << 21 ) |
			  ((buffer[7] & 0x7f ) << 14 ) |
		  	((buffer[8] & 0x7f ) << 7 ) |
		  	(buffer[9] & 0x7f );
	    	lseek(fd, len, SEEK_CUR);

	    	DEBUG_STREAM("ID3V2 Header found" << endl );
		return len;
	}
   }
   DEBUG_STREAM("ID3V2 Header not found" << endl );
   //with other case 
   //go back to the begining of the file
   lseek(fd, 0, SEEK_SET);
   return 0;
  }
  
  /**
  * Node: this function only skips the wav header if exists
  * it doesn't parse the header, for format chechking the parseHeader
  * function is used
  */
  int MP3ReadNode::isWav() {
    int   bytes_read;
    bool  data_found;
    char  buffer[4];
    int   mp3pos = 0;

    bytes_read = read( fd, buffer, 4 );

    if ( bytes_read != 4 ) return(0);

    if ( strncmp( buffer, "RIFF", 4 ) == 0 ) {
      mp3pos = 4;

      DEBUG_STREAM("RIFF Header found" << endl );

      data_found = false;

      /* look for data chunk and set mp3pos to end of data chunk */
      while( !data_found ) {
        lseek( fd, mp3pos, SEEK_SET );
        bytes_read = read( fd, buffer, 4 );
        if ( bytes_read == 4 ) {
          mp3pos += 4;
          if ( strncmp( buffer, "data", 4 ) == 0 ) {

            data_found = true;
            //skip data chunk
            mp3pos += 4;

            DEBUG_STREAM("DATA chunk found, set fileposition to: " << mp3pos << endl);

          } else mp3pos -= 3;
        }
      }

    }
    else mp3pos = 0;

    lseek(fd, mp3pos, SEEK_SET);
    return( mp3pos );

  }

  bool MP3ReadNode::parseHeader(char* _buffer) {
    /* MP3 Header 32 Bit
       example: 1111 1111  1111 1011  1011 0000  0000 0100
                |-b0----|  |-b1----|  |-b2----|  |-b3----|
    */

    char b  = 0;
    char b0 = *(_buffer + 0); // verbose(b0); cout << " ";
    char b1 = *(_buffer + 1); // verbose(b1); cout << " "; 
    char b2 = *(_buffer + 2); // verbose(b2); cout << " ";
    char b3 = *(_buffer + 3); // verbose(b3); cerr << endl;


    /* Bit 31-21 : Sync, should be 1111 1111 111 */
    if ( ((b0 & 0xff) != 0xff) || ((b1 & 0xe0) != 0xe0) ) {
      DEBUG_STREAM("No sync found" << endl);
      return false;
    } 

    /* Bit 20,19 : MPEG Audio version ID */
    switch ((b1 & 0x18) >> 3) {
    case 0 : fmpeg = 25; break; /* 00 = MPEG Version 2.5 */
    case 1 : fmpeg =  0; break; /* 01 = reserved */
    case 2 : fmpeg =  2; break; /* 10 = MPEG Version 2 (ISO/IEC 13818-3) */
    case 3 : fmpeg =  1; break; /* 11 = MPEG Version 1 (ISO/IEC 11172-3) */
    default: return false; //no correct header
    }
    ERROR_STREAM("MPEG Version : " << (int) fmpeg << endl);

    
    /* Bit 18,17 : Layer description */
    switch((b1 & 0x06) >> 1) {
    case 0 : flayer = 0; break; /* 00 = reserved */
    case 1 : flayer = 3; break; /* 01 = Layer III  */ 
    case 2 : flayer = 2; break; /* 10 = Layer II   */
    case 3 : flayer = 1; break; /* 11 = Layer I    */
    default: return false; //no correct header
    }
    ERROR_STREAM("Layer : " << (int) flayer << endl);

    /* Bit 16 : Protection bit */
    fprotbit = b1 & 0x01;
    DEBUG_STREAM("Protection Bit : " << (int) fprotbit << endl);
    switch(fprotbit) {
    case 0: 
    case 1: 
      break;
    default: 
      return false; //no correct header
    }

    /* Bit 15,14,13,12 : Bitrate */
    static int bitrate_table[2][3][16] =  /* values in kBit */
    { 
      { 
        {0, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448, -1}, /* MPEG 1 - Layer 1 */
        {0, 32, 48, 56,  64,  80,  96, 112, 128, 160, 192, 224, 256, 320, 384, -1}, /* MPEG 1 - Layer 2 */
        {0, 32, 40, 48,  56,  64,  80,  96, 112, 128, 160, 192, 224, 256, 320, -1}  /* MPEG 1 - Layer 3 */
      }, 
      { 
        {0, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448, -1}, /* MPEG 2 - Layer 1 */ 
        {0, 32, 48, 56,  64,  80,  96, 112, 128, 160, 192, 224, 256, 320, 384, -1}, /* MPEG 2 - Layer 2 */ 
        {0,  8, 16, 24,  32,  64,  80,  56,  64, 128, 160, 112, 128, 256, 320, -1}  /* MPEG 2 - Layer 3 */
      }  
    };
    b = (b2 & 0xF0) >> 4;
    if ((b >= 0) && (b <= 15)) {
      fbitrate = bitrate_table[fmpeg-1][flayer-1][b];
      if (fbitrate==0||fbitrate==-1) return false;
    } else {
      return false;
    }


    /* Bit 11,10: Frequenz */
    b = (b2 & 0x0c) >> 2;
    switch (fmpeg) {
    case 1: /* MPEG-1 */
      switch (b) {
      case 0: fsample_rate = 44100; break;
      case 1: fsample_rate = 48000; break;
      case 2: fsample_rate = 32000; break;
      case 3: fsample_rate =     0; break;
      }
      break; 
    case 2: /* MPEG-2 */
      switch (b) { 
      case 0: fsample_rate = 22050; break;
      case 1: fsample_rate = 24000; break;
      case 2: fsample_rate = 16000; break;
      case 3: fsample_rate =     0; break;
      }
      break;
    case 25: /* MPEG-2.5 */
      switch (b) { 
      case 0: fsample_rate = 11025; break;
      case 1: fsample_rate = 12000; break;
      case 2: fsample_rate =  8000; break;
      case 3: fsample_rate =     0; break;
      }
      break;
    default:
      return false;
    }
    if (0 >= fsample_rate) return false;

    ERROR_STREAM("fsample_rate : " << (int) fsample_rate << endl);

    /* Bit 9 : Padding Bit */
    fpadding = (b2 & 0x02) >> 1;
    switch (fpadding) {
    case 0:
    case 1:
      break;
    default:
      return false;
    }

    /* Bit 7,6 : Number of channels */
    switch((b3 & 0xc0) >> 6) {
    case 3:
      fnumchannels = 1;
      break;
    case 0:
    case 1:
    case 2:
      fnumchannels = 2;
      break;
    default:
      return false;
      break;
    }

     /* Bit 3: Copyright
	Not used - but checked to verify complete header */
     switch ((b3&0x08) >> 2) {
     case 0:
     case 1:
       break;
     default:
       return false;
     }

     /* Bit 0/1: Emphasis
	Not used - but checked to verify complete header */
     switch (b3&0x03) {
     case 0:
     case 1:
     case 2:
     case 3:
       break;
     default:
       return false;
     }

    return true;
  }

  /******************** ISeekable interface ********************/
  bool MP3ReadNode::hasPercentSeek() const {
    return true;
  }

  void MP3ReadNode::seekPercent(const Rational& percent) {
    if (filename.substr(0,4) == string("/dev")) {
      // no seeking on devices!
      return;
    }

    Rational internal_percent(percent);
    if (internal_percent > Rational(1)) {
      ERROR_STREAM("Seeking more than 100% not possible, defaulting to end!" << endl);
      internal_percent = Rational(1);
    }
    else if (internal_percent < Rational(-1)) {
      ERROR_STREAM("Seeking less than -100% not possible, defaulting to beginning!" << endl);
      internal_percent = Rational(-1);
    }

    Rational rat_seekto = Rational(filesize) * internal_percent;
    seekto = rat_seekto.getNumerator() / rat_seekto.getDenominator();
    DEBUG_STREAM(seekto << " = " << internal_percent << " * " << filesize << endl);
  }
   
  void MP3ReadNode::seekPercentTo(const Rational& percent) {
    seekPercent(percent - Rational(filepos, filesize));
  }

  bool MP3ReadNode::hasBytesSeek() const {
    return true;
  }

  void MP3ReadNode::seekBytes(const int64_t& offset) {
    if ((offset < 0) && (labs(offset) > static_cast<int64_t>(filepos))) {
      ERROR_STREAM("Seeking to negativ byte-offset not possible, defaulting to beginning of file!" << endl);
      seekBytesTo(0);
    } else {
      seekBytesTo(filepos + offset);
    }
  }

  void MP3ReadNode::seekBytesTo(const u_int64_t& bytes) {
    u_int64_t internal_bytes(bytes);

    if (internal_bytes > filesize) {
      ERROR_STREAM("Seeking over EOF not possible, defaulting to EOF!" << endl);
      internal_bytes = filesize;
    }

    seekto = internal_bytes - filepos;

    DEBUG_STREAM(seekto << " = " << internal_bytes << " - " << filepos << endl);
  }

  bool MP3ReadNode::hasTimeSeek() const {
    return false;
  }

  void MP3ReadNode::seekTime(const Interval&) {
  }

  void MP3ReadNode::seekTimeTo(const Time&) {
  }
} /* namespace */
