/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "TimedBufferDropNode.hpp"
#include "nmm/interfaces/base/sync/ISyncReset.hpp"
#include "nmm/interfaces/file/ITrack.hpp"
#include "nmm/base/format/Property.hpp"
#include "nmm/base/memorymgr/Buffer.hpp"
#include "nmm/base/registry/Plugin.hpp"

NMM_REGISTER_PLUGIN(NMM::TimedBufferDropNode, "TimedBufferDropNode")

namespace NMM {

  TimedBufferDropNode::TimedBufferDropNode(const char *name, StreamQueue::Mode qmode, int qsize) :
    GenericFilterNode(name, qmode, qsize),
    first_time(true),
    currently_dropping(true),
    relative_inout_points(false),
    first_ts_interval(),
    sent_endtrack_event(false)
  {
  }

  void TimedBufferDropNode::useRelativeInOutPoints(const bool& relative)
  {
    relative_inout_points = relative;
  }

  bool TimedBufferDropNode::isRelativeInOutPoints()
  {
    return relative_inout_points;
  }

  void TimedBufferDropNode::addInOutPoint(const unsigned int& in_secs, const unsigned int& in_millisecs, 
				     const unsigned int& out_secs, const unsigned int& out_millisecs)
  {
    inout_point_t inout_point;
    inout_point.first.sync_time.sec  = in_secs;
    inout_point.first.sync_time.nsec = in_millisecs * 1000000;
    inout_point.second.sync_time.sec  = out_secs;
    inout_point.second.sync_time.nsec = out_millisecs * 1000000;
    inout_points.push(inout_point);
  }

  Result TimedBufferDropNode::doInit() 
  {
    Format* any_format_i = new Format("*/*");
    getOwnInputProperty()->addNewFormat(any_format_i);
    
    Format* any_format_o = new Format("*/*");
    getOwnOutputProperty()->addNewFormat(any_format_o);
    
    return SUCCESS;
  }
  
  Result TimedBufferDropNode::doInitOutput() {
    /* get input format */
    if (!getInputFormat()) {
      GLOBAL_ERROR_STREAM("No Input Format! FAILURE" << endl);
      return FAILURE;
    }

    Format* i_format = new Format(*getInputFormat());
    getOwnWorkingInputProperty()->addNewFormat(i_format);
    getOwnWorkingInputProperty()->setDefaultFormat(i_format);
    
    Format* o_format = new Format(*i_format);
    getOwnWorkingOutputProperty()->addNewFormat(o_format);
    getOwnWorkingOutputProperty()->setDefaultFormat(o_format);
    
    return SUCCESS;
  }
  
  Message* TimedBufferDropNode::processBuffer(Buffer *in) 
  {
    // save the first time stamp, all other times are then offset by this time
    if(relative_inout_points && first_time) {
      if(in->getTimestamp().is_valid) {
	first_time = false;
	first_ts_interval.sec = in->getTimestamp().sync_time.sec;
	first_ts_interval.nsec = in->getTimestamp().sync_time.nsec;
	DEBUG_STREAM("First buffer " << first_ts_interval << endl);
      }
    }

    if(!inout_points.empty()) {
      if(currently_dropping) {
	if(in->getTimestamp().is_valid) {
	  if(inout_points.top().first.sync_time > (in->getTimestamp().sync_time - first_ts_interval) ) {
	    DEBUG_STREAM("CURRENTLY Dropping " << inout_points.top().first.sync_time << " > " \
			 << (in->getTimestamp().sync_time - first_ts_interval) << endl);
	    return 0;
	  }
	  else { 
	    // .. stop dropping
	    currently_dropping = false;
	    DEBUG_STREAM("NO MORE Dropping " << inout_points.top().first.sync_time << " <= " \
			 << (in->getTimestamp().sync_time - first_ts_interval) << endl);
	    return in;
	  }
	}
	else { // invalid ts in in buffer -> drop it
	  return 0;
	}
      }
      else { // (currently_dropping == false)
	if(in->getTimestamp().is_valid) {
	  if(inout_points.top().second.sync_time < (in->getTimestamp().sync_time - first_ts_interval) ) {
	    currently_dropping = true;
	    DEBUG_STREAM("START Dropping again " << inout_points.top().second.sync_time << " < " \
			 << (in->getTimestamp().sync_time - first_ts_interval) << endl);
	    inout_points.pop();
	    CEvent* ce = new FlushEvent();
	    ce->insertEvent(ce->size(), ISyncReset::create_syncReset());
	    return ce;
	  }
	  else {
	    DEBUG_STREAM("NOT Dropping " << inout_points.top().second.sync_time << " >=" \
			 << (in->getTimestamp().sync_time - first_ts_interval) << endl);
	    return in;
	  }
	}
	else { // invalid ts in in buffer -> do not drop it this time because currently_dropping == false
	  return in;
	}
      }
    }
    else { // inout_points.empty() -> send "end_track" once, then only send 0-Buffer
      if(!sent_endtrack_event) {
	sent_endtrack_event = true;
	return new CEvent(Event(ITrack::endTrack_event));
      }
      else {
	return 0;
      }
    }
      
  }
}
