/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Christian Gerstner <gerstner@graphics.cs.uni-sb.de>
 *                    Eric Peters <epeters@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include <iostream>
#include <math.h>
#include "RFBProtoController.hpp"

namespace NMM {
  
    RFBProtoController::RFBProtoController(IVNC::Mode mode) : NamedObject("RFBProtoController") {
	is_connected = false;
	tolerance = 0.4;
	interval = 1.0;
	appData.shareDesktop = true;
	appData.useBGR233 = false;
	shared = false;
	endianTest = 1;
	this->mode = mode;
	requested = false;
	appData.passwordFile = 0;
	appData.enc_password = new unsigned char[CHALLENGESIZE];
	updateBuffer = 0;
    
	s_stream = 0;
	vncimage = 0;
    }

    RFBProtoController::~RFBProtoController() {
	if (s_stream) delete s_stream;
	if (vncimage) {
	    delete[] vncimage->data;
	    delete vncimage;
	}
	if (updateBuffer) {
	    delete[] updateBuffer;
	    delete[] appData.enc_password;
	}
    }

    void RFBProtoController::setHostAndPort(string hostname, int port) {
	int vncServerPort = port;
	if (vncServerPort < 100)
	    vncServerPort += SERVER_PORT_OFFSET;
	addr.setPort(vncServerPort);
	addr.setAddress(hostname.c_str());
	addr.setTransportProtocol("tcp");
    }

    void RFBProtoController::setHostname(string hostname) {
	addr.setAddress(hostname.c_str());
    }

    void RFBProtoController::setPort(int port) {
	int vncServerPort = port;
	if (vncServerPort < 100)
	    vncServerPort += SERVER_PORT_OFFSET;
	addr.setPort(vncServerPort);
    }

    bool RFBProtoController::isConnected() {
	return is_connected;
    }

    Result RFBProtoController::connectToRFBServer() {
	try {
	    s_stream = s_connect.connect(addr);
	}
	catch (const ConnectionException& ce) {
	    GLOBAL_ERROR_STREAM("Connection to VNC server failed:" << ce << endl);
	    return FAILURE;
	}
	if (s_stream <= 0) {
	    GLOBAL_ERROR_STREAM("Unable to connect to VNC server" << endl);
	    return FAILURE;
	}
    
	is_connected = true;
	return SUCCESS;
    }

    Result RFBProtoController::disconnectFromRFBServer() {
	try {
	    s_stream->close();
	} catch (...) {
	    GLOBAL_ERROR_STREAM("Error while disconnecting from RFBServer" << endl);
	    return FAILURE;
	}
	return SUCCESS;
    }

    Result RFBProtoController::initialiseRFBConnection() {
	rfbProtocolVersionMsg pv;
	int major,minor;
	CARD32 authScheme, reasonLen, authResult;
	char* reason=0;
	CARD8 challenge[CHALLENGESIZE];
	char* passwd=0;
	char* desktopName=0;
	rfbClientInitMsg ci;
    
	read(pv, sz_rfbProtocolVersionMsg);
    
	pv[sz_rfbProtocolVersionMsg] = 0;
    
	if (sscanf(pv,rfbProtocolVersionFormat,&major,&minor) != 2) {
	    GLOBAL_ERROR_STREAM("Not a valid VNC server" << endl);
	    return FAILURE;
	}
    
	GLOBAL_DEBUG_STREAM("VNC server supports protocol version " << major << "." << minor << " (viewer " 
			    << rfbProtocolMajorVersion << "." << rfbProtocolMinorVersion << ")" << endl);
    
	major = rfbProtocolMajorVersion;
	minor = rfbProtocolMinorVersion;
    
	sprintf(pv,rfbProtocolVersionFormat,major,minor);
    
	write(pv, sz_rfbProtocolVersionMsg);
    
	read((char*)&authScheme, 4);
    
	authScheme = swap32IfLE(authScheme);
    
	switch (authScheme) {
      
	    case rfbConnFailed:
		read((char*)&reasonLen, 4);
		reasonLen = swap32IfLE(reasonLen);
      
		reason = new char[reasonLen];
      
		read(reason, reasonLen);
      
		GLOBAL_ERROR_STREAM("VNC connection failed: " << reason << endl);
		delete[] reason;
		reason = 0;
		return FAILURE;
      
	    case rfbNoAuth:
		GLOBAL_DEBUG_STREAM("No authentication needed" << endl);
		break;
      
	    case rfbVncAuth:
		read((char*)challenge, CHALLENGESIZE);
		if (appData.passwordFile) {
		    passwd = decryptPasswdFromFile((char*)appData.passwordFile);
	
		    if (!passwd) {
			GLOBAL_ERROR_STREAM("Cannot read valid password from file \"" << appData.passwordFile << "\"" << endl);
			return FAILURE;
		    }
		}
      
		else {
		    if (appData.enc_password != reinterpret_cast<unsigned char*>(const_cast<char*>(""))) {
			DEBUG_STREAM("rfbvncauth: appData.enc_passwd=" << appData.enc_password << endl);
			passwd =  decryptPasswdFromMemory(appData.enc_password);
			if (!passwd) {
			    GLOBAL_ERROR_STREAM("Cannot read valid password from Memory" << endl);
			    return FAILURE;
			}
		    }
		}
      
		if ((!passwd) || (strlen(passwd) == 0)) {
		    GLOBAL_ERROR_STREAM("Reading password failed" << endl);
		    return FAILURE;
		}
		if (strlen(passwd) > 8) {
		    passwd[8] = '\0';
		}
      
		encryptBytes(challenge, passwd);
      
		/* Lose the password from memory */
		for (int i = strlen(passwd); i >= 0; i--) {
		    passwd[i] = '\0';
		}
		write((char*)&challenge, CHALLENGESIZE);
      
		read((char*)&authResult, 4);
      
		authResult = swap32IfLE(authResult);
      
		switch (authResult) {
		    case rfbVncAuthOK:
			GLOBAL_DEBUG_STREAM("VNC authentication succeeded" << endl);
			is_connected = true;
			break;
		    case rfbVncAuthFailed:
			GLOBAL_ERROR_STREAM("VNC authentication failed - wrong password?" << endl);
			return FAILURE;
		    case rfbVncAuthTooMany:
			GLOBAL_ERROR_STREAM("VNC authentication failed - too many tries" << endl);
			return FAILURE;
		    default:
			GLOBAL_ERROR_STREAM("Unknown VNC authentication result: " << (int)authResult << endl);
			return FAILURE;
		}
		break;
      
	    default:
		GLOBAL_ERROR_STREAM("Unknown authentication scheme from VNC server: " << (int)authScheme << endl);
		return FAILURE;
      
	}
    
	ci.shared = (appData.shareDesktop ? 1 : 0); //desktop should be always shared
    
	write((char*)&ci, sz_rfbClientInitMsg);
    
	read((char*)&si, sz_rfbServerInitMsg);
    
	si.framebufferWidth    = swap16IfLE(si.framebufferWidth);
	si.framebufferHeight   = swap16IfLE(si.framebufferHeight);
	si.format.redMax       = swap16IfLE(si.format.redMax);
	si.format.greenMax     = swap16IfLE(si.format.greenMax);
	si.format.blueMax      = swap16IfLE(si.format.blueMax);
	si.nameLength = swap32IfLE(si.nameLength);
	desktopName = new char[si.nameLength + 1];
    
	read(desktopName, si.nameLength);
    
	desktopName[si.nameLength] = 0;
    
	GLOBAL_DEBUG_STREAM("Desktop name \"" << desktopName << "\"" << endl);
	GLOBAL_DEBUG_STREAM("Connected to VNC server, using protocol version " << rfbProtocolMajorVersion << "." << rfbProtocolMinorVersion << endl);
	GLOBAL_DEBUG_STREAM("VNC server default format:" << endl);

	delete[] desktopName;
	printPixelFormat(&si.format);
	return SUCCESS;
    }

    void RFBProtoController::printPixelFormat(rfbPixelFormat* format) {
	if (format->bitsPerPixel == 1) {
	    GLOBAL_DEBUG_STREAM("Single bit per pixel." << endl);
	    GLOBAL_DEBUG_STREAM((format->bigEndian ? "Most" : "Least") <<  " significant bit in each byte is leftmost on the screen." << endl);
	} else {
	    GLOBAL_DEBUG_STREAM((int)format->bitsPerPixel << " bits per pixel." << endl);
	    if (format->bitsPerPixel != 8) {
		GLOBAL_DEBUG_STREAM((format->bigEndian ? "Most" : "Least") << " significant byte first in each pixel." << endl);
	    }
	    if (format->trueColour) {
		GLOBAL_DEBUG_STREAM("True colour: max red " << (int)format->redMax << " green " << (int)format->greenMax << " blue " << (int)format->blueMax 
				    << ", shift red " << (int)format->redShift << " green " << (int)format->greenShift << " blue " << (int)format->blueShift 
				    << endl);
	    } else {
		GLOBAL_DEBUG_STREAM("Colour map (not true colour)." << endl);
	    }
	}
    }

    void RFBProtoController::initDesktopAndColor() {
	setVisualAndCmap();
	desktopInitBeforeRealization();
    }

    void RFBProtoController::setVisualAndCmap() {
	if (!appData.useBGR233) {
	    myFormat.bitsPerPixel = si.format.bitsPerPixel;
	    myFormat.depth = si.format.depth;
	    myFormat.trueColour = 1;
	    myFormat.bigEndian = si.format.bigEndian;
	    myFormat.redShift = si.format.redShift;
	    myFormat.greenShift = si.format.greenShift;
	    myFormat.blueShift = si.format.blueShift;
	    myFormat.redMax = si.format.redMax;
	    myFormat.greenMax = si.format.greenMax;
	    myFormat.blueMax = si.format.blueMax;
      
	    GLOBAL_DEBUG_STREAM("Using default colormap which is TrueColor.  Pixel format:" << endl);
	    printPixelFormat(&myFormat);
	    return;
	}
    
	appData.useBGR233 = true;
    
	myFormat.bitsPerPixel = 8;
	myFormat.depth = 8;
	myFormat.trueColour = 1;
	myFormat.bigEndian = 0;
	myFormat.redMax = 7;
	myFormat.greenMax = 7;
	myFormat.blueMax = 3;
	myFormat.redShift = 0;
	myFormat.greenShift = 3;
	myFormat.blueShift = 6;
    
	GLOBAL_DEBUG_STREAM("Using default colormap and translating from BGR233.  Pixel format:" << endl);
	printPixelFormat(&myFormat);
    }

    void RFBProtoController::desktopInitBeforeRealization() {
	vncimage = 0;
	updateBuffer = 0;
    
	if (!vncimage) {
	    vncimage = createImage(si.framebufferHeight, si.framebufferWidth, si.format.depth, si.format.bitsPerPixel, NULL, 32);
      
	    // will be deleted in destructor
	    vncimage->data = new char[vncimage->bytes_per_line * vncimage->height];
	    updateBuffer = new char[vncimage->bytes_per_line * vncimage->height];
            
	    if (!vncimage->data) {
		GLOBAL_ERROR_STREAM("Error while allocating memory in VNCImage" << endl);
	    }
	}
    }

    int RFBProtoController::round(int nbytes, int pad) { return ((((nbytes) + ((pad)-1)) / (pad)) * ((pad)>>3)); }

    VncImage* RFBProtoController::createImage(int height, int width, int depth, int bits_per_pixel, char *data, int pad) {
	// will be deleted in destructor
	VncImage* image = new VncImage(); 
    
	if (depth == 0 || depth > 32)
	    return 0;
	if (!image) 
	    return 0;
    
	image->width = width;
	image->height = height;
	image->bits_per_pixel = bits_per_pixel;
	image->bitmap_pad = pad;
	image->depth = depth;
	image->data = data;
    
	image->bytes_per_line = round((bits_per_pixel * width), pad);
	image->hasNewImage = false;
    
	return image;
    }

    Result RFBProtoController::setFormatAndEncodings() {
	rfbSetPixelFormatMsg spf;
	char buf[sz_rfbSetEncodingsMsg + MAX_ENCODINGS * 4];
	rfbSetEncodingsMsg *se = (rfbSetEncodingsMsg *)buf;
	CARD32 *encs = (CARD32 *)(&buf[sz_rfbSetEncodingsMsg]);
	int len = 0;
    
	spf.type = rfbSetPixelFormat;
	spf.format = myFormat;
	spf.format.redMax = swap16IfLE(spf.format.redMax);
	spf.format.greenMax = swap16IfLE(spf.format.greenMax);
	spf.format.blueMax = swap16IfLE(spf.format.blueMax);
	write((char*) &spf, sz_rfbSetPixelFormatMsg);
    
	se->type = rfbSetEncodings;se->nEncodings = 0;
    
	/*if (appData.encodingsString) {
	  char *encStr = appData.encodingsString;
	  int encStrLen;
	  do {
	  char *nextEncStr = strchr(encStr, ' ');
	  if (nextEncStr) {
	  encStrLen = nextEncStr - encStr;
	  nextEncStr++;
	  } else {
	  encStrLen = strlen(encStr);
	  }
      
	  if (strncasecmp(encStr,"raw",encStrLen) == 0) {
	  encs[se->nEncodings++] = Swap32IfLE(rfbEncodingRaw);
	  } else if (strncasecmp(encStr,"copyrect",encStrLen) == 0) {
	  encs[se->nEncodings++] = Swap32IfLE(rfbEncodingRaw);//CopyRect
	  } else if (strncasecmp(encStr,"hextile",encStrLen) == 0) {
	  encs[se->nEncodings++] = Swap32IfLE(rfbEncodingRaw);//Hextile
	  } else if (strncasecmp(encStr,"corre",encStrLen) == 0) {
	  encs[se->nEncodings++] = Swap32IfLE(rfbEncodingRaw);//CoRRE
	  } else if (strncasecmp(encStr,"rre",encStrLen) == 0) {
	  encs[se->nEncodings++] = Swap32IfLE(rfbEncodingRaw);//RRE
	  } else {
	  fprintf(stderr,"Unknown encoding '%.*s'\n",encStrLen,encStr);
	  }

	  encStr = nextEncStr;
	  } while (encStr && se->nEncodings < MAX_ENCODINGS);
	  } else {
	  if (SameMachine(rfbsock)) {
	  fprintf(stderr,"Same machine: preferring raw encoding\n");*/
	encs[se->nEncodings++] = swap32IfLE(rfbEncodingRaw);
	//}
    
	//encs[se->nEncodings++] = Swap32IfLE(rfbEncodingRaw);//CopyRect
	//encs[se->nEncodings++] = Swap32IfLE(rfbEncodingRaw);//Hextile
	//encs[se->nEncodings++] = Swap32IfLE(rfbEncodingRaw);//CoRRE
	//encs[se->nEncodings++] = Swap32IfLE(rfbEncodingRaw);//RRE
	//}
    
	len = sz_rfbSetEncodingsMsg + se->nEncodings * 4;
	se->nEncodings = swap16IfLE(se->nEncodings);
    
	write(buf, len);
    
	return SUCCESS;
    }

    int RFBProtoController::getFramebufferWidth() {
	return si.framebufferWidth;
    }

    int RFBProtoController::getFramebufferHeight() {
	return si.framebufferHeight;
    }

    int RFBProtoController::getBitsPerPixel() {
	return si.format.bitsPerPixel;
    }

    int RFBProtoController::getColordepth() {
	return si.format.depth;
    }

    int RFBProtoController::getColorspace() {
    
	/* e.g. redMax is 2^n-1 where n is the number of pixels used for the red pixel */
        double redtmp = si.format.redMax+1;
        double redpixels   = log(redtmp)/log(2.0);
        double greentmp = si.format.greenMax+1;
	double greenpixels = log(greentmp)/log(2.0);
        double bluetmp = si.format.blueMax+1;
	double bluepixels  = log(bluetmp)/log(2.0);
    
	switch (si.format.depth) {
	    case 16:
		return rgb565;
		break;
	    case 24:
		return rgb24;
		break;
	    case 32:
		if (redpixels == 11 && greenpixels == 11 && bluepixels == 10) {
		    return rgb32;
		} else if (redpixels == 8 && greenpixels == 8 && bluepixels == 8) {
		    return rgb24;
		} else {
		    return rgb00;
		}
		break;
	    default:
		return rgb00;
		break;
	}
	return rgb00;
    }

    int RFBProtoController::getChannelorder() {
	if (si.format.redShift > si.format.blueShift) {
	    return rgb;
	} else if (si.format.blueShift > si.format.redShift) {
	    return bgr;
	}
	return -1;
    }

    Result RFBProtoController::sendIncrementalFramebufferUpdateRequest() {
	return sendFramebufferUpdateRequest(0, 0, si.framebufferWidth, si.framebufferHeight, true);
    }

    Result RFBProtoController::sendFramebufferUpdateRequest(int x, int y, int w, int h, bool incremental) {
	rfbFramebufferUpdateRequestMsg fur;
    
	fur.type = rfbFramebufferUpdateRequest;
	fur.incremental = incremental ? 1 : 0;
	fur.x = swap16IfLE(x);
	fur.y = swap16IfLE(y);
	fur.w = swap16IfLE(w);
	fur.h = swap16IfLE(h);
    
	write((char*) &fur, sz_rfbFramebufferUpdateRequestMsg);
    
	return SUCCESS;
    }

    void RFBProtoController::setTolerance(float tolerance) {
	this->tolerance = tolerance;
	GLOBAL_DEBUG_STREAM("setting tolerance " << tolerance << endl);
    }

    void RFBProtoController::setInterval(float interval) {
	this->interval = interval;
	GLOBAL_DEBUG_STREAM("setting interval " << interval << endl);
    }

    void RFBProtoController::setShared(bool isShared) {
	appData.shareDesktop = isShared;
    }

    void RFBProtoController::setPasswdPath(const char* passwdpath)
    {
	appData.passwordFile = passwdpath;
	GLOBAL_DEBUG_STREAM("setting passwordpath to " << passwdpath << endl);
    }

    void RFBProtoController::setPasswd(const string passwd)
    {
	for(int i=0; i<CHALLENGESIZE;i++)
	    appData.enc_password[i]= (reinterpret_cast<unsigned char*>(const_cast<char*>(passwd.c_str())))[i];    
    }

    float RFBProtoController::getTolerance() const {
	return tolerance;
    }

    float RFBProtoController::getInterval() const {
	return interval;
    }

    bool RFBProtoController::isShared() {
	return shared;
    }
  
    Result RFBProtoController::handleRFBServerMessage(char* /*in_buffer*/) {
	int result;
	struct pollfd pfd[1];
	pfd[0].fd     = s_stream->getSocket();
	pfd[0].events = POLLIN;
	rfbServerToClientMsg msg;
	bool isBigEnough = false;
	int interval_int = static_cast<int>(interval*1000);
	result = poll(pfd,1,interval_int);
	switch(result){
	    case -1:
		cerr << "Error while reading from RFBSocket" << endl;
		return FAILURE;
	    case 0:
		timeIsUp = true;
		return SUCCESS;
	    default:
		if (!(pfd[0].revents & POLLIN)){
		    return FAILURE;
		}
	}
	read((char*) &msg, 1);
    
	switch (msg.type) {

	    case rfbSetColourMapEntries:
	    {
		int i;
		CARD16 rgb[3];
		
		read(((char *)&msg) + 1, sz_rfbSetColourMapEntriesMsg - 1);
      
		msg.scme.nColours = swap16IfLE(msg.scme.nColours);

		for (i = 0; i < msg.scme.nColours; i++) {
		    read((char *)rgb, 6);
		}
		break;
	    }

	    case rfbFramebufferUpdate:
	    {
		rfbFramebufferUpdateRectHeader rect;
		int bytesPerLine;
		int i;
		int nupdates = 0;
	
		read(((char *)&msg.fu) + 1,sz_rfbFramebufferUpdateMsg - 1);
	      
		msg.fu.nRects = swap16IfLE(msg.fu.nRects);
	
		for (i = 0; i < msg.fu.nRects; i++) {
	
		    read((char *)&rect, sz_rfbFramebufferUpdateRectHeader);

		    rect.r.x = swap16IfLE(rect.r.x);
		    rect.r.y = swap16IfLE(rect.r.y);
		    rect.r.w = swap16IfLE(rect.r.w);
		    rect.r.h = swap16IfLE(rect.r.h);
	  
		    rect.encoding = swap32IfLE(rect.encoding);

		    if ((rect.r.x + rect.r.w > si.framebufferWidth) ||
			(rect.r.y + rect.r.h > si.framebufferHeight)) {
			GLOBAL_DEBUG_STREAM("Rect too large: " << rect.r.w << "x" << rect.r.h 
					    <<  " at (" << rect.r.x << ", " << rect.r.y << ")" << endl);
			return FAILURE;
		    }

		    if ((rect.r.h * rect.r.w) == 0) {
			GLOBAL_DEBUG_STREAM("Zero size rect - ignoring" << endl);
			continue;
		    }
	  
		    switch (rect.encoding) {
			case rfbEncodingRaw:
			{
			    bytesPerLine = rect.r.w * myFormat.bitsPerPixel / 8;

			    switch (mode) {
				
				case IVNC::REAL_TIME:
				    if (nupdates % 10 == 0) {
					isBigEnough = true;
				    } else {
					isBigEnough = false;
				    }
				    read(updateBuffer,bytesPerLine * rect.r.h);
				    copyDataToScreen(updateBuffer, rect.r.x, rect.r.y, rect.r.w, rect.r.h);
				    break;
				    
				case IVNC::REQUESTED:
				    if ((rect.r.h >= si.framebufferHeight*tolerance) ||
					(rect.r.w >= si.framebufferWidth*tolerance)){ 
					// Framebufferupdate bigger than tolerance-Value
					isBigEnough = true;
				    }
				    readIntoVNCImage(rect.r.x, rect.r.y, rect.r.w, rect.r.h);
//		 /* the image is directly read into the buffer */
// 		if (requested && in_buffer && 
// 		    (rect.r.h == si.framebufferHeight) &&
// 		    (rect.r.w == si.framebufferWidth)) {
// 		  requested = false;
// 		  read(in_buffer, bytesPerLine * si.framebufferHeight);
// 		  cout << "handleRFBServerMessage - read(in_buffer, bytesPerLine * si.framebufferHeight);" << endl;
		  
// 		} else {
// 		  if (((rect.r.h >= si.framebufferHeight * tolerance) &&
// 		       (rect.r.w >= si.framebufferWidth * tolerance)) ){
// 		    requested = true;
// 		    read(updateBuffer,bytesPerLine * rect.r.h);
// 		    sendFramebufferUpdateRequest(0, 0, si.framebufferWidth, si.framebufferHeight, false);
// 		    cout << "handleRFBServerMessage - sendFramebufferUpdateRequest()" << endl;
// 		  }
// 		  cout << "not requested" << endl;
// 		  read(updateBuffer,bytesPerLine * rect.r.h);
// 		  //copyDataToScreen(buffer, rect.r.x, rect.r.y, rect.r.w, rect.r.h /*linesToRead*/);
// 		}
		
				    break;
				
				case IVNC::BLANK_SHEET:
				    read(updateBuffer,bytesPerLine * rect.r.h);
				    copyDataToScreen(updateBuffer, rect.r.x, rect.r.y, rect.r.w, rect.r.h);
				    break;
				
				default:
				    GLOBAL_ERROR_STREAM("RFBProtoController is in an unknown mode" << endl);
				    break;
			    }
	      
			    break;
			}	    
			case rfbEncodingCopyRect:{
			    rfbCopyRect cr;
	    
			    read((char *)&cr, sz_rfbCopyRect);
	    
			    // CopyRectangle-Encoding not supported yet but the data has to be moved from the socket
	    
			    //cr.srcX = swap16IfLE(cr.srcX);
			    //cr.srcY = swap16IfLE(cr.srcY);
	      
			    /*if (appData.copyRectDelay != 0) {
			      XFillRectangle(dpy, desktopWin, srcGC, cr.srcX, cr.srcY,
			      rect.r.w, rect.r.h);
			      XFillRectangle(dpy, desktopWin, dstGC, rect.r.x, rect.r.y,
			      rect.r.w, rect.r.h);
			      XSync(dpy,false);usleep(appData.copyRectDelay * 1000);
			      XFillRectangle(dpy, desktopWin, dstGC, rect.r.x, rect.r.y,
			      rect.r.w, rect.r.h);
			      XFillRectangle(dpy, desktopWin, srcGC, cr.srcX, cr.srcY,
			      rect.r.w, rect.r.h);
			      }*/
			    /*
			      CopyRectOnScreen(cr.srcX, cr.srcY, rect.r.w, rect.r.h, rect.r.x, rect.r.y);
			      sendFramebufferUpdateRequest(cr.srcX, cr.srcY, rect.r.w, rect.r.h, true);*/
			    //	XCopyArea(dpy, desktopWin, desktopWin, gc, cr.srcX, cr.srcY,
			    //	  rect.r.w, rect.r.h, rect.r.x, rect.r.y);
			
			    break;
			}
	
			default:
			    cerr << "Unknown rect encoding " << (int)rect.encoding << endl;
			    return FAILURE;
		    }
		}
	    
		if (sendIncrementalFramebufferUpdateRequest() == FAILURE)
		    return FAILURE;

		nupdates++;

		break;
	    }

	    case rfbBell:
		break;
      
	    case rfbServerCutText:
		//Case has no more effects, but Message has to be removed from socket
	    {
		char* serverCutText;
	
		read(((char *)&msg) + 1, sz_rfbServerCutTextMsg - 1);

		msg.sct.length = swap32IfLE(msg.sct.length);

		serverCutText = new char[msg.sct.length+1];

		read(serverCutText, msg.sct.length);

		serverCutText[msg.sct.length] = 0;
	
		delete[] serverCutText;
		break;
	    }

	    default:
		GLOBAL_ERROR_STREAM("Unknown message type " << msg.type << " from VNC server" << endl);
		return FAILURE;
	}

	if (isBigEnough) {
	    vncimage->hasNewImage = true;
	}
	return SUCCESS;
    }

    void RFBProtoController::copyDataToScreen(char *buf, int x, int y, int width, int height) {
  
	int h;
	int widthInBytes = width * myFormat.bitsPerPixel / 8;
	int scrWidthInBytes = si.framebufferWidth * myFormat.bitsPerPixel / 8;
    
	char *scr2 = (vncimage->data + y * scrWidthInBytes + x * myFormat.bitsPerPixel /8);
    
	for (h = 0; h < height; h++) {
	    memcpy(scr2, buf, widthInBytes);
	    buf += widthInBytes;
	    scr2 += scrWidthInBytes;
	}
    }

    void RFBProtoController::readIntoVNCImage(int x, int y, int width, int height) {
	int h;
	int widthInBytes = width * myFormat.bitsPerPixel / 8;
	int scrWidthInBytes = si.framebufferWidth * myFormat.bitsPerPixel / 8;
	char *scr2 = (vncimage->data + y * scrWidthInBytes + x * myFormat.bitsPerPixel /8);
	try {
	    for (h = 0; h < height; h++) {
		s_stream->readStream(scr2, widthInBytes);
		scr2 += scrWidthInBytes;
	    }
	} catch (Exception e) {
	    is_connected = false;
	    GLOBAL_ERROR_STREAM("Error while reading from socket: " << e << endl);
	    GLOBAL_ERROR_STREAM("perhaps VNC server closed connection" << endl);
	}
    }

    void RFBProtoController::read(char* buffer, int length) {
	try {
	    s_stream->readStream(buffer, length);
	} catch (Exception e) {
	    is_connected = false;
	    GLOBAL_ERROR_STREAM("Error while reading from socket: " << e << endl);
	    GLOBAL_ERROR_STREAM("perhaps VNC server closed connection" << endl);
	}
    }

    void RFBProtoController::write(char* buffer, int length) {
	try {
	    s_stream->writeStream(buffer, length);
	} catch (Exception e) {
	    is_connected = false;
	    GLOBAL_ERROR_STREAM("Error while reading from socket: " << e << endl);
	    GLOBAL_ERROR_STREAM("perhaps VNC server closed connection" << endl);
	}
    }

    void RFBProtoController::sendPointerEvent(int x, int y, int buttonMask) {
	rfbPointerEventMsg pe;
    
	pe.type = rfbPointerEvent;
	pe.buttonMask = buttonMask;
	if (x < 0) x = 0;
	if (y < 0) y = 0;
	pe.x = swap16IfLE(x);
	pe.y = swap16IfLE(y);
	write((char*) &pe, sz_rfbPointerEventMsg);
    }

    void RFBProtoController::sendKeyEvent(CARD32 key, bool down) {
	rfbKeyEventMsg ke;
    
	ke.type = rfbKeyEvent;
	ke.down = down ? 1 : 0;
	ke.key = swap32IfLE(key);
	write((char*) &ke, sz_rfbKeyEventMsg);
    }

    bool RFBProtoController::hasNewImage() {
	return vncimage->hasNewImage;
    }

    void RFBProtoController::setHasNewImage(bool isNew) {
	vncimage->hasNewImage = isNew;
    }

    bool RFBProtoController::getTimeIsUp() {
	return timeIsUp;
    }

    void RFBProtoController::setTimeIsUp(bool tiu) {
	timeIsUp = tiu;
    }

    VncImage* RFBProtoController::getVncImage() {
	return vncimage;
    }

    int RFBProtoController::swap16IfLE(int s) {
	if (endianTest)
	    return ((((s) & 0xff) << 8) | (((s) >> 8) & 0xff));
	return s;
    }

    int RFBProtoController::swap32IfLE(int l) { 
	if (endianTest) {
	    return ((((l) & 0xff000000) >> 24) | 
		    (((l) & 0x00ff0000) >> 8)  | 
		    (((l) & 0x0000ff00) << 8)  | 
		    (((l) & 0x000000ff) << 24));
	}
	return l;
    }

    IVNC::Mode RFBProtoController::getMode(){
	return mode;
    }

    void RFBProtoController::setMode(IVNC::Mode mode){
	this->mode = mode;
    }

    bool RFBProtoController::hasRequested() {
	return requested;
    }

    void RFBProtoController::setRequested(bool _requested) {
	requested = _requested;
    }
  
} // end namespace
