/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "GLDisplay.hpp"

#include "nmm/base/memorymgr/Buffer.hpp"
#include "nmm/base/Node.hpp"

#include <GL/glu.h>

namespace NMM {

  GLDisplay::GLDisplay(Node* parent,
		       int width, int height,
		       string title,
                       bool rgb_channel_order)
    : XVirtualImg(parent, width, height, title),
      __gl_visual_info(0),
      _ctx(0),
      rgb_channel_order(rgb_channel_order),
      tex_img_width(1024),
      tex_img_height(1024) {
  }

  GLDisplay::~GLDisplay()
  {}
  
  void GLDisplay::initVisual() {
    int vi_features[] = 
      {
	GLX_RGBA,
	GLX_RED_SIZE, 8,
	GLX_GREEN_SIZE, 8,
	GLX_BLUE_SIZE, 8,
	GLX_DEPTH_SIZE, 16,
	GLX_DOUBLEBUFFER,
	None
      };
    
    __gl_visual_info = glXChooseVisual(getDisplay(), DefaultScreen(getDisplay()), vi_features);
    setVisualInfo(__gl_visual_info);
    
    if(!__gl_visual_info) {
      NODE_DEBUG_STREAM(getParent(), "GLDisplay> cannot find visual" << endl);
      return;
    }
    
    NODE_DEBUG_STREAM(getParent(), "GLDisplay> visual=0x" << hex << __gl_visual_info->visualid << dec << endl);
    NODE_DEBUG_STREAM(getParent(), "renderer='" << glGetString(GL_RENDERER) <<"'" << endl);
  }
  
  void GLDisplay::deinitVisual() {
    if(__gl_visual_info) {
      XFree(__gl_visual_info);
    }
  }
  
  void GLDisplay::initContext() {
    _ctx = glXCreateContext(getDisplay(), __gl_visual_info, None, True);
    if(!_ctx) {
      NODE_DEBUG_STREAM(getParent(), "GLDisplay> cannot create GLX context" << endl);
      return;
    }
  }
  
  void GLDisplay::deinitContext() {
    if(!_ctx) {
      glXDestroyContext(getDisplay(), _ctx);
    }
  }
  
  bool GLDisplay::initImage() {
    if(!glXQueryExtension(getDisplay(), NULL, NULL)) {
      NODE_ERROR_STREAM(getParent(), "GLDisplay> no GLX extension!!!" << endl);
      return false;
    }

    glXMakeCurrent(getDisplay(), getWindow(), _ctx);

    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    
    glEnable(GL_TEXTURE_2D);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,
		    GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
		    GL_LINEAR);
    
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, tex_img_width,
		 tex_img_height, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);

    resizeGL();

    return true;
  }

  void GLDisplay::deinitImage() {
    // unbind the thread from window and context:
    // this is important to avoid a segfault in the libGL when
    // a GLDisplay is created, deleted, and created again
    glXMakeCurrent(getDisplay(), None, NULL);
  }

  void GLDisplay::prepareOutput(Buffer* in_buffer) {
    // write the image in a GL texture
    glTexSubImage2D(GL_TEXTURE_2D, 0,
		    0, 0, getWidth(), getHeight(),
		    GL_RGBA, GL_UNSIGNED_BYTE, in_buffer->getData());


    float s = getWidth() / (float) tex_img_width;
    float t = getHeight() / (float) tex_img_height;
    
    glBegin(GL_QUADS);
    glTexCoord2f(0, t);
    glVertex2i(0, 0);
    glTexCoord2f(s, t);
    glVertex2i(getWidth(), 0);
    glTexCoord2f(s, 0);
    glVertex2i(getWidth(), getHeight());
    glTexCoord2f(0, 0);
    glVertex2i(0, getHeight());
    glEnd();
  }

  void GLDisplay::update(Buffer*) {
    // really make the image visible
    glFlush();
    resizeGL();

    glXSwapBuffers(getDisplay(), getWindow()); 
  }

  void GLDisplay::resizeGL() {
    glViewport(0, 0, getWidth(), getHeight());
  
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(0, getWidth(), 0, getHeight());
  
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();  
  }

// errror 20030507: as "checkResize() does not work yet, I comment it out here for the moment
//   void GLDisplay::checkResize(int width, int height) {
//     _width = width;
//     _height = height;
//     //   XEvent report;
//     //   bool change = false;
//     //   while (XCheckTypedWindowEvent(getDisplay(), getWindow(), ConfigureNotify, &report)) {
//     //     _width = report.xconfigure.width;
//     //     _height = report.xconfigure.height;
//     //     change = true;
//     //   }
//     //   if (change) {
//     resizeGL();
//     //GLOBAL_DEBUG_STREAM("GLDisplay resize: " << _width << " x " << _height << endl);
//     //   }
//   }

}
