/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "VideoGrabNode.hpp"
#include "nmm/base/memorymgr/B_Allocator.hpp"
#include "nmm/base/format/Property.hpp"
#include "nmm/base/registry/Plugin.hpp"
#include "nmm/utils/utils.hpp"
#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <iostream>

NMM_REGISTER_PLUGIN(NMM::VideoGrabNode, "VideoGrabNode")
 
namespace NMM {

  VideoGrabNode::VideoGrabNode(const char* name /* = "VideoGrabNode" */,
                                 StreamQueue::Mode mode /* = StreamQueue::MODE_SUSPEND */,
                                 int queue_size /* = StreamQueue::DEFAULT_SIZE */)
    : GenericSourceNode(name, mode, queue_size), 
      x_offset(0),
      y_offset(0),
      x_range(320),
      y_range(240),
      __last_timer(0), 
      __condition(__cond_mutex)
  {
    stimer = new StreamTimer(this, StreamTimer::CONST_RATE);
  }
  
  VideoGrabNode::~VideoGrabNode() {
    delete stimer;
    if (__last_timer)
      destroyLoopTimer(__last_timer);
  }
  
  void VideoGrabNode::setCrop(const int& xOffset, const int& yOffset, const int& xRange, const int& yRange)
  {
    x_offset = xOffset;
    y_offset = yOffset;
    x_range  = xRange;
    y_range  = yRange;
  } 

  void VideoGrabNode::addSupportedFormats(Property* property) {
    NMM::Format *myOutputFormat;
    myOutputFormat=property->addNewFormat("video/raw");
    myOutputFormat->addIntParamValue(Format::x_resolution_param, x_range);
    myOutputFormat->addIntParamValue(Format::y_resolution_param, y_range);
    myOutputFormat->addIntParamValue(Format::ratio_x_param, 320);
    myOutputFormat->addIntParamValue(Format::ratio_y_param, 240);
    myOutputFormat->addStringParamValue(Format::colorspace_param,"rgb24");
    myOutputFormat->addStringParamValue(Format::channelorder_param,"rgb");
    myOutputFormat->addIntParamValue(Format::bitperpixel_param,32);
    myOutputFormat->addStringParamValue(Format::endian_param, "little");
    myOutputFormat->addStringParamValue(Format::format_param, "packet");
    myOutputFormat->addFloatParamValue(Format::framerate_param, 15.0);
    property->setDefaultFormat(myOutputFormat);
  }
  
  Result VideoGrabNode::doInit() {
    addSupportedFormats(getOwnOutputProperty());
    return SUCCESS;
  }
  
  Result VideoGrabNode::doDeinitOutput() {
    XShmDetach(_display, &_shminfo);
    shmdt(_shminfo.shmaddr);
    shmctl(_shminfo.shmid, IPC_RMID, 0);
    XDestroyImage(_ximage);

    return SUCCESS;
  }

  Result VideoGrabNode::doInitOutput() {
    addSupportedFormats(getOwnWorkingOutputProperty());
    return SUCCESS;
  }
 
  Result VideoGrabNode::doActivate() {
    if (getOutputFormat() == NULL) { return FAILURE; }
    image_width = getOutputFormat()->getIntValue(Format::x_resolution_param);
    image_height = getOutputFormat()->getIntValue(Format::y_resolution_param);
    image_bpp = getOutputFormat()->getIntValue(Format::bitperpixel_param);
    image_fps = getOutputFormat()->getFloatValue(Format::framerate_param);
    DEBUG_STREAM("fps is " << image_fps << endl);
    DEBUG_STREAM("bitsperpixel is " << image_bpp << endl);
    
    stimer->reset();
    stimer->setRate(image_fps);
    
    _display = XOpenDisplay(NULL);
    if(!_display) { 
      return FAILURE; 
    }

    if(!XShmQueryExtension(_display)) {
      return FAILURE;
    }

    _root = RootWindow(_display, 0);

    _ximage = XShmCreateImage(_display,
			      DefaultVisual(_display, DefaultScreen(_display)),
			      DefaultDepth(_display, DefaultScreen(_display)),
			      ZPixmap,
			      NULL,
			      &_shminfo,
			      image_width,
			      image_height);

    if(!_ximage) { return FAILURE; }

    _shminfo.shmid = shmget(IPC_PRIVATE, 
			    _ximage->bytes_per_line * _ximage->height,
			    IPC_CREAT|0777);

    if(_shminfo.shmid < 0) { return FAILURE; }

    _shminfo.shmaddr = _ximage->data = (char*)(void*)shmat(_shminfo.shmid, 0, 0);
    _shminfo.readOnly = False;

    if(!XShmAttach(_display, &_shminfo)) { return FAILURE; } 

    XSync(_display, False);

    shmctl(_shminfo.shmid, IPC_RMID, 0);

    return SUCCESS;
  }

  Result VideoGrabNode::doStop() {
    destroyLoopTimer(__last_timer);
    __last_timer = NULL;
    return SUCCESS;
  }
  
  Message *VideoGrabNode::processBuffer(Buffer*) {
    if (!__last_timer)  {
      /* create the timer */
      __last_timer = createLoopTimer();
      /* initialize the timer with current time */
      startLoopTimer(__last_timer);
    }

    Buffer* out_buffer =  getNewBuffer((_ximage->bytes_per_line) * _ximage->height);
    
    XShmGetImage(_display, _root, _ximage, x_offset, y_offset, AllPlanes);

    memcpy(out_buffer->getData(), _ximage->data, out_buffer->getSize());

    /* wait until next frame must be send */
    waitToSendBuffer();

    /* set the new timestamp */
    stimer->setTimestamp(out_buffer);

    return(out_buffer);

  }

  void VideoGrabNode::waitToSendBuffer() {
    /* Begin> Calculate the Timestamp of the next Buffer */
    __last_timer->tv_usec += static_cast<long>(1000000000L/image_fps);
    if(__last_timer->tv_usec >=  1000000000L) {
      __last_timer->tv_sec++;
      __last_timer->tv_usec -= 1000000000L;
    }
    /* End< */

    /* wait until next frame */
    __condition.timedWait(__last_timer->tv_sec, __last_timer->tv_usec);
  }

}
