/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        NMM work group
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef _NMM_XDISPLAYNODE_HPP_
#define _NMM_XDISPLAYNODE_HPP_

#include "nmm/base/memorymgr/ShmBufferManager.hpp"
#include "nmm/base/sync/GenericUSyncSinkNode.hpp"
#include "nmm/interfaces/video/display/IVideoDisplayImpl.hpp"
#include "nmm/interfaces/video/display/IXDisplayEventSenderImpl.hpp"
#include "nmm/interfaces/video/display/IXWindowIDImpl.hpp"
#include "nmm/interfaces/video/display/IMouseCursorImpl.hpp"
#include "nmm/interfaces/av/mpeg/IStreamJumpImpl.hpp"
#include "nmm/utils/Benchmark.hpp"
#include "nmm/misc/gui/XVirtualImg.hpp"

namespace NMM {
  
  class XDisplayNode : public GenericUSyncSinkNode,
                       public IVideoDisplayImpl,
                       public IXDisplayEventSenderImpl,
		       public IXWindowIDImpl,
                       public IStreamJumpImpl,
		       public IMouseCursorImpl {
  public:
    
    /**
     * @param name name of the node
     * @param queue_mode the way how buffers are queued (or discarded)
     * @param queue_size size of the queue
     */
    XDisplayNode(const char* name = "XDisplayNode",
		 StreamQueue::Mode queue_mode = StreamQueue::MODE_SUSPEND,
		 int queue_size = 20);
    
    virtual ~XDisplayNode();
    
    Result getDepth( int& );
    
    /**
     * Refresh display.
     * I don't know what this does but I comment it anyway to be consistent.
     */
    void refresh();
    
    /**
     * \param shared_pages if value > 1 then a ShmBufferManager is used with the given number of display pages
     */
    void useSharedMemoryManager( unsigned int shared_pages );
    
  protected:
    virtual Result doInit();
    virtual Result doInitOutput();
    virtual Result doStart();
    virtual Result doStop();
    virtual Result doDeinit();
    virtual Result doDeactivate();
    virtual Result doFlush();
    
    virtual void prepareBuffer(Buffer *in_buffer);
    virtual Message* presentBuffer(Buffer *in_buffer);
    
    /** IDisplayEventSender interface. */
    virtual void enableDisplayEvents(const string& config_file);
    virtual void disableDisplayEvents();

    /** IXDisplayEventSender interface. */
    virtual void setUsingKeySymKeyEvents(const bool& ukke);
    virtual bool isUsingKeySymKeyEvents();
    virtual void setSendingMouseEventsDirect(const bool& smed);
    virtual bool isSendingMouseEventsDirect();
    virtual void setXAutoRepeat(const bool& on);
    virtual bool isXAutoRepeat();
    
    /** IXWindowID interface. */
    virtual unsigned long int getWindowID(); 

    /** IVideoDisplay interface. */
    virtual void setFullscreen( const bool& fullscreen );
    virtual bool getFullscreen() const;
    virtual bool hasFullscreen() const;
    virtual bool isFullscreenOnly() const;
    virtual bool makeScreenshot(const string &) const;
    virtual void setLeftWindowAlignment();
    virtual void setCenterWindowAlignment();
    virtual void setRightWindowAlignment();

   /** IStreamJump interface. */
    virtual Result fastForward( const int& value );
    virtual Result streamJump( const int& value );
    virtual Result sendIFramesOnly( const int& count );
    
    /** IMouseCursor interface. */
    virtual void useDotCursor(const bool& value);
    virtual void useTriangleCursor(const bool& value);
    virtual void useStandardCursor();
    
    virtual Result setFormat(const Format& format);

  private:
    string colorspace;
    int width, height;
    int depth;
    
    char *frame_title;
    
    int title_update_rate;
    
    bool fullscreen;
    int  ratio_x, ratio_y;
    bool xshm_supported;
    bool display_local;

    XVirtualImg *image;
    Buffer *last_in_buffer;
    /* holds the shared memory buffer manager, NULL if default buffer manager is used */
    ShmBufferManager* buffermanager;
    
    /* number of (display) pages for shared memory, <=1 if ShmBufferManager is not used */
    unsigned int      shared_memory_pages;
    
    Benchmark benchmark;

    ///! flag to store if display events were enabled when
    ///  deactivating, to enable the on activation again.
    bool display_events_enabled;
    ///! if display events were enabled last time, the Node was activ,
    ///  this string stores the config file parameter given on
    ///  enabling. Its needed to reenable display events on
    ///  reactivation.
    string display_events_config;

    enum WindowAlignment {
      LEFT = 0,
      CENTER = 1,
      RIGHT = 2
    } __align;

    Result allocate();
    Result setResolution(int x, int y);
    void setRatio(int x, int y);
  };
}

#endif //_NMM_XDISPLAYNODE_HPP_
