/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:         
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 */

/*
 * Code is based on spu_mixer.c that is part of Ogle, a free linux DVD player.
 * Copyrighted 2000-2001 by Bjrn Englund, Hkan Hjort
 */

class SPUDecoder {

  public:
    SPUDecoder();
    ~SPUDecoder();

    enum color_mode { YV12A, RGBA };

    /**
     * SPU Decoder
     * \param packet reference to a pointer of a spu packet, after executing the function
     *        this pointer points at the end of the decoded data
     * \param size the size of the packet, after executing the function
     *        it contains 0, if all data could be decoded
     *        it is > 0, if 'size bytes' undecoded data are in the buffer
     *        it is < 0, if not enough data is available to decode the packet,
     *        abs(size) bytes are needed to decode the buffer
     * \param palette the color palette that is used to decode the image, it is made up 16 entries
     *        each containing a 32 bit value, that describes the RGB or YUV value depending on the mode
     * \returns the decoded image (in rgba or yv12a format), NULL if the image could not be decoded
     */
    unsigned char* decode( unsigned char** packet, int* size, unsigned int* palette );

    /**
     * decodes only the command sequence of a packet and extracts informations like width and heigth of the image
     * \param packet pointer to encoded data
     * \param size size of the packet data
     * \retuns 0, if the packet contains enough data to decode the image
     *         it is >0, if more data as needed is in the packet, the surplus number of bytes are returned
     *         it is <0, if too few data are in the packet, the bytes*(-1) that are needed are returned
     */
    int decodeCS( unsigned char* packet, unsigned int size );

    /**
     * decodes the image
     * \param packet pointer to encoded data
     * \palette (see decode)
     * \image pointer where the decoded image is stored (be sure enough data is reserved, depending on the mode)
     */
    void decodeBitmap( unsigned char* packet, unsigned int* palette, unsigned char* image );

    /**
     * \returns width of the image (valid after a decodeCS call)
     */
    unsigned int getWidth();

    /**
     * \returns height of the image (valid after a decodeCS call)
     */
    unsigned int getHeight();

    /**
     * \returns x position where image should be displayed
     */
    unsigned int getXPosition();

    /**
     * \returns y position where image should be displayed
     */
    unsigned int getYPosition();

    /**
     * \returns true, if last decodeCS call was successful (image could be decoded)
     */
    bool isCSOk();

    /**
     * \returns true, if last decodeCS call found out that the spu packet has menu data
     */
    bool isMenu();

    /**
     * set color mode
     * \param m the color model in which the image is decoded, can be RGBA or YV12A
     */
    void setMode( color_mode m );

    /**
     * \returns color mode
     */
    color_mode getMode();

    /**
     * Initialize color scheme
     * \param scheme_nr number of color scheme 0-3
     * \param state_nr 0 = scheme for selection state, 1 = scheme for activation state
     * \param value the upper 16 bits describe 4 alpha values(range 0-15)
     *              the lower ones descibe 4 color values(range 0-15 = color out of the palette)
     */
    void initColorScheme( unsigned int scheme_nr, unsigned int state_nr, unsigned int value );

    /**
     * set color scheme that should be used for decoding image
     * (must be called before decodeBitmap
     * \param scheme_nr number of color scheme 0-3
     * \param state_nr 0 = scheme for selection state, 1 = scheme for activation state
     */
    void setColorScheme( unsigned int scheme_nr, unsigned int state_nr );

  private:
    /* color mode */
    color_mode mode;

    /* x and y position of the image */
    unsigned int x_start;
    unsigned int y_start;

    /* width and height of the image */
    unsigned int width;
    unsigned int height;

    /* real width and height of the image (needed for image in yv12 format, there are only even width and height */
    unsigned int real_width;
    unsigned int real_height;

    /* pointer to memory where decoded image should be placed */
    unsigned char* image;

    /**
      * 4*2 color schemes
      * a scheme is a 32-bit value: the upper 16 bits describe 4 alpha values(range 0-15),
      * the lower ones descibe 4 color values(range 0-15 = color out of the palette)
      * we have 4 color schemes for the selection buttons and 4 color schemes for the action buttons
      * note: the dvd standard has only 3*2 schemes, entries 0,0 and 0,1 are unused at the moment
      */
    unsigned int color_scheme[4][2];

    /* contains 4 colors out of the 16 color palette */
    unsigned int color[4];

    /* contains 4 alpha values */
    unsigned int alpha[4];

    /* indicates if lower or upper nibble should be used */
    bool aligned;

    /* indicates if last deocdeCS call was successful */
    bool cs_ok;

    /* indicates if last deocdeCS call found a menu */
    bool is_menu;

    /* since the encoded image is interlaced, we must save the actual field */
    unsigned int field;

    /* pointer to the two half frames */
    unsigned int fieldoffset[2];

    /* saves a byte for getNibble() */
    unsigned char next;

    /* actual byte position in the encoded stream */
    unsigned char* byte_pos;

    /* size of the spu */
    unsigned int spu_size;

    /* offset to the command sequence section */
    unsigned int cs_offset;

    /* allocates memory for decoded image */
    bool allocateMemory();

    /* set byte position for getBytes call */
    void setByte( unsigned char* byte );

    /**
     * gets up to 4 bytes out of the stream and put it in a 32 bit value
     * \param num number of bytes to get (up to 4)
     * \returns 32 bit value that is composed of the given number of bytes
     */
    unsigned int getBytes( unsigned int num );

    /**
     * gets alternatly the lower or upper half of a byte out of the stream (byte_pos),
     * increases byte_pos
     */
    unsigned char getNibble();

    /**
     * draws a part of a line in rgba format
     * \param addr pointer to the rgba line
     * \param length number of pixel that should be drawn
     * \param color  rgb value of the pixel
     * \param alpha  alpha value of the pixel
     */
    void drawRGBA( unsigned int* addr, unsigned int length, unsigned int color, unsigned int alpha );

    /**
     * draws a part of a line in yv12a format
     * \param plane1 pointer to the y plane of the line
     * \param plane2 pointer to the u plane of the line
     * \param plane3 pointer to the v plane of the line
     * \param plane4 pointer to the alpha plane of the line
     * \param field even (0) or odd (1) field
     * \param length number of pixel that should be drawn
     * \param color  yuv value of the pixel
     * \param alpha  alpha value of the pixel
     */
    void drawYV12A( unsigned char* plane1, unsigned char* plane2, unsigned char* plane3, unsigned char* plane4,
		    unsigned int field, unsigned int length, unsigned int color, unsigned int alpha );

};
