/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:         
 *                    Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef NMM_SOCKCONNECTOR_HPP
#define NMM_SOCKCONNECTOR_HPP NMM_SOCKCONNECTOR_HPP

#include "Sock.hpp"

namespace NMM {

  /** This wrapper class can be used to establish a new connection and is a wrapper class
   * for the connect and disconnect method.
   * With tcp, this is a wrapper for the socket(2) and connect(2) functions.
   * With udp, a dummy implementation is provided which simply creates a udp socket and sets its
   * peer address to the server address. No emulation of tcp functionality is done and no requests
   * are actually sent to the server udp socket.
   */
  class SockConnector {
  public:
    /** @name Constructor & Destructor
     */
    //@{
    /** Default Constructor.
     */
    SockConnector();

    /** Destructor.
     */
    ~SockConnector();
    //@}

    /** @name Connection Methods
     */
    //@{
    /** This method establish a connection to the given address.
     * This method sets the peer address of the Sock object if the connection request was successful.
     * If the transport protocol is UDP, the connection always trivially succeeds.
     * @param addr The address to connecto to.
     * @param proto The transport layer protocol that is used for the connection. The current implementation
     *        allows "tcp" or "udp"
     * @return Returns a pointer to a new Sock or SockStream that can be used to transport or receive data
     */
    Sock* connect(const INetAddr& addr, const char* proto = "tcp")
      throw (const ConnectionException&);
    
    /** This method establish a connection to the given address and stores the socket in the given Sock.
     * Unlike the other connect method, it doesn't create a new Sock object but simply sets the socket of
     * an existing Sock object.
     * This method sets the peer address of the Sock object if the connection request was successful.
     * If the transport protocol is UDP, the connection always trivially succeeds.
     * @param Sock Reference to the Sock which is responsible for the new connection
     * @param addr The address to connecto to.
     * @param proto The transport layer protocol that is used for the connection. The current implementation
     *        allows "tcp" or "udp"
     */
    void connect(Sock& sock, const INetAddr& addr, const char* proto = "tcp")
      throw (const ConnectionException&);
    //@}
  private:
    /** @name Exception methods
     * These methods create exceptions from an errno value
     */
    //@{
    void createConnectionException(int);
    void createSocketException(int);
    void createShutdownException(int);
  };
  //@}

}

#endif
