/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:         
 *                    Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef NMM_SOCKDGRAM_HPP
#define NMM_SOCKDGRAM_HPP NMM_SOCKDGRAM_HPP

#include "Sock.hpp"

namespace NMM {

  /** This class provide methods to transport data over a established connection
   */
  class SockDGram : public Sock {
  public:
    /** @name Constructor & Destructor
     */
    //@{
    /** Creates a SockDGram for a certain protocol and request a socket for it.
     * @param owner True, if this socket should be responsible for the created socket
     */
    SockDGram(bool owner = true);

    /** Creates SockDGram for an already existing socket.
     * @param socket Socket to assign to this SockDGram
     * @param owner If true, this socket is responsible for the socket
     */
    SockDGram(int socket, bool owner);

    /** Destructor. Releases the socket if SockDGram is owner of it.
     */
    ~SockDGram();
    //@}
    
    /** @name Addresses
     */
    //@{
    /** Set address of peer.
     * This method should only be called by a SockConnector or SockAcceptor.
     * Changes the address where UDP datagrams should be sent to.
     * @addr Pointer to peer address
     */
    void setPeerAddress(const INetAddr& addr);
    
    /** Returns the peer address of a socket.
     * This method gets the correct peer address only if the socket is connected.
     * @return Peer address of this socket
     */
    const INetAddr& getPeerAddress() const { return __peer_addr; } 
    
    /** Returns the local address of the socket.
     * @return Local address of this socket
     */
    INetAddr getLocalAddress() const;
    
    /** @name Socket Handling
     */
    //@{
    /** Assigns a new socket. The old socket is closed if Sock is owner of it
     * @param socket The new socket
     * @param owner Specify if Sock is responsible for the new socket
     */
    void setSocket(int socket, bool owner = false);
    
    /** returns the current socket.
     * @return current socket
     */
    inline int  getSocket()  const    { return __socket; }
    
    /** This method closes a connection.
     * In this case no difference is made if the SockStream is owner of the socket or not.
     * This implementation does nothing and is only needed for completeness of the interface.
     */
    void close();
    //@}
    
    /** @name Transmission Methods
     */
    //@{
    /** Receives data from a connection. This method uses the recvfrom(2) system call.
     * If the datagram does not come from the current peer address, a TransmissionException is thrown
     * @param buffer Buffer where the received data should be stored
     * @param len Maximum number of bytes to receive
     * @param flags Flags, as specified in recvfrom(2)
     * @return Actual number of bytes received
     * @exception TransmissionException if the transmission fails or a socket error occurs
     */
    ssize_t recv (void* buffer,size_t len, int flags) throw (const TransmissionException&);
    
    /** Sends data over a connection. This method uses the sendto(2) system call.
     * The datagram is sent to the current peer address
     * @param buffer Buffer containing the data to be sent
     * @param len Maximum number of bytes to send
     * @param flags Flags, as specified in sendto(2)
     * @return Actual number of bytes sent
     * @exception TransmissionException if the transmission fails or a socket error occurs
     */
    ssize_t send (const void* buffer,size_t len, int flags) throw (const TransmissionException&);

    /** Reads data from the socket.
     * This method seems to be obsolete and was only used in the WM-Client example
     * @param buffer Buffer where the read data should be stored
     * @param len Maximum number of bytes to read
     * @return Actual number of bytes read
     * @exception TransmissionException if the transmission fails or a socket error occurs
     */
    ssize_t read (char* buffer,size_t len) throw (const TransmissionException&);
    
    /** Writes data to the socket.
     * This method seems to be obsolete and was only used in the WM-Client example
     * @param buffer Buffer containing the data to be written
     * @param len Maximum number of bytes to write
     * @return Actual number of bytes written
     * @exception TransmissionException if the transmission fails or a socket error occurs
     */
    ssize_t write(const char* buffer, size_t len) throw (const TransmissionException&);
    
    /** Read data from the socket, treating it like a stream.
     * This method reads data until len bytes have been read or the connection has been closed.
     * @param buffer Buffer where the read data should be stored
     * @param len Maximum number of bytes to read
     * @param max_packet_size Maximum size of a single packet, 0 to use protocol default
     * @exception TransmissionException if the transmission fails or a socket error occurs
     * @exception ConnectionClosed if the connection was closed before len bytes have been read
     */
    void readStream(char* buffer, size_t len,  size_t max_packet_size = 0)
      throw (const TransmissionException&, const ConnectionClosed&);

    /** Write data from the socket, treating it like a stream.
     * This method tries to write exactly len bytes of data.
     * @param buffer Buffer containing the data to be written
     * @param len Maximum number of bytes to write
     * @param max_packet_size Maximum size of a single packet, 0 to use protocol default
     * @exception TransmissionException if the transmission fails or a socket error occurs
     * @exception ConnectionClosed if the connection was closed before len bytes have been written
     */
    void writeStream(const char* buffer, size_t length, size_t max_packet_size = 0)
      throw (const TransmissionException&, const ConnectionClosed&);
    
  private:
    int  __socket;                  ///< Stores the datagram socket
    bool __own_socket;              ///< If true, this object is responsible for the socket
    INetAddr __peer_addr;           ///< Stores the peer address to send to and receive from
  };

}

#endif //NMM_SOCKDGRAM_HPP
