/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        NMM work group,
 *                    http://www.networkmultimedia.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "utils.hpp"

#include <errno.h>
#include <stdio.h>
#include <sys/time.h> // for gettimeofday()
#include <unistd.h> // for gethostname(), getdomainname()
#include <math.h>  // for log10()
#include <sstream>
#include <fstream>

namespace NMM {

  void suspendThread(long millis) {
    struct ::timeval tv;
    if (millis > 0) {
      tv.tv_sec = millis / 1000;
      tv.tv_usec = (millis % 1000) * 1000;
      select(0, 0, 0, 0, &tv);
    }
  }
  
  looptimer createLoopTimer() {
    return ((looptimer) malloc(sizeof(struct ::timeval)));
  }

  void destroyLoopTimer(looptimer timer) {
    free(timer);
  }

  long millisDifference(const looptimer first, const looptimer last) {
    long ret;
    ret = (last->tv_sec - first->tv_sec) * 1000;
    ret += (last->tv_usec - first->tv_usec) / 1000;
    return ret;
  }

  void startLoopTimer(looptimer timer) {
    gettimeofday(timer, 0);
  }

  void delayLoopTimer(looptimer timer, long millis) {
    struct timeval timer2;
    gettimeofday(&timer2, 0);
    long difference = millisDifference(&timer2, timer);
    // wait another millis - difference milliseconds, negatives
    // will be ignored.
    suspendThread(millis - difference);
  }

  string getHostname() throw (const Exception&) {
    vector<char> tmp(255, '0');
    memset(&tmp[0], 0x0, 255);
    
    if (gethostname(&tmp[0], 255) == -1) {
      switch(errno) {
      case EINVAL        : throw Exception("length of hostname is  negative  or,  for  sethostname,  length is larger than the maximum allowed size, or length is smaller than the actual size.");

      case EPERM         : throw Exception("For sethostname, the caller was not the superuser.");

      case EFAULT        : throw Exception("name is an invalid address.");

      case ENAMETOOLONG  : throw Exception("length is larger than the maximum allowed size, or length is smaller than the actual size.");

      default            : throw Exception("Unkown error");
      }
    }
    return string(&tmp[0]);
  }

  string getDomainname() {
    vector<char> tmp(255, '0');
    memset(&tmp[0], 0x0, 255);

    if (getdomainname(&tmp[0], 255) == -1) 
      return string();
    return string(&tmp[0]);
  }

  /********************************************************************************/
  /* FileLocation stuff */
  string getNMMEnvVar(string envvarname) {
    /* get the envvarname environment variable */
    char* envvar = getenv(envvarname.c_str());
    if (NULL != envvar) {
      /* Variable is set, use value */
      return string(envvar);
    } else {
      /* Variable is not set */
      return string("");
    }
  }

  string getNMMPrefixDir() {
    /* this return the define */
    return string(INSTALL_PREFIX);
  }

  string getNMMPkgdataDir() {
    /* this return the define */
    return string(INSTALL_PKGDATADIR);
  }

  string getNMMLibDir() {
    /* this return the define */
    return string(INSTALL_LIBDIR);
  }

  string ensureTrailingSlash(const string val) {
    if ('/' != val[val.length()-1]) {
      return val + string("/");
    }
    return val;
  }

  string getNMMDir() {
    string retval = getNMMEnvVar("NMM_DEV_DIR");
    if ("" == retval) {
      retval = getNMMEnvVar("NMM_DIR");
      if ("" == retval) {
	retval = getNMMPrefixDir();
      }
    }
    
    return ensureTrailingSlash(retval);
  }

  string getNMMResourceDir() {
    string retval = getNMMEnvVar("NMM_DEV_DIR");
    if ("" != retval) {
      return ensureTrailingSlash(retval) + string("resources/");
    }


    retval = getNMMEnvVar("NMM_DIR");
    if ("" != retval) {
      return ensureTrailingSlash(retval) + string("share/nmm/");
    }

    return ensureTrailingSlash(getNMMPkgdataDir());
  }


  string getNMMResource(const string res) {
    return getNMMResourceDir() + res;
  }

  string getNMMPluginDir() {
    string retval = getNMMEnvVar("NMM_DEVLIB_DIR");
    if ("" != retval) {
      return ensureTrailingSlash(retval);
    }

    retval = getNMMEnvVar("NMM_DEV_DIR");
    if ("" != retval) {
      return ensureTrailingSlash(retval) + string("dev-lib/");
    }


    retval = getNMMEnvVar("NMM_DIR");
    if ("" != retval) {
      return ensureTrailingSlash(retval) + string("lib/");
    }

    return ensureTrailingSlash(getNMMLibDir()) + string("nmm/dev-lib");
  }

  string getNMMProjectPluginDir() {
    string retval = getNMMEnvVar("NMM_PROJECT_DIR");
    if ("" != retval) {
      return ensureTrailingSlash(retval);
    }

    return ensureTrailingSlash(getNMMPluginDir());
  }

  string getNMMLocaleDir() {
    string retval = getNMMEnvVar("NMM_DEV_DIR");
    if ("" != retval) {
      // if NMM_DEV_DIR is set use $NMM_DEV_DIR/resources/locale/
      return ensureTrailingSlash(retval) + string("resources/locale/");
    }


    retval = getNMMEnvVar("NMM_DIR");
    if ("" != retval) {
      // if NMM_DIR is set use $NMM_DIR/share/nmm/locale/
      // warning: if you change the installation dir in configure in
      // (defined there in localedir variable), you have to adapt the
      // string below (and the comment above :-) )
      return ensureTrailingSlash(retval) + string("share/nmm/locale/");
    }

    // there is no function, that simply returns the define!!!
    return ensureTrailingSlash(string(INSTALL_LOCALEDIR));
  }
  /********************************************************************************/

  bool isFilenameWithinDirectories(const string& dir_plus_filename, const string& directories)
  {
    // no restrictions -> filename is okay
    if(directories.size() == 0) {
      return true;
    }

    // try to get directory only from directory plus filename
    size_t pos_last_slash = dir_plus_filename.rfind("/", dir_plus_filename.size() - 1);
    string directory = "";
    // if a slash was found, take sub-string only
    if(pos_last_slash != string::npos) {
      directory = ensureTrailingSlash(dir_plus_filename.substr(0, pos_last_slash));
    }
    // else directory is the current working directory
    else {
      directory = ensureTrailingSlash(string(getenv("PWD")));
    }

    // iterate over all given directories
    size_t curr_pos_in_directories = 0;
    while(true) {
      // test if there is more then one directory in directories (separated with a ':')
      size_t pos_curr_dir_end = directories.find(":", curr_pos_in_directories);
      // check only the current directory
      if(pos_curr_dir_end != string::npos) {
	string curr_directory = ensureTrailingSlash(directories.substr(curr_pos_in_directories, pos_curr_dir_end - curr_pos_in_directories));
	if(directory.find(curr_directory) == 0) {
	  return true;
	}
	else { // otherwise, move to next entry and check if we are done
	  curr_pos_in_directories = pos_curr_dir_end + 1;
	  if(curr_pos_in_directories >= directories.size()) {
	    return false;
	  }
	}
      }
      else { // only one directory in directories OR last directory
	string curr_directory = ensureTrailingSlash(directories.substr(curr_pos_in_directories, directories.size() - curr_pos_in_directories));
	if(directory.find(curr_directory) == 0) {
	  return true;
	}
	else {
	  return false;
	}
      }
    }
  }

  string getAbsolutePath(const string& filename)
  {
    string tmp = filename;
    if(tmp.substr(0,2) == "./") {
      char* working_dir = new char[PATH_MAX];
      getcwd(working_dir, PATH_MAX);
      tmp = working_dir + tmp.substr(1, tmp.size() - 1);
      delete working_dir;
    }
    else if(tmp.substr(0,1) != "/") {
      char* working_dir = new char[PATH_MAX];
      getcwd(working_dir, PATH_MAX);
      tmp = working_dir + string("/") + tmp;
      delete working_dir;
    }
    return tmp;    
  }

  /********************************************************************************/

  void saveYV12BufferAsPGM(const unsigned char* buf,
                            const int wrap,
                            const int xsize,
                            const int ysize,
                            const string& filename)
  {
    ofstream f(filename.c_str());
    f << "P5" << endl;
    f << xsize << " " << ysize << endl;
    f << "255" << endl;
    for(int i=0; i < ysize; i++) {
      f.write(reinterpret_cast<const char*>(buf) + i * wrap, xsize);
    }
    f.close();
  }

  void saveRGBX32BufferAsPGM(const unsigned char* buf,
                            const int wrap,
                            const int xsize,
                            const int ysize,
                            const string& filename)
  {
    ofstream f(filename.c_str());
    f << "P5" << endl;
    f << xsize << " " << ysize << endl;
    f << "255" << endl;
    for(int i=0; i < ysize; i++) {
      for (int j=0; j < xsize; j++) {
	f.write(reinterpret_cast<const char*>(buf) + i * 4 * wrap + j * 4, 1);
      }
    }
    f.close();
  }

  void saveRGB24BufferAsPGM(const unsigned char* buf,
                            const int wrap,
                            const int xsize,
                            const int ysize,
                            const string& filename)
  {
    ofstream f(filename.c_str());
    f << "P5" << endl;
    f << xsize << " " << ysize << endl;
    f << "255" << endl;
    for(int i=0; i < ysize; i++) {
      for (int j=0; j < xsize; j++) {
	f.write(reinterpret_cast<const char *>(buf) + i * 3 * wrap + j * 3, 1);
      }
    }
    f.close();
  }

  string operator+(const string str, const int& i) {
    stringstream s;
    s<<str<<i;
    return s.str();
  }

  string itoa(const int& i) {
    stringstream s;
    s<<i;
    return s.str();
  }
    
  
  unsigned int yuv2rgb( int Y, int Cb, int Cr ) {
    int Ey, Epb, Epr;
    int Eg, Eb, Er;
    unsigned int result;

    Ey  = (Y-16);
    Epb = (Cb-128);
    Epr = (Cr-128);

    Eg = (298*Ey - 100*Epb - 208*Epr)/256;
    Eb = (298*Ey + 516*Epb)/256;
    Er = (298*Ey + 408*Epr)/256;

    if(Eg > 255)
      Eg = 255;
    if(Eg < 0)
      Eg = 0;

    if(Eb > 255)
      Eb = 255;
    if(Eb < 0)
      Eb = 0;

    if(Er > 255)
      Er = 255;
    if(Er < 0)
      Er = 0;

    result = (Eb << 16) | (Eg << 8) | Er;

    return result;
  }

  void convertYV12ToRGB24(unsigned char* rgb_image, unsigned char* yv12_image,
			unsigned int width, unsigned int height, bool use_alpha) {
    unsigned int framesize, x, y;

    unsigned char* y_plane;
    unsigned char* u_plane;
    unsigned char* v_plane;
    unsigned char* a_plane;
    unsigned int rgb;
    unsigned width_h = width>>1;
    unsigned int offset;
    unsigned char* yc;
    unsigned char* uc;
    unsigned char* vc;
    unsigned char* ac;
    unsigned char* dest = rgb_image;
    
    framesize = width*height;
    y_plane = yv12_image;
    u_plane = yv12_image+framesize;
    v_plane = u_plane+(framesize>>2);
    if ( use_alpha ) {
      a_plane = v_plane+(framesize>>2);
    }

    yc = y_plane;
    ac = a_plane;

    for( y=0; y<height; y++ ) {
      offset = width_h*(y>>1);
      uc = u_plane + offset;
      vc = v_plane + offset;
      for( x=0; x<width; x++ ) {

        rgb = yuv2rgb( *yc, *uc, *vc );
	if (use_alpha) {
          
	  unsigned char alpha = *ac;
	  *dest++ = ((rgb & 0xff) * alpha) / 255;
	  *dest++ = (((rgb>>8) & 0xff) * alpha) / 255;
	  *dest++ = (((rgb>>16) & 0xff) * alpha) / 255;
	  ac++;
        } else {
	  *dest++ = rgb & 0xff;
	  *dest++ = (rgb>>8) & 0xff;
	  *dest++ = (rgb>>16) & 0xff;
	}

	yc++;
	if (x%2) { uc++; vc++; }
      }
    }
  }

  void convertYV12ToRGB32(unsigned char* rgb_image, unsigned char* yv12_image,
			unsigned int width, unsigned int height, bool use_alpha) {
    unsigned int framesize, x, y;

    unsigned char* y_plane;
    unsigned char* u_plane;
    unsigned char* v_plane;
    unsigned char* a_plane;
    unsigned int rgb;
    unsigned width_h = width>>1;
    unsigned int offset;
    unsigned char* yc;
    unsigned char* uc;
    unsigned char* vc;
    unsigned char* ac;
    unsigned char* dest = rgb_image;
    
    framesize = width*height;
    y_plane = yv12_image;
    u_plane = yv12_image+framesize;
    v_plane = u_plane+(framesize>>2);
    if ( use_alpha ) {
      a_plane = v_plane+(framesize>>2);
    }

    yc = y_plane;
    ac = a_plane;

    for( y=0; y<height; y++ ) {
      offset = width_h*(y>>1);
      uc = u_plane + offset;
      vc = v_plane + offset;
      for( x=0; x<width; x++ ) {

        rgb = yuv2rgb( *yc, *uc, *vc );
	if (use_alpha) {
	  unsigned char alpha = *ac;
	  *dest++ = ((rgb & 0xff) * alpha) / 255;
	  *dest++ = (((rgb>>8) & 0xff) * alpha) / 255;
	  *dest++ = (((rgb>>16) & 0xff) * alpha) / 255;
          *dest++;
	  ac++;
        }else{
	  *dest++ = rgb & 0xff;
	  *dest++ = (rgb>>8) & 0xff;
	  *dest++ = (rgb>>16) & 0xff;
          *dest++;
	}

	yc++;
	if (x%2) { uc++; vc++; }
      }
    }
  }

  void convertRGBX32toRGB24InPlace(unsigned char* image,
				   unsigned int width,
				   unsigned int height ) {
    unsigned char* sp = image;
    unsigned char* dp = image;

    for (unsigned int i = 0; i < height; i++) {
      for (unsigned int j = 0; j < width; j++) {
	*dp++ = *sp++;
	*dp++ = *sp++;
	*dp++ = *sp++;
	sp++;
      }
    }
  }

  time_t stringToTime(string s){
    try{
      time_t t = time(NULL);
      struct tm *abs_time = localtime(&t);
      int min, h, day, mon, year;

      h    = atoi( (s.substr(8,2)).c_str() );
      min  = atoi( (s.substr(10,2)).c_str());
      day  = atoi( (s.substr(6,2)).c_str() );
      mon  = atoi( (s.substr(4,2)).c_str() );
      year = atoi( (s.substr(0,4)).c_str() );

      abs_time->tm_sec  = 0;
      abs_time->tm_min  = min;
      abs_time->tm_hour = h;
      abs_time->tm_mday = day;
      abs_time->tm_mon  = mon-1;
      abs_time->tm_year = year-1900;

      int d = abs_time->tm_isdst;

      t = mktime( abs_time );

      //abs_time = localtime(&t);
      t += (d - abs_time->tm_isdst)*3600;
      //cerr << d << " " << abs_time->tm_isdst  << endl;
      return t;      

    }catch(const Exception &e){
      return 0;
    }catch(...){
      return 0;
    }
  }

}
