/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        NMM work group,
 *                    http://www.networkmultimedia.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


/* This file (and the corresponding .cpp file) declare and define
 * several utility functions that might be useful in several
 * cases. The functions are usually described in the comment
 * blocks immidiately preceding them.
 */

#ifndef _NMM_UTILS_HPP_
#define _NMM_UTILS_HPP_

#include "nmm/NMMTypes.hpp"
#include <sys/time.h> // for gettimeofday()


namespace NMM {

  enum useAlpha_e{NO_ALPHA=0, DISCARD_ALPHA, USE_ALPHA};

  /**
   * This function can suspend a thread (or process) for a given
   * number of milliseconds. The time granularity may be more or
   * less than one millisecond, and this call is neither guaranteed
   * not to return before the given time, nor not to return after.
   * This function simply gives a "best effort" delay.
   * \param millis The number of milliseconds to sleep.
   */
  void suspendThread(long millis);

  /**
   * looptimer is merely a shortcut for a pointer to a timeval struct.
   * It is needed for the loop timer functions.
   */
  typedef struct ::timeval* looptimer;

  /**
   * This function allocates and returns a new loop timer.
   */
  looptimer createLoopTimer();

  /**
   * This function frees a previously allocated loop timer.
   * \param timer The timer to be destroyed.
   */
  void destroyLoopTimer(looptimer timer);

  /**
   * This function is the first of the two loop timer functions. They
   * can be used to safely, intelligently, efficiently and portably
   * delay a loop so that the beginning of a loop is not executed
   * more often than every x milliseconds. To use this, first 
   * create a new looptimer with createLoopTimer(). At the beginning of
   * the loop, call startLoopTimer(). At the end of the loop, call
   * delayLoopTimer with the number of milliseconds the loop should
   * (at least) take to complete. For example, to make sure a calculation
   * is not executed faster than in 5 times a second, call startLoopTimer(t),
   * then perform the calculation, then call delayLoopTimer(t, 200).
   * \param timer A previously allocated looptimer.
   */
  void startLoopTimer(looptimer timer);

  /**
   * This is the second loop timer function. See startLoopTimer() for
   * details.
   * \param timer The looptimer passed to startLoopTimer().
   * \param millis The time (in milliseconds) the entire loop should
   * take at the least.
   */
  void delayLoopTimer(looptimer timer, long millis);

  /**
   * Returns the difference in milliseconds between looptimer last
   * and looptimer first
   * \param first older looptimer
   * \param last newer looptimer
   */
  long millisDifference(const looptimer first, const looptimer last);

  string getHostname() throw (const Exception&);
  string getDomainname();

  /* FileLocation stuff */
  /**
   * \returns path where NMM is installed or source dir where NMM is
   * build if NMM_DEV_DIR evironment variable is set
   */
  string getNMMDir();

  /**
   * \returns path where NMM resources are found
   */
  string getNMMResourceDir();

  /**
   * \param res Resource to search
   * \returns complete path where NMM resource res is found
   */
  string getNMMResource(const string res);

  /**
   * \returns path where NMM plugins are found
   */
  string getNMMPluginDir();

  /**
   * \returns path where NMM Project plugins are found or
   * getNMMPluginDir() if environment variable NMM_PROJECT_DIR not
   * set.
   */
  string getNMMProjectPluginDir();

  /**
   * \returns path where NMM plugins are found
   */
  string getNMMLocaleDir();

  /**
   * \returns abolute path and filename for given filename
   */
  string getAbsolutePath(const string& filename);

  /**
   * \returns true when given filename (may include full path) is within the set of given directories; also true when set is empty
   * \param dir_plus_filename the filename to check
   * \param directories the directories separated with ':' 
   */
  bool isFilenameWithinDirectories(const string& dir_plus_filename, const string& directories);

  /**
   * Writes an image buffer in YV12 planar format to a file in PGM format
   * useful for testing decoding of compressed video data
   * \param image A Pointer to the image data to save.
   * \param wrap The width of one row of the image including padding.
   * \param width The width of of row to be used.
   * \param height The height of the image.
   * \param filename The name of the file to save to.
   * note: as PGM is only grayscaled, only the Y-component is stored.
   */
  void saveYV12BufferAsPGM(const unsigned char* image,
                           const int wrap,
                           const int width,
                           const int height,
                           const string& filename);

  /**
   * Writes an image buffer in RGBX32 packet format (rgb24 + 8 bit padding)
   * to a file in PGM format by just taking the red value
   * useful for testing decoding of compressed video data
   * \param image A Pointer to the image data to save.
   * \param wrap The width of one row of the image including padding.
   * \param width The width of of row to be used.
   * \param height The height of the image.
   * \param filename The name of the file to save to.
   */
  void saveRGBX32BufferAsPGM(const unsigned char* image,
                             const int wrap,
                             const int width,
                             const int height,
                             const string& filename);

  /**
   * Writes an image buffer in rgb24 packet format
   * to a file in PGM format by just taking the red value
   * useful for testing decoding of compressed video data
   * \param image A Pointer to the image data to save.
   * \param wrap The width of one row of the image including padding.
   * \param width The width of of row to be used.
   * \param height The height of the image.
   * \param filename The name of the file to save to.
   */
  void saveRGB24BufferAsPGM(const unsigned char* image,
                            const int wrap,
                            const int width,
                            const int height,
                            const string& filename);

  /**
   * Helper method for convertYV12ToRGB
   */
  unsigned int yuv2rgb(int Y, int Cb, int Cr);

  /** 
   * Converts an image in YV12 or YV12A format into an RGB image with colorspace rgb24,
   * 24 bits per pixel (no padding bits!)
   */
  void convertYV12ToRGB24(unsigned char* rgb_image, unsigned char* yv12_image, unsigned int width, unsigned int height, bool use_alpha);

  /** 
   * Converts an image in YV12 or YV12A format into an RGB image with colorspace rgb32,
   * 32 bits per pixel (no padding bits!)
   */
  void convertYV12ToRGB32(unsigned char* rgb_image, unsigned char* yv12_image, unsigned int width, unsigned int height, bool use_alpha);


  /** 
   * Converts an image in RGBX32 format (rgb24 + 8bit padding) into an RGB image in RGB24 format,
   * 24 bits per pixel (no padding bits!)
   * The conversion is done in place, no new memory is allocated
   */
  void convertRGBX32toRGB24InPlace(unsigned char*, unsigned int, unsigned int);

  /**
   * adds the possibility to do concatenate strings and ints
   */
  string operator+(const string, const int&);

  /**
   * Convert Int to String
   */
  string itoa(const int&);

  /**
   * Converts a string oft he form %Y%m%d%H%M to time_t
   */
  time_t stringToTime(string s);

}

#endif //_NMM_UTILS_HPP_
